<!DOCTYPE html>
<html lang="{{ page.lang | default: site.lang | default: "en" }}">

{%- include head.html -%}

<link rel="icon" type="image/svg+xml" href="/assets/images/favicon.svg">
<link rel="icon" type="image/png" href="/assets/images/favicon.png">

<style>
    main {
        margin: 0 50px;
    }

    #top-description {
        font-size: xx-large;
        margin-top: 3em;
        margin-left: 1em;
    }

    #intro {
        font-size: x-large;
        font-weight: lighter;
        margin-top: 2em;
        margin-left: 2em;
    }

    .dlrow {
        display: grid;
        align-items: start;
        grid-template-columns: repeat(auto-fit, minmax(22em, 1fr));
        width: 100%;
    }

    p:before, #top-description:before {
        content: "";
        width: 6em;
        display: block;
        overflow: hidden;
    }

    dl {
        display: grid;
        grid-template-rows: auto auto;
        grid-auto-flow: column;
    }
    dl dt {
        font-weight: normal;
        font-size: x-large;
        text-align: center;
        border-bottom: 1px solid #8d8;
        margin: 20px 30px;
        height: 2em;
    }
    dl dd {
        margin: 0 30px 20px 30px;
        min-height: 3em;
        color: #828282;
        font-size: medium;
    }
</style>

<body>

<a href="https://github.com/tstack/lnav">
    <img loading="lazy" width="149" style="position: absolute; z-index: 100" height="149" src="https://github.blog/wp-content/uploads/2008/12/forkme_left_green_007200.png?resize=149%2C149" class="attachment-full size-full" alt="Fork me on GitHub" data-recalc-dims="1">
</a>

{%- include header.html -%}

<main class="page-content" aria-label="Content">
    {{ content }}
</main>

{%- include footer.html -%}

</body>

</html>
