/**
 * Copyright (c) 2021, Timothy Stack
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * * Neither the name of Timothy Stack nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "xml_util.hh"

#include "config.h"
#include "fmtlib/fmt/format.h"

namespace lnav {
namespace pugixml {

std::string
get_actual_path(const pugi::xml_node& node)
{
    std::string retval;
    auto curr = node;

    while (curr) {
        switch (curr.type()) {
            case pugi::node_null:
                break;
            case pugi::node_pcdata:
                retval += "text()";
                break;
            default: {
                auto name = std::string(curr.name());

                if (curr.previous_sibling(curr.name())
                    || curr.next_sibling(curr.name())) {
                    auto sibling = curr;
                    int index = 0;

                    while (sibling) {
                        index += 1;
                        sibling = sibling.previous_sibling(curr.name());
                    }

                    name = fmt::format(FMT_STRING("{}[{}]"), name, index);
                }
                if (retval.empty()) {
                    retval = name;
                } else {
                    retval = fmt::format(FMT_STRING("{}/{}"), name, retval);
                }
                break;
            }
        }
        curr = curr.parent();
    }

    return retval;
}

}  // namespace pugixml
}  // namespace lnav
