/* override table width restrictions */
@media screen and (min-width: 767px) {

    .wy-table-responsive table td {
        /* !important prevents the common CSS stylesheets from overriding
           this as on RTD they are loaded after this stylesheet */
        white-space: normal !important;
    }

}

th p {
    margin-bottom: 0px;
}

.wy-nav-content {
    max-width: 900px;
    overflow-x: auto;
}

table.query-results p {
    font-size: 0.9em !important;
}


DL DT {
    border-bottom: 1px solid #333 !important;
}
