/* Generated by re2c 3.1 on Mon Nov 18 20:39:03 2024 */
#line 1 "../../lnav/src/data_scanner_re.re"
/**
 * Copyright (c) 2015, Timothy Stack
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * * Neither the name of Timothy Stack nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include "base/date_time_scanner.hh"
#include "config.h"
#include "data_scanner.hh"

#line 41 "../../lnav/src/data_scanner_re.cc"
enum YYCONDTYPE {
	yycbol,
	yycinit,
	yycdbldocstring,
	yycsdocstring,
	yycunits
};
#line 38 "../../lnav/src/data_scanner_re.re"


std::optional<data_scanner::tokenize_result> data_scanner::tokenize_int(text_format_t tf)
{
    data_token_t token_out = DT_INVALID;
    capture_t cap_all;
    capture_t cap_inner;
#   define YYCTYPE unsigned char
#   define CAPTURE(tok) { \
        if (YYCURSOR.val == EMPTY) { \
            this->ds_next_offset = this->ds_input.length(); \
        } else { \
            this->ds_next_offset = YYCURSOR.val - this->ds_input.udata(); \
        } \
        cap_all.c_end = this->ds_next_offset; \
        cap_inner.c_end = this->ds_next_offset; \
        token_out = tok; \
    }

#   define RET(tok) { \
        CAPTURE(tok); \
        return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()}; \
    }
    static const unsigned char *EMPTY = (const unsigned char *) "";

    struct _YYCURSOR {
        YYCTYPE operator*() const {
            if (this->val < this->lim) {
                return *val;
            }
            return '\0';
        }

        operator const YYCTYPE *() const {
            if (this->val < this->lim) {
                return this->val;
            }
            return EMPTY;
        }

        const YYCTYPE *operator=(const YYCTYPE *rhs) {
            this->val = rhs;
            return rhs;
        }

        const YYCTYPE *operator+(int rhs) {
            return this->val + rhs;
        }

        const _YYCURSOR *operator-=(int rhs) {
            this->val -= rhs;
            return this;
        }

        _YYCURSOR& operator++() {
            this->val += 1;
            return *this;
        }

        const YYCTYPE *val{nullptr};
        const YYCTYPE *lim{nullptr};
    } YYCURSOR;
    YYCURSOR = (const unsigned char *) this->ds_input.udata() + this->ds_next_offset;
    _YYCURSOR yyt1;
    _YYCURSOR yyt2;
    _YYCURSOR yyt3;
    _YYCURSOR yyt4;
    _YYCURSOR hunk_heading;
    _YYCURSOR table_start;
    _YYCURSOR table_end;
    const YYCTYPE *YYLIMIT = (const unsigned char *) this->ds_input.end();
    const YYCTYPE *YYMARKER = YYCURSOR;
    class _yycond {
    public:
        int operator()() const {
            return this->val;
        }

        void operator=(int v) {
            this->val = v;
        }

        int val{yycinit};
    } c;

    if (this->ds_bol) {
        c = yycbol;
    }
    this->ds_bol = false;
    if (this->ds_units) {
        c = yycunits;
    }
    this->ds_units = false;

    YYCURSOR.lim = YYLIMIT;

    cap_all.c_begin = this->ds_next_offset;
    cap_all.c_end = this->ds_next_offset;
    cap_inner.c_begin = this->ds_next_offset;
    cap_inner.c_end = this->ds_next_offset;

    
#line 152 "../../lnav/src/data_scanner_re.cc"
{
	uint8_t yych;
	unsigned int yyaccept = 0;
	static const unsigned char yybm_bol[] = {
		/* table 1 .. 8: 0 */
		  0, 124, 124, 124, 124, 124, 124, 124, 
		124, 253, 124, 124, 124, 125, 124, 124, 
		124, 124, 124, 124, 124, 124, 124, 124, 
		124, 124, 124, 124, 124, 124, 124, 124, 
		253, 124,  12, 124, 124, 124, 124,  12, 
		124, 124,  88, 124, 124,  73, 124, 110, 
		109, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 124, 124, 124,  96, 110, 
		124, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 124,  12, 124, 124, 109, 
		124, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 109, 109, 109, 109, 109, 
		109, 109, 109, 124, 124, 124, 124, 124, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 9 .. 16: 256 */
		  0, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253,  73, 253, 253, 253, 253, 144, 
		253, 253, 224, 253, 253, 226, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253,   1, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 252, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 17 .. 24: 512 */
		  0, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254,  92, 254, 254, 255, 255, 186, 
		254, 254, 118, 254, 254,   7, 254, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 142, 254, 254, 255, 142, 143, 
		254, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 254,  24, 254, 254, 142, 
		254, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 254, 254, 254, 254, 254, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 25 .. 32: 768 */
		  0, 144, 144, 144, 144, 144, 144, 144, 
		144, 128,   0, 144, 144, 128, 144, 144, 
		144, 144, 144, 144, 144, 144, 128, 144, 
		144, 144, 144, 128, 144, 144, 144, 144, 
		128, 144, 128, 144, 144, 128, 128, 128, 
		128, 128, 152, 144, 144, 128, 148, 128, 
		160, 160, 160, 160, 160, 160, 160, 160, 
		160, 160, 176, 144, 144, 128, 144, 128, 
		144, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 224, 224, 224, 224, 224, 
		224, 224, 224, 128, 144, 129, 144, 176, 
		144, 226, 226, 226, 226, 226, 226, 226, 
		226, 226, 226, 226, 226, 226, 226, 226, 
		226, 226, 226, 226, 226, 226, 226, 226, 
		226, 226, 226, 128, 144, 128, 144, 144, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 33 .. 40: 1024 */
		  0,  16,  16,  16,  16,  16,  16,  16, 
		 16,  16,  16,  16,  16,  16,  16,  16, 
		 16,  16,  16,  16,  16,  16,  16,  16, 
		 16,  16,  16,  16,  16,  16,  16,  16, 
		 16,  16,  16,  16,  16,  16,  16,  16, 
		 16,  16,   0,  16,  16, 241,  16,  16, 
		255, 255, 255, 255, 255, 255, 255, 255, 
		251, 251,  48,  16,  16,  16,  16,  16, 
		 16, 249, 249, 249, 249, 249, 249, 241, 
		241, 241, 241, 241, 241, 241, 241, 241, 
		241, 241, 241, 241, 241, 241, 241, 241, 
		241, 241, 241,  16,  16,  16,  16, 240, 
		 16, 185, 185, 185, 185, 185, 185, 177, 
		177, 177, 177, 177, 177, 177, 177, 177, 
		177, 177, 177, 177, 177, 177, 177, 177, 
		177, 177, 177,  16,  16,  16,  16,  16, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 41 .. 48: 1280 */
		  0, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 128, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		192, 192, 192, 192, 192, 192, 192, 192, 
		198, 193, 192, 193, 193, 193, 193, 200, 
		193, 193, 192, 192, 192, 197, 193, 193, 
		245, 245, 245, 245, 245, 245, 245, 245, 
		245, 245, 192, 192, 192, 192, 192, 192, 
		193, 245, 245, 245, 245, 245, 245, 245, 
		245, 245, 245, 245, 245, 245, 245, 245, 
		245, 245, 245, 245, 245, 245, 245, 245, 
		245, 245, 245, 192, 193,   0, 193, 213, 
		192, 201, 201, 201, 201, 201, 201, 201, 
		201, 201, 201, 201, 201, 201, 201, 201, 
		201, 201, 201, 201, 201, 201, 201, 201, 
		201, 201, 201, 192, 192, 192, 193, 192, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 49 .. 56: 1536 */
		  0,  36,  36,  36,  36,  36,  36,  36, 
		 36,  37,   0,  36,  36,  37,  36,  36, 
		 36,  36,  36,  36,  36,  36,   0,  36, 
		 36,  36,  36,   0,  36,  36,  36,  36, 
		 37, 164,  32, 164, 164, 180, 164,   4, 
		164, 164, 164,  52,  36, 180, 180, 164, 
		246, 246, 246, 246, 246, 246, 246, 246, 
		246, 246,  36,  38,  36,  46,  36,  38, 
		164, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244,  36, 128,  36, 164, 180, 
		 36, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244, 244, 244, 244, 244, 244, 
		244, 244, 244,  36,  36,  36, 164,  36, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 57 .. 58: 1792 */
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,  64,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		128, 128, 128, 128, 128, 128, 128, 128, 
		128, 128,   0,   0,   0,   0,   0,   0, 
		  0, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128,   0,   0,   0,   0, 128, 
		  0, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128, 128, 128, 128, 128, 128, 
		128, 128, 128,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
	};
	static const unsigned char yybm_init[] = {
		/* table 1 .. 8: 0 */
		  0, 252, 252, 252, 252, 252, 252, 252, 
		252, 253, 124, 252, 252, 253, 252, 252, 
		252, 252, 252, 252, 252, 252, 252, 252, 
		252, 252, 252, 252, 252, 252, 252, 252, 
		253, 252, 140, 252, 252, 252, 252, 140, 
		252, 252, 216, 252, 252, 201, 252, 238, 
		237, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 252, 252, 252, 224, 238, 
		252, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 252, 140, 252, 252, 237, 
		252, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 237, 237, 237, 237, 237, 
		237, 237, 237, 252, 252, 252, 252, 252, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 9 .. 16: 256 */
		  0, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253,  73, 253, 253, 253, 253, 144, 
		253, 253, 224, 253, 253, 226, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253,   1, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 252, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		253, 253, 253, 253, 253, 253, 253, 253, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 17 .. 24: 512 */
		  0, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254, 254, 254, 254, 254, 254, 254, 
		254, 254,  92, 254, 254, 255, 255, 186, 
		254, 254, 118, 254, 254,   7, 254, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 142, 254, 254, 255, 142, 143, 
		254, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 254,  24, 254, 254, 142, 
		254, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 143, 143, 143, 143, 143, 
		143, 143, 143, 254, 254, 254, 254, 254, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 25 .. 32: 768 */
		  0,  32,  32,  32,  32,  32,  32,  32, 
		 32,   0,   0,  32,  32,   0,  32,  32, 
		 32,  32,  32,  32,  32,  32,   0,  32, 
		 32,  32,  32,   0,  32,  32,  32,  32, 
		  0,  32,   0,  32,  32,   0,   0,   0, 
		  0,   0,  48,  32,  32,   3,  40,   0, 
		 67,  67,  67,  67,  67,  67,  67,  67, 
		 67,  67,  97,  32,  32,   0,  32,   0, 
		 32, 195, 195, 195, 195, 195, 195, 195, 
		195, 195, 195, 195, 195, 195, 195, 195, 
		195, 195, 195, 195, 195, 195, 195, 195, 
		195, 195, 195,   0,  32,   0,  32,  99, 
		 32, 199, 199, 199, 199, 199, 199, 199, 
		199, 199, 199, 199, 199, 199, 199, 199, 
		199, 199, 199, 199, 199, 199, 199, 199, 
		199, 199, 199,   0,  32,   0,  32,  32, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 33 .. 40: 1024 */
		  0, 132, 132, 132, 132, 132, 132, 132, 
		132, 132, 132, 132, 132, 132, 132, 132, 
		132, 132, 132, 132, 132, 132, 132, 132, 
		132, 132, 132, 132, 132, 132, 132, 132, 
		133, 132, 132, 132, 132, 132, 132, 134, 
		132, 132,   4, 132, 132, 140, 132, 132, 
		252, 252, 252, 252, 252, 252, 252, 252, 
		220, 220, 132, 132, 132, 132, 132, 132, 
		132, 204, 204, 204, 204, 204, 204, 140, 
		140, 140, 140, 140, 140, 140, 140, 140, 
		140, 140, 140, 140, 140, 140, 140, 140, 
		140, 140, 140, 132, 132, 128, 132, 132, 
		132, 206, 206, 206, 206, 206, 206, 142, 
		142, 142, 142, 142, 142, 142, 142, 142, 
		142, 142, 142, 142, 142, 142, 142, 142, 
		142, 142, 142, 132, 132, 132, 132, 132, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 41 .. 48: 1280 */
		  0,  18,  18,  18,  18,  18,  18,  18, 
		 18,  18,   0,  18,  18,  18,  18,  18, 
		 18,  18,  18,  18,  18,  18,   0,  18, 
		 18,  18,  18,   0,  18,  18,  18,  18, 
		 18, 210,  16, 210, 210, 218, 210,   2, 
		210, 210,  82,  26,  18, 218, 218, 210, 
		251, 251, 251, 251, 251, 251, 251, 251, 
		251, 251,  18,  19,  18,  23,  18,  19, 
		210, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250,  18, 192,  18, 210, 218, 
		 18, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250, 250, 250, 250, 250, 250, 
		250, 250, 250,  18,  18,  18, 210,  18, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		/* table 49 .. 51: 1536 */
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0, 128,   0,   0,   0, 128,   0,   0, 
		  0,   0,   0,   0,   0,   0,  32,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		128,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		 64,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,   0,   0,   0,   0,   0,   0, 
		  0,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,   0,   0,   0,   0,  64, 
		  0,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,  64,  64,  64,  64,  64, 
		 64,  64,  64,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
		  0,   0,   0,   0,   0,   0,   0,   0, 
	};
	if (c() < 2) {
		if (c() < 1) {
			goto yyc_bol;
		} else {
			goto yyc_init;
		}
	} else {
		if (c() < 3) {
			goto yyc_dbldocstring;
		} else {
			if (c() < 4) {
				goto yyc_sdocstring;
			} else {
				goto yyc_units;
			}
		}
	}
/* *********************************** */
yyc_bol:
	yych = *YYCURSOR;
	if (yybm_bol[1792+yych] & 64) {
		goto yy8;
	}
	if (yych <= 'C') {
		if (yych <= ')') {
			if (yych <= 0x1F) {
				if (yych <= '\n') {
					if (yych <= 0x00) goto yy1;
					if (yych <= 0x08) goto yy2;
					if (yych <= '\t') goto yy4;
					goto yy6;
				} else {
					if (yych <= '\r') {
						if (yych <= '\f') goto yy2;
						goto yy7;
					} else {
						if (yych == 0x1B) goto yy9;
						goto yy2;
					}
				}
			} else {
				if (yych <= '$') {
					if (yych <= '!') {
						if (yych <= ' ') goto yy4;
						goto yy2;
					} else {
						if (yych <= '"') goto yy10;
						if (yych <= '#') goto yy11;
						goto yy2;
					}
				} else {
					if (yych <= '&') {
						if (yych <= '%') goto yy12;
						goto yy2;
					} else {
						if (yych <= '\'') goto yy13;
						if (yych <= '(') goto yy14;
						goto yy15;
					}
				}
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					if (yych <= '+') {
						if (yych <= '*') goto yy2;
						goto yy16;
					} else {
						if (yych <= ',') goto yy17;
						if (yych <= '-') goto yy18;
						goto yy19;
					}
				} else {
					if (yych <= '0') {
						if (yych <= '/') goto yy21;
						goto yy23;
					} else {
						if (yych <= '1') goto yy25;
						if (yych <= '2') goto yy26;
						goto yy27;
					}
				}
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						if (yych <= ':') goto yy28;
						goto yy30;
					} else {
						if (yych <= '<') goto yy31;
						if (yych <= '=') goto yy33;
						goto yy34;
					}
				} else {
					if (yych <= '@') {
						if (yych <= '?') goto yy2;
						goto yy35;
					} else {
						if (yych <= 'A') goto yy36;
						if (yych <= 'B') goto yy38;
						goto yy40;
					}
				}
			}
		}
	} else {
		if (yych <= 'f') {
			if (yych <= '[') {
				if (yych <= 'M') {
					if (yych <= 'F') {
						if (yych <= 'E') goto yy38;
						goto yy41;
					} else {
						if (yych == 'I') goto yy43;
						goto yy42;
					}
				} else {
					if (yych <= 'S') {
						if (yych <= 'N') goto yy44;
						goto yy42;
					} else {
						if (yych <= 'T') goto yy45;
						if (yych <= 'Z') goto yy42;
						goto yy46;
					}
				}
			} else {
				if (yych <= '`') {
					if (yych <= ']') {
						if (yych <= '\\') goto yy48;
						goto yy49;
					} else {
						if (yych == '_') goto yy50;
						goto yy2;
					}
				} else {
					if (yych <= 'c') {
						if (yych <= 'a') goto yy52;
						goto yy53;
					} else {
						if (yych <= 'd') goto yy54;
						if (yych <= 'e') goto yy53;
						goto yy55;
					}
				}
			}
		} else {
			if (yych <= '|') {
				if (yych <= 's') {
					if (yych <= 'n') {
						if (yych <= 'm') goto yy56;
						goto yy57;
					} else {
						if (yych == 'r') goto yy58;
						goto yy56;
					}
				} else {
					if (yych <= 'u') {
						if (yych <= 't') goto yy59;
						goto yy60;
					} else {
						if (yych <= 'z') goto yy56;
						if (yych <= '{') goto yy61;
						goto yy2;
					}
				}
			} else {
				if (yych <= 0xE0) {
					if (yych <= 0x7F) {
						if (yych <= '}') goto yy62;
						goto yy2;
					} else {
						if (yych <= 0xC1) goto yy63;
						if (yych <= 0xDF) goto yy65;
						goto yy66;
					}
				} else {
					if (yych <= 0xF0) {
						if (yych <= 0xEF) goto yy67;
						goto yy68;
					} else {
						if (yych <= 0xF3) goto yy69;
						if (yych <= 0xF4) goto yy70;
						goto yy63;
					}
				}
			}
		}
	}
yy1:
	++YYCURSOR;
#line 174 "../../lnav/src/data_scanner_re.re"
	{ return std::nullopt; }
#line 847 "../../lnav/src/data_scanner_re.cc"
yy2:
	++YYCURSOR;
yy3:
#line 491 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_GARBAGE); }
#line 853 "../../lnav/src/data_scanner_re.cc"
yy4:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy72;
	if (yych <= '9') goto yy73;
	if (yych <= ':') goto yy75;
	goto yy72;
yy5:
#line 488 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WHITE); }
#line 864 "../../lnav/src/data_scanner_re.cc"
yy6:
	++YYCURSOR;
#line 484 "../../lnav/src/data_scanner_re.re"
	{
           this->ds_bol = true;
           RET(DT_LINE);
       }
#line 872 "../../lnav/src/data_scanner_re.cc"
yy7:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '\n') goto yy6;
		goto yy72;
	} else {
		if (yych <= '9') goto yy73;
		if (yych <= ':') goto yy75;
		goto yy72;
	}
yy8:
	yych = *++YYCURSOR;
	if (yybm_bol[1792+yych] & 64) {
		goto yy8;
	}
#line 176 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_ZERO_WIDTH_SPACE);
       }
#line 893 "../../lnav/src/data_scanner_re.cc"
yy9:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '[') goto yy76;
	goto yy3;
yy10:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1A) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy3;
			if (yych <= '\t') goto yy78;
			goto yy3;
		} else {
			if (yych == 0x16) goto yy3;
			goto yy78;
		}
	} else {
		if (yych <= '"') {
			if (yych <= 0x1B) goto yy3;
			if (yych <= '!') goto yy78;
			goto yy79;
		} else {
			if (yych <= 0x7F) goto yy78;
			if (yych <= 0xC1) goto yy3;
			if (yych <= 0xF4) goto yy78;
			goto yy3;
		}
	}
yy11:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '[') goto yy88;
	goto yy3;
yy12:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy3;
		} else {
			if (yych == ',') goto yy3;
			if (yych <= '.') goto yy90;
			goto yy3;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy3;
			if (yych <= 'Z') goto yy90;
			goto yy3;
		} else {
			if (yych == '`') goto yy3;
			if (yych <= 'z') goto yy90;
			goto yy3;
		}
	}
yy13:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1A) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy3;
			if (yych <= '\t') goto yy93;
			goto yy3;
		} else {
			if (yych == 0x16) goto yy3;
			goto yy93;
		}
	} else {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy3;
			if (yych <= '&') goto yy93;
			goto yy94;
		} else {
			if (yych <= 0x7F) goto yy93;
			if (yych <= 0xC1) goto yy3;
			if (yych <= 0xF4) goto yy93;
			goto yy3;
		}
	}
yy14:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy102;
#line 395 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LPAREN); }
#line 981 "../../lnav/src/data_scanner_re.cc"
yy15:
	++YYCURSOR;
#line 396 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RPAREN); }
#line 986 "../../lnav/src/data_scanner_re.cc"
yy16:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '0') {
		if (yych <= '+') {
			if (yych == '%') goto yy108;
			if (yych <= '*') goto yy3;
			goto yy104;
		} else {
			if (yych <= ',') goto yy3;
			if (yych <= '.') goto yy108;
			if (yych <= '/') goto yy3;
			goto yy105;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy106;
			if (yych <= '?') goto yy3;
			if (yych <= 'Z') goto yy108;
			goto yy3;
		} else {
			if (yych == '`') goto yy3;
			if (yych <= 'z') goto yy108;
			goto yy3;
		}
	}
yy17:
	++YYCURSOR;
#line 387 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COMMA); }
#line 1017 "../../lnav/src/data_scanner_re.cc"
yy18:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy108;
			if (yych <= '*') goto yy3;
			goto yy108;
		} else {
			if (yych <= ',') goto yy3;
			if (yych <= '-') goto yy109;
			if (yych <= '.') goto yy108;
			goto yy3;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '0') goto yy110;
			if (yych <= '9') goto yy111;
			if (yych <= '?') goto yy3;
			goto yy108;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy3;
				goto yy108;
			} else {
				if (yych <= '`') goto yy3;
				if (yych <= 'z') goto yy108;
				goto yy3;
			}
		}
	}
yy19:
	yyaccept = 2;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy90;
			if (yych >= '+') goto yy90;
		} else {
			if (yych <= ',') goto yy20;
			if (yych <= '-') goto yy90;
			if (yych <= '.') goto yy112;
			goto yy113;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy20;
			if (yych <= 'Z') goto yy90;
		} else {
			if (yych == '`') goto yy20;
			if (yych <= 'z') goto yy90;
		}
	}
yy20:
#line 489 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_DOT); }
#line 1075 "../../lnav/src/data_scanner_re.cc"
yy21:
	yych = *++YYCURSOR;
	if (yych == '*') goto yy116;
	if (yych == 'P') goto yy117;
	goto yy115;
yy22:
#line 300 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PATH); }
#line 1084 "../../lnav/src/data_scanner_re.cc"
yy23:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy123;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7': goto yy125;
		case '8':
		case '9': goto yy127;
		case ':': goto yy129;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy131;
		case 'B':
		case 'b': goto yy132;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		case 'x': goto yy142;
		default: goto yy24;
	}
yy24:
#line 444 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_NUMBER);
           auto sf = this->to_string_fragment(cap_all);
           if (isalpha(sf.back())) {
               while (isalpha(sf.back())) {
                   sf.pop_back();
               }
               cap_all.c_end = sf.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = sf.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               this->ds_units = true;
           }
           return tokenize_result{DT_NUMBER, cap_all, cap_inner, this->ds_input.data()};
       }
#line 1180 "../../lnav/src/data_scanner_re.cc"
yy25:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy123;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy143;
		case ':': goto yy129;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy131;
		case 'B':
		case 'b': goto yy132;
		case 'E':
		case 'e': goto yy144;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy26:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy123;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy143;
		case '5': goto yy145;
		case '6':
		case '7':
		case '8':
		case '9': goto yy146;
		case ':': goto yy129;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy131;
		case 'B':
		case 'b': goto yy132;
		case 'E':
		case 'e': goto yy144;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy27:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy123;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy146;
		case ':': goto yy129;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy131;
		case 'B':
		case 'b': goto yy132;
		case 'E':
		case 'e': goto yy144;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy28:
	yyaccept = 4;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy147;
yy29:
#line 385 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COLON); }
#line 1425 "../../lnav/src/data_scanner_re.cc"
yy30:
	++YYCURSOR;
#line 388 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_SEMI); }
#line 1430 "../../lnav/src/data_scanner_re.cc"
yy31:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '>') {
		if (yych <= '-') {
			if (yych == '!') goto yy148;
			if (yych >= '-') goto yy149;
		} else {
			if (yych <= '.') goto yy32;
			if (yych <= '/') goto yy150;
			if (yych <= ':') goto yy149;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy151;
			if (yych <= '@') goto yy32;
			if (yych <= 'Z') goto yy149;
		} else {
			if (yych == '`') goto yy32;
			if (yych <= 'z') goto yy149;
		}
	}
yy32:
#line 397 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LANGLE); }
#line 1456 "../../lnav/src/data_scanner_re.cc"
yy33:
	++YYCURSOR;
#line 386 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EQUALS); }
#line 1461 "../../lnav/src/data_scanner_re.cc"
yy34:
	++YYCURSOR;
#line 398 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RANGLE); }
#line 1466 "../../lnav/src/data_scanner_re.cc"
yy35:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '@') goto yy152;
	goto yy3;
yy36:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '&') {
			if (yych <= ' ') {
				if (yych >= ' ') goto yy165;
			} else {
				if (yych == '%') goto yy165;
			}
		} else {
			if (yych <= '+') {
				if (yych <= '\'') goto yy155;
				if (yych >= '+') goto yy165;
			} else {
				if (yych <= ',') goto yy37;
				if (yych <= '/') goto yy165;
				goto yy160;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy161;
			} else {
				if (yych <= '@') goto yy165;
				if (yych <= 'F') goto yy162;
				goto yy163;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy165;
			} else {
				if (yych <= 'f') goto yy166;
				if (yych <= 'z') goto yy167;
			}
		}
	}
yy37:
#line 466 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WORD); }
#line 1513 "../../lnav/src/data_scanner_re.cc"
yy38:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '&') {
			if (yych != '\n') goto yy165;
		} else {
			if (yych <= '\'') goto yy155;
			if (yych <= '/') goto yy165;
			if (yych <= '9') goto yy160;
			goto yy161;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy165;
			if (yych <= 'F') goto yy162;
			goto yy163;
		} else {
			if (yych <= '`') goto yy165;
			if (yych <= 'f') goto yy166;
			if (yych <= 'z') goto yy167;
			goto yy165;
		}
	}
yy39:
#line 476 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 1543 "../../lnav/src/data_scanner_re.cc"
yy40:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '\'') {
			if (yych == '\n') goto yy39;
			if (yych <= '&') goto yy165;
			goto yy155;
		} else {
			if (yych <= '/') goto yy165;
			if (yych <= '9') goto yy160;
			if (yych <= ':') goto yy161;
			goto yy165;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= 'F') goto yy162;
			if (yych == 'R') goto yy168;
			goto yy163;
		} else {
			if (yych <= '`') goto yy165;
			if (yych <= 'f') goto yy166;
			if (yych <= 'z') goto yy167;
			goto yy165;
		}
	}
yy41:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '\'') {
			if (yych == '\n') goto yy39;
			if (yych <= '&') goto yy165;
			goto yy155;
		} else {
			if (yych <= '/') goto yy165;
			if (yych <= '9') goto yy160;
			if (yych <= ':') goto yy161;
			goto yy165;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'A') goto yy169;
			if (yych <= 'F') goto yy162;
			if (yych <= 'Z') goto yy163;
			goto yy165;
		} else {
			if (yych <= 'a') goto yy170;
			if (yych <= 'f') goto yy166;
			if (yych <= 'z') goto yy167;
			goto yy165;
		}
	}
yy42:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '\n') {
			if (yych <= '\t') goto yy172;
			goto yy39;
		} else {
			if (yych == '\'') goto yy155;
			goto yy172;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy173;
			if (yych <= '@') goto yy172;
			goto yy163;
		} else {
			if (yych <= '`') goto yy172;
			if (yych <= 'z') goto yy167;
			goto yy172;
		}
	}
yy43:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '%') {
			if (yych == ' ') goto yy172;
			if (yych <= '$') goto yy37;
			goto yy172;
		} else {
			if (yych <= '\'') {
				if (yych <= '&') goto yy37;
				goto yy155;
			} else {
				if (yych == '+') goto yy172;
				goto yy37;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy172;
				goto yy173;
			} else {
				if (yych <= '?') goto yy37;
				if (yych <= '@') goto yy172;
				goto yy163;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy37;
				goto yy172;
			} else {
				if (yych <= '`') goto yy37;
				if (yych <= 'z') goto yy167;
				goto yy37;
			}
		}
	}
yy44:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '&') {
			if (yych == '\n') goto yy39;
			goto yy172;
		} else {
			if (yych <= '\'') goto yy155;
			if (yych == ':') goto yy173;
			goto yy172;
		}
	} else {
		if (yych <= '`') {
			if (yych == 'U') goto yy174;
			if (yych <= 'Z') goto yy163;
			goto yy172;
		} else {
			if (yych == 'o') goto yy175;
			if (yych <= 'z') goto yy167;
			goto yy172;
		}
	}
yy45:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '&') {
			if (yych == '\n') goto yy39;
			goto yy172;
		} else {
			if (yych <= '\'') goto yy155;
			if (yych == ':') goto yy173;
			goto yy172;
		}
	} else {
		if (yych <= '`') {
			if (yych == 'R') goto yy176;
			if (yych <= 'Z') goto yy163;
			goto yy172;
		} else {
			if (yych == 'r') goto yy177;
			if (yych <= 'z') goto yy167;
			goto yy172;
		}
	}
yy46:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '\\') {
		if (yych <= 0x00) goto yy47;
		if (yych != '\n') goto yy179;
	} else {
		if (yych <= 0x7F) {
			if (yych <= ']') goto yy180;
			goto yy179;
		} else {
			if (yych <= 0xC1) goto yy47;
			if (yych <= 0xF4) goto yy179;
		}
	}
yy47:
#line 393 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LSQUARE); }
#line 1721 "../../lnav/src/data_scanner_re.cc"
yy48:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '[') {
			if (yych == '\n') goto yy3;
			goto yy187;
		} else {
			if (yych <= '\\') goto yy113;
			if (yych == 'n') goto yy6;
			goto yy187;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy3;
			if (yych <= 0xDF) goto yy188;
			if (yych <= 0xE0) goto yy189;
			goto yy190;
		} else {
			if (yych <= 0xF0) goto yy191;
			if (yych <= 0xF3) goto yy192;
			if (yych <= 0xF4) goto yy193;
			goto yy3;
		}
	}
yy49:
	++YYCURSOR;
#line 394 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RSQUARE); }
#line 1751 "../../lnav/src/data_scanner_re.cc"
yy50:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy51:
	if (yybm_bol[1792+yych] & 128) {
		goto yy50;
	}
	switch (yych) {
		case '%':
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy159;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		case ':': goto yy194;
		case '@': goto yy130;
		default: goto yy39;
	}
yy52:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy37;
				goto yy51;
			} else {
				if (yych == '\'') goto yy155;
				goto yy37;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy51;
				goto yy37;
			} else {
				if (yych <= '/') goto yy51;
				if (yych <= '9') goto yy195;
				goto yy196;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy37;
				goto yy51;
			} else {
				if (yych <= 'F') goto yy197;
				if (yych <= 'Z') goto yy198;
				goto yy37;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy51;
				goto yy37;
			} else {
				if (yych <= 'f') goto yy166;
				if (yych <= 'z') goto yy167;
				goto yy37;
			}
		}
	}
yy53:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '\'') goto yy155;
			goto yy51;
		} else {
			if (yych <= '9') goto yy195;
			if (yych <= ':') goto yy196;
			goto yy51;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy197;
			if (yych <= 'Z') goto yy198;
			goto yy51;
		} else {
			if (yych <= 'f') goto yy166;
			if (yych <= 'z') goto yy167;
			goto yy51;
		}
	}
yy54:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= '/') {
			if (yych == '\'') goto yy155;
			goto yy51;
		} else {
			if (yych <= '9') goto yy195;
			if (yych <= ':') goto yy196;
			if (yych <= '@') goto yy51;
			goto yy197;
		}
	} else {
		if (yych <= 'f') {
			if (yych <= 'Z') goto yy198;
			if (yych <= '`') goto yy51;
			goto yy166;
		} else {
			if (yych == 'i') goto yy199;
			if (yych <= 'z') goto yy167;
			goto yy51;
		}
	}
yy55:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '&') {
			if (yych == '"') goto yy200;
			goto yy51;
		} else {
			if (yych <= '\'') goto yy201;
			if (yych <= '/') goto yy51;
			if (yych <= '9') goto yy195;
			goto yy196;
		}
	} else {
		if (yych <= '`') {
			if (yych <= '@') goto yy51;
			if (yych <= 'F') goto yy197;
			if (yych <= 'Z') goto yy198;
			goto yy51;
		} else {
			if (yych <= 'a') goto yy170;
			if (yych <= 'f') goto yy166;
			if (yych <= 'z') goto yy167;
			goto yy51;
		}
	}
yy56:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych == '\'') goto yy155;
		goto yy203;
	} else {
		if (yych <= 'Z') goto yy198;
		if (yych <= '`') goto yy203;
		if (yych <= 'z') goto yy167;
		goto yy203;
	}
yy57:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych == '\'') goto yy155;
		if (yych <= '@') goto yy203;
		goto yy198;
	} else {
		if (yych <= 't') {
			if (yych <= '`') goto yy203;
			goto yy167;
		} else {
			if (yych <= 'u') goto yy205;
			if (yych <= 'z') goto yy167;
			goto yy203;
		}
	}
yy58:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '"') {
			if (yych <= '!') goto yy203;
			goto yy200;
		} else {
			if (yych == '\'') goto yy201;
			goto yy203;
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'Z') goto yy198;
			if (yych <= '`') goto yy203;
			goto yy167;
		} else {
			if (yych <= 'e') goto yy206;
			if (yych <= 'z') goto yy167;
			goto yy203;
		}
	}
yy59:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych == '\'') goto yy155;
		if (yych <= '@') goto yy203;
		goto yy198;
	} else {
		if (yych <= 'q') {
			if (yych <= '`') goto yy203;
			goto yy167;
		} else {
			if (yych <= 'r') goto yy177;
			if (yych <= 'z') goto yy167;
			goto yy203;
		}
	}
yy60:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '\'') {
		if (yych == '"') goto yy200;
		if (yych <= '&') goto yy203;
		goto yy201;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy203;
			goto yy198;
		} else {
			if (yych <= '`') goto yy203;
			if (yych <= 'z') goto yy167;
			goto yy203;
		}
	}
yy61:
	yych = *++YYCURSOR;
	if (yych == '}') goto yy102;
#line 391 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LCURLY); }
#line 1985 "../../lnav/src/data_scanner_re.cc"
yy62:
	++YYCURSOR;
#line 392 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RCURLY); }
#line 1990 "../../lnav/src/data_scanner_re.cc"
yy63:
	++YYCURSOR;
yy64:
#line 175 "../../lnav/src/data_scanner_re.re"
	{ return std::nullopt; }
#line 1996 "../../lnav/src/data_scanner_re.cc"
yy65:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy64;
	if (yych <= 0xBF) goto yy2;
	goto yy64;
yy66:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x9F) goto yy64;
	if (yych <= 0xBF) goto yy207;
	goto yy64;
yy67:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy64;
	if (yych <= 0xBF) goto yy207;
	goto yy64;
yy68:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x8F) goto yy64;
	if (yych <= 0xBF) goto yy208;
	goto yy64;
yy69:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy64;
	if (yych <= 0xBF) goto yy208;
	goto yy64;
yy70:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy64;
	if (yych <= 0x8F) goto yy208;
	goto yy64;
yy71:
	yych = *++YYCURSOR;
yy72:
	if (yybm_bol[1536+yych] & 1) {
		goto yy71;
	}
	goto yy5;
yy73:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy209;
yy74:
	YYCURSOR = YYMARKER;
	if (yyaccept <= 23) {
		if (yyaccept <= 11) {
			if (yyaccept <= 5) {
				if (yyaccept <= 2) {
					if (yyaccept <= 1) {
						if (yyaccept == 0) {
							goto yy5;
						} else {
							goto yy3;
						}
					} else {
						goto yy20;
					}
				} else {
					if (yyaccept <= 4) {
						if (yyaccept == 3) {
							goto yy24;
						} else {
							goto yy29;
						}
					} else {
						goto yy32;
					}
				}
			} else {
				if (yyaccept <= 8) {
					if (yyaccept <= 7) {
						if (yyaccept == 6) {
							goto yy37;
						} else {
							goto yy39;
						}
					} else {
						goto yy47;
					}
				} else {
					if (yyaccept <= 10) {
						if (yyaccept == 9) {
							goto yy64;
						} else {
							goto yy22;
						}
					} else {
						goto yy121;
					}
				}
			}
		} else {
			if (yyaccept <= 17) {
				if (yyaccept <= 14) {
					if (yyaccept <= 13) {
						if (yyaccept == 12) {
							goto yy126;
						} else {
							goto yy128;
						}
					} else {
						goto yy134;
					}
				} else {
					if (yyaccept <= 16) {
						if (yyaccept == 15) {
							yyt3 = YYCURSOR;
							goto yy157;
						} else {
							goto yy103;
						}
					} else {
						goto yy80;
					}
				}
			} else {
				if (yyaccept <= 20) {
					if (yyaccept <= 19) {
						if (yyaccept == 18) {
							goto yy214;
						} else {
							goto yy222;
						}
					} else {
						goto yy231;
					}
				} else {
					if (yyaccept <= 22) {
						if (yyaccept == 21) {
							goto yy243;
						} else {
							goto yy233;
						}
					} else {
						goto yy276;
					}
				}
			}
		}
	} else {
		if (yyaccept <= 35) {
			if (yyaccept <= 29) {
				if (yyaccept <= 26) {
					if (yyaccept <= 25) {
						if (yyaccept == 24) {
							goto yy157;
						} else {
							goto yy220;
						}
					} else {
						goto yy308;
					}
				} else {
					if (yyaccept <= 28) {
						if (yyaccept == 27) {
							yyt2 = yyt1;
							goto yy308;
						} else {
							yyt1 = YYCURSOR;
							goto yy499;
						}
					} else {
						goto yy513;
					}
				}
			} else {
				if (yyaccept <= 32) {
					if (yyaccept <= 31) {
						if (yyaccept == 30) {
							goto yy394;
						} else {
							goto yy536;
						}
					} else {
						goto yy548;
					}
				} else {
					if (yyaccept <= 34) {
						if (yyaccept == 33) {
							goto yy564;
						} else {
							goto yy461;
						}
					} else {
						goto yy499;
					}
				}
			}
		} else {
			if (yyaccept <= 41) {
				if (yyaccept <= 38) {
					if (yyaccept <= 37) {
						if (yyaccept == 36) {
							goto yy684;
						} else {
							goto yy328;
						}
					} else {
						yyt2 = yyt4;
						goto yy308;
					}
				} else {
					if (yyaccept <= 40) {
						if (yyaccept == 39) {
							goto yy901;
						} else {
							goto yy466;
						}
					} else {
						goto yy995;
					}
				}
			} else {
				if (yyaccept <= 44) {
					if (yyaccept <= 43) {
						if (yyaccept == 42) {
							goto yy1006;
						} else {
							goto yy1028;
						}
					} else {
						goto yy1411;
					}
				} else {
					if (yyaccept <= 46) {
						if (yyaccept == 45) {
							goto yy1459;
						} else {
							goto yy1541;
						}
					} else {
						goto yy1657;
					}
				}
			}
		}
	}
yy75:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy210;
	goto yy74;
yy76:
	yych = *++YYCURSOR;
	if (yybm_bol[1536+yych] & 2) {
		goto yy76;
	}
	if (yych <= '@') goto yy74;
	if (yych <= 'Z') goto yy211;
	if (yych <= '`') goto yy74;
	if (yych <= 'z') goto yy211;
	goto yy74;
yy77:
	yych = *++YYCURSOR;
yy78:
	if (yybm_bol[1536+yych] & 4) {
		goto yy77;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x1B) goto yy74;
			goto yy212;
		} else {
			if (yych <= '\\') goto yy81;
			if (yych <= 0xC1) goto yy74;
			goto yy82;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy83;
			if (yych <= 0xEF) goto yy84;
			goto yy85;
		} else {
			if (yych <= 0xF3) goto yy86;
			if (yych <= 0xF4) goto yy87;
			goto yy74;
		}
	}
yy79:
	yych = *++YYCURSOR;
	if (yych == '"') goto yy213;
yy80:
#line 180 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 2296 "../../lnav/src/data_scanner_re.cc"
yy81:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy77;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy83;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy84;
			goto yy85;
		} else {
			if (yych <= 0xF3) goto yy86;
			if (yych <= 0xF4) goto yy87;
			goto yy74;
		}
	}
yy82:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy77;
	goto yy74;
yy83:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy82;
	goto yy74;
yy84:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy82;
	goto yy74;
yy85:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy84;
	goto yy74;
yy86:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy84;
	goto yy74;
yy87:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy84;
	goto yy74;
yy88:
	yych = *++YYCURSOR;
	if (yybm_bol[1536+yych] & 8) {
		goto yy88;
	}
	if (yych == '[') goto yy215;
	goto yy74;
yy89:
	yych = *++YYCURSOR;
yy90:
	if (yybm_bol[1536+yych] & 16) {
		goto yy89;
	}
	if (yych != '@') goto yy74;
yy91:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy216;
	goto yy217;
yy92:
	yych = *++YYCURSOR;
yy93:
	if (yybm_bol[1536+yych] & 32) {
		goto yy92;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy74;
			goto yy218;
		} else {
			if (yych <= '\\') goto yy95;
			if (yych <= 0xC1) goto yy74;
			goto yy96;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy97;
			if (yych <= 0xEF) goto yy98;
			goto yy99;
		} else {
			if (yych <= 0xF3) goto yy100;
			if (yych <= 0xF4) goto yy101;
			goto yy74;
		}
	}
yy94:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= 'R') {
			if (yych == '\'') goto yy221;
			yyt2 = YYCURSOR;
			goto yy219;
		} else {
			if (yych <= 'S') goto yy74;
			if (yych == 's') goto yy74;
			yyt2 = YYCURSOR;
			goto yy219;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy223;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy224;
			}
			yyt2 = YYCURSOR;
			goto yy225;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy226;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy227;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy228;
			}
			goto yy74;
		}
	}
yy95:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy92;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy97;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy98;
			goto yy99;
		} else {
			if (yych <= 0xF3) goto yy100;
			if (yych <= 0xF4) goto yy101;
			goto yy74;
		}
	}
yy96:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy92;
	goto yy74;
yy97:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy96;
	goto yy74;
yy98:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy96;
	goto yy74;
yy99:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy98;
	goto yy74;
yy100:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy98;
	goto yy74;
yy101:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy98;
	goto yy74;
yy102:
	++YYCURSOR;
yy103:
#line 390 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMPTY_CONTAINER); }
#line 2488 "../../lnav/src/data_scanner_re.cc"
yy104:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= ',') goto yy74;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy74;
			goto yy229;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy74;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy229;
			goto yy74;
		} else {
			if (yych <= '_') goto yy89;
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy229;
			goto yy74;
		}
	}
yy105:
	yych = *++YYCURSOR;
	if (yybm_bol[1536+yych] & 64) {
		goto yy107;
	}
	if (yych <= '\r') {
		if (yych <= 0x08) {
			if (yych <= 0x00) goto yy232;
			goto yy90;
		} else {
			if (yych <= '\n') goto yy232;
			if (yych <= '\f') goto yy90;
			goto yy232;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy90;
			goto yy232;
		} else {
			if (yych == '.') goto yy234;
			goto yy90;
		}
	}
yy106:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy234;
	goto yy236;
yy107:
	yych = *++YYCURSOR;
yy108:
	if (yybm_bol[1536+yych] & 64) {
		goto yy107;
	}
	if (yych <= '%') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy232;
			if (yych <= 0x08) goto yy74;
			if (yych <= '\n') goto yy232;
			goto yy74;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy232;
				goto yy74;
			} else {
				if (yych <= ' ') goto yy232;
				if (yych <= '$') goto yy74;
				goto yy89;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych == '+') goto yy89;
			if (yych <= ',') goto yy74;
			goto yy89;
		} else {
			if (yych <= '@') {
				if (yych <= '?') goto yy74;
				goto yy91;
			} else {
				if (yych == '_') goto yy89;
				goto yy74;
			}
		}
	}
yy109:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy244;
				}
				yyt1 = YYCURSOR;
				goto yy242;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy244;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy242;
				}
				goto yy245;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy244;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy242;
				}
				yyt1 = YYCURSOR;
				goto yy246;
			} else {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy242;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy247;
				}
				yyt1 = YYCURSOR;
				goto yy246;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy244;
				}
				yyt1 = YYCURSOR;
				goto yy242;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy246;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy242;
				}
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy248;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy249;
				}
				yyt1 = YYCURSOR;
				goto yy250;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy251;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy252;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy253;
				}
				goto yy74;
			}
		}
	}
yy110:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 4) {
		goto yy256;
	}
	if (yych <= '.') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy232;
				goto yy90;
			} else {
				if (yych <= '\n') goto yy232;
				if (yych <= '\f') goto yy90;
				goto yy232;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy254;
				goto yy90;
			} else {
				if (yych <= '%') goto yy120;
				if (yych <= '-') goto yy90;
				goto yy255;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '/') goto yy90;
				goto yy257;
			} else {
				if (yych <= '@') goto yy90;
				if (yych <= 'F') goto yy258;
				goto yy107;
			}
		} else {
			if (yych <= 'w') {
				if (yych <= '`') goto yy90;
				if (yych <= 'f') goto yy258;
				goto yy107;
			} else {
				if (yych <= 'x') goto yy260;
				if (yych <= 'z') goto yy107;
				goto yy90;
			}
		}
	}
yy111:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy255;
	goto yy262;
yy112:
	yych = *++YYCURSOR;
	if (yych != '/') goto yy90;
yy113:
	yych = *++YYCURSOR;
	if (yych == 'P') goto yy117;
	goto yy115;
yy114:
	yych = *++YYCURSOR;
yy115:
	if (yybm_bol[1536+yych] & 128) {
		goto yy114;
	}
	goto yy22;
yy116:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 1) {
		goto yy116;
	}
	if (yych <= 0xDF) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy22;
			if (yych <= '\'') goto yy265;
			goto yy266;
		} else {
			if (yych <= 0x7F) goto yy265;
			if (yych <= 0xC1) goto yy22;
			goto yy267;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy268;
			if (yych <= 0xEF) goto yy269;
			goto yy270;
		} else {
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy22;
		}
	}
yy117:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy273;
	goto yy115;
yy118:
	yych = *++YYCURSOR;
yy119:
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych == '%') goto yy274;
	goto yy74;
yy120:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
		} else {
			if (yych == ',') goto yy121;
			if (yych <= '.') goto yy90;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy121;
			if (yych <= 'Z') goto yy90;
		} else {
			if (yych == '`') goto yy121;
			if (yych <= 'z') goto yy90;
		}
	}
yy121:
#line 443 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PERCENTAGE); }
#line 2804 "../../lnav/src/data_scanner_re.cc"
yy122:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= ',') goto yy74;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy74;
			goto yy138;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy74;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy74;
		} else {
			if (yych == '`') goto yy74;
			if (yych <= 'z') goto yy138;
			goto yy74;
		}
	}
yy123:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy90;
			if (yych <= '0') goto yy275;
			goto yy277;
		} else {
			if (yych <= '2') goto yy278;
			if (yych <= '9') goto yy275;
			goto yy90;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy138;
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy124:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy279;
		if (yych <= '@') goto yy74;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy125:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych >= '%') goto yy120;
			} else {
				if (yych == '+') goto yy89;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy281;
				goto yy282;
			} else {
				if (yych <= '/') goto yy283;
				if (yych <= '7') goto yy284;
				goto yy285;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy286;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy287;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
			} else {
				if (yych <= 'f') goto yy287;
				if (yych <= 'z') goto yy135;
			}
		}
	}
yy126:
#line 442 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_OCTAL_NUMBER); }
#line 2914 "../../lnav/src/data_scanner_re.cc"
yy127:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych >= '%') goto yy120;
			} else {
				if (yych == '+') goto yy89;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy281;
				goto yy282;
			} else {
				if (yych <= '/') goto yy283;
				if (yych <= '9') goto yy285;
				goto yy286;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych >= '@') goto yy130;
			} else {
				if (yych <= 'F') goto yy287;
				if (yych <= 'Z') goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
			} else {
				if (yych <= 'f') goto yy287;
				if (yych <= 'z') goto yy135;
			}
		}
	}
yy128:
#line 458 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_HEX_NUMBER); }
#line 2958 "../../lnav/src/data_scanner_re.cc"
yy129:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy289;
		goto yy290;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy291;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy130:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy74;
			goto yy216;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy292;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy292;
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy292;
			goto yy74;
		}
	}
yy131:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy281;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy287;
				goto yy293;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy287;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy287;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy132:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy24;
				goto yy281;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy287;
				goto yy293;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy24;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy287;
				if (yych <= 'Z') goto yy136;
				goto yy24;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy287;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy133:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy137;
		goto yy136;
	} else {
		if (yych <= 'b') goto yy137;
		if (yych == 'i') goto yy295;
		goto yy136;
	}
yy134:
#line 480 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_ID);
       }
#line 3092 "../../lnav/src/data_scanner_re.cc"
yy135:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy136:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy134;
			goto yy89;
		} else {
			if (yych <= ',') goto yy134;
			if (yych <= '.') goto yy122;
			if (yych <= '/') goto yy124;
			goto yy135;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy294;
			if (yych <= '?') goto yy134;
			if (yych <= '@') goto yy130;
			goto yy135;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy134;
				goto yy138;
			} else {
				if (yych <= '`') goto yy134;
				if (yych <= 'z') goto yy135;
				goto yy134;
			}
		}
	}
yy137:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy136;
			goto yy24;
		} else {
			if (yych == ',') goto yy24;
			goto yy136;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy136;
			goto yy24;
		}
	}
yy138:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy139:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy134;
			goto yy89;
		} else {
			if (yych <= ',') goto yy134;
			if (yych <= '.') goto yy122;
			if (yych <= '/') goto yy124;
			goto yy138;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy134;
			if (yych <= '@') goto yy130;
			goto yy138;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy134;
				goto yy138;
			} else {
				if (yych <= '`') goto yy134;
				if (yych <= 'z') goto yy138;
				goto yy134;
			}
		}
	}
yy140:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy24;
				goto yy136;
			} else {
				if (yych == '+') goto yy136;
				goto yy24;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy136;
				if (yych <= '?') goto yy24;
				goto yy136;
			} else {
				if (yych <= 'B') goto yy137;
				if (yych <= 'R') goto yy136;
				goto yy137;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy136;
				if (yych <= '^') goto yy24;
				goto yy136;
			} else {
				if (yych <= '`') goto yy24;
				if (yych <= 'a') goto yy136;
				goto yy137;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy297;
				goto yy136;
			} else {
				if (yych <= 's') goto yy137;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy141:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy137;
	if (yych == 's') goto yy137;
	goto yy136;
yy142:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy299;
			goto yy134;
		} else {
			if (yych == ',') goto yy134;
			goto yy299;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy134;
			if (yych <= 'Z') goto yy299;
			goto yy134;
		} else {
			if (yych == '`') goto yy134;
			if (yych <= 'z') goto yy299;
			goto yy134;
		}
	}
yy143:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy281;
		case '.': goto yy300;
		case '/': goto yy283;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy301;
		case ':': goto yy286;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy287;
		case 'B':
		case 'b': goto yy302;
		case 'E':
		case 'e': goto yy303;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy144:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy304;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy305;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy287;
				goto yy293;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy287;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy287;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy145:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy281;
		case '.': goto yy300;
		case '/': goto yy283;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5': goto yy301;
		case '6':
		case '7':
		case '8':
		case '9': goto yy306;
		case ':': goto yy286;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy287;
		case 'B':
		case 'b': goto yy302;
		case 'E':
		case 'e': goto yy303;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy146:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy281;
		case '.': goto yy300;
		case '/': goto yy283;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy306;
		case ':': goto yy286;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy287;
		case 'B':
		case 'b': goto yy302;
		case 'E':
		case 'e': goto yy303;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy147:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy309;
			} else {
				if (yych <= '1') goto yy310;
				if (yych <= '2') goto yy311;
				goto yy309;
			}
		} else {
			if (yych <= 'E') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy313;
			} else {
				if (yych <= 'F') goto yy314;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'e') goto yy313;
				if (yych <= 'f') goto yy314;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy148:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy321;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= ':') goto yy322;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy322;
			if (yych <= '^') goto yy74;
			goto yy322;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy322;
			goto yy74;
		}
	}
yy149:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 32) {
		goto yy325;
	}
	goto yy324;
yy150:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy74;
		if (yych <= '\f') goto yy330;
		goto yy74;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy330;
			goto yy74;
		} else {
			if (yych == '>') goto yy74;
			goto yy330;
		}
	}
yy151:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy331;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= ':') goto yy331;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy331;
			if (yych <= '^') goto yy74;
			goto yy331;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy331;
			goto yy74;
		}
	}
yy152:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy332;
	goto yy74;
yy153:
	yych = *++YYCURSOR;
yy154:
	if (yybm_bol[1280+yych] & 4) {
		goto yy153;
	}
	if (yych == '\n') goto yy333;
	goto yy74;
yy155:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
yy156:
	if (yybm_bol[1280+yych] & 8) {
		goto yy155;
	}
	if (yych <= '&') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				yyt3 = YYCURSOR;
				goto yy157;
			}
			if (yych <= '\n') {
				yyt3 = YYCURSOR;
				goto yy334;
			}
			if (yych >= '\r') {
				yyt3 = YYCURSOR;
				goto yy334;
			}
			yyt3 = YYCURSOR;
		} else {
			if (yych <= ' ') {
				if (yych >= ' ') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych <= '!') {
					yyt3 = YYCURSOR;
					goto yy335;
				}
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '+') {
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy335;
				}
				if (yych >= '.') {
					yyt3 = YYCURSOR;
					goto yy336;
				}
				yyt3 = YYCURSOR;
			}
		} else {
			if (yych <= ';') {
				if (yych >= ':') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych == '?') {
					yyt3 = YYCURSOR;
					goto yy335;
				}
				yyt3 = YYCURSOR;
			}
		}
	}
yy157:
	YYCURSOR = yyt3;
#line 464 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WORD); }
#line 3758 "../../lnav/src/data_scanner_re.cc"
yy158:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '$') {
			if (yych <= '\n') {
				if (yych <= '\t') goto yy74;
				goto yy333;
			} else {
				if (yych == ' ') goto yy153;
				goto yy74;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '%') goto yy89;
				goto yy74;
			} else {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy74;
				goto yy337;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '/') {
				if (yych <= '.') goto yy89;
				goto yy74;
			} else {
				if (yych <= '9') goto yy338;
				if (yych <= '?') goto yy74;
				goto yy91;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy338;
				if (yych <= '^') goto yy74;
				goto yy338;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy138;
				goto yy74;
			}
		}
	}
yy159:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= ',') goto yy74;
			if (yych <= '-') goto yy340;
			if (yych <= '.') goto yy89;
			goto yy74;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy341;
			if (yych <= '?') goto yy74;
			if (yych <= '@') goto yy91;
			goto yy341;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy74;
				goto yy341;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy341;
				goto yy74;
			}
		}
	}
yy160:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy172;
			goto yy342;
		} else {
			if (yych <= '/') goto yy172;
			if (yych <= '9') goto yy343;
			goto yy344;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy172;
			goto yy343;
		} else {
			if (yych <= '`') goto yy172;
			if (yych <= 'f') goto yy345;
			goto yy172;
		}
	}
yy161:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych <= '.') goto yy74;
			goto yy288;
		} else {
			if (yych <= '9') goto yy291;
			if (yych <= ':') goto yy346;
			goto yy74;
		}
	} else {
		if (yych <= '\\') {
			if (yych <= 'F') goto yy291;
			if (yych <= '[') goto yy74;
			goto yy113;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy162:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '-') goto yy342;
			goto yy165;
		} else {
			if (yych <= '9') goto yy343;
			if (yych <= ':') goto yy344;
			goto yy165;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy347;
			if (yych <= 'Z') goto yy348;
			goto yy165;
		} else {
			if (yych <= 'f') goto yy349;
			if (yych <= 'z') goto yy350;
			goto yy165;
		}
	}
yy163:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy172;
	if (yych <= 'Z') goto yy348;
	if (yych <= '`') goto yy172;
	if (yych <= 'z') goto yy350;
	goto yy172;
yy164:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy165:
	if (yybm_bol[1280+yych] & 16) {
		goto yy164;
	}
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych == '\n') goto yy333;
			if (yych <= 0x1F) goto yy39;
			goto yy153;
		} else {
			if (yych <= '%') {
				if (yych <= '$') goto yy39;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy39;
			}
		}
	} else {
		if (yych <= ':') {
			if (yych <= '-') goto yy158;
			if (yych <= '.') goto yy159;
			if (yych <= '/') goto yy124;
			goto yy194;
		} else {
			if (yych <= '@') {
				if (yych <= '?') goto yy39;
				goto yy130;
			} else {
				if (yych <= '`') goto yy39;
				if (yych <= 'z') goto yy50;
				goto yy39;
			}
		}
	}
yy166:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy281;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy345;
				yyt3 = YYCURSOR;
				goto yy352;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy349;
				if (yych <= 'Z') goto yy350;
				goto yy156;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy353;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy167:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy156;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy355;
			}
			if (yych <= '?') goto yy156;
			if (yych <= '@') goto yy130;
			goto yy350;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy156;
				goto yy50;
			} else {
				if (yych <= '`') goto yy156;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy168:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'E') {
		if (yych <= '@') goto yy172;
		if (yych <= 'D') goto yy348;
		goto yy356;
	} else {
		if (yych <= 'Z') goto yy348;
		if (yych <= '`') goto yy172;
		if (yych <= 'z') goto yy350;
		goto yy172;
	}
yy169:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= '/') {
			if (yych == '-') goto yy342;
			goto yy165;
		} else {
			if (yych <= '9') goto yy343;
			if (yych <= ':') goto yy344;
			if (yych <= '@') goto yy165;
			goto yy347;
		}
	} else {
		if (yych <= 'Z') {
			if (yych == 'L') goto yy357;
			goto yy348;
		} else {
			if (yych <= '`') goto yy165;
			if (yych <= 'f') goto yy349;
			if (yych <= 'z') goto yy350;
			goto yy165;
		}
	}
yy170:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '/') {
				if (yych <= '-') goto yy281;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				goto yy124;
			} else {
				if (yych <= '9') goto yy345;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy352;
				}
				goto yy156;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'F') {
				if (yych <= '@') goto yy130;
				goto yy349;
			} else {
				if (yych <= 'Z') goto yy350;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '`') goto yy156;
				if (yych <= 'f') goto yy353;
				goto yy354;
			} else {
				if (yych <= 'l') goto yy358;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy171:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy172:
	if (yybm_bol[1280+yych] & 32) {
		goto yy171;
	}
	if (yych <= '-') {
		if (yych <= '$') {
			if (yych <= '\n') {
				if (yych <= '\t') goto yy39;
				goto yy333;
			} else {
				if (yych == ' ') goto yy153;
				goto yy39;
			}
		} else {
			if (yych <= '*') {
				if (yych <= '%') goto yy89;
				goto yy39;
			} else {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy39;
				goto yy158;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '/') {
				if (yych <= '.') goto yy159;
				goto yy124;
			} else {
				if (yych <= ':') goto yy204;
				if (yych <= '?') goto yy39;
				goto yy130;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy39;
				goto yy164;
			} else {
				if (yych <= '`') goto yy39;
				if (yych <= 'z') goto yy202;
				goto yy39;
			}
		}
	}
yy173:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == '/') goto yy288;
		goto yy74;
	} else {
		if (yych <= ':') goto yy359;
		if (yych == '\\') goto yy113;
		goto yy74;
	}
yy174:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'L') {
		if (yych <= '@') goto yy172;
		if (yych <= 'K') goto yy348;
		goto yy360;
	} else {
		if (yych <= 'Z') goto yy348;
		if (yych <= '`') goto yy172;
		if (yych <= 'z') goto yy350;
		goto yy172;
	}
yy175:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'Z') goto yy350;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 'm') {
				if (yych <= '`') goto yy156;
				goto yy354;
			} else {
				if (yych <= 'n') goto yy361;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy176:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'U') {
		if (yych <= '@') goto yy172;
		if (yych <= 'T') goto yy348;
		goto yy362;
	} else {
		if (yych <= 'Z') goto yy348;
		if (yych <= '`') goto yy172;
		if (yych <= 'z') goto yy350;
		goto yy172;
	}
yy177:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'Z') goto yy350;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 't') {
				if (yych <= '`') goto yy156;
				goto yy354;
			} else {
				if (yych <= 'u') goto yy361;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy178:
	yych = *++YYCURSOR;
yy179:
	if (yybm_bol[1280+yych] & 64) {
		goto yy178;
	}
	if (yych <= 0xE0) {
		if (yych <= ']') {
			if (yych <= '\n') goto yy74;
			goto yy363;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy181;
			goto yy182;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy183;
			goto yy184;
		} else {
			if (yych <= 0xF3) goto yy185;
			if (yych <= 0xF4) goto yy186;
			goto yy74;
		}
	}
yy180:
	yyaccept = 16;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '\n') {
		if (yych <= 0x00) goto yy103;
		if (yych <= '\t') goto yy179;
		goto yy103;
	} else {
		if (yych <= 0x7F) goto yy179;
		if (yych <= 0xC1) goto yy103;
		if (yych <= 0xF4) goto yy179;
		goto yy103;
	}
yy181:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy178;
	goto yy74;
yy182:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy181;
	goto yy74;
yy183:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy181;
	goto yy74;
yy184:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy183;
	goto yy74;
yy185:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy183;
	goto yy74;
yy186:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy183;
	goto yy74;
yy187:
	++YYCURSOR;
#line 490 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_ESCAPED_CHAR); }
#line 4364 "../../lnav/src/data_scanner_re.cc"
yy188:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy187;
	goto yy74;
yy189:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy188;
	goto yy74;
yy190:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy188;
	goto yy74;
yy191:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy190;
	goto yy74;
yy192:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy190;
	goto yy74;
yy193:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy190;
	goto yy74;
yy194:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy359;
	goto yy74;
yy195:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy203;
			goto yy281;
		} else {
			if (yych <= '/') goto yy203;
			if (yych <= '9') goto yy345;
			goto yy344;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy203;
			goto yy345;
		} else {
			if (yych <= '`') goto yy203;
			if (yych <= 'f') goto yy345;
			goto yy203;
		}
	}
yy196:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy291;
		goto yy346;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy291;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy197:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '-') goto yy281;
			goto yy51;
		} else {
			if (yych <= '9') goto yy345;
			if (yych <= ':') goto yy344;
			goto yy51;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy349;
			if (yych <= 'Z') goto yy350;
			goto yy51;
		} else {
			if (yych <= 'f') goto yy349;
			if (yych <= 'z') goto yy350;
			goto yy51;
		}
	}
yy198:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy203;
	if (yych <= 'Z') goto yy350;
	if (yych <= '`') goto yy203;
	if (yych <= 'z') goto yy350;
	goto yy203;
yy199:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'Z') goto yy350;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 'e') {
				if (yych <= '`') goto yy156;
				goto yy354;
			} else {
				if (yych <= 'f') goto yy364;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy200:
	yych = *++YYCURSOR;
	if (yych == '"') goto yy79;
	goto yy78;
yy201:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 2) {
		goto yy369;
	}
	if (yych <= '"') {
		if (yych <= 0x15) {
			if (yych <= '\t') {
				if (yych <= 0x00) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0x08) goto yy93;
				yyt3 = YYCURSOR;
				goto yy365;
			} else {
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych == '\r') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				goto yy93;
			}
		} else {
			if (yych <= 0x1B) {
				if (yych <= 0x16) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0x1A) goto yy93;
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= 0x1F) goto yy93;
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				yyt3 = YYCURSOR;
				goto yy365;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych <= '&') goto yy93;
				if (yych <= '\'') goto yy367;
				yyt3 = YYCURSOR;
				goto yy365;
			} else {
				if (yych == ',') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				if (yych <= '-') goto yy93;
				yyt3 = YYCURSOR;
				goto yy368;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy93;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				if (yych <= '>') goto yy93;
				yyt3 = YYCURSOR;
				goto yy366;
			} else {
				if (yych <= 0x7F) goto yy93;
				if (yych <= 0xC1) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0xF4) goto yy93;
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	}
yy202:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy203:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy39;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy39;
				goto yy122;
			} else {
				if (yych <= '.') goto yy159;
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy204;
			if (yych <= '?') goto yy39;
			if (yych <= '@') goto yy130;
			goto yy202;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy39;
				goto yy50;
			} else {
				if (yych <= '`') goto yy39;
				if (yych <= 'z') goto yy202;
				goto yy39;
			}
		}
	}
yy204:
	yych = *++YYCURSOR;
	if (yych == '/') goto yy288;
	if (yych == ':') goto yy359;
	goto yy74;
yy205:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'Z') goto yy350;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '`') goto yy156;
				goto yy354;
			} else {
				if (yych <= 'l') goto yy370;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy206:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy156;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy156;
				goto yy371;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy355;
			}
			if (yych <= '?') goto yy156;
			if (yych <= '@') goto yy130;
			goto yy350;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy156;
				goto yy50;
			} else {
				if (yych <= '`') goto yy156;
				if (yych <= 'z') goto yy354;
				goto yy156;
			}
		}
	}
yy207:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy2;
	goto yy74;
yy208:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy207;
	goto yy74;
yy209:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy372;
	goto yy74;
yy210:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy373;
	goto yy74;
yy211:
	++YYCURSOR;
#line 381 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_CSI);
       }
#line 4755 "../../lnav/src/data_scanner_re.cc"
yy212:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '"') goto yy77;
	goto yy80;
yy213:
	yyaccept = 18;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych >= 0x01) goto yy78;
		} else {
			if (yych <= '\n') goto yy214;
			if (yych <= 0x15) goto yy78;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych != 0x1B) goto yy78;
		} else {
			if (yych <= 0xC1) goto yy214;
			if (yych <= 0xF4) goto yy78;
		}
	}
yy214:
#line 193 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 3;
           goto yyc_dbldocstring;
       }
#line 4793 "../../lnav/src/data_scanner_re.cc"
yy215:
	yych = *++YYCURSOR;
	if (yybm_bol[1280+yych] & 128) {
		goto yy215;
	}
	if (yych <= 0xE0) {
		if (yych <= ']') {
			if (yych <= 0x00) goto yy74;
			goto yy374;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy375;
			goto yy376;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy377;
			goto yy378;
		} else {
			if (yych <= 0xF3) goto yy379;
			if (yych <= 0xF4) goto yy380;
			goto yy74;
		}
	}
yy216:
	yych = *++YYCURSOR;
yy217:
	if (yybm_bol[1024+yych] & 1) {
		goto yy216;
	}
	if (yych <= ',') goto yy74;
	if (yych <= '.') goto yy381;
	goto yy74;
yy218:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= 'R') {
			if (yych == '\'') {
				yyt1 = YYCURSOR;
				goto yy382;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'S') goto yy74;
			if (yych == 's') goto yy74;
			yyt2 = YYCURSOR;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy223;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy224;
			}
			yyt2 = YYCURSOR;
			goto yy225;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy226;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy227;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy228;
			}
			goto yy74;
		}
	}
yy219:
	++YYCURSOR;
	yyt1 = yyt2;
yy220:
	YYCURSOR = yyt1;
#line 278 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           if (tf == text_format_t::TF_RUST) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when([](char ch) { return ch != '\'' && !isalnum(ch); });
               cap_all.c_end = split_res.first.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = split_res.first.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               return tokenize_result{DT_SYMBOL, cap_all, cap_inner, this->ds_input.data()};
           }
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 4897 "../../lnav/src/data_scanner_re.cc"
yy221:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych >= 0x01) goto yy93;
		} else {
			if (yych <= '\n') goto yy222;
			if (yych <= 0x15) goto yy93;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych != 0x1B) goto yy93;
		} else {
			if (yych <= 0xC1) goto yy222;
			if (yych <= 0xF4) goto yy93;
		}
	}
yy222:
#line 221 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 3;
           goto yyc_sdocstring;
       }
#line 4930 "../../lnav/src/data_scanner_re.cc"
yy223:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy219;
	goto yy74;
yy224:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy223;
	goto yy74;
yy225:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy223;
	goto yy74;
yy226:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy225;
	goto yy74;
yy227:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy225;
	goto yy74;
yy228:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy225;
	goto yy74;
yy229:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
yy230:
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych >= '+') goto yy89;
		} else {
			if (yych <= ',') goto yy231;
			if (yych <= '-') goto yy104;
			if (yych <= '.') goto yy89;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy229;
			if (yych <= '?') goto yy231;
			if (yych <= '@') goto yy91;
			goto yy229;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy89;
			} else {
				if (yych <= '`') goto yy231;
				if (yych <= 'z') goto yy229;
			}
		}
	}
yy231:
#line 468 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 4994 "../../lnav/src/data_scanner_re.cc"
yy232:
	++YYCURSOR;
yy233:
	YYCURSOR -= 1;
#line 472 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 5003 "../../lnav/src/data_scanner_re.cc"
yy234:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy383;
	goto yy90;
yy235:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy236:
	if (yybm_bol[1024+yych] & 2) {
		goto yy235;
	}
	switch (yych) {
		case 0x00:
		case '\t':
		case '\n':
		case '\r':
		case ' ': goto yy232;
		case '%':
		case '+':
		case '-':
		case '_': goto yy89;
		case '.': goto yy384;
		case '@': goto yy91;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy107;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy237;
		case 'E':
		case 'e': goto yy238;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy239;
		case 'm': goto yy240;
		case 'p':
		case 'u': goto yy241;
		default: goto yy24;
	}
yy237:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy108;
				goto yy24;
			} else {
				if (yych <= '\n') goto yy108;
				if (yych <= '\f') goto yy24;
				goto yy108;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy108;
				goto yy24;
			} else {
				if (yych <= '%') goto yy108;
				if (yych <= '*') goto yy24;
				goto yy108;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '.') {
				if (yych <= ',') goto yy24;
				goto yy108;
			} else {
				if (yych <= '/') goto yy24;
				if (yych <= '9') goto yy108;
				goto yy24;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy108;
				if (yych <= '^') goto yy24;
				goto yy108;
			} else {
				if (yych <= '`') goto yy24;
				if (yych <= 'z') goto yy108;
				goto yy24;
			}
		}
	}
yy238:
	yych = *++YYCURSOR;
	if (yybm_bol[1536+yych] & 64) {
		goto yy107;
	}
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy232;
			if (yych <= 0x08) goto yy90;
			goto yy232;
		} else {
			if (yych == '\r') goto yy232;
			goto yy90;
		}
	} else {
		if (yych <= '+') {
			if (yych <= ' ') goto yy232;
			if (yych <= '*') goto yy90;
			goto yy304;
		} else {
			if (yych == '-') goto yy304;
			goto yy90;
		}
	}
yy239:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy232;
				goto yy90;
			} else {
				if (yych <= '\n') goto yy232;
				if (yych <= '\f') goto yy90;
				goto yy232;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy90;
				goto yy232;
			} else {
				if (yych <= '/') goto yy90;
				if (yych <= '9') goto yy107;
				goto yy90;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'B') {
				if (yych <= 'A') goto yy107;
				goto yy237;
			} else {
				if (yych <= 'Z') goto yy107;
				if (yych <= '`') goto yy90;
				goto yy107;
			}
		} else {
			if (yych <= 'h') {
				if (yych <= 'b') goto yy237;
				goto yy107;
			} else {
				if (yych <= 'i') goto yy385;
				if (yych <= 'z') goto yy107;
				goto yy90;
			}
		}
	}
yy240:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy108;
				if (yych <= 0x08) goto yy24;
				if (yych <= '\n') goto yy108;
				goto yy24;
			} else {
				if (yych <= '\r') goto yy108;
				if (yych == ' ') goto yy108;
				goto yy24;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') goto yy108;
				if (yych == '+') goto yy108;
				goto yy24;
			} else {
				if (yych == '/') goto yy24;
				if (yych <= '9') goto yy108;
				goto yy24;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'S') {
				if (yych == 'B') goto yy237;
				if (yych <= 'R') goto yy108;
				goto yy237;
			} else {
				if (yych <= 'Z') goto yy108;
				if (yych == '_') goto yy108;
				goto yy24;
			}
		} else {
			if (yych <= 'i') {
				if (yych == 'b') goto yy237;
				if (yych <= 'h') goto yy108;
				goto yy386;
			} else {
				if (yych == 's') goto yy237;
				if (yych <= 'z') goto yy108;
				goto yy24;
			}
		}
	}
yy241:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy232;
			if (yych <= 0x08) goto yy90;
			if (yych <= '\n') goto yy232;
			goto yy90;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy232;
				goto yy90;
			} else {
				if (yych <= ' ') goto yy232;
				if (yych <= '/') goto yy90;
				goto yy107;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy90;
			if (yych == 'S') goto yy237;
			goto yy107;
		} else {
			if (yych <= 'r') {
				if (yych <= '`') goto yy90;
				goto yy107;
			} else {
				if (yych <= 's') goto yy237;
				if (yych <= 'z') goto yy107;
				goto yy90;
			}
		}
	}
yy242:
	++YYCURSOR;
yy243:
	YYCURSOR = yyt1;
#line 389 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMDASH); }
#line 5283 "../../lnav/src/data_scanner_re.cc"
yy244:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy243;
		} else {
			if (yych == ',') goto yy243;
			if (yych <= '.') goto yy90;
			goto yy243;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy243;
			if (yych <= 'Z') goto yy90;
			goto yy243;
		} else {
			if (yych == '`') goto yy243;
			if (yych <= 'z') goto yy90;
			goto yy243;
		}
	}
yy245:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy387;
	goto yy90;
yy246:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy230;
			goto yy243;
		} else {
			if (yych == ',') goto yy243;
			if (yych <= '.') goto yy230;
			goto yy243;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy230;
			if (yych <= '?') goto yy243;
			if (yych <= 'Z') goto yy230;
			goto yy243;
		} else {
			if (yych == '`') goto yy243;
			if (yych <= 'z') goto yy230;
			goto yy243;
		}
	}
yy247:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ',') goto yy243;
		if (yych == '/') goto yy243;
		goto yy216;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy243;
			goto yy216;
		} else {
			if (yych <= '`') goto yy243;
			if (yych <= 'z') goto yy216;
			goto yy243;
		}
	}
yy248:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy242;
	goto yy74;
yy249:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy248;
	goto yy74;
yy250:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy248;
	goto yy74;
yy251:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy250;
	goto yy74;
yy252:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy250;
	goto yy74;
yy253:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy250;
	goto yy74;
yy254:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych == '%') goto yy274;
	goto yy233;
yy255:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy388;
	goto yy90;
yy256:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1024+yych] & 4) {
		goto yy256;
	}
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy232;
				if (yych <= 0x08) goto yy126;
				goto yy232;
			} else {
				if (yych == '\r') goto yy232;
				goto yy126;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy254;
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				if (yych <= ',') goto yy126;
				goto yy89;
			}
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') {
				if (yych <= '.') goto yy389;
				if (yych <= '/') goto yy126;
			} else {
				if (yych <= '?') goto yy126;
				if (yych <= '@') goto yy91;
				goto yy258;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy107;
				if (yych <= '^') goto yy126;
				goto yy89;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'f') goto yy258;
				if (yych <= 'z') goto yy107;
				goto yy126;
			}
		}
	}
yy257:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy232;
				if (yych <= 0x08) goto yy128;
				goto yy232;
			} else {
				if (yych == '\r') goto yy232;
				goto yy128;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy254;
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				if (yych <= ',') goto yy128;
				goto yy89;
			}
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') {
				if (yych <= '.') goto yy389;
				if (yych <= '/') goto yy128;
				goto yy257;
			} else {
				if (yych <= '?') goto yy128;
				if (yych <= '@') goto yy91;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy107;
				if (yych <= '^') goto yy128;
				goto yy89;
			} else {
				if (yych <= '`') goto yy128;
				if (yych <= 'f') goto yy258;
				if (yych <= 'z') goto yy107;
				goto yy128;
			}
		}
	}
yy258:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy259:
	if (yybm_bol[1024+yych] & 8) {
		goto yy258;
	}
	if (yych <= '*') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy232;
				goto yy128;
			} else {
				if (yych <= '\n') goto yy232;
				if (yych <= '\f') goto yy128;
				goto yy232;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy128;
				goto yy232;
			} else {
				if (yych == '%') goto yy89;
				goto yy128;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ',') {
				if (yych <= '+') goto yy89;
				goto yy128;
			} else {
				if (yych <= '.') goto yy89;
				if (yych <= '?') goto yy128;
				goto yy91;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy107;
				if (yych <= '^') goto yy128;
				goto yy89;
			} else {
				if (yych <= '`') goto yy128;
				if (yych <= 'z') goto yy107;
				goto yy128;
			}
		}
	}
yy260:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 8) {
		goto yy258;
	}
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy232;
			if (yych <= 0x08) goto yy90;
			goto yy232;
		} else {
			if (yych == '\r') goto yy232;
			goto yy90;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ' ') goto yy232;
			if (yych <= '@') goto yy90;
			goto yy107;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy107;
			goto yy90;
		}
	}
yy261:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy262:
	switch (yych) {
		case 0x00:
		case '\t':
		case '\n':
		case '\r': goto yy232;
		case ' ': goto yy254;
		case '%': goto yy120;
		case '+':
		case '-':
		case '_': goto yy89;
		case '.': goto yy390;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy261;
		case '@': goto yy91;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy258;
		case 'B':
		case 'b': goto yy263;
		case 'E':
		case 'e': goto yy264;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy239;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy107;
		case 'S':
		case 's': goto yy237;
		case 'm': goto yy240;
		case 'p':
		case 'u': goto yy241;
		default: goto yy24;
	}
yy263:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy259;
				goto yy24;
			} else {
				if (yych <= '\n') goto yy259;
				if (yych <= '\f') goto yy24;
				goto yy259;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy259;
				goto yy24;
			} else {
				if (yych <= '%') goto yy259;
				if (yych <= '*') goto yy24;
				goto yy259;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '.') {
				if (yych <= ',') goto yy24;
				goto yy259;
			} else {
				if (yych <= '/') goto yy24;
				if (yych <= '9') goto yy259;
				goto yy24;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy259;
				if (yych <= '^') goto yy24;
				goto yy259;
			} else {
				if (yych <= '`') goto yy24;
				if (yych <= 'z') goto yy259;
				goto yy24;
			}
		}
	}
yy264:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy304;
	if (yych == '-') goto yy304;
	goto yy259;
yy265:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 16) {
		goto yy265;
	}
	if (yych <= 0xE0) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			goto yy391;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy267;
			goto yy268;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy269;
			goto yy270;
		} else {
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy74;
		}
	}
yy266:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 8) {
		goto yy392;
	}
	if (yych <= '\\') {
		if (yych <= ')') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy22;
				if (yych <= ' ') goto yy265;
				goto yy116;
			} else {
				if (yych <= '"') goto yy265;
				if (yych == '\'') goto yy265;
				goto yy116;
			}
		} else {
			if (yych <= '9') {
				if (yych <= ',') goto yy265;
				if (yych == '/') goto yy393;
				goto yy116;
			} else {
				if (yych <= '?') goto yy265;
				if (yych == '[') goto yy265;
				goto yy116;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '`') {
				if (yych <= ']') goto yy265;
				if (yych <= '_') goto yy116;
				goto yy265;
			} else {
				if (yych <= 'z') goto yy116;
				if (yych == '~') goto yy116;
				goto yy265;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy22;
				if (yych <= 0xDF) goto yy267;
				if (yych <= 0xE0) goto yy268;
				goto yy269;
			} else {
				if (yych <= 0xF0) goto yy270;
				if (yych <= 0xF3) goto yy271;
				if (yych <= 0xF4) goto yy272;
				goto yy22;
			}
		}
	}
yy267:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy265;
	goto yy74;
yy268:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy267;
	goto yy74;
yy269:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy267;
	goto yy74;
yy270:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy269;
	goto yy74;
yy271:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy269;
	goto yy74;
yy272:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy269;
	goto yy74;
yy273:
	yych = *++YYCURSOR;
	if (yych == 'o') goto yy395;
	goto yy115;
yy274:
	++YYCURSOR;
	goto yy121;
yy275:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy398;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy401;
		case 'E':
		case 'e': goto yy402;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy403;
		case 'm': goto yy404;
		case 'p':
		case 'u': goto yy405;
		default: goto yy276;
	}
yy276:
#line 419 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_NUMBER);
           auto sf = this->to_string_fragment(cap_all);
           if (isalpha(sf.back())) {
               while (isalpha(sf.back())) {
                   sf.pop_back();
               }
               cap_all.c_end = sf.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = sf.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               this->ds_units = true;
           }
           return tokenize_result{DT_NUMBER, cap_all, cap_inner, this->ds_input.data()};
       }
#line 5906 "../../lnav/src/data_scanner_re.cc"
yy277:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy275;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy401;
		case 'E':
		case 'e': goto yy402;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy403;
		case 'm': goto yy404;
		case 'p':
		case 'u': goto yy405;
		default: goto yy276;
	}
yy278:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy275;
		case '5': goto yy406;
		case '6':
		case '7':
		case '8':
		case '9': goto yy398;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy401;
		case 'E':
		case 'e': goto yy402;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy403;
		case 'm': goto yy404;
		case 'p':
		case 'u': goto yy405;
		default: goto yy276;
	}
yy279:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy280:
	if (yych <= '@') {
		if (yych <= '9') {
			if (yych <= ',') goto yy134;
			if (yych <= '/') goto yy124;
			goto yy279;
		} else {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy134;
			goto yy124;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy279;
			if (yych <= '^') goto yy134;
			goto yy279;
		} else {
			if (yych <= '`') goto yy134;
			if (yych <= 'z') goto yy279;
			goto yy134;
		}
	}
yy281:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy74;
				goto yy89;
			} else {
				if (yych <= '/') goto yy74;
				if (yych <= '9') goto yy407;
				goto yy74;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy91;
			if (yych <= 'F') goto yy407;
			if (yych <= 'Z') goto yy138;
			goto yy74;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy74;
			} else {
				if (yych <= 'f') goto yy407;
				if (yych <= 'z') goto yy138;
				goto yy74;
			}
		}
	}
yy282:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy90;
			if (yych <= '0') goto yy408;
			goto yy409;
		} else {
			if (yych <= '2') goto yy410;
			if (yych <= '9') goto yy408;
			goto yy90;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy138;
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy283:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy279;
		if (yych <= '@') goto yy74;
		goto yy411;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy411;
			goto yy74;
		}
	}
yy284:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy126;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '7') goto yy413;
				goto yy414;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy415;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy416;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy416;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy285:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy128;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy414;
				goto yy415;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy416;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy416;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy286:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy417;
		goto yy290;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy418;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy418;
			goto yy74;
		}
	}
yy287:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy128;
				goto yy136;
			} else {
				if (yych <= '9') goto yy416;
				if (yych <= ':') goto yy415;
				goto yy128;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy416;
			if (yych <= 'Z') goto yy136;
			goto yy128;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy416;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy288:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= ',') {
			if (yych <= '$') goto yy420;
			if (yych <= '&') goto yy419;
			goto yy420;
		} else {
			if (yych == '.') goto yy420;
			if (yych <= '9') goto yy419;
			goto yy420;
		}
	} else {
		if (yych <= '@') {
			if (yych == '>') goto yy420;
			if (yych <= '?') goto yy419;
			goto yy420;
		} else {
			if (yych <= 'Z') goto yy419;
			if (yych <= '`') goto yy420;
			if (yych <= 'z') goto yy419;
			goto yy420;
		}
	}
yy289:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy427;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy429;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy429;
		goto yy74;
	}
yy290:
	yych = *++YYCURSOR;
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy430;
			} else {
				if (yych <= '1') goto yy431;
				if (yych <= '2') goto yy432;
				goto yy430;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy433;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy433;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy291:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy429;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy429;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy429;
		goto yy74;
	}
yy292:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy134;
			if (yych <= '-') goto yy435;
			goto yy436;
		} else {
			if (yych <= '/') goto yy124;
			if (yych <= '9') goto yy292;
			if (yych <= ':') goto yy296;
			goto yy134;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy124;
			if (yych <= 'Z') goto yy292;
			goto yy134;
		} else {
			if (yych <= '_') goto yy279;
			if (yych <= '`') goto yy134;
			if (yych <= 'z') goto yy292;
			goto yy134;
		}
	}
yy293:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy418;
		goto yy290;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy418;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy418;
			goto yy74;
		}
	}
yy294:
	yych = *++YYCURSOR;
	if (yych == '/') goto yy288;
	if (yych == ':') goto yy124;
	goto yy74;
yy295:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy137;
	if (yych == 'b') goto yy137;
	goto yy136;
yy296:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy124;
	goto yy74;
yy297:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy137;
		goto yy136;
	} else {
		if (yych <= 'b') goto yy137;
		if (yych == 'n') goto yy137;
		goto yy136;
	}
yy298:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy299:
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy128;
			goto yy89;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy128;
				goto yy122;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy298;
				goto yy294;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy298;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy298;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy300:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy90;
			if (yych <= '0') goto yy437;
			goto yy438;
		} else {
			if (yych <= '2') goto yy439;
			if (yych <= '9') goto yy437;
			goto yy90;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy138;
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy301:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy300;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy440;
		case ':': goto yy415;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy416;
		case 'B':
		case 'b': goto yy441;
		case 'E':
		case 'e': goto yy442;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy302:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy136;
			} else {
				if (yych <= '9') goto yy416;
				if (yych <= ':') goto yy415;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy416;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy416;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy303:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy443;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy416;
				goto yy415;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy416;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy416;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy304:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy444;
	goto yy90;
yy305:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy445;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy407;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy407;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy306:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy446;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy440;
		case ':': goto yy415;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy416;
		case 'B':
		case 'b': goto yy441;
		case 'E':
		case 'e': goto yy442;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy307:
	++YYCURSOR;
yy308:
	YYCURSOR = yyt2;
#line 319 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_IPV6_ADDRESS); }
#line 6832 "../../lnav/src/data_scanner_re.cc"
yy309:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy448;
				yyt1 = YYCURSOR;
				goto yy449;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy310:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy451;
				yyt1 = YYCURSOR;
				goto yy449;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy311:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy451;
				goto yy452;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy448;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy449;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy450;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy450;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy312:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy307;
	goto yy74;
yy313:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy450;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy449;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy450;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy450;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy314:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy449;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy450;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy453;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'e') goto yy450;
				goto yy453;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych >= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy315:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy307;
	goto yy74;
yy316:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy315;
	goto yy74;
yy317:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy315;
	goto yy74;
yy318:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy317;
	goto yy74;
yy319:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy317;
	goto yy74;
yy320:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy317;
	goto yy74;
yy321:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy455;
			goto yy458;
		} else {
			if (yych <= '/') goto yy455;
			if (yych <= ':') goto yy459;
			goto yy455;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy459;
			if (yych <= '^') goto yy455;
			goto yy459;
		} else {
			if (yych <= '`') goto yy455;
			if (yych <= 'z') goto yy459;
			goto yy455;
		}
	}
yy322:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy455;
			goto yy459;
		} else {
			if (yych <= '/') goto yy455;
			if (yych <= ':') goto yy459;
			goto yy455;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy459;
			if (yych <= '^') goto yy455;
			goto yy459;
		} else {
			if (yych <= '`') goto yy455;
			if (yych <= 'z') goto yy459;
			goto yy455;
		}
	}
yy323:
	yych = *++YYCURSOR;
yy324:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy323;
			} else {
				if (yych == '\r') goto yy323;
				goto yy74;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy323;
				goto yy74;
			} else {
				if (yych <= '-') goto yy462;
				if (yych <= '.') goto yy74;
				goto yy326;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy462;
				goto yy74;
			} else {
				if (yych <= '>') goto yy327;
				if (yych <= '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy462;
				if (yych <= '^') goto yy74;
				goto yy462;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy462;
				goto yy74;
			}
		}
	}
yy325:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 32) {
		goto yy325;
	}
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy463;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy463;
			if (yych <= 0x1F) goto yy74;
			goto yy463;
		}
	} else {
		if (yych <= '<') {
			if (yych != '/') goto yy74;
		} else {
			if (yych <= '=') goto yy464;
			if (yych <= '>') goto yy327;
			if (yych >= '@') goto yy74;
		}
	}
yy326:
	yych = *++YYCURSOR;
	if (yych == '>') goto yy465;
	goto yy74;
yy327:
	++YYCURSOR;
yy328:
#line 329 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_OPEN_TAG);
       }
#line 7338 "../../lnav/src/data_scanner_re.cc"
yy329:
	yych = *++YYCURSOR;
yy330:
	if (yych <= '/') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy467;
			if (yych <= '\f') goto yy74;
			goto yy467;
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy74;
				goto yy467;
			} else {
				if (yych == '-') goto yy329;
				goto yy74;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '=') {
				if (yych <= ':') goto yy329;
				goto yy74;
			} else {
				if (yych <= '>') goto yy468;
				if (yych <= '@') goto yy74;
				goto yy329;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy74;
				goto yy329;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy329;
				goto yy74;
			}
		}
	}
yy331:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy470;
			goto yy471;
		} else {
			if (yych <= '/') goto yy470;
			if (yych <= ':') goto yy471;
			goto yy470;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy471;
			if (yych <= '^') goto yy470;
			goto yy471;
		} else {
			if (yych <= '`') goto yy470;
			if (yych <= 'z') goto yy471;
			goto yy470;
		}
	}
yy332:
	yych = *++YYCURSOR;
	if (yych == '-') goto yy472;
	goto yy74;
yy333:
	++YYCURSOR;
#line 337 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_H1);
           cap_inner.c_end -= 1;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 7412 "../../lnav/src/data_scanner_re.cc"
yy334:
	++YYCURSOR;
	goto yy157;
yy335:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy334;
		goto yy157;
	} else {
		if (yych <= '\r') goto yy334;
		if (yych == ' ') goto yy334;
		goto yy157;
	}
yy336:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy334;
		goto yy74;
	} else {
		if (yych <= '\r') goto yy334;
		if (yych == ' ') goto yy334;
		goto yy74;
	}
yy337:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 64) {
		goto yy337;
	}
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy74;
		} else {
			if (yych <= ' ') goto yy153;
			if (yych == '%') goto yy89;
			goto yy74;
		}
	} else {
		if (yych <= '?') {
			if (yych == ',') goto yy74;
			if (yych <= '.') goto yy89;
			goto yy74;
		} else {
			if (yych <= '@') goto yy91;
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy89;
			goto yy74;
		}
	}
yy338:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy339:
	if (yych <= '.') {
		if (yych <= '$') {
			if (yych <= '\n') {
				if (yych <= '\t') goto yy134;
				goto yy333;
			} else {
				if (yych == ' ') goto yy153;
				goto yy134;
			}
		} else {
			if (yych <= '+') {
				if (yych <= '%') goto yy89;
				if (yych <= '*') goto yy134;
				goto yy89;
			} else {
				if (yych <= ',') goto yy134;
				if (yych <= '-') goto yy158;
				goto yy122;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '9') {
				if (yych <= '/') goto yy124;
				goto yy338;
			} else {
				if (yych <= ':') goto yy296;
				if (yych <= '?') goto yy134;
				goto yy130;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy338;
				if (yych <= '^') goto yy134;
				goto yy338;
			} else {
				if (yych <= '`') goto yy134;
				if (yych <= 'z') goto yy138;
				goto yy134;
			}
		}
	}
yy340:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1024+yych] & 128) {
		goto yy340;
	}
	if (yych <= ',') {
		if (yych <= '%') {
			if (yych <= '$') goto yy39;
			goto yy89;
		} else {
			if (yych == '+') goto yy89;
			goto yy39;
		}
	} else {
		if (yych <= ':') {
			if (yych <= '.') goto yy473;
			if (yych <= '/') goto yy39;
			goto yy474;
		} else {
			if (yych == '@') goto yy91;
			goto yy39;
		}
	}
yy341:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy39;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy39;
				goto yy475;
			} else {
				if (yych <= '.') goto yy159;
				if (yych <= '/') goto yy124;
				goto yy341;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy194;
			if (yych <= '?') goto yy39;
			if (yych <= '@') goto yy130;
			goto yy341;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy39;
				goto yy341;
			} else {
				if (yych <= '`') goto yy39;
				if (yych <= 'z') goto yy341;
				goto yy39;
			}
		}
	}
yy342:
	yych = *++YYCURSOR;
	if (yych <= '.') {
		if (yych <= '$') {
			if (yych <= '\n') {
				if (yych <= '\t') goto yy74;
				goto yy333;
			} else {
				if (yych == ' ') goto yy153;
				goto yy74;
			}
		} else {
			if (yych <= '+') {
				if (yych <= '%') goto yy89;
				if (yych <= '*') goto yy74;
				goto yy89;
			} else {
				if (yych <= ',') goto yy74;
				if (yych <= '-') goto yy337;
				goto yy89;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '/') goto yy74;
				if (yych <= '9') goto yy476;
				goto yy74;
			} else {
				if (yych <= '@') goto yy91;
				if (yych <= 'F') goto yy476;
				goto yy338;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy338;
				goto yy74;
			} else {
				if (yych <= 'f') goto yy407;
				if (yych <= 'z') goto yy138;
				goto yy74;
			}
		}
	}
yy343:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy172;
		if (yych <= '9') goto yy477;
		if (yych <= ':') goto yy196;
		goto yy172;
	} else {
		if (yych <= 'F') goto yy477;
		if (yych <= '`') goto yy172;
		if (yych <= 'f') goto yy478;
		goto yy172;
	}
yy344:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy418;
		goto yy346;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy418;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy418;
			goto yy74;
		}
	}
yy345:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy203;
		if (yych <= '9') goto yy478;
		if (yych <= ':') goto yy196;
		goto yy203;
	} else {
		if (yych <= 'F') goto yy478;
		if (yych <= '`') goto yy203;
		if (yych <= 'f') goto yy478;
		goto yy203;
	}
yy346:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= '2') {
			if (yych <= '/') {
				if (yych == '-') {
					yyt2 = YYCURSOR;
					goto yy479;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '0') goto yy480;
				if (yych <= '1') goto yy481;
				goto yy482;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy480;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy483;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy483;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy347:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy172;
			goto yy487;
		} else {
			if (yych <= '/') goto yy172;
			if (yych <= '9') goto yy477;
			goto yy196;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy172;
			goto yy477;
		} else {
			if (yych <= '`') goto yy172;
			if (yych <= 'f') goto yy478;
			goto yy172;
		}
	}
yy348:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy487;
	goto yy172;
yy349:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy203;
			goto yy488;
		} else {
			if (yych <= '/') goto yy203;
			if (yych <= '9') goto yy478;
			goto yy196;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy203;
			goto yy478;
		} else {
			if (yych <= '`') goto yy203;
			if (yych <= 'f') goto yy478;
			goto yy203;
		}
	}
yy350:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy488;
	goto yy203;
yy351:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy334;
			if (yych <= '\f') goto yy90;
			goto yy334;
		} else {
			if (yych == ' ') goto yy334;
			if (yych <= ',') goto yy90;
			goto yy340;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') goto yy90;
			if (yych <= '9') goto yy341;
			if (yych <= '@') goto yy90;
			goto yy341;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy90;
				goto yy341;
			} else {
				if (yych <= '`') goto yy90;
				if (yych <= 'z') goto yy341;
				goto yy90;
			}
		}
	}
yy352:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') goto yy157;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy418;
		goto yy346;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy157;
			goto yy418;
		} else {
			if (yych <= '`') goto yy157;
			if (yych <= 'f') goto yy418;
			goto yy157;
		}
	}
yy353:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy478;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy490;
				}
				goto yy156;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy478;
				goto yy202;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy491;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy354:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy89;
				goto yy156;
			} else {
				if (yych <= '-') goto yy122;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				goto yy124;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
			} else {
				if (yych <= '?') goto yy156;
				if (yych <= '@') goto yy130;
				goto yy202;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy156;
				goto yy50;
			} else {
				if (yych <= '`') goto yy156;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy355:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy288;
	if (yych == ':') goto yy359;
	goto yy157;
yy356:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy487;
	if (yych == 'A') goto yy493;
	goto yy172;
yy357:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy487;
	if (yych == 'S') goto yy494;
	goto yy172;
yy358:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy355;
				}
				goto yy156;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= 'r') {
				if (yych <= '_') goto yy50;
				if (yych <= '`') goto yy156;
				goto yy492;
			} else {
				if (yych <= 's') goto yy495;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy359:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy496;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy484;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy484;
			if (yych <= '^') goto yy74;
			goto yy484;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy484;
			goto yy74;
		}
	}
yy360:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy487;
	if (yych == 'L') goto yy498;
	goto yy172;
yy361:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy355;
				}
				goto yy156;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= 'd') {
				if (yych <= '_') goto yy50;
				if (yych <= '`') goto yy156;
				goto yy492;
			} else {
				if (yych <= 'e') goto yy500;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy362:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy487;
	if (yych == 'E') goto yy498;
	goto yy172;
yy363:
	yych = *++YYCURSOR;
	if (yybm_bol[1280+yych] & 64) {
		goto yy178;
	}
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			goto yy501;
		} else {
			if (yych <= ']') goto yy363;
			if (yych <= 0xC1) goto yy74;
			goto yy181;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy182;
			if (yych <= 0xEF) goto yy183;
			goto yy184;
		} else {
			if (yych <= 0xF3) goto yy185;
			if (yych <= 0xF4) goto yy186;
			goto yy74;
		}
	}
yy364:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy355;
				}
				goto yy156;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= 'e') {
				if (yych <= '_') goto yy50;
				if (yych <= '`') goto yy156;
				goto yy492;
			} else {
				if (yych <= 'f') goto yy502;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy365:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych <= 0x00) goto yy157;
			goto yy93;
		} else {
			if (yych <= '\n') goto yy157;
			if (yych <= 0x15) goto yy93;
			goto yy157;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych == 0x1B) goto yy157;
			goto yy93;
		} else {
			if (yych <= 0xC1) goto yy157;
			if (yych <= 0xF4) goto yy93;
			goto yy157;
		}
	}
yy366:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy157;
			if (yych <= 0x08) goto yy93;
			if (yych <= '\t') goto yy365;
			goto yy157;
		} else {
			if (yych == '\r') goto yy365;
			if (yych <= 0x15) goto yy93;
			goto yy157;
		}
	} else {
		if (yych <= ' ') {
			if (yych == 0x1B) goto yy157;
			if (yych <= 0x1F) goto yy93;
			goto yy365;
		} else {
			if (yych <= 0x7F) goto yy93;
			if (yych <= 0xC1) goto yy157;
			if (yych <= 0xF4) goto yy93;
			goto yy157;
		}
	}
yy367:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '&') {
				if (yych == '!') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				yyt2 = YYCURSOR;
				goto yy219;
			} else {
				if (yych <= '\'') goto yy504;
				if (yych <= '+') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt1 = yyt3 = YYCURSOR;
				goto yy503;
			}
		} else {
			if (yych <= '>') {
				if (yych == '.') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				yyt2 = YYCURSOR;
				goto yy219;
			} else {
				if (yych <= '?') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				if (yych <= 'R') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 's') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				if (yych <= 'r') {
					yyt1 = YYCURSOR;
					goto yy505;
				}
				goto yy155;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy505;
				}
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy223;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy224;
				}
				yyt2 = YYCURSOR;
				goto yy225;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy226;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy227;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy228;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	}
yy368:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy365;
		goto yy93;
	} else {
		if (yych <= '\r') goto yy365;
		if (yych == ' ') goto yy365;
		goto yy93;
	}
yy369:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 2) {
		goto yy369;
	}
	if (yych <= '+') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0x08) goto yy92;
				if (yych <= '\t') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				yyt3 = YYCURSOR;
				goto yy334;
			} else {
				if (yych <= '\r') {
					if (yych <= '\f') goto yy92;
					yyt3 = YYCURSOR;
					goto yy365;
				} else {
					if (yych == 0x16) {
						yyt3 = YYCURSOR;
						goto yy157;
					}
					goto yy92;
				}
			}
		} else {
			if (yych <= '!') {
				if (yych <= 0x1B) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0x1F) goto yy92;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				yyt3 = YYCURSOR;
				goto yy366;
			} else {
				if (yych <= '&') {
					if (yych <= '"') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				} else {
					if (yych <= '\'') goto yy506;
					if (yych <= '*') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				}
			}
		}
	} else {
		if (yych <= '\\') {
			if (yych <= '9') {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				if (yych == '.') {
					yyt3 = YYCURSOR;
					goto yy368;
				}
				goto yy92;
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				} else {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy366;
					}
					if (yych <= '[') goto yy92;
					goto yy95;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy92;
				if (yych <= 0xC1) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 0xDF) goto yy96;
				goto yy97;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy98;
					goto yy99;
				} else {
					if (yych <= 0xF3) goto yy100;
					if (yych <= 0xF4) goto yy101;
					yyt3 = YYCURSOR;
					goto yy157;
				}
			}
		}
	}
yy370:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy489;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy156;
				goto yy122;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy355;
				}
				goto yy156;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '_') goto yy50;
				if (yych <= '`') goto yy156;
				goto yy492;
			} else {
				if (yych <= 'l') goto yy500;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy371:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy138;
		if (yych <= '@') goto yy90;
		goto yy507;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy507;
			goto yy90;
		}
	}
yy372:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy508;
	goto yy74;
yy373:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy509;
	goto yy74;
yy374:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '<') goto yy215;
			goto yy510;
		} else {
			if (yych == ']') goto yy511;
			if (yych <= 0x7F) goto yy215;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy375;
			if (yych <= 0xE0) goto yy376;
			goto yy377;
		} else {
			if (yych <= 0xF0) goto yy378;
			if (yych <= 0xF3) goto yy379;
			if (yych <= 0xF4) goto yy380;
			goto yy74;
		}
	}
yy375:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy215;
	goto yy74;
yy376:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy375;
	goto yy74;
yy377:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy375;
	goto yy74;
yy378:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy377;
	goto yy74;
yy379:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy377;
	goto yy74;
yy380:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy377;
	goto yy74;
yy381:
	yych = *++YYCURSOR;
	if (yybm_bol[768+yych] & 4) {
		goto yy381;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy216;
		if (yych <= '/') goto yy74;
		goto yy216;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy74;
			goto yy512;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy512;
			goto yy74;
		}
	}
yy382:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1536+yych] & 32) {
		goto yy92;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy220;
			goto yy218;
		} else {
			if (yych <= '\\') goto yy95;
			if (yych <= 0xC1) goto yy220;
			goto yy96;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy97;
			if (yych <= 0xEF) goto yy98;
			goto yy99;
		} else {
			if (yych <= 0xF3) goto yy100;
			if (yych <= 0xF4) goto yy101;
			goto yy220;
		}
	}
yy383:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy383;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy514;
		case 'E':
		case 'e': goto yy515;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy516;
		case 'm': goto yy517;
		case 'p':
		case 'u': goto yy518;
		default: goto yy276;
	}
yy384:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy519;
	goto yy90;
yy385:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy232;
			if (yych <= 0x08) goto yy90;
			if (yych <= '\n') goto yy232;
			goto yy90;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy232;
				goto yy90;
			} else {
				if (yych <= ' ') goto yy232;
				if (yych <= '/') goto yy90;
				goto yy107;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy90;
			if (yych == 'B') goto yy237;
			goto yy107;
		} else {
			if (yych <= 'a') {
				if (yych <= '`') goto yy90;
				goto yy107;
			} else {
				if (yych <= 'b') goto yy237;
				if (yych <= 'z') goto yy107;
				goto yy90;
			}
		}
	}
yy386:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy232;
				goto yy90;
			} else {
				if (yych <= '\n') goto yy232;
				if (yych <= '\f') goto yy90;
				goto yy232;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy90;
				goto yy232;
			} else {
				if (yych <= '/') goto yy90;
				if (yych <= '9') goto yy107;
				goto yy90;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'B') {
				if (yych <= 'A') goto yy107;
				goto yy237;
			} else {
				if (yych <= 'Z') goto yy107;
				if (yych <= '`') goto yy90;
				goto yy107;
			}
		} else {
			if (yych <= 'm') {
				if (yych <= 'b') goto yy237;
				goto yy107;
			} else {
				if (yych <= 'n') goto yy237;
				if (yych <= 'z') goto yy107;
				goto yy90;
			}
		}
	}
yy387:
	yych = *++YYCURSOR;
	if (yych == '\n') goto yy74;
	goto yy521;
yy388:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy388;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy514;
		case 'E':
		case 'e': goto yy515;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy516;
		case 'm': goto yy517;
		case 'p':
		case 'u': goto yy518;
		default: goto yy276;
	}
yy389:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy528;
	goto yy90;
yy390:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy529;
	goto yy90;
yy391:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ')') goto yy265;
			goto yy530;
		} else {
			if (yych == '/') goto yy531;
			if (yych <= 0x7F) goto yy265;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy267;
			if (yych <= 0xE0) goto yy268;
			goto yy269;
		} else {
			if (yych <= 0xF0) goto yy270;
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy74;
		}
	}
yy392:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 8) {
		goto yy392;
	}
	if (yych <= '\\') {
		if (yych <= ')') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy22;
				if (yych <= ' ') goto yy265;
				goto yy116;
			} else {
				if (yych <= '"') goto yy265;
				if (yych == '\'') goto yy265;
				goto yy116;
			}
		} else {
			if (yych <= '9') {
				if (yych <= ',') goto yy265;
				if (yych == '/') goto yy532;
				goto yy116;
			} else {
				if (yych <= '?') goto yy265;
				if (yych == '[') goto yy265;
				goto yy116;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '`') {
				if (yych <= ']') goto yy265;
				if (yych <= '_') goto yy116;
				goto yy265;
			} else {
				if (yych <= 'z') goto yy116;
				if (yych == '~') goto yy116;
				goto yy265;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy22;
				if (yych <= 0xDF) goto yy267;
				if (yych <= 0xE0) goto yy268;
				goto yy269;
			} else {
				if (yych <= 0xF0) goto yy270;
				if (yych <= 0xF3) goto yy271;
				if (yych <= 0xF4) goto yy272;
				goto yy22;
			}
		}
	}
yy393:
	yych = *++YYCURSOR;
	if (yybm_bol[1536+yych] & 128) {
		goto yy114;
	}
yy394:
#line 244 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 8955 "../../lnav/src/data_scanner_re.cc"
yy395:
	yych = *++YYCURSOR;
	if (yych == 'g') goto yy533;
	goto yy115;
yy396:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy534;
		if (yych <= '@') goto yy90;
		goto yy534;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy534;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy534;
			goto yy90;
		}
	}
yy397:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy90;
			if (yych <= '0') goto yy537;
			goto yy538;
		} else {
			if (yych <= '2') goto yy539;
			if (yych <= '9') goto yy537;
			goto yy90;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy138;
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy398:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	goto yy541;
yy399:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy400:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy134;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy134;
				goto yy396;
			} else {
				if (yych <= '.') goto yy542;
				if (yych <= '/') goto yy124;
				goto yy399;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy134;
			if (yych <= '@') goto yy130;
			goto yy399;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy134;
				goto yy399;
			} else {
				if (yych <= '`') goto yy134;
				if (yych <= 'z') goto yy399;
				goto yy134;
			}
		}
	}
yy401:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy400;
			goto yy276;
		} else {
			if (yych == ',') goto yy276;
			goto yy400;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy276;
			if (yych <= 'Z') goto yy400;
			goto yy276;
		} else {
			if (yych == '`') goto yy276;
			if (yych <= 'z') goto yy400;
			goto yy276;
		}
	}
yy402:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy543;
	if (yych == '-') goto yy544;
	goto yy400;
yy403:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy401;
		goto yy400;
	} else {
		if (yych <= 'b') goto yy401;
		if (yych == 'i') goto yy545;
		goto yy400;
	}
yy404:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy276;
				goto yy400;
			} else {
				if (yych == '+') goto yy400;
				goto yy276;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy400;
				if (yych <= '?') goto yy276;
				goto yy400;
			} else {
				if (yych <= 'B') goto yy401;
				if (yych <= 'R') goto yy400;
				goto yy401;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy400;
				if (yych <= '^') goto yy276;
				goto yy400;
			} else {
				if (yych <= '`') goto yy276;
				if (yych <= 'a') goto yy400;
				goto yy401;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy546;
				goto yy400;
			} else {
				if (yych <= 's') goto yy401;
				if (yych <= 'z') goto yy400;
				goto yy276;
			}
		}
	}
yy405:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy401;
	if (yych == 's') goto yy401;
	goto yy400;
yy406:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	if (yych <= '/') goto yy541;
	if (yych <= '5') goto yy398;
	goto yy541;
yy407:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy134;
			goto yy89;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy134;
				goto yy122;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy547;
				goto yy296;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy134;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy547;
				if (yych <= 'Z') goto yy138;
				goto yy134;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy134;
			} else {
				if (yych <= 'f') goto yy547;
				if (yych <= 'z') goto yy138;
				goto yy134;
			}
		}
	}
yy408:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy400;
	} else {
		if (yych <= '.') goto yy397;
		if (yych <= '/') goto yy400;
		if (yych <= '9') goto yy549;
		goto yy400;
	}
yy409:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy400;
	} else {
		if (yych <= '.') goto yy397;
		if (yych <= '/') goto yy400;
		if (yych <= '9') goto yy408;
		goto yy400;
	}
yy410:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy408;
		case '5': goto yy550;
		case '6':
		case '7':
		case '8':
		case '9': goto yy549;
		default: goto yy400;
	}
yy411:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy280;
	if (yych <= 'Z') goto yy551;
	if (yych <= '`') goto yy280;
	if (yych <= 'z') goto yy551;
	goto yy280;
yy412:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy552;
		if (yych <= '@') goto yy90;
		goto yy138;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy413:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '7') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy126;
				goto yy554;
			} else {
				if (yych == '%') goto yy120;
				goto yy126;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy126;
				goto yy555;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy556;
				goto yy557;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '9') goto yy558;
				if (yych <= ':') goto yy559;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy560;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy560;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy414:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy128;
				goto yy554;
			} else {
				if (yych == '%') goto yy120;
				goto yy128;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy89;
				if (yych <= ',') goto yy128;
				goto yy555;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy556;
				goto yy558;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy559;
				goto yy128;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy560;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy560;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy415:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy74;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy291;
		goto yy290;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy291;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy416:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy128;
				goto yy136;
			} else {
				if (yych <= '9') goto yy560;
				if (yych <= ':') goto yy559;
				goto yy128;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy560;
			if (yych <= 'Z') goto yy136;
			goto yy128;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy560;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy417:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy561;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy562;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy562;
		goto yy74;
	}
yy418:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy562;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy562;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy562;
		goto yy74;
	}
yy419:
	yych = *++YYCURSOR;
yy420:
	if (yybm_bol[768+yych] & 16) {
		goto yy419;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy74;
			if (yych <= '&') goto yy563;
			goto yy74;
		} else {
			if (yych <= 'Z') goto yy563;
			if (yych <= ']') goto yy74;
			if (yych <= 'z') goto yy563;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy421;
			if (yych <= 0xE0) goto yy422;
			goto yy423;
		} else {
			if (yych <= 0xF0) goto yy424;
			if (yych <= 0xF3) goto yy425;
			if (yych <= 0xF4) goto yy426;
			goto yy74;
		}
	}
yy421:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy419;
	goto yy74;
yy422:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy421;
	goto yy74;
yy423:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy421;
	goto yy74;
yy424:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy423;
	goto yy74;
yy425:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy423;
	goto yy74;
yy426:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy423;
	goto yy74;
yy427:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy565;
		if (yych <= ':') goto yy566;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy565;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy565;
		goto yy74;
	}
yy428:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy567;
		if (yych <= ':') goto yy568;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy567;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy567;
		goto yy74;
	}
yy429:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy565;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy565;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy565;
		goto yy74;
	}
yy430:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy571;
				yyt2 = YYCURSOR;
				goto yy572;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy573;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy573;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy431:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy574;
				yyt2 = YYCURSOR;
				goto yy572;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy573;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy573;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy432:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '4') goto yy574;
				goto yy575;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy571;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy573;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy573;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy433:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy573;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy573;
				goto yy279;
			} else {
				if (yych != '_') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy573;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy434:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= ',') goto yy308;
		if (yych <= ':') goto yy280;
		if (yych <= '?') goto yy308;
		goto yy280;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy308;
			goto yy280;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy280;
			goto yy308;
		}
	}
yy435:
	yych = *++YYCURSOR;
	if (yybm_bol[768+yych] & 4) {
		goto yy381;
	}
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy216;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy292;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy292;
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy292;
			goto yy74;
		}
	}
yy436:
	yych = *++YYCURSOR;
	if (yybm_bol[768+yych] & 4) {
		goto yy381;
	}
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy216;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy292;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy576;
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy576;
			goto yy74;
		}
	}
yy437:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy577;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy578;
		case 'E':
		case 'e': goto yy579;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy580;
		case 'm': goto yy581;
		case 'p':
		case 'u': goto yy582;
		default: goto yy24;
	}
yy438:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy437;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy578;
		case 'E':
		case 'e': goto yy579;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy580;
		case 'm': goto yy581;
		case 'p':
		case 'u': goto yy582;
		default: goto yy24;
	}
yy439:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy400;
		case '.': goto yy397;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy437;
		case '5': goto yy583;
		case '6':
		case '7':
		case '8':
		case '9': goto yy577;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy578;
		case 'E':
		case 'e': goto yy579;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy580;
		case 'm': goto yy581;
		case 'p':
		case 'u': goto yy582;
		default: goto yy24;
	}
yy440:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case ' ': goto yy554;
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy555;
		case '.': goto yy446;
		case '/': goto yy556;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy584;
		case ':': goto yy559;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy560;
		case 'B':
		case 'b': goto yy585;
		case 'E':
		case 'e': goto yy586;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy441:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy136;
			} else {
				if (yych <= '9') goto yy560;
				if (yych <= ':') goto yy559;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy560;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy560;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy442:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy304;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy560;
				goto yy559;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy560;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy560;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy443:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy587;
		if (yych <= '@') goto yy90;
		goto yy138;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy444:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy444;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy589;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy590;
		case 'm': goto yy591;
		case 'p':
		case 'u': goto yy592;
		default: goto yy24;
	}
yy445:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy139;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy593;
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy547;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy594;
		case 'S':
		case 's': goto yy595;
		case 'm': goto yy596;
		case 'p':
		case 'u': goto yy597;
		default: goto yy24;
	}
yy446:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy598;
		if (yych <= '@') goto yy90;
		goto yy138;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy447:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy308;
		}
		if (yych <= '0') goto yy600;
		goto yy601;
	} else {
		if (yych <= '2') goto yy602;
		if (yych <= '9') goto yy600;
		yyt2 = yyt1;
		goto yy308;
	}
yy448:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy603;
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy449:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') {
			yyt2 = yyt1;
			goto yy307;
		}
		if (yych <= '/') goto yy74;
		goto yy604;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy604;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy604;
			goto yy74;
		}
	}
yy450:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy603;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy449;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy603;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy603;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy451:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy605;
				yyt1 = YYCURSOR;
				goto yy449;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy452:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy605;
				goto yy603;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy449;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy603;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy603;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy453:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy449;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy603;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy606;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'e') goto yy603;
				goto yy606;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy454:
	yych = *++YYCURSOR;
yy455:
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy454;
			} else {
				if (yych == '\r') goto yy454;
				goto yy74;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy454;
				if (yych <= '!') goto yy74;
			} else {
				if (yych == '\'') goto yy457;
				goto yy74;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy607;
				if (yych <= '/') goto yy74;
				goto yy607;
			} else {
				if (yych == '>') goto yy460;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy607;
				if (yych <= '^') goto yy74;
				goto yy607;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy607;
				goto yy74;
			}
		}
	}
yy456:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 2) {
		goto yy608;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '"') goto yy74;
			goto yy609;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy610;
			goto yy611;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF3) goto yy614;
			if (yych <= 0xF4) goto yy615;
			goto yy74;
		}
	}
yy457:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 4) {
		goto yy616;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '\'') goto yy74;
			goto yy617;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy618;
			goto yy619;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy620;
			goto yy621;
		} else {
			if (yych <= 0xF3) goto yy622;
			if (yych <= 0xF4) goto yy623;
			goto yy74;
		}
	}
yy458:
	yych = *++YYCURSOR;
	if (yybm_bol[768+yych] & 32) {
		goto yy458;
	}
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy624;
				goto yy626;
			} else {
				if (yych == '\r') goto yy626;
				goto yy624;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy626;
				if (yych <= '!') goto yy624;
				goto yy627;
			} else {
				if (yych == '\'') goto yy628;
				if (yych <= ')') goto yy624;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '<') {
				if (yych == '-') goto yy629;
				goto yy624;
			} else {
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy631;
				if (yych <= 0x7F) goto yy624;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy632;
				if (yych <= 0xE0) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) goto yy635;
				if (yych <= 0xF3) goto yy636;
				if (yych <= 0xF4) goto yy637;
				goto yy74;
			}
		}
	}
yy459:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy638;
				goto yy74;
			} else {
				if (yych <= '\r') goto yy638;
				if (yych <= 0x1F) goto yy74;
				goto yy638;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy456;
				goto yy74;
			} else {
				if (yych <= '\'') goto yy457;
				if (yych <= ',') goto yy74;
				goto yy459;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy74;
				if (yych <= ':') goto yy459;
				goto yy74;
			} else {
				if (yych <= '=') goto yy639;
				if (yych >= '?') goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy459;
				if (yych <= '^') goto yy74;
				goto yy459;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy459;
				goto yy74;
			}
		}
	}
yy460:
	++YYCURSOR;
yy461:
#line 321 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_DECL_TAG);
       }
#line 11079 "../../lnav/src/data_scanner_re.cc"
yy462:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy640;
			} else {
				if (yych == '\r') goto yy640;
				goto yy74;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy640;
				if (yych <= ',') goto yy74;
				goto yy462;
			} else {
				if (yych <= '.') goto yy74;
				if (yych <= '/') goto yy326;
				goto yy462;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy74;
				goto yy464;
			} else {
				if (yych <= '>') goto yy327;
				if (yych <= '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy462;
				if (yych <= '^') goto yy74;
				goto yy462;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy462;
				goto yy74;
			}
		}
	}
yy463:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy463;
			} else {
				if (yych == '\r') goto yy463;
				goto yy74;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy463;
				if (yych <= ',') goto yy74;
				goto yy462;
			} else {
				if (yych <= '.') goto yy74;
				if (yych <= '/') goto yy326;
				goto yy462;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy74;
			} else {
				if (yych <= '>') goto yy327;
				if (yych <= '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy462;
				if (yych <= '^') goto yy74;
				goto yy462;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy462;
				goto yy74;
			}
		}
	}
yy464:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy641;
			} else {
				if (yych <= '\t') goto yy642;
				if (yych <= '\f') goto yy641;
				goto yy642;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy642;
				goto yy641;
			} else {
				if (yych <= '"') goto yy643;
				if (yych <= '&') goto yy641;
				goto yy644;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy641;
				goto yy74;
			} else {
				if (yych <= 0x7F) goto yy641;
				if (yych <= 0xC1) goto yy74;
				goto yy645;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy646;
				if (yych <= 0xEF) goto yy647;
				goto yy648;
			} else {
				if (yych <= 0xF3) goto yy649;
				if (yych <= 0xF4) goto yy650;
				goto yy74;
			}
		}
	}
yy465:
	++YYCURSOR;
yy466:
#line 325 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_EMPTY_TAG);
       }
#line 11218 "../../lnav/src/data_scanner_re.cc"
yy467:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy467;
		if (yych <= '\f') goto yy74;
		goto yy467;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy74;
			goto yy467;
		} else {
			if (yych != '>') goto yy74;
		}
	}
yy468:
	++YYCURSOR;
#line 333 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_CLOSE_TAG);
       }
#line 11239 "../../lnav/src/data_scanner_re.cc"
yy469:
	yych = *++YYCURSOR;
yy470:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy469;
			} else {
				if (yych == '\r') goto yy469;
				goto yy74;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy469;
				goto yy74;
			} else {
				if (yych <= '-') goto yy651;
				if (yych <= '.') goto yy74;
				goto yy326;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy651;
				goto yy74;
			} else {
				if (yych <= '?') goto yy326;
				if (yych <= '@') goto yy74;
				goto yy651;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy74;
				goto yy651;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy651;
				goto yy74;
			}
		}
	}
yy471:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy652;
			} else {
				if (yych == '\r') goto yy652;
				goto yy74;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy652;
				if (yych <= ',') goto yy74;
				goto yy471;
			} else {
				if (yych <= '.') goto yy74;
				if (yych <= '/') goto yy326;
				goto yy471;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy74;
				goto yy653;
			} else {
				if (yych == '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy471;
				if (yych <= '^') goto yy74;
				goto yy471;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy471;
				goto yy74;
			}
		}
	}
yy472:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy654;
	goto yy74;
yy473:
	yych = *++YYCURSOR;
	if (yybm_bol[1024+yych] & 128) {
		goto yy340;
	}
	if (yych <= '+') {
		if (yych == '%') goto yy89;
		if (yych <= '*') goto yy74;
		goto yy89;
	} else {
		if (yych <= '.') {
			if (yych <= ',') goto yy74;
			goto yy89;
		} else {
			if (yych == '@') goto yy91;
			goto yy74;
		}
	}
yy474:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy655;
	goto yy74;
yy475:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy39;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy39;
				goto yy340;
			} else {
				if (yych <= '.') goto yy473;
				if (yych <= '/') goto yy39;
				goto yy341;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy474;
			if (yych <= '?') goto yy39;
			if (yych <= '@') goto yy91;
			goto yy341;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy39;
				goto yy341;
			} else {
				if (yych <= '`') goto yy39;
				if (yych <= 'z') goto yy341;
				goto yy39;
			}
		}
	}
yy476:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '%') {
			if (yych <= 0x1F) {
				if (yych == '\n') goto yy333;
				goto yy134;
			} else {
				if (yych <= ' ') goto yy153;
				if (yych <= '$') goto yy134;
				goto yy89;
			}
		} else {
			if (yych <= ',') {
				if (yych == '+') goto yy89;
				goto yy134;
			} else {
				if (yych <= '-') goto yy158;
				if (yych <= '.') goto yy122;
				goto yy124;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '9') goto yy656;
				if (yych <= ':') goto yy296;
				goto yy134;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy656;
				goto yy338;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy338;
				goto yy134;
			} else {
				if (yych <= 'f') goto yy547;
				if (yych <= 'z') goto yy138;
				goto yy134;
			}
		}
	}
yy477:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy172;
		if (yych <= '9') goto yy657;
		if (yych <= ':') goto yy658;
		goto yy172;
	} else {
		if (yych <= 'F') goto yy657;
		if (yych <= '`') goto yy172;
		if (yych <= 'f') goto yy659;
		goto yy172;
	}
yy478:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy203;
		if (yych <= '9') goto yy659;
		if (yych <= ':') goto yy658;
		goto yy203;
	} else {
		if (yych <= 'F') goto yy659;
		if (yych <= '`') goto yy203;
		if (yych <= 'f') goto yy659;
		goto yy203;
	}
yy479:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy308;
			goto yy497;
		} else {
			if (yych <= '/') goto yy308;
			if (yych <= ':') goto yy497;
			goto yy308;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy497;
			if (yych <= '^') goto yy308;
			goto yy497;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy497;
			goto yy308;
		}
	}
yy480:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy662;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy664;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy664;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy481:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy665;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy664;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy664;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy482:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '4') goto yy665;
				goto yy666;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy662;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy663;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy664;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy664;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy483:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy667;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy664;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy664;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy664;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy484:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy485:
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy39;
			if (yych <= '-') goto yy668;
			goto yy359;
		} else {
			if (yych <= '/') goto yy124;
			if (yych <= '9') goto yy484;
			if (yych <= ':') goto yy194;
			goto yy39;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy124;
			if (yych <= 'Z') goto yy484;
			goto yy39;
		} else {
			if (yych == '`') goto yy39;
			if (yych <= 'z') goto yy484;
			goto yy39;
		}
	}
yy486:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= ',') goto yy308;
		if (yych <= ':') goto yy485;
		if (yych <= '?') goto yy308;
		goto yy485;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy308;
			goto yy485;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy485;
			goto yy308;
		}
	}
yy487:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy669;
	if (yych <= '/') goto yy154;
	if (yych <= '9') goto yy669;
	goto yy154;
yy488:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy670;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy670;
	goto yy74;
yy489:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy670;
	if (yych <= '/') goto yy157;
	if (yych <= '9') goto yy670;
	goto yy157;
yy490:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') goto yy157;
		if (yych <= '/') goto yy288;
		if (yych <= '9') goto yy291;
		goto yy346;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy157;
			goto yy291;
		} else {
			if (yych <= '`') goto yy157;
			if (yych <= 'f') goto yy291;
			goto yy157;
		}
	}
yy491:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy659;
				yyt3 = YYCURSOR;
				goto yy671;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy659;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy672;
				if (yych >= '{') goto yy156;
			}
		}
	}
yy492:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych <= 0x1F) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy335;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych == '%') goto yy89;
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '\'') goto yy155;
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych <= '+') goto yy89;
				yyt3 = YYCURSOR;
				goto yy335;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= '-') goto yy122;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				goto yy124;
			} else {
				if (yych <= '9') goto yy202;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy355;
				}
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt3 = YYCURSOR;
					goto yy335;
				}
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy202;
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '_') goto yy50;
				if (yych <= '`') {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 'z') goto yy492;
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	}
yy493:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'T') goto yy673;
	goto yy172;
yy494:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'E') goto yy498;
	goto yy172;
yy495:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'Z') goto yy202;
				if (yych <= '^') goto yy156;
				goto yy50;
			}
		} else {
			if (yych <= 'd') {
				if (yych <= '`') goto yy156;
				goto yy492;
			} else {
				if (yych <= 'e') goto yy500;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy496:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy497:
	if (yych <= ':') {
		if (yych <= '.') {
			if (yych <= ',') goto yy39;
			if (yych <= '-') goto yy496;
			goto yy655;
		} else {
			if (yych <= '/') goto yy39;
			if (yych <= '9') goto yy496;
			goto yy474;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy39;
			if (yych <= 'Z') goto yy496;
			goto yy39;
		} else {
			if (yych == '`') goto yy39;
			if (yych <= 'z') goto yy496;
			goto yy39;
		}
	}
yy498:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\n') {
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= '\t') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				goto yy172;
			} else {
				if (yych == '\r') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				if (yych >= ' ') {
					yyt1 = YYCURSOR;
					goto yy675;
				}
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= '$') {
				if (yych <= '!') {
					yyt1 = YYCURSOR;
					goto yy676;
				}
				if (yych <= '"') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				yyt1 = YYCURSOR;
			} else {
				if (yych <= '%') goto yy172;
				if (yych <= '&') {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				goto yy172;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy676;
				}
				if (yych <= '-') goto yy172;
				yyt1 = YYCURSOR;
				goto yy677;
			} else {
				if (yych <= '9') goto yy172;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy678;
				}
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy676;
				}
				if (yych <= 'Z') goto yy172;
				yyt1 = YYCURSOR;
			} else {
				if (yych == '`') {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= 'z') goto yy172;
				yyt1 = YYCURSOR;
			}
		}
	}
yy499:
	YYCURSOR = yyt1;
#line 462 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_CONSTANT); }
#line 12238 "../../lnav/src/data_scanner_re.cc"
yy500:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				yyt1 = YYCURSOR;
				goto yy499;
			} else {
				if (yych <= '\r') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				if (yych <= 0x1F) {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= ' ') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				yyt1 = YYCURSOR;
				goto yy676;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '"') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				goto yy203;
			} else {
				if (yych <= '&') {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy679;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				goto yy203;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy676;
				}
				if (yych <= '-') goto yy203;
				yyt1 = YYCURSOR;
				goto yy677;
			} else {
				if (yych <= '9') goto yy203;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy678;
				}
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy674;
				}
				yyt1 = YYCURSOR;
				goto yy499;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy676;
				}
				if (yych <= 'Z') goto yy203;
				yyt1 = YYCURSOR;
				goto yy499;
			} else {
				if (yych <= '_') goto yy203;
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy499;
				}
				if (yych <= 'z') goto yy492;
				yyt1 = YYCURSOR;
				goto yy499;
			}
		}
	}
yy501:
	++YYCURSOR;
#line 344 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_H1);
           cap_inner.c_end -= 1;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 12351 "../../lnav/src/data_scanner_re.cc"
yy502:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy156;
				yyt3 = YYCURSOR;
				goto yy680;
			} else {
				if (yych == '%') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy89;
				goto yy156;
			} else {
				if (yych <= '-') goto yy122;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				goto yy124;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy202;
				yyt3 = YYCURSOR;
				goto yy355;
			} else {
				if (yych <= '?') goto yy156;
				if (yych <= '@') goto yy130;
				goto yy202;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy156;
				goto yy50;
			} else {
				if (yych <= '`') goto yy156;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy503:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy334;
		goto yy220;
	} else {
		if (yych <= '\r') goto yy334;
		if (yych == ' ') goto yy334;
		goto yy220;
	}
yy504:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 2) {
		goto yy369;
	}
	if (yych <= '"') {
		if (yych <= 0x15) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy222;
				if (yych <= 0x08) goto yy93;
				yyt3 = YYCURSOR;
				goto yy365;
			} else {
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych == '\r') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				goto yy93;
			}
		} else {
			if (yych <= 0x1B) {
				if (yych <= 0x16) goto yy222;
				if (yych <= 0x1A) goto yy93;
				goto yy222;
			} else {
				if (yych <= 0x1F) goto yy93;
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				yyt3 = YYCURSOR;
				goto yy365;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych <= '&') goto yy93;
				if (yych <= '\'') goto yy506;
				yyt3 = YYCURSOR;
				goto yy365;
			} else {
				if (yych == ',') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				if (yych <= '-') goto yy93;
				yyt3 = YYCURSOR;
				goto yy368;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy93;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				if (yych <= '>') goto yy93;
				yyt3 = YYCURSOR;
				goto yy366;
			} else {
				if (yych <= 0x7F) goto yy93;
				if (yych <= 0xC1) goto yy222;
				if (yych <= 0xF4) goto yy93;
				goto yy222;
			}
		}
	}
yy505:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) goto yy220;
			if (yych <= '\n') goto yy156;
			if (yych <= '\f') goto yy220;
			goto yy156;
		} else {
			if (yych <= '"') {
				if (yych <= 0x1F) goto yy220;
				goto yy156;
			} else {
				if (yych <= '&') goto yy220;
				if (yych <= '*') goto yy156;
				goto yy220;
			}
		}
	} else {
		if (yych <= ';') {
			if (yych <= '-') {
				if (yych <= ',') goto yy156;
				goto yy220;
			} else {
				if (yych <= '.') goto yy156;
				if (yych <= '9') goto yy220;
				goto yy156;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '>') goto yy220;
				goto yy156;
			} else {
				if (yych <= '`') goto yy220;
				if (yych <= 'z') goto yy156;
				goto yy220;
			}
		}
	}
yy506:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '&') {
				if (yych == '!') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				yyt2 = YYCURSOR;
				goto yy219;
			} else {
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy681;
				}
				if (yych <= '+') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt1 = yyt3 = YYCURSOR;
				goto yy503;
			}
		} else {
			if (yych <= '>') {
				if (yych == '.') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				yyt2 = YYCURSOR;
				goto yy219;
			} else {
				if (yych <= '?') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy503;
				}
				if (yych <= 'R') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 's') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				if (yych <= 'r') {
					yyt1 = YYCURSOR;
					goto yy505;
				}
				goto yy155;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy505;
				}
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy219;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy223;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy224;
				}
				yyt2 = YYCURSOR;
				goto yy225;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy226;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy227;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy228;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	}
yy507:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '\'') goto yy155;
	if (yych <= '`') goto yy139;
	if (yych <= 'z') goto yy682;
	goto yy139;
yy508:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy683;
			}
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy683;
			}
			if (yych <= 0xC1) goto yy74;
			yyt1 = YYCURSOR;
			goto yy685;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy686;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy687;
			}
			yyt1 = YYCURSOR;
			goto yy688;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy689;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy690;
			}
			goto yy74;
		}
	}
yy509:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy691;
	goto yy74;
yy510:
	yych = *++YYCURSOR;
	if (yych == '=') goto yy510;
	if (yych != ']') goto yy74;
yy511:
	++YYCURSOR;
#line 266 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 12693 "../../lnav/src/data_scanner_re.cc"
yy512:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 4) {
		goto yy381;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy216;
		if (yych >= '0') goto yy216;
	} else {
		if (yych <= 'Z') {
			if (yych >= 'A') goto yy512;
		} else {
			if (yych <= '`') goto yy513;
			if (yych <= 'z') goto yy512;
		}
	}
yy513:
#line 460 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMAIL); }
#line 12714 "../../lnav/src/data_scanner_re.cc"
yy514:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy276;
		} else {
			if (yych == ',') goto yy276;
			if (yych <= '.') goto yy90;
			goto yy276;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy276;
			if (yych <= 'Z') goto yy90;
			goto yy276;
		} else {
			if (yych == '`') goto yy276;
			if (yych <= 'z') goto yy90;
			goto yy276;
		}
	}
yy515:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy543;
	if (yych == '-') goto yy543;
	goto yy90;
yy516:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy514;
		goto yy90;
	} else {
		if (yych <= 'b') goto yy514;
		if (yych == 'i') goto yy692;
		goto yy90;
	}
yy517:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') goto yy90;
				goto yy276;
			} else {
				if (yych == ',') goto yy276;
				goto yy90;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '/') goto yy276;
				if (yych <= '9') goto yy90;
				goto yy276;
			} else {
				if (yych == 'B') goto yy514;
				goto yy90;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'S') goto yy514;
				if (yych <= 'Z') goto yy90;
				goto yy276;
			} else {
				if (yych == '`') goto yy276;
				goto yy90;
			}
		} else {
			if (yych <= 'i') {
				if (yych <= 'b') goto yy514;
				if (yych <= 'h') goto yy90;
				goto yy693;
			} else {
				if (yych == 's') goto yy514;
				if (yych <= 'z') goto yy90;
				goto yy276;
			}
		}
	}
yy518:
	yych = *++YYCURSOR;
	if (yych == 'S') goto yy514;
	if (yych == 's') goto yy514;
	goto yy90;
yy519:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy519;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy589;
		case 'E':
		case 'e': goto yy694;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy590;
		case 'm': goto yy591;
		case 'p':
		case 'u': goto yy592;
		default: goto yy24;
	}
yy520:
	yych = *++YYCURSOR;
yy521:
	if (yybm_bol[768+yych] & 128) {
		goto yy520;
	}
	if (yych <= 0xE0) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			goto yy695;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy523;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy524;
			goto yy525;
		} else {
			if (yych <= 0xF3) goto yy526;
			if (yych <= 0xF4) goto yy527;
			goto yy74;
		}
	}
yy522:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy520;
	goto yy74;
yy523:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy522;
	goto yy74;
yy524:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy522;
	goto yy74;
yy525:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy524;
	goto yy74;
yy526:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy524;
	goto yy74;
yy527:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy524;
	goto yy74;
yy528:
	yych = *++YYCURSOR;
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= ',') goto yy74;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy74;
			goto yy528;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy74;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy89;
			goto yy74;
		} else {
			if (yych == '`') goto yy74;
			if (yych <= 'z') goto yy89;
			goto yy74;
		}
	}
yy529:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy529;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy589;
		case 'E':
		case 'e': goto yy694;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy590;
		case 'm': goto yy591;
		case 'p':
		case 'u': goto yy592;
		default: goto yy24;
	}
yy530:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ')') goto yy265;
			goto yy530;
		} else {
			if (yych == '/') goto yy696;
			if (yych <= 0x7F) goto yy265;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy267;
			if (yych <= 0xE0) goto yy268;
			goto yy269;
		} else {
			if (yych <= 0xF0) goto yy270;
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy74;
		}
	}
yy531:
	++YYCURSOR;
	goto yy394;
yy532:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 1) {
		goto yy116;
	}
	if (yych <= 0xDF) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy394;
			if (yych <= '\'') goto yy265;
			goto yy266;
		} else {
			if (yych <= 0x7F) goto yy265;
			if (yych <= 0xC1) goto yy394;
			goto yy267;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy268;
			if (yych <= 0xEF) goto yy269;
			goto yy270;
		} else {
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy394;
		}
	}
yy533:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy697;
	goto yy115;
yy534:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
yy535:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych >= '+') goto yy89;
		} else {
			if (yych <= ',') goto yy536;
			if (yych <= '.') goto yy122;
			if (yych <= '/') goto yy124;
			goto yy534;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy536;
			if (yych <= '@') goto yy130;
			goto yy534;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy534;
			} else {
				if (yych <= '`') goto yy536;
				if (yych <= 'z') goto yy534;
			}
		}
	}
yy536:
#line 434 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_VERSION_NUMBER);
       }
#line 13128 "../../lnav/src/data_scanner_re.cc"
yy537:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy698;
	if (yych <= '/') goto yy701;
	if (yych <= '9') goto yy699;
	goto yy701;
yy538:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy698;
	if (yych <= '/') goto yy701;
	if (yych <= '9') goto yy537;
	goto yy701;
yy539:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy698;
		goto yy701;
	} else {
		if (yych <= '4') goto yy537;
		if (yych <= '5') goto yy702;
		if (yych <= '9') goto yy699;
		goto yy701;
	}
yy540:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
yy541:
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy396;
		case '.': goto yy542;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy540;
		case ':': goto yy296;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy399;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy401;
		case 'E':
		case 'e': goto yy402;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy403;
		case 'm': goto yy404;
		case 'p':
		case 'u': goto yy405;
		default: goto yy276;
	}
yy542:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy74;
			goto yy89;
		} else {
			if (yych <= ',') goto yy74;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy74;
			goto yy700;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy74;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy74;
		} else {
			if (yych == '`') goto yy74;
			if (yych <= 'z') goto yy138;
			goto yy74;
		}
	}
yy543:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy90;
	if (yych <= '9') goto yy703;
	goto yy90;
yy544:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy704;
		if (yych <= '@') goto yy90;
		goto yy534;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy534;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy534;
			goto yy90;
		}
	}
yy545:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy401;
	if (yych == 'b') goto yy401;
	goto yy400;
yy546:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy401;
		goto yy400;
	} else {
		if (yych <= 'b') goto yy401;
		if (yych == 'n') goto yy401;
		goto yy400;
	}
yy547:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych >= '+') goto yy89;
		} else {
			if (yych <= '-') {
				if (yych >= '-') goto yy281;
			} else {
				if (yych <= '.') goto yy122;
				if (yych <= '/') goto yy124;
				goto yy138;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy705;
			if (yych <= '?') goto yy548;
			if (yych <= '@') goto yy130;
			goto yy138;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy138;
			} else {
				if (yych <= '`') goto yy548;
				if (yych <= 'z') goto yy138;
			}
		}
	}
yy548:
#line 303 "../../lnav/src/data_scanner_re.re"
	{
           if ((YYCURSOR.val - (this->ds_input.udata() + this->ds_next_offset)) == 17) {
               RET(DT_MAC_ADDRESS);
           } else {
               RET(DT_HEX_DUMP);
           }
       }
#line 13339 "../../lnav/src/data_scanner_re.cc"
yy549:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	goto yy553;
yy550:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	if (yych <= '/') goto yy553;
	if (yych <= '5') goto yy549;
	goto yy553;
yy551:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy280;
	if (yych <= 'Z') goto yy706;
	if (yych <= '`') goto yy280;
	if (yych <= 'z') goto yy706;
	goto yy280;
yy552:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy553:
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy134;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy134;
				goto yy396;
			} else {
				if (yych <= '.') goto yy542;
				if (yych <= '/') goto yy124;
				goto yy552;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy134;
			if (yych <= '@') goto yy130;
			goto yy399;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy134;
				goto yy399;
			} else {
				if (yych <= '`') goto yy134;
				if (yych <= 'z') goto yy399;
				goto yy134;
			}
		}
	}
yy554:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy119;
	if (yych <= '9') goto yy707;
	goto yy119;
yy555:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy708;
		if (yych <= '@') goto yy90;
		goto yy138;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy556:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy709;
		if (yych <= '@') goto yy74;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy557:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy126;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '7') goto yy710;
				goto yy711;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy294;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy712;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy712;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy558:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy128;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy711;
				goto yy294;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy712;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy712;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy559:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy74;
			goto yy713;
		} else {
			if (yych <= '.') goto yy74;
			if (yych <= '/') goto yy288;
			goto yy291;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy714;
			if (yych <= '@') goto yy74;
			goto yy291;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy560:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy128;
				goto yy136;
			} else {
				if (yych <= '9') goto yy712;
				if (yych <= ':') goto yy136;
				goto yy128;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy712;
			if (yych <= 'Z') goto yy136;
			goto yy128;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy712;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy561:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '-') {
				yyt1 = YYCURSOR;
				goto yy715;
			}
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy683;
			}
			yyt1 = YYCURSOR;
			goto yy716;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy717;
				yyt1 = YYCURSOR;
				goto yy683;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy716;
				}
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy683;
				}
				yyt1 = YYCURSOR;
				goto yy716;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy683;
			}
			if (yych <= 0xC1) goto yy548;
			if (yych <= 0xDF) {
				yyt1 = YYCURSOR;
				goto yy685;
			}
			yyt1 = YYCURSOR;
			goto yy686;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt1 = YYCURSOR;
					goto yy687;
				}
				yyt1 = YYCURSOR;
				goto yy688;
			} else {
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy689;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy690;
				}
				goto yy548;
			}
		}
	}
yy562:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy565;
			goto yy719;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy565;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy565;
			goto yy548;
		}
	}
yy563:
	yyaccept = 33;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 16) {
		goto yy419;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy564;
			if (yych <= '&') goto yy563;
		} else {
			if (yych <= 'Z') goto yy563;
			if (yych <= ']') goto yy564;
			if (yych <= 'z') goto yy563;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy421;
			if (yych <= 0xE0) goto yy422;
			goto yy423;
		} else {
			if (yych <= 0xF0) goto yy424;
			if (yych <= 0xF3) goto yy425;
			if (yych <= 0xF4) goto yy426;
		}
	}
yy564:
#line 299 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_URL); }
#line 13705 "../../lnav/src/data_scanner_re.cc"
yy565:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy720;
		if (yych <= ':') goto yy428;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy720;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy720;
		goto yy74;
	}
yy566:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy721;
		if (yych <= ':') goto yy568;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy567;
		if (yych <= '`') goto yy74;
		if (yych >= 'g') goto yy74;
	}
yy567:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy722;
		if (yych <= ':') goto yy723;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy722;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy722;
		goto yy74;
	}
yy568:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy724;
			} else {
				if (yych <= '1') goto yy725;
				if (yych <= '2') goto yy726;
				goto yy724;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy727;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy727;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy569:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy308;
		if (yych <= '9') goto yy279;
		if (yych <= '@') goto yy308;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy308;
			goto yy279;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy279;
			goto yy308;
		}
	}
yy570:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy308;
			if (yych <= '0') goto yy728;
			goto yy729;
		} else {
			if (yych <= '2') goto yy730;
			if (yych <= '9') goto yy728;
			goto yy308;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy279;
			if (yych <= '^') goto yy308;
			goto yy279;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy279;
			goto yy308;
		}
	}
yy571:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy731;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy731;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy731;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy572:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy74;
			goto yy307;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy732;
			goto yy124;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy732;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy732;
			goto yy74;
		}
	}
yy573:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy731;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy731;
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy731;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy574:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy733;
				yyt2 = YYCURSOR;
				goto yy572;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy731;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy731;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy575:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '5') goto yy733;
				goto yy731;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'F') goto yy731;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy731;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy576:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy513;
			if (yych <= '-') goto yy435;
			goto yy436;
		} else {
			if (yych <= '/') goto yy124;
			if (yych <= '9') goto yy292;
			if (yych <= ':') goto yy296;
			goto yy513;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy124;
			if (yych <= 'Z') goto yy576;
			goto yy513;
		} else {
			if (yych <= '_') goto yy279;
			if (yych <= '`') goto yy513;
			if (yych <= 'z') goto yy576;
			goto yy513;
		}
	}
yy577:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	goto yy599;
yy578:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy400;
			goto yy24;
		} else {
			if (yych == ',') goto yy24;
			goto yy400;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy400;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy400;
			goto yy24;
		}
	}
yy579:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy304;
	goto yy400;
yy580:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy578;
		goto yy400;
	} else {
		if (yych <= 'b') goto yy578;
		if (yych == 'i') goto yy734;
		goto yy400;
	}
yy581:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy24;
				goto yy400;
			} else {
				if (yych == '+') goto yy400;
				goto yy24;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy400;
				if (yych <= '?') goto yy24;
				goto yy400;
			} else {
				if (yych <= 'B') goto yy578;
				if (yych <= 'R') goto yy400;
				goto yy578;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy400;
				if (yych <= '^') goto yy24;
				goto yy400;
			} else {
				if (yych <= '`') goto yy24;
				if (yych <= 'a') goto yy400;
				goto yy578;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy735;
				goto yy400;
			} else {
				if (yych <= 's') goto yy578;
				if (yych <= 'z') goto yy400;
				goto yy24;
			}
		}
	}
yy582:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy578;
	if (yych == 's') goto yy578;
	goto yy400;
yy583:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy397;
	if (yych <= '/') goto yy599;
	if (yych <= '5') goto yy577;
	goto yy599;
yy584:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy446;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy736;
		case ':': goto yy294;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy712;
		case 'B':
		case 'b': goto yy737;
		case 'E':
		case 'e': goto yy738;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy585:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy136;
			} else {
				if (yych <= '9') goto yy712;
				if (yych <= ':') goto yy136;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy712;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy712;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy586:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy304;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy443;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy712;
				goto yy136;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy712;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy712;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy587:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy588:
	switch (yych) {
		case '%':
		case '+': goto yy89;
		case '-':
		case '.': goto yy122;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy587;
		case ':': goto yy296;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy138;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy595;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy594;
		case 'm': goto yy596;
		case 'p':
		case 'u': goto yy597;
		default: goto yy24;
	}
yy589:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy24;
		} else {
			if (yych == ',') goto yy24;
			if (yych <= '.') goto yy90;
			goto yy24;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy90;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy90;
			goto yy24;
		}
	}
yy590:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy589;
		goto yy90;
	} else {
		if (yych <= 'b') goto yy589;
		if (yych == 'i') goto yy739;
		goto yy90;
	}
yy591:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') goto yy90;
				goto yy24;
			} else {
				if (yych == ',') goto yy24;
				goto yy90;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '/') goto yy24;
				if (yych <= '9') goto yy90;
				goto yy24;
			} else {
				if (yych == 'B') goto yy589;
				goto yy90;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'S') goto yy589;
				if (yych <= 'Z') goto yy90;
				goto yy24;
			} else {
				if (yych == '`') goto yy24;
				goto yy90;
			}
		} else {
			if (yych <= 'i') {
				if (yych <= 'b') goto yy589;
				if (yych <= 'h') goto yy90;
				goto yy740;
			} else {
				if (yych == 's') goto yy589;
				if (yych <= 'z') goto yy90;
				goto yy24;
			}
		}
	}
yy592:
	yych = *++YYCURSOR;
	if (yych == 'S') goto yy589;
	if (yych == 's') goto yy589;
	goto yy90;
yy593:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych == '%') goto yy588;
			goto yy548;
		} else {
			if (yych <= '+') goto yy588;
			if (yych <= ',') goto yy548;
			if (yych <= '-') goto yy281;
			goto yy588;
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') goto yy705;
			if (yych <= '?') goto yy548;
			if (yych <= 'Z') goto yy588;
			goto yy548;
		} else {
			if (yych == '`') goto yy548;
			if (yych <= 'z') goto yy588;
			goto yy548;
		}
	}
yy594:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych != 'B') goto yy139;
	} else {
		if (yych <= 'b') goto yy595;
		if (yych == 'i') goto yy741;
		goto yy139;
	}
yy595:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy139;
			goto yy24;
		} else {
			if (yych == ',') goto yy24;
			goto yy139;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy139;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy139;
			goto yy24;
		}
	}
yy596:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy24;
				goto yy139;
			} else {
				if (yych == '+') goto yy139;
				goto yy24;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy139;
				if (yych <= '?') goto yy24;
				goto yy139;
			} else {
				if (yych <= 'B') goto yy595;
				if (yych <= 'R') goto yy139;
				goto yy595;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy139;
				if (yych <= '^') goto yy24;
				goto yy139;
			} else {
				if (yych <= '`') goto yy24;
				if (yych <= 'a') goto yy139;
				goto yy595;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy742;
				goto yy139;
			} else {
				if (yych <= 's') goto yy595;
				if (yych <= 'z') goto yy139;
				goto yy24;
			}
		}
	}
yy597:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy595;
	if (yych == 's') goto yy595;
	goto yy139;
yy598:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy599:
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy396;
		case '.': goto yy542;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy598;
		case ':': goto yy296;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy399;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy578;
		case 'E':
		case 'e': goto yy579;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy580;
		case 'm': goto yy581;
		case 'p':
		case 'u': goto yy582;
		default: goto yy24;
	}
yy600:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy743;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy744;
	goto yy74;
yy601:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy743;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy600;
	goto yy74;
yy602:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy743;
		goto yy74;
	} else {
		if (yych <= '4') goto yy600;
		if (yych <= '5') goto yy745;
		if (yych <= '9') goto yy744;
		goto yy74;
	}
yy603:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy746;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy449;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy746;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy746;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy604:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy747;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy748;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy747;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy747;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy605:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy746;
				yyt1 = YYCURSOR;
				goto yy449;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy746;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy746;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy606:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy746;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy449;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy746;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy749;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'e') goto yy746;
				goto yy749;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy607:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy750;
				goto yy74;
			} else {
				if (yych <= '\r') goto yy750;
				if (yych <= 0x1F) goto yy74;
				goto yy750;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy456;
				goto yy74;
			} else {
				if (yych <= '\'') goto yy457;
				if (yych <= ',') goto yy74;
				goto yy607;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy74;
				if (yych <= ':') goto yy607;
				goto yy74;
			} else {
				if (yych <= '=') goto yy639;
				if (yych <= '>') goto yy460;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy607;
				if (yych <= '^') goto yy74;
				goto yy607;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy607;
				goto yy74;
			}
		}
	}
yy608:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 2) {
		goto yy608;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			goto yy751;
		} else {
			if (yych <= '\\') goto yy609;
			if (yych <= 0xC1) goto yy74;
			goto yy610;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy611;
			if (yych <= 0xEF) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF3) goto yy614;
			if (yych <= 0xF4) goto yy615;
			goto yy74;
		}
	}
yy609:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy608;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy611;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF3) goto yy614;
			if (yych <= 0xF4) goto yy615;
			goto yy74;
		}
	}
yy610:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy608;
	goto yy74;
yy611:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy610;
	goto yy74;
yy612:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy610;
	goto yy74;
yy613:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy612;
	goto yy74;
yy614:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy612;
	goto yy74;
yy615:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy612;
	goto yy74;
yy616:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 4) {
		goto yy616;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			goto yy751;
		} else {
			if (yych <= '\\') goto yy617;
			if (yych <= 0xC1) goto yy74;
			goto yy618;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy619;
			if (yych <= 0xEF) goto yy620;
			goto yy621;
		} else {
			if (yych <= 0xF3) goto yy622;
			if (yych <= 0xF4) goto yy623;
			goto yy74;
		}
	}
yy617:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy616;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy619;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy620;
			goto yy621;
		} else {
			if (yych <= 0xF3) goto yy622;
			if (yych <= 0xF4) goto yy623;
			goto yy74;
		}
	}
yy618:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy616;
	goto yy74;
yy619:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy618;
	goto yy74;
yy620:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy618;
	goto yy74;
yy621:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy620;
	goto yy74;
yy622:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy620;
	goto yy74;
yy623:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy620;
	goto yy74;
yy624:
	yych = *++YYCURSOR;
yy625:
	if (yybm_bol[512+yych] & 8) {
		goto yy624;
	}
	if (yych <= 0xE0) {
		if (yych <= '-') {
			if (yych <= '*') goto yy74;
			goto yy752;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy632;
			goto yy633;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy634;
			goto yy635;
		} else {
			if (yych <= 0xF3) goto yy636;
			if (yych <= 0xF4) goto yy637;
			goto yy74;
		}
	}
yy626:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy626;
				goto yy624;
			} else {
				if (yych <= '\r') goto yy626;
				if (yych == ' ') goto yy626;
				goto yy624;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy627;
				if (yych == '\'') goto yy628;
				goto yy624;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy74;
					goto yy624;
				} else {
					if (yych <= '-') goto yy753;
					if (yych <= '/') goto yy624;
					goto yy754;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '<') goto yy624;
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy631;
				goto yy624;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy754;
					goto yy624;
				} else {
					if (yych == '`') goto yy624;
					goto yy754;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy624;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				goto yy633;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy634;
					goto yy635;
				} else {
					if (yych <= 0xF3) goto yy636;
					if (yych <= 0xF4) goto yy637;
					goto yy74;
				}
			}
		}
	}
yy627:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 128) {
		goto yy755;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy624;
			goto yy608;
		} else {
			if (yych <= '-') goto yy756;
			if (yych <= '\\') goto yy757;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy74;
		}
	}
yy628:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 1) {
		goto yy764;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy624;
			goto yy616;
		} else {
			if (yych <= '-') goto yy765;
			if (yych <= '\\') goto yy766;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy74;
		}
	}
yy629:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 2) {
		goto yy773;
	}
	if (yych <= '=') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy624;
				goto yy626;
			} else {
				if (yych == '\r') goto yy626;
				if (yych <= 0x1F) goto yy624;
				goto yy626;
			}
		} else {
			if (yych <= '\'') {
				if (yych == '"') goto yy627;
				if (yych <= '&') goto yy624;
				goto yy628;
			} else {
				if (yych <= '/') goto yy624;
				if (yych <= ':') goto yy458;
				if (yych <= '<') goto yy624;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy631;
				if (yych <= '@') goto yy624;
				if (yych <= 'Z') goto yy458;
				goto yy624;
			} else {
				if (yych == '`') goto yy624;
				if (yych <= 'z') goto yy458;
				goto yy624;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				if (yych <= 0xE0) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) goto yy635;
				if (yych <= 0xF3) goto yy636;
				if (yych <= 0xF4) goto yy637;
				goto yy74;
			}
		}
	}
yy630:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 4) {
		goto yy774;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy775;
			goto yy776;
		} else {
			if (yych <= '*') goto yy777;
			if (yych <= '-') goto yy778;
			if (yych <= '>') goto yy624;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy779;
			if (yych <= 0xE0) goto yy780;
			goto yy781;
		} else {
			if (yych <= 0xF0) goto yy782;
			if (yych <= 0xF3) goto yy783;
			if (yych <= 0xF4) goto yy784;
			goto yy74;
		}
	}
yy631:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '*') {
		if (yych <= 0x00) goto yy461;
		if (yych <= ')') goto yy625;
		goto yy461;
	} else {
		if (yych <= 0x7F) goto yy625;
		if (yych <= 0xC1) goto yy461;
		if (yych <= 0xF4) goto yy625;
		goto yy461;
	}
yy632:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy624;
	goto yy74;
yy633:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy632;
	goto yy74;
yy634:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy632;
	goto yy74;
yy635:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy634;
	goto yy74;
yy636:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy634;
	goto yy74;
yy637:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy634;
	goto yy74;
yy638:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy638;
				goto yy74;
			} else {
				if (yych <= '\r') goto yy638;
				if (yych <= 0x1F) goto yy74;
				goto yy638;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy456;
				goto yy74;
			} else {
				if (yych <= '\'') goto yy457;
				if (yych <= ',') goto yy74;
				goto yy607;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy74;
				if (yych <= ':') goto yy607;
				goto yy74;
			} else {
				if (yych <= '=') goto yy639;
				if (yych <= '>') goto yy460;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy607;
				if (yych <= '^') goto yy74;
				goto yy607;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy607;
				goto yy74;
			}
		}
	}
yy639:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 32) {
		goto yy777;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			goto yy785;
		} else {
			if (yych <= '\'') goto yy786;
			if (yych <= 0xC1) goto yy74;
			goto yy787;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy788;
			if (yych <= 0xEF) goto yy789;
			goto yy790;
		} else {
			if (yych <= 0xF3) goto yy791;
			if (yych <= 0xF4) goto yy792;
			goto yy74;
		}
	}
yy640:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy640;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy640;
			if (yych <= 0x1F) goto yy74;
			goto yy640;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy326;
			goto yy74;
		} else {
			if (yych <= '=') goto yy464;
			if (yych <= '>') goto yy327;
			if (yych <= '?') goto yy326;
			goto yy74;
		}
	}
yy641:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 16) {
		goto yy641;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '-') goto yy793;
			goto yy794;
		} else {
			if (yych <= ':') goto yy793;
			if (yych <= '>') goto yy327;
			if (yych <= '?') goto yy794;
			goto yy793;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy645;
			if (yych <= 0xE0) goto yy646;
			goto yy647;
		} else {
			if (yych <= 0xF0) goto yy648;
			if (yych <= 0xF3) goto yy649;
			if (yych <= 0xF4) goto yy650;
			goto yy74;
		}
	}
yy642:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 2) {
		goto yy794;
	}
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy642;
				goto yy641;
			} else {
				if (yych <= '\r') goto yy642;
				if (yych == ' ') goto yy642;
				goto yy641;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy643;
				if (yych == '\'') goto yy644;
				goto yy641;
			} else {
				if (yych == '.') goto yy641;
				if (yych <= ':') goto yy793;
				goto yy641;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy327;
				if (yych <= '@') goto yy641;
				if (yych <= 'Z') goto yy793;
				goto yy641;
			} else {
				if (yych == '`') goto yy641;
				if (yych <= 'z') goto yy793;
				goto yy641;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy645;
				if (yych <= 0xE0) goto yy646;
				goto yy647;
			} else {
				if (yych <= 0xF0) goto yy648;
				if (yych <= 0xF3) goto yy649;
				if (yych <= 0xF4) goto yy650;
				goto yy74;
			}
		}
	}
yy643:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 32) {
		goto yy643;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy641;
			if (yych <= '-') goto yy795;
			goto yy796;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy795;
				goto yy797;
			} else {
				if (yych <= '?') goto yy796;
				if (yych <= 'Z') goto yy795;
				goto yy798;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy795;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy799;
			goto yy800;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy801;
				goto yy802;
			} else {
				if (yych <= 0xF3) goto yy803;
				if (yych <= 0xF4) goto yy804;
				goto yy74;
			}
		}
	}
yy644:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 64) {
		goto yy644;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy641;
			if (yych <= '-') goto yy805;
			goto yy806;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy805;
				goto yy807;
			} else {
				if (yych <= '?') goto yy806;
				if (yych <= 'Z') goto yy805;
				goto yy808;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy805;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy809;
			goto yy810;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy811;
				goto yy812;
			} else {
				if (yych <= 0xF3) goto yy813;
				if (yych <= 0xF4) goto yy814;
				goto yy74;
			}
		}
	}
yy645:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy641;
	goto yy74;
yy646:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy645;
	goto yy74;
yy647:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy645;
	goto yy74;
yy648:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy647;
	goto yy74;
yy649:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy647;
	goto yy74;
yy650:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy647;
	goto yy74;
yy651:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy815;
			} else {
				if (yych == '\r') goto yy815;
				goto yy74;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy815;
				if (yych <= ',') goto yy74;
				goto yy651;
			} else {
				if (yych <= '.') goto yy74;
				if (yych <= '/') goto yy326;
				goto yy651;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy74;
				goto yy653;
			} else {
				if (yych == '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy651;
				if (yych <= '^') goto yy74;
				goto yy651;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy651;
				goto yy74;
			}
		}
	}
yy652:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy652;
			} else {
				if (yych == '\r') goto yy652;
				goto yy74;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy652;
				if (yych <= ',') goto yy74;
				goto yy651;
			} else {
				if (yych <= '.') goto yy74;
				if (yych <= '/') goto yy326;
				goto yy651;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy74;
			} else {
				if (yych == '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy651;
				if (yych <= '^') goto yy74;
				goto yy651;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy651;
				goto yy74;
			}
		}
	}
yy653:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy816;
			} else {
				if (yych <= '\t') goto yy817;
				if (yych <= '\f') goto yy816;
				goto yy817;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy817;
				goto yy816;
			} else {
				if (yych <= '"') goto yy818;
				if (yych <= '&') goto yy816;
				goto yy819;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy816;
				goto yy74;
			} else {
				if (yych <= 0x7F) goto yy816;
				if (yych <= 0xC1) goto yy74;
				goto yy820;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy821;
				if (yych <= 0xEF) goto yy822;
				goto yy823;
			} else {
				if (yych <= 0xF3) goto yy824;
				if (yych <= 0xF4) goto yy825;
				goto yy74;
			}
		}
	}
yy654:
	yych = *++YYCURSOR;
	if (yych == ',') goto yy826;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy654;
	goto yy74;
yy655:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy74;
			goto yy496;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy496;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy496;
			if (yych <= '^') goto yy74;
			goto yy496;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy496;
			goto yy74;
		}
	}
yy656:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') {
		if (yych <= '$') {
			if (yych <= '\n') {
				if (yych <= '\t') goto yy548;
				goto yy333;
			} else {
				if (yych == ' ') goto yy153;
				goto yy548;
			}
		} else {
			if (yych <= '+') {
				if (yych <= '%') goto yy89;
				if (yych <= '*') goto yy548;
				goto yy89;
			} else {
				if (yych <= ',') goto yy548;
				if (yych <= '-') goto yy342;
				goto yy122;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '9') {
				if (yych <= '/') goto yy124;
				goto yy338;
			} else {
				if (yych <= ':') goto yy705;
				if (yych <= '?') goto yy548;
				goto yy130;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy338;
				if (yych <= '^') goto yy548;
				goto yy338;
			} else {
				if (yych <= '`') goto yy548;
				if (yych <= 'z') goto yy138;
				goto yy548;
			}
		}
	}
yy657:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy172;
		if (yych <= '9') goto yy827;
		goto yy172;
	} else {
		if (yych <= 'F') goto yy827;
		if (yych <= '`') goto yy172;
		if (yych <= 'f') goto yy828;
		goto yy172;
	}
yy658:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy74;
			goto yy713;
		} else {
			if (yych <= '.') goto yy74;
			if (yych <= '/') goto yy288;
			goto yy291;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy829;
			if (yych <= '@') goto yy74;
			goto yy291;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy291;
			goto yy74;
		}
	}
yy659:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy203;
		if (yych <= '9') goto yy828;
		goto yy203;
	} else {
		if (yych <= 'F') goto yy828;
		if (yych <= '`') goto yy203;
		if (yych <= 'f') goto yy828;
		goto yy203;
	}
yy660:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych <= ',') goto yy308;
			if (yych <= '.') goto yy497;
			goto yy308;
		} else {
			if (yych <= '9') goto yy484;
			if (yych <= ':') goto yy497;
			goto yy308;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy484;
			if (yych <= '^') goto yy308;
			goto yy484;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy484;
			goto yy308;
		}
	}
yy661:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy496;
			goto yy308;
		} else {
			if (yych <= '0') goto yy830;
			if (yych <= '1') goto yy831;
			if (yych <= '2') goto yy832;
			goto yy830;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy308;
			if (yych <= 'Z') goto yy484;
			goto yy308;
		} else {
			if (yych == '`') goto yy308;
			if (yych <= 'z') goto yy484;
			goto yy308;
		}
	}
yy662:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy833;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy833;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy833;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy663:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy74;
			goto yy307;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy732;
			goto yy359;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy732;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy732;
			goto yy74;
		}
	}
yy664:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy667;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy833;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy833;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy833;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy665:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy834;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy833;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy833;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy666:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '5') goto yy834;
				goto yy833;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy663;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'F') goto yy833;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy833;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy667:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy308;
			goto yy496;
		} else {
			if (yych <= '/') goto yy308;
			if (yych <= '9') goto yy484;
			goto yy308;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy484;
			if (yych <= '^') goto yy308;
			goto yy484;
		} else {
			if (yych <= '`') goto yy308;
			if (yych <= 'z') goto yy484;
			goto yy308;
		}
	}
yy668:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') {
			if (yych <= ',') goto yy39;
			if (yych <= '-') goto yy496;
			goto yy655;
		} else {
			if (yych <= '/') goto yy39;
			if (yych <= '9') goto yy484;
			goto yy474;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy39;
			if (yych <= 'Z') goto yy484;
			goto yy39;
		} else {
			if (yych == '`') goto yy39;
			if (yych <= 'z') goto yy484;
			goto yy39;
		}
	}
yy669:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy154;
	if (yych <= '9') goto yy835;
	goto yy154;
yy670:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy836;
	goto yy74;
yy671:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy157;
			goto yy713;
		} else {
			if (yych <= '.') goto yy157;
			if (yych <= '/') goto yy288;
			goto yy291;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy829;
			if (yych <= '@') goto yy157;
			goto yy291;
		} else {
			if (yych <= '`') goto yy157;
			if (yych <= 'f') goto yy291;
			goto yy157;
		}
	}
yy672:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy828;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy828;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy837;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy673:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'E') goto yy838;
	goto yy172;
yy674:
	++YYCURSOR;
	goto yy499;
yy675:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 4) {
		goto yy153;
	}
	if (yych == '\n') goto yy333;
	goto yy499;
yy676:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy674;
		goto yy499;
	} else {
		if (yych <= '\r') goto yy674;
		if (yych == ' ') goto yy674;
		goto yy499;
	}
yy677:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy674;
			if (yych <= '\f') goto yy90;
			goto yy674;
		} else {
			if (yych == ' ') goto yy674;
			if (yych <= ',') goto yy90;
			goto yy340;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') goto yy90;
			if (yych <= '9') goto yy341;
			if (yych <= '@') goto yy90;
			goto yy341;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy90;
				goto yy341;
			} else {
				if (yych <= '`') goto yy90;
				if (yych <= 'z') goto yy341;
				goto yy90;
			}
		}
	}
yy678:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy288;
	if (yych == ':') goto yy359;
	goto yy499;
yy679:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) goto yy499;
			if (yych <= '\n') goto yy156;
			if (yych <= '\f') goto yy499;
			goto yy156;
		} else {
			if (yych <= '"') {
				if (yych <= 0x1F) goto yy499;
				goto yy156;
			} else {
				if (yych <= '&') goto yy499;
				if (yych <= '*') goto yy156;
				goto yy499;
			}
		}
	} else {
		if (yych <= ';') {
			if (yych <= '-') {
				if (yych <= ',') goto yy156;
				goto yy499;
			} else {
				if (yych <= '.') goto yy156;
				if (yych <= '9') goto yy499;
				goto yy156;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '>') goto yy499;
				goto yy156;
			} else {
				if (yych <= '`') goto yy499;
				if (yych <= 'z') goto yy156;
				goto yy499;
			}
		}
	}
yy680:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy839;
	goto yy157;
yy681:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[768+yych] & 2) {
		goto yy369;
	}
	if (yych <= '+') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy220;
				if (yych <= 0x08) goto yy92;
				if (yych <= '\t') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				yyt3 = YYCURSOR;
				goto yy334;
			} else {
				if (yych <= '\r') {
					if (yych <= '\f') goto yy92;
					yyt3 = YYCURSOR;
					goto yy365;
				} else {
					if (yych == 0x16) goto yy220;
					goto yy92;
				}
			}
		} else {
			if (yych <= '!') {
				if (yych <= 0x1B) goto yy220;
				if (yych <= 0x1F) goto yy92;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy365;
				}
				yyt3 = YYCURSOR;
				goto yy366;
			} else {
				if (yych <= '&') {
					if (yych <= '"') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				} else {
					if (yych <= '\'') goto yy506;
					if (yych <= '*') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				}
			}
		}
	} else {
		if (yych <= '\\') {
			if (yych <= '9') {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy366;
				}
				if (yych == '.') {
					yyt3 = YYCURSOR;
					goto yy368;
				}
				goto yy92;
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						yyt3 = YYCURSOR;
						goto yy365;
					}
					goto yy92;
				} else {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy366;
					}
					if (yych <= '[') goto yy92;
					goto yy95;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy92;
				if (yych <= 0xC1) goto yy220;
				if (yych <= 0xDF) goto yy96;
				goto yy97;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy98;
					goto yy99;
				} else {
					if (yych <= 0xF3) goto yy100;
					if (yych <= 0xF4) goto yy101;
					goto yy220;
				}
			}
		}
	}
yy682:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych <= 0x1F) {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy335;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych == '%') goto yy89;
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '\'') goto yy155;
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				if (yych <= '+') goto yy89;
				yyt3 = YYCURSOR;
				goto yy335;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= '-') goto yy122;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy840;
				}
				goto yy124;
			} else {
				if (yych <= '9') goto yy138;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy841;
				}
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy334;
				}
				yyt3 = YYCURSOR;
				goto yy157;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt3 = YYCURSOR;
					goto yy335;
				}
				if (yych <= '@') goto yy130;
				if (yych <= 'Z') goto yy138;
				yyt3 = YYCURSOR;
				goto yy157;
			} else {
				if (yych <= '_') goto yy138;
				if (yych <= '`') {
					yyt3 = YYCURSOR;
					goto yy157;
				}
				if (yych <= 'z') goto yy682;
				yyt3 = YYCURSOR;
				goto yy157;
			}
		}
	}
yy683:
	++YYCURSOR;
yy684:
	YYCURSOR = yyt1;
#line 301 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 16986 "../../lnav/src/data_scanner_re.cc"
yy685:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy683;
	goto yy74;
yy686:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy685;
	goto yy74;
yy687:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy685;
	goto yy74;
yy688:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy687;
	goto yy74;
yy689:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy687;
	goto yy74;
yy690:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy687;
	goto yy74;
yy691:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy842;
	goto yy74;
yy692:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy514;
	if (yych == 'b') goto yy514;
	goto yy90;
yy693:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy514;
		goto yy90;
	} else {
		if (yych <= 'b') goto yy514;
		if (yych == 'n') goto yy514;
		goto yy90;
	}
yy694:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy304;
	if (yych == '-') goto yy304;
	goto yy90;
yy695:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy843;
	goto yy74;
yy696:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1024+yych] & 16) {
		goto yy265;
	}
	if (yych <= 0xE0) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy394;
			goto yy391;
		} else {
			if (yych <= 0xC1) goto yy394;
			if (yych <= 0xDF) goto yy267;
			goto yy268;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy269;
			goto yy270;
		} else {
			if (yych <= 0xF3) goto yy271;
			if (yych <= 0xF4) goto yy272;
			goto yy394;
		}
	}
yy697:
	yych = *++YYCURSOR;
	if (yych == 'a') goto yy844;
	goto yy115;
yy698:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy90;
			if (yych <= '0') goto yy845;
			goto yy846;
		} else {
			if (yych <= '2') goto yy847;
			if (yych <= '9') goto yy845;
			goto yy90;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy138;
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy699:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy698;
	goto yy701;
yy700:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
yy701:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy536;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy536;
				goto yy396;
			} else {
				if (yych <= '.') goto yy542;
				if (yych <= '/') goto yy124;
				goto yy700;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy296;
			if (yych <= '?') goto yy536;
			if (yych <= '@') goto yy130;
			goto yy700;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy536;
				goto yy700;
			} else {
				if (yych <= '`') goto yy536;
				if (yych <= 'z') goto yy700;
				goto yy536;
			}
		}
	}
yy702:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy698;
	if (yych <= '/') goto yy701;
	if (yych <= '5') goto yy699;
	goto yy701;
yy703:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy89;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy703;
		case '@': goto yy91;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy514;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy516;
		case 'm': goto yy517;
		case 'p':
		case 'u': goto yy518;
		default: goto yy276;
	}
yy704:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+': goto yy89;
		case '-':
		case '.': goto yy122;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy704;
		case ':': goto yy296;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy534;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy848;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy849;
		case 'm': goto yy850;
		case 'p':
		case 'u': goto yy851;
		default: goto yy276;
	}
yy705:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy852;
		if (yych <= ':') goto yy124;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy852;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy852;
		goto yy74;
	}
yy706:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy853;
	goto yy280;
yy707:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy854;
	goto yy74;
yy708:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy855;
	if (yych <= '/') goto yy139;
	if (yych <= '9') goto yy856;
	goto yy139;
yy709:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') goto yy280;
	if (yych <= '/') goto yy857;
	if (yych <= '9') goto yy858;
	goto yy280;
yy710:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy126;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '7') goto yy859;
				goto yy860;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy294;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy861;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy861;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy711:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy128;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy860;
				goto yy294;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy861;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy861;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy712:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy128;
				goto yy136;
			} else {
				if (yych <= '9') goto yy861;
				if (yych <= ':') goto yy136;
				goto yy128;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy861;
			if (yych <= 'Z') goto yy136;
			goto yy128;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy861;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy713:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy862;
		goto yy74;
	} else {
		if (yych <= 'Z') goto yy862;
		if (yych <= '`') goto yy74;
		if (yych <= 'z') goto yy862;
		goto yy74;
	}
yy714:
	yych = *++YYCURSOR;
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy863;
			} else {
				if (yych <= '1') goto yy864;
				if (yych <= '2') goto yy865;
				goto yy863;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy866;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy866;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy715:
	yyaccept = 36;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy684;
		if (yych <= '9') goto yy852;
		goto yy684;
	} else {
		if (yych <= 'F') goto yy852;
		if (yych <= '`') goto yy684;
		if (yych <= 'f') goto yy852;
		goto yy684;
	}
yy716:
	yyaccept = 36;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy684;
		if (yych <= '9') goto yy720;
		if (yych <= ':') goto yy428;
		goto yy684;
	} else {
		if (yych <= 'F') goto yy720;
		if (yych <= '`') goto yy684;
		if (yych <= 'f') goto yy720;
		goto yy684;
	}
yy717:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy867;
		if (yych <= ':') goto yy568;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy868;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy868;
		goto yy74;
	}
yy718:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy852;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy852;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy852;
		goto yy74;
	}
yy719:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy868;
		if (yych <= ':') goto yy568;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy868;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy868;
		goto yy74;
	}
yy720:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy428;
	goto yy74;
yy721:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy869;
		if (yych <= ':') goto yy723;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy722;
		if (yych <= '`') goto yy74;
		if (yych >= 'g') goto yy74;
	}
yy722:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy870;
		if (yych >= ';') goto yy74;
	} else {
		if (yych <= 'F') goto yy870;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy870;
		goto yy74;
	}
yy723:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy871;
		if (yych <= ':') goto yy872;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy871;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy871;
		goto yy74;
	}
yy724:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy874;
				yyt4 = YYCURSOR;
				goto yy875;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy876;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy876;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy725:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy877;
				yyt4 = YYCURSOR;
				goto yy875;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy876;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy876;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy726:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy877;
				goto yy878;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy874;
				if (yych <= ':') {
					yyt4 = YYCURSOR;
					goto yy875;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy876;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy876;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy727:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy876;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy875;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy876;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy876;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy728:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy879;
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy880;
	goto yy280;
yy729:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy879;
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy728;
	goto yy280;
yy730:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy879;
		goto yy280;
	} else {
		if (yych <= '4') goto yy728;
		if (yych <= '5') goto yy881;
		if (yych <= '9') goto yy880;
		goto yy280;
	}
yy731:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy882;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy882;
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy882;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy732:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy883;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy884;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy883;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy883;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy733:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				yyt2 = YYCURSOR;
				goto yy570;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy882;
				yyt2 = YYCURSOR;
				goto yy572;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy882;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy882;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy734:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy578;
	if (yych == 'b') goto yy578;
	goto yy400;
yy735:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy578;
		goto yy400;
	} else {
		if (yych <= 'b') goto yy578;
		if (yych == 'n') goto yy578;
		goto yy400;
	}
yy736:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy446;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy885;
		case ':': goto yy294;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy861;
		case 'B':
		case 'b': goto yy886;
		case 'E':
		case 'e': goto yy887;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy737:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy136;
			} else {
				if (yych <= '9') goto yy861;
				if (yych <= ':') goto yy136;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy861;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy861;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy738:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy304;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy443;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy861;
				goto yy136;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy861;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy861;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy739:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy589;
	if (yych == 'b') goto yy589;
	goto yy90;
yy740:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy589;
		goto yy90;
	} else {
		if (yych <= 'b') goto yy589;
		if (yych == 'n') goto yy589;
		goto yy90;
	}
yy741:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy595;
	if (yych == 'b') goto yy595;
	goto yy139;
yy742:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy595;
		goto yy139;
	} else {
		if (yych <= 'b') goto yy595;
		if (yych == 'n') goto yy595;
		goto yy139;
	}
yy743:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy74;
		if (yych <= '0') goto yy888;
		goto yy889;
	} else {
		if (yych <= '2') goto yy890;
		if (yych <= '9') goto yy888;
		goto yy74;
	}
yy744:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy743;
	goto yy74;
yy745:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy743;
	if (yych <= '/') goto yy74;
	if (yych <= '5') goto yy744;
	goto yy74;
yy746:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy449;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy747:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy891;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy891;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy891;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy748:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy732;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy732;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy732;
			goto yy74;
		}
	}
yy749:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy892;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy750:
	yych = *++YYCURSOR;
	if (yych <= '!') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy750;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy750;
			if (yych == ' ') goto yy750;
			goto yy74;
		}
	} else {
		if (yych <= '\'') {
			if (yych <= '"') goto yy456;
			if (yych <= '&') goto yy74;
			goto yy457;
		} else {
			if (yych <= '<') goto yy74;
			if (yych <= '=') goto yy639;
			if (yych <= '>') goto yy460;
			goto yy74;
		}
	}
yy751:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy893;
			} else {
				if (yych == '\r') goto yy893;
				goto yy74;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy893;
				if (yych <= '!') goto yy74;
				goto yy456;
			} else {
				if (yych == '\'') goto yy457;
				goto yy74;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy607;
				if (yych <= '/') goto yy74;
				goto yy607;
			} else {
				if (yych == '>') goto yy460;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy607;
				if (yych <= '^') goto yy74;
				goto yy607;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy607;
				goto yy74;
			}
		}
	}
yy752:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ',') goto yy624;
			goto yy894;
		} else {
			if (yych <= 0x7F) goto yy624;
			if (yych <= 0xC1) goto yy74;
			goto yy632;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy633;
			if (yych <= 0xEF) goto yy634;
			goto yy635;
		} else {
			if (yych <= 0xF3) goto yy636;
			if (yych <= 0xF4) goto yy637;
			goto yy74;
		}
	}
yy753:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy895;
				goto yy624;
			} else {
				if (yych <= '\r') goto yy895;
				if (yych == ' ') goto yy895;
				goto yy624;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy627;
				if (yych == '\'') goto yy628;
				goto yy624;
			} else {
				if (yych <= '-') goto yy896;
				if (yych <= '/') goto yy624;
				if (yych >= ';') goto yy624;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy631;
				if (yych <= '@') goto yy624;
			} else {
				if (yych == '_') goto yy754;
				if (yych <= '`') goto yy624;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy624;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				goto yy633;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy634;
					goto yy635;
				} else {
					if (yych <= 0xF3) goto yy636;
					if (yych <= 0xF4) goto yy637;
					goto yy74;
				}
			}
		}
	}
yy754:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy895;
				goto yy624;
			} else {
				if (yych <= '\r') goto yy895;
				if (yych == ' ') goto yy895;
				goto yy624;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy627;
				if (yych == '\'') goto yy628;
				goto yy624;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy74;
					goto yy624;
				} else {
					if (yych <= '-') goto yy753;
					if (yych <= '/') goto yy624;
					goto yy754;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '<') goto yy624;
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy631;
				goto yy624;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy754;
					goto yy624;
				} else {
					if (yych == '`') goto yy624;
					goto yy754;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy624;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				goto yy633;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy634;
					goto yy635;
				} else {
					if (yych <= 0xF3) goto yy636;
					if (yych <= 0xF4) goto yy637;
					goto yy74;
				}
			}
		}
	}
yy755:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 128) {
		goto yy755;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy897;
			goto yy608;
		} else {
			if (yych <= '-') goto yy756;
			if (yych <= '\\') goto yy757;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy74;
		}
	}
yy756:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy897;
			goto yy755;
		} else {
			if (yych <= '-') goto yy898;
			if (yych != '\\') goto yy755;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy74;
		}
	}
yy757:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ')') goto yy755;
			goto yy608;
		} else {
			if (yych == '-') goto yy756;
			if (yych <= 0x7F) goto yy755;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy74;
		}
	}
yy758:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy755;
	goto yy74;
yy759:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy758;
	goto yy74;
yy760:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy758;
	goto yy74;
yy761:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy760;
	goto yy74;
yy762:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy760;
	goto yy74;
yy763:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy760;
	goto yy74;
yy764:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 1) {
		goto yy764;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy897;
			goto yy616;
		} else {
			if (yych <= '-') goto yy765;
			if (yych <= '\\') goto yy766;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy74;
		}
	}
yy765:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy897;
			goto yy764;
		} else {
			if (yych <= '-') goto yy899;
			if (yych != '\\') goto yy764;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy74;
		}
	}
yy766:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ')') goto yy764;
			goto yy616;
		} else {
			if (yych == '-') goto yy765;
			if (yych <= 0x7F) goto yy764;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy74;
		}
	}
yy767:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy764;
	goto yy74;
yy768:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy767;
	goto yy74;
yy769:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy767;
	goto yy74;
yy770:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy769;
	goto yy74;
yy771:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy769;
	goto yy74;
yy772:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy769;
	goto yy74;
yy773:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 2) {
		goto yy773;
	}
	if (yych <= '=') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy624;
				goto yy626;
			} else {
				if (yych == '\r') goto yy626;
				if (yych <= 0x1F) goto yy624;
				goto yy626;
			}
		} else {
			if (yych <= '\'') {
				if (yych == '"') goto yy627;
				if (yych <= '&') goto yy624;
				goto yy628;
			} else {
				if (yych <= '/') goto yy624;
				if (yych <= ':') goto yy458;
				if (yych <= '<') goto yy624;
				goto yy630;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy900;
				if (yych <= '@') goto yy624;
				if (yych <= 'Z') goto yy458;
				goto yy624;
			} else {
				if (yych == '`') goto yy624;
				if (yych <= 'z') goto yy458;
				goto yy624;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				if (yych <= 0xE0) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) goto yy635;
				if (yych <= 0xF3) goto yy636;
				if (yych <= 0xF4) goto yy637;
				goto yy74;
			}
		}
	}
yy774:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 4) {
		goto yy774;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych >= '#') goto yy776;
		} else {
			if (yych <= '*') goto yy777;
			if (yych <= '-') goto yy778;
			if (yych <= '>') goto yy631;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy779;
			if (yych <= 0xE0) goto yy780;
			goto yy781;
		} else {
			if (yych <= 0xF0) goto yy782;
			if (yych <= 0xF3) goto yy783;
			if (yych <= 0xF4) goto yy784;
			goto yy74;
		}
	}
yy775:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 8) {
		goto yy775;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy902;
			goto yy785;
		} else {
			if (yych <= '-') goto yy903;
			if (yych <= '>') goto yy904;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy74;
		}
	}
yy776:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 16) {
		goto yy776;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy902;
			goto yy786;
		} else {
			if (yych <= '-') goto yy911;
			if (yych <= '>') goto yy912;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy913;
			if (yych <= 0xE0) goto yy914;
			goto yy915;
		} else {
			if (yych <= 0xF0) goto yy916;
			if (yych <= 0xF3) goto yy917;
			if (yych <= 0xF4) goto yy918;
			goto yy74;
		}
	}
yy777:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 32) {
		goto yy777;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy785;
			goto yy786;
		} else {
			if (yych <= '>') goto yy460;
			if (yych <= 0xC1) goto yy74;
			goto yy787;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy788;
			if (yych <= 0xEF) goto yy789;
			goto yy790;
		} else {
			if (yych <= 0xF3) goto yy791;
			if (yych <= 0xF4) goto yy792;
			goto yy74;
		}
	}
yy778:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy775;
			goto yy774;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy776;
				goto yy774;
			} else {
				if (yych <= '-') goto yy919;
				if (yych <= '=') goto yy774;
				goto yy631;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy774;
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy780;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy781;
				goto yy782;
			} else {
				if (yych <= 0xF3) goto yy783;
				if (yych <= 0xF4) goto yy784;
				goto yy74;
			}
		}
	}
yy779:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy774;
	goto yy74;
yy780:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy779;
	goto yy74;
yy781:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy779;
	goto yy74;
yy782:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy781;
	goto yy74;
yy783:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy781;
	goto yy74;
yy784:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy781;
	goto yy74;
yy785:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 64) {
		goto yy785;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			goto yy920;
		} else {
			if (yych <= '>') goto yy921;
			if (yych <= 0xC1) goto yy74;
			goto yy922;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy923;
			if (yych <= 0xEF) goto yy924;
			goto yy925;
		} else {
			if (yych <= 0xF3) goto yy926;
			if (yych <= 0xF4) goto yy927;
			goto yy74;
		}
	}
yy786:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 128) {
		goto yy786;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			goto yy920;
		} else {
			if (yych <= '>') goto yy928;
			if (yych <= 0xC1) goto yy74;
			goto yy929;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy930;
			if (yych <= 0xEF) goto yy931;
			goto yy932;
		} else {
			if (yych <= 0xF3) goto yy933;
			if (yych <= 0xF4) goto yy934;
			goto yy74;
		}
	}
yy787:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy777;
	goto yy74;
yy788:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy787;
	goto yy74;
yy789:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy787;
	goto yy74;
yy790:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy789;
	goto yy74;
yy791:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy789;
	goto yy74;
yy792:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy789;
	goto yy74;
yy793:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 1) {
		goto yy793;
	}
	if (yych <= 0x7F) {
		if (yych <= '<') {
			if (yych <= 0x00) goto yy74;
			if (yych != '/') goto yy641;
		} else {
			if (yych <= '=') goto yy642;
			if (yych <= '>') goto yy327;
			if (yych >= '@') goto yy641;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy645;
			if (yych <= 0xE0) goto yy646;
			goto yy647;
		} else {
			if (yych <= 0xF0) goto yy648;
			if (yych <= 0xF3) goto yy649;
			if (yych <= 0xF4) goto yy650;
			goto yy74;
		}
	}
yy794:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 16) {
		goto yy641;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '-') goto yy793;
			goto yy794;
		} else {
			if (yych <= ':') goto yy793;
			if (yych <= '>') goto yy465;
			if (yych <= '?') goto yy794;
			goto yy793;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy645;
			if (yych <= 0xE0) goto yy646;
			goto yy647;
		} else {
			if (yych <= 0xF0) goto yy648;
			if (yych <= 0xF3) goto yy649;
			if (yych <= 0xF4) goto yy650;
			goto yy74;
		}
	}
yy795:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy795;
				goto yy643;
			} else {
				if (yych <= '\r') goto yy795;
				if (yych == ' ') goto yy795;
				goto yy643;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy641;
				if (yych == '-') goto yy795;
				goto yy643;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy795;
				} else {
					if (yych <= '<') goto yy643;
					if (yych <= '=') goto yy935;
					goto yy797;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy796;
				if (yych <= '@') goto yy643;
				if (yych <= 'Z') goto yy795;
				goto yy643;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy798;
					goto yy643;
				} else {
					if (yych == '`') goto yy643;
					goto yy795;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy643;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy799;
				goto yy800;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy801;
					goto yy802;
				} else {
					if (yych <= 0xF3) goto yy803;
					if (yych <= 0xF4) goto yy804;
					goto yy74;
				}
			}
		}
	}
yy796:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 32) {
		goto yy643;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy641;
			if (yych <= '-') goto yy795;
			goto yy796;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy795;
				goto yy936;
			} else {
				if (yych <= '?') goto yy796;
				if (yych <= 'Z') goto yy795;
				goto yy798;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy795;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy799;
			goto yy800;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy801;
				goto yy802;
			} else {
				if (yych <= 0xF3) goto yy803;
				if (yych <= 0xF4) goto yy804;
				goto yy74;
			}
		}
	}
yy797:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy328;
			if (yych <= '!') goto yy937;
			goto yy938;
		} else {
			if (yych == '\\') goto yy939;
			if (yych <= 0x7F) goto yy937;
			goto yy328;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy940;
			if (yych <= 0xE0) goto yy941;
			goto yy942;
		} else {
			if (yych <= 0xF0) goto yy943;
			if (yych <= 0xF3) goto yy944;
			if (yych <= 0xF4) goto yy945;
			goto yy328;
		}
	}
yy798:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy643;
				goto yy795;
			} else {
				if (yych <= '.') goto yy643;
				if (yych <= '/') goto yy796;
				goto yy795;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy643;
				if (yych <= '>') goto yy797;
				goto yy796;
			} else {
				if (yych <= '@') goto yy643;
				if (yych <= 'Z') goto yy795;
				goto yy643;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy643;
				goto yy795;
			} else {
				if (yych <= 0x7F) goto yy643;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy800;
				if (yych <= 0xEF) goto yy801;
				goto yy802;
			} else {
				if (yych <= 0xF3) goto yy803;
				if (yych <= 0xF4) goto yy804;
				goto yy74;
			}
		}
	}
yy799:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy643;
	goto yy74;
yy800:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy799;
	goto yy74;
yy801:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy799;
	goto yy74;
yy802:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy801;
	goto yy74;
yy803:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy801;
	goto yy74;
yy804:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy801;
	goto yy74;
yy805:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy805;
				goto yy644;
			} else {
				if (yych <= '\r') goto yy805;
				if (yych == ' ') goto yy805;
				goto yy644;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy641;
				if (yych == '-') goto yy805;
				goto yy644;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy805;
				} else {
					if (yych <= '<') goto yy644;
					if (yych <= '=') goto yy946;
					goto yy807;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy806;
				if (yych <= '@') goto yy644;
				if (yych <= 'Z') goto yy805;
				goto yy644;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy808;
					goto yy644;
				} else {
					if (yych == '`') goto yy644;
					goto yy805;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy644;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy809;
				goto yy810;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy811;
					goto yy812;
				} else {
					if (yych <= 0xF3) goto yy813;
					if (yych <= 0xF4) goto yy814;
					goto yy74;
				}
			}
		}
	}
yy806:
	yych = *++YYCURSOR;
	if (yybm_bol[512+yych] & 64) {
		goto yy644;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy641;
			if (yych <= '-') goto yy805;
			goto yy806;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy805;
				goto yy947;
			} else {
				if (yych <= '?') goto yy806;
				if (yych <= 'Z') goto yy805;
				goto yy808;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy805;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy809;
			goto yy810;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy811;
				goto yy812;
			} else {
				if (yych <= 0xF3) goto yy813;
				if (yych <= 0xF4) goto yy814;
				goto yy74;
			}
		}
	}
yy807:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy328;
			if (yych <= '&') goto yy948;
			goto yy938;
		} else {
			if (yych == '\\') goto yy949;
			if (yych <= 0x7F) goto yy948;
			goto yy328;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy950;
			if (yych <= 0xE0) goto yy951;
			goto yy952;
		} else {
			if (yych <= 0xF0) goto yy953;
			if (yych <= 0xF3) goto yy954;
			if (yych <= 0xF4) goto yy955;
			goto yy328;
		}
	}
yy808:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy644;
				goto yy805;
			} else {
				if (yych <= '.') goto yy644;
				if (yych <= '/') goto yy806;
				goto yy805;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy644;
				if (yych <= '>') goto yy807;
				goto yy806;
			} else {
				if (yych <= '@') goto yy644;
				if (yych <= 'Z') goto yy805;
				goto yy644;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy644;
				goto yy805;
			} else {
				if (yych <= 0x7F) goto yy644;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy810;
				if (yych <= 0xEF) goto yy811;
				goto yy812;
			} else {
				if (yych <= 0xF3) goto yy813;
				if (yych <= 0xF4) goto yy814;
				goto yy74;
			}
		}
	}
yy809:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy644;
	goto yy74;
yy810:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy809;
	goto yy74;
yy811:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy809;
	goto yy74;
yy812:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy811;
	goto yy74;
yy813:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy811;
	goto yy74;
yy814:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy811;
	goto yy74;
yy815:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy815;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy815;
			if (yych <= 0x1F) goto yy74;
			goto yy815;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy326;
			goto yy74;
		} else {
			if (yych <= '=') goto yy653;
			if (yych == '?') goto yy326;
			goto yy74;
		}
	}
yy816:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy816;
				goto yy956;
			} else {
				if (yych <= '.') goto yy816;
				if (yych <= '/') goto yy957;
				goto yy956;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy816;
				if (yych <= '>') goto yy74;
				goto yy957;
			} else {
				if (yych <= '@') goto yy816;
				if (yych <= 'Z') goto yy956;
				goto yy816;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy816;
				goto yy956;
			} else {
				if (yych <= 0x7F) goto yy816;
				if (yych <= 0xC1) goto yy74;
				goto yy820;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy821;
				if (yych <= 0xEF) goto yy822;
				goto yy823;
			} else {
				if (yych <= 0xF3) goto yy824;
				if (yych <= 0xF4) goto yy825;
				goto yy74;
			}
		}
	}
yy817:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy817;
				goto yy816;
			} else {
				if (yych <= '\r') goto yy817;
				if (yych == ' ') goto yy817;
				goto yy816;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy818;
				if (yych == '\'') goto yy819;
				goto yy816;
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy956;
					goto yy816;
				} else {
					if (yych <= '/') goto yy957;
					if (yych <= ':') goto yy956;
					goto yy816;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy74;
				if (yych <= '?') goto yy957;
				if (yych <= '@') goto yy816;
				goto yy956;
			} else {
				if (yych == '_') goto yy956;
				if (yych <= '`') goto yy816;
				goto yy956;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy816;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy820;
				goto yy821;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy822;
					goto yy823;
				} else {
					if (yych <= 0xF3) goto yy824;
					if (yych <= 0xF4) goto yy825;
					goto yy74;
				}
			}
		}
	}
yy818:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '!') goto yy818;
				goto yy816;
			} else {
				if (yych == '-') goto yy958;
				if (yych <= '.') goto yy818;
				goto yy959;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy958;
				if (yych <= '=') goto yy818;
				goto yy960;
			} else {
				if (yych <= '?') goto yy959;
				if (yych <= '@') goto yy818;
				if (yych <= 'Z') goto yy958;
				goto yy818;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy961;
				if (yych <= '^') goto yy818;
				goto yy958;
			} else {
				if (yych <= '`') goto yy818;
				if (yych <= 'z') goto yy958;
				if (yych <= 0x7F) goto yy818;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy962;
				if (yych <= 0xE0) goto yy963;
				goto yy964;
			} else {
				if (yych <= 0xF0) goto yy965;
				if (yych <= 0xF3) goto yy966;
				if (yych <= 0xF4) goto yy967;
				goto yy74;
			}
		}
	}
yy819:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '&') goto yy819;
				goto yy816;
			} else {
				if (yych == '-') goto yy968;
				if (yych <= '.') goto yy819;
				goto yy969;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy968;
				if (yych <= '=') goto yy819;
				goto yy970;
			} else {
				if (yych <= '?') goto yy969;
				if (yych <= '@') goto yy819;
				if (yych <= 'Z') goto yy968;
				goto yy819;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy971;
				if (yych <= '^') goto yy819;
				goto yy968;
			} else {
				if (yych <= '`') goto yy819;
				if (yych <= 'z') goto yy968;
				if (yych <= 0x7F) goto yy819;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy972;
				if (yych <= 0xE0) goto yy973;
				goto yy974;
			} else {
				if (yych <= 0xF0) goto yy975;
				if (yych <= 0xF3) goto yy976;
				if (yych <= 0xF4) goto yy977;
				goto yy74;
			}
		}
	}
yy820:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy816;
	goto yy74;
yy821:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy820;
	goto yy74;
yy822:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy820;
	goto yy74;
yy823:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy822;
	goto yy74;
yy824:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy822;
	goto yy74;
yy825:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy822;
	goto yy74;
yy826:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy978;
	goto yy74;
yy827:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy172;
		if (yych <= '9') goto yy979;
		goto yy172;
	} else {
		if (yych <= 'F') goto yy979;
		if (yych <= '`') goto yy172;
		if (yych <= 'f') goto yy980;
		goto yy172;
	}
yy828:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy203;
		if (yych <= '9') goto yy980;
		goto yy203;
	} else {
		if (yych <= 'F') goto yy980;
		if (yych <= '`') goto yy203;
		if (yych <= 'f') goto yy980;
		goto yy203;
	}
yy829:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= '2') {
			if (yych <= '/') {
				if (yych == '-') {
					yyt2 = YYCURSOR;
					goto yy479;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '0') goto yy981;
				if (yych <= '1') goto yy982;
				goto yy983;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy981;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy984;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy984;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy830:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy985;
	if (yych <= '/') goto yy485;
	if (yych <= '9') goto yy986;
	goto yy485;
yy831:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy985;
	if (yych <= '/') goto yy485;
	if (yych <= '9') goto yy830;
	goto yy485;
yy832:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy985;
		goto yy485;
	} else {
		if (yych <= '4') goto yy830;
		if (yych <= '5') goto yy987;
		if (yych <= '9') goto yy986;
		goto yy485;
	}
yy833:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy667;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy988;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy988;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy988;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy834:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy661;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy988;
				yyt2 = YYCURSOR;
				goto yy663;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy988;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy988;
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy835:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy989;
	goto yy154;
yy836:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy990;
	goto yy74;
yy837:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy980;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy980;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy991;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy838:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy992;
	goto yy172;
yy839:
	yych = *++YYCURSOR;
	if (yych == '-') goto yy993;
	goto yy74;
yy840:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy334;
			goto yy90;
		} else {
			if (yych <= '\r') goto yy334;
			if (yych == ' ') goto yy334;
			goto yy90;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy138;
			if (yych <= '@') goto yy90;
			if (yych <= 'Z') goto yy138;
			goto yy90;
		} else {
			if (yych == '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy841:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy124;
	goto yy157;
yy842:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy996;
			}
			yyt2 = YYCURSOR;
			goto yy994;
		} else {
			if (yych <= ':') goto yy74;
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy997;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy998;
			}
			yyt2 = YYCURSOR;
			goto yy999;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1000;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1001;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1002;
			}
			goto yy74;
		}
	}
yy843:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy1003;
	goto yy74;
yy844:
	yych = *++YYCURSOR;
	if (yych == 'm') goto yy1004;
	goto yy115;
yy845:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1008;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1009;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1010;
				}
				goto yy1011;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1012;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1013;
				}
				yyt1 = YYCURSOR;
				goto yy1014;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				goto yy536;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1015;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1016;
				}
				yyt1 = YYCURSOR;
				goto yy1017;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1018;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1019;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1020;
				}
				goto yy536;
			}
		}
	}
yy846:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1008;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1009;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1010;
				}
				goto yy845;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1012;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1013;
				}
				yyt1 = YYCURSOR;
				goto yy1014;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				goto yy536;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1015;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1016;
				}
				yyt1 = YYCURSOR;
				goto yy1017;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1018;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1019;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1020;
				}
				goto yy536;
			}
		}
	}
yy847:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy1008;
				}
				yyt1 = YYCURSOR;
				goto yy1009;
			}
		} else {
			if (yych <= '5') {
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1010;
				}
				if (yych <= '4') goto yy845;
				goto yy1021;
			} else {
				if (yych <= '9') goto yy1011;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1012;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1013;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1014;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				goto yy536;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1015;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1016;
				}
				yyt1 = YYCURSOR;
				goto yy1017;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1018;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1019;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1020;
				}
				goto yy536;
			}
		}
	}
yy848:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy535;
			goto yy276;
		} else {
			if (yych == ',') goto yy276;
			goto yy535;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy276;
			if (yych <= 'Z') goto yy535;
			goto yy276;
		} else {
			if (yych == '`') goto yy276;
			if (yych <= 'z') goto yy535;
			goto yy276;
		}
	}
yy849:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy848;
		goto yy535;
	} else {
		if (yych <= 'b') goto yy848;
		if (yych == 'i') goto yy1022;
		goto yy535;
	}
yy850:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy276;
				goto yy535;
			} else {
				if (yych == '+') goto yy535;
				goto yy276;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy535;
				if (yych <= '?') goto yy276;
				goto yy535;
			} else {
				if (yych <= 'B') goto yy848;
				if (yych <= 'R') goto yy535;
				goto yy848;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy535;
				if (yych <= '^') goto yy276;
				goto yy535;
			} else {
				if (yych <= '`') goto yy276;
				if (yych <= 'a') goto yy535;
				goto yy848;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy1023;
				goto yy535;
			} else {
				if (yych <= 's') goto yy848;
				if (yych <= 'z') goto yy535;
				goto yy276;
			}
		}
	}
yy851:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy848;
	if (yych == 's') goto yy848;
	goto yy535;
yy852:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1024;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1024;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1024;
		goto yy74;
	}
yy853:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1025;
		if (yych <= '@') goto yy74;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy854:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1026;
	goto yy74;
yy855:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy90;
		if (yych <= '9') goto yy1027;
		if (yych <= '@') goto yy90;
		goto yy138;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy90;
			goto yy138;
		} else {
			if (yych <= '`') goto yy90;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy856:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy855;
	goto yy139;
yy857:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1029;
		if (yych <= '@') goto yy74;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy858:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy857;
	goto yy280;
yy859:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy126;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '7') goto yy1030;
				goto yy1031;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy294;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy1032;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy1032;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy860:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy128;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy1031;
				goto yy294;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy1032;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy1032;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy861:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy128;
				goto yy136;
			} else {
				if (yych <= '9') goto yy1032;
				if (yych <= ':') goto yy136;
				goto yy128;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy1032;
			if (yych <= 'Z') goto yy136;
			goto yy128;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy1032;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy862:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy862;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy862;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy862;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy863:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1034;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1035;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1036;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1036;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy864:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1037;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1035;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1036;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1036;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy865:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '4') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1037;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy1038;
				if (yych <= '9') goto yy1034;
				yyt1 = YYCURSOR;
				goto yy1035;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1036;
				goto yy279;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'f') goto yy1036;
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy866:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy1036;
				yyt1 = YYCURSOR;
				goto yy1035;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1036;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1036;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy867:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1039;
		if (yych <= ':') goto yy723;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1040;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1040;
		goto yy74;
	}
yy868:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1040;
		if (yych <= ':') goto yy723;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1040;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1040;
		goto yy74;
	}
yy869:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy996;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			yyt1 = YYCURSOR;
			goto yy1041;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy723;
				yyt2 = YYCURSOR;
				goto yy994;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy1041;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy994;
				}
				yyt1 = YYCURSOR;
				goto yy1041;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy997;
			}
			yyt2 = YYCURSOR;
			goto yy998;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy999;
				}
				yyt2 = YYCURSOR;
				goto yy1000;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1001;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1002;
				}
				goto yy74;
			}
		}
	}
yy870:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1042;
		if (yych <= ':') goto yy723;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1042;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1042;
		goto yy74;
	}
yy871:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1043;
		if (yych <= ':') goto yy1044;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1043;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1043;
		goto yy74;
	}
yy872:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1045;
			} else {
				if (yych <= '1') goto yy1046;
				if (yych <= '2') goto yy1047;
				goto yy1045;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1048;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1048;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy873:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') {
			yyt2 = yyt4;
			goto yy308;
		}
		if (yych <= '0') goto yy600;
		goto yy601;
	} else {
		if (yych <= '2') goto yy602;
		if (yych <= '9') goto yy600;
		yyt2 = yyt4;
		goto yy308;
	}
yy874:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1049;
				yyt4 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1049;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1049;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy875:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') {
			yyt2 = yyt4;
			goto yy307;
		}
		if (yych <= '/') goto yy74;
		goto yy1048;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1048;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1048;
			goto yy74;
		}
	}
yy876:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1049;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy875;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1049;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1049;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy877:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1050;
				yyt4 = YYCURSOR;
				goto yy875;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1049;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1049;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy878:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1050;
				goto yy1049;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt4 = YYCURSOR;
					goto yy875;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1049;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1049;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy879:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy74;
			if (yych <= '0') goto yy1051;
			goto yy1052;
		} else {
			if (yych <= '2') goto yy1053;
			if (yych <= '9') goto yy1051;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy279;
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy880:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy879;
	goto yy280;
yy881:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy879;
	if (yych <= '/') goto yy280;
	if (yych <= '5') goto yy880;
	goto yy280;
yy882:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy279;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy572;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy134;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy883:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1054;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1054;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1054;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy884:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1048;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1048;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1048;
			goto yy74;
		}
	}
yy885:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy446;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1055;
		case ':': goto yy294;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1032;
		case 'B':
		case 'b': goto yy1056;
		case 'E':
		case 'e': goto yy1057;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy886:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy24;
			goto yy136;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy136;
			} else {
				if (yych <= '9') goto yy1032;
				if (yych <= ':') goto yy136;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy136;
			if (yych <= 'F') goto yy1032;
			if (yych <= 'Z') goto yy136;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy1032;
				if (yych <= 'z') goto yy136;
				goto yy24;
			}
		}
	}
yy887:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy128;
			goto yy304;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy128;
				goto yy443;
			} else {
				if (yych <= '/') goto yy136;
				if (yych <= '9') goto yy1032;
				goto yy136;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy136;
			} else {
				if (yych <= 'F') goto yy1032;
				if (yych <= 'Z') goto yy136;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy136;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy1032;
				if (yych <= 'z') goto yy136;
				goto yy128;
			}
		}
	}
yy888:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy1058;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1059;
	goto yy74;
yy889:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy1058;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy888;
	goto yy74;
yy890:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy1058;
		goto yy74;
	} else {
		if (yych <= '4') goto yy888;
		if (yych <= '5') goto yy1060;
		if (yych <= '9') goto yy1059;
		goto yy74;
	}
yy891:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1061;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy748;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1061;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1061;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy892:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy307;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1062;
			if (yych <= '1') goto yy1063;
			goto yy1064;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy1065;
			if (yych <= '@') goto yy74;
			goto yy604;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy604;
			goto yy74;
		}
	}
yy893:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy893;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy893;
			if (yych <= 0x1F) goto yy74;
			goto yy893;
		}
	} else {
		if (yych <= '&') {
			if (yych == '"') goto yy456;
			goto yy74;
		} else {
			if (yych <= '\'') goto yy457;
			if (yych == '>') goto yy460;
			goto yy74;
		}
	}
yy894:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ',') goto yy624;
			goto yy894;
		} else {
			if (yych == '>') goto yy900;
			if (yych <= 0x7F) goto yy624;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy632;
			if (yych <= 0xE0) goto yy633;
			goto yy634;
		} else {
			if (yych <= 0xF0) goto yy635;
			if (yych <= 0xF3) goto yy636;
			if (yych <= 0xF4) goto yy637;
			goto yy74;
		}
	}
yy895:
	yych = *++YYCURSOR;
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy624;
				goto yy895;
			} else {
				if (yych == '\r') goto yy895;
				goto yy624;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy895;
				if (yych <= '!') goto yy624;
				goto yy627;
			} else {
				if (yych == '\'') goto yy628;
				if (yych <= ')') goto yy624;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '<') {
				if (yych == '-') goto yy752;
				goto yy624;
			} else {
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy631;
				if (yych <= 0x7F) goto yy624;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy632;
				if (yych <= 0xE0) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) goto yy635;
				if (yych <= 0xF3) goto yy636;
				if (yych <= 0xF4) goto yy637;
				goto yy74;
			}
		}
	}
yy896:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy895;
				goto yy624;
			} else {
				if (yych <= '\r') goto yy895;
				if (yych == ' ') goto yy895;
				goto yy624;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy627;
				if (yych == '\'') goto yy628;
				goto yy624;
			} else {
				if (yych <= '-') goto yy896;
				if (yych <= '/') goto yy624;
				if (yych <= ':') goto yy754;
				goto yy624;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy630;
				if (yych <= '>') goto yy900;
				if (yych <= '@') goto yy624;
				goto yy754;
			} else {
				if (yych == '_') goto yy754;
				if (yych <= '`') goto yy624;
				goto yy754;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy624;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				goto yy633;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy634;
					goto yy635;
				} else {
					if (yych <= 0xF3) goto yy636;
					if (yych <= 0xF4) goto yy637;
					goto yy74;
				}
			}
		}
	}
yy897:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1066;
				goto yy624;
			} else {
				if (yych <= '\r') goto yy1066;
				if (yych == ' ') goto yy1066;
				goto yy624;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy627;
				if (yych == '\'') goto yy628;
				goto yy624;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy74;
					goto yy624;
				} else {
					if (yych <= '-') goto yy753;
					if (yych <= '/') goto yy624;
					goto yy754;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych == '>') goto yy631;
				if (yych <= '@') goto yy624;
				goto yy754;
			} else {
				if (yych == '_') goto yy754;
				if (yych <= '`') goto yy624;
				goto yy754;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy624;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy632;
				goto yy633;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy634;
					goto yy635;
				} else {
					if (yych <= 0xF3) goto yy636;
					if (yych <= 0xF4) goto yy637;
					goto yy74;
				}
			}
		}
	}
yy898:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy897;
			goto yy755;
		} else {
			if (yych <= '=') {
				if (yych <= '-') goto yy898;
				goto yy755;
			} else {
				if (yych <= '>') goto yy1067;
				if (yych <= '[') goto yy755;
				goto yy757;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy755;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy758;
			goto yy759;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy760;
				goto yy761;
			} else {
				if (yych <= 0xF3) goto yy762;
				if (yych <= 0xF4) goto yy763;
				goto yy74;
			}
		}
	}
yy899:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy897;
			goto yy764;
		} else {
			if (yych <= '=') {
				if (yych <= '-') goto yy899;
				goto yy764;
			} else {
				if (yych <= '>') goto yy1068;
				if (yych <= '[') goto yy764;
				goto yy766;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy764;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy767;
			goto yy768;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy769;
				goto yy770;
			} else {
				if (yych <= 0xF3) goto yy771;
				if (yych <= 0xF4) goto yy772;
				goto yy74;
			}
		}
	}
yy900:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[512+yych] & 8) {
		goto yy624;
	}
	if (yych <= 0xE0) {
		if (yych <= '-') {
			if (yych >= '+') goto yy752;
		} else {
			if (yych <= 0xC1) goto yy901;
			if (yych <= 0xDF) goto yy632;
			goto yy633;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy634;
			goto yy635;
		} else {
			if (yych <= 0xF3) goto yy636;
			if (yych <= 0xF4) goto yy637;
		}
	}
yy901:
#line 255 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 22747 "../../lnav/src/data_scanner_re.cc"
yy902:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 4) {
		goto yy902;
	}
	if (yych <= 0xDF) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '*') goto yy920;
			goto yy1069;
		} else {
			if (yych <= '>') goto yy1070;
			if (yych <= 0xC1) goto yy74;
			goto yy1071;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1072;
			if (yych <= 0xEF) goto yy1073;
			goto yy1074;
		} else {
			if (yych <= 0xF3) goto yy1075;
			if (yych <= 0xF4) goto yy1076;
			goto yy74;
		}
	}
yy903:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy902;
			goto yy775;
		} else {
			if (yych <= '-') goto yy1077;
			if (yych != '>') goto yy775;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy74;
		}
	}
yy904:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[512+yych] & 128) {
		goto yy755;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy461;
			if (yych <= '"') goto yy897;
			goto yy608;
		} else {
			if (yych <= '-') goto yy756;
			if (yych <= '\\') goto yy757;
			goto yy461;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy461;
		}
	}
yy905:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy775;
	goto yy74;
yy906:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy905;
	goto yy74;
yy907:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy905;
	goto yy74;
yy908:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy907;
	goto yy74;
yy909:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy907;
	goto yy74;
yy910:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy907;
	goto yy74;
yy911:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy902;
			goto yy776;
		} else {
			if (yych <= '-') goto yy1078;
			if (yych != '>') goto yy776;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy913;
			if (yych <= 0xE0) goto yy914;
			goto yy915;
		} else {
			if (yych <= 0xF0) goto yy916;
			if (yych <= 0xF3) goto yy917;
			if (yych <= 0xF4) goto yy918;
			goto yy74;
		}
	}
yy912:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[256+yych] & 1) {
		goto yy764;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy461;
			if (yych <= '\'') goto yy897;
			goto yy616;
		} else {
			if (yych <= '-') goto yy765;
			if (yych <= '\\') goto yy766;
			goto yy461;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy461;
		}
	}
yy913:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy776;
	goto yy74;
yy914:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy913;
	goto yy74;
yy915:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy913;
	goto yy74;
yy916:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy915;
	goto yy74;
yy917:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy915;
	goto yy74;
yy918:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy915;
	goto yy74;
yy919:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy775;
			goto yy774;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy776;
				goto yy774;
			} else {
				if (yych <= '-') goto yy919;
				if (yych <= '=') goto yy774;
				goto yy900;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy774;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy779;
			goto yy780;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy781;
				goto yy782;
			} else {
				if (yych <= 0xF3) goto yy783;
				if (yych <= 0xF4) goto yy784;
				goto yy74;
			}
		}
	}
yy920:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 8) {
		goto yy920;
	}
	if (yych <= 0xE0) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy74;
			goto yy1079;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1080;
			goto yy1081;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1082;
			goto yy1083;
		} else {
			if (yych <= 0xF3) goto yy1084;
			if (yych <= 0xF4) goto yy1085;
			goto yy74;
		}
	}
yy921:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[512+yych] & 2) {
		goto yy608;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy461;
			goto yy751;
		} else {
			if (yych <= '\\') goto yy609;
			if (yych <= 0xC1) goto yy461;
			goto yy610;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy611;
			if (yych <= 0xEF) goto yy612;
			goto yy613;
		} else {
			if (yych <= 0xF3) goto yy614;
			if (yych <= 0xF4) goto yy615;
			goto yy461;
		}
	}
yy922:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy785;
	goto yy74;
yy923:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy922;
	goto yy74;
yy924:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy922;
	goto yy74;
yy925:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy924;
	goto yy74;
yy926:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy924;
	goto yy74;
yy927:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy924;
	goto yy74;
yy928:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[512+yych] & 4) {
		goto yy616;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy461;
			goto yy751;
		} else {
			if (yych <= '\\') goto yy617;
			if (yych <= 0xC1) goto yy461;
			goto yy618;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy619;
			if (yych <= 0xEF) goto yy620;
			goto yy621;
		} else {
			if (yych <= 0xF3) goto yy622;
			if (yych <= 0xF4) goto yy623;
			goto yy461;
		}
	}
yy929:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy786;
	goto yy74;
yy930:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy929;
	goto yy74;
yy931:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy929;
	goto yy74;
yy932:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy931;
	goto yy74;
yy933:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy931;
	goto yy74;
yy934:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy931;
	goto yy74;
yy935:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy935;
				goto yy643;
			} else {
				if (yych <= '\r') goto yy935;
				if (yych == ' ') goto yy935;
				goto yy643;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy1086;
				if (yych == '-') goto yy795;
				goto yy643;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy796;
					goto yy795;
				} else {
					if (yych <= '=') goto yy643;
					if (yych <= '>') goto yy797;
					goto yy796;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy643;
				if (yych <= 'Z') goto yy795;
				if (yych <= '[') goto yy643;
				goto yy798;
			} else {
				if (yych == '_') goto yy795;
				if (yych <= '`') goto yy643;
				goto yy795;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy643;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy799;
				goto yy800;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy801;
					goto yy802;
				} else {
					if (yych <= 0xF3) goto yy803;
					if (yych <= 0xF4) goto yy804;
					goto yy74;
				}
			}
		}
	}
yy936:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy466;
			if (yych >= '"') goto yy938;
		} else {
			if (yych == '\\') goto yy939;
			if (yych >= 0x80) goto yy466;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy940;
			if (yych <= 0xE0) goto yy941;
			goto yy942;
		} else {
			if (yych <= 0xF0) goto yy943;
			if (yych <= 0xF3) goto yy944;
			if (yych <= 0xF4) goto yy945;
			goto yy466;
		}
	}
yy937:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '!') goto yy937;
		} else {
			if (yych == '\\') goto yy939;
			if (yych <= 0x7F) goto yy937;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy940;
			if (yych <= 0xE0) goto yy941;
			goto yy942;
		} else {
			if (yych <= 0xF0) goto yy943;
			if (yych <= 0xF3) goto yy944;
			if (yych <= 0xF4) goto yy945;
			goto yy74;
		}
	}
yy938:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy1087;
			} else {
				if (yych == '\r') goto yy1087;
				goto yy74;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy1087;
				goto yy74;
			} else {
				if (yych <= '-') goto yy462;
				if (yych <= '.') goto yy74;
				goto yy326;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy462;
				goto yy74;
			} else {
				if (yych <= '>') goto yy327;
				if (yych <= '?') goto yy326;
				goto yy74;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy462;
				if (yych <= '^') goto yy74;
				goto yy462;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy462;
				goto yy74;
			}
		}
	}
yy939:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy937;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy941;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy942;
			goto yy943;
		} else {
			if (yych <= 0xF3) goto yy944;
			if (yych <= 0xF4) goto yy945;
			goto yy74;
		}
	}
yy940:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy937;
	goto yy74;
yy941:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy940;
	goto yy74;
yy942:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy940;
	goto yy74;
yy943:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy942;
	goto yy74;
yy944:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy942;
	goto yy74;
yy945:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy942;
	goto yy74;
yy946:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy946;
				goto yy644;
			} else {
				if (yych <= '\r') goto yy946;
				if (yych == ' ') goto yy946;
				goto yy644;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy1086;
				if (yych == '-') goto yy805;
				goto yy644;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy806;
					goto yy805;
				} else {
					if (yych <= '=') goto yy644;
					if (yych <= '>') goto yy807;
					goto yy806;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy644;
				if (yych <= 'Z') goto yy805;
				if (yych <= '[') goto yy644;
				goto yy808;
			} else {
				if (yych == '_') goto yy805;
				if (yych <= '`') goto yy644;
				goto yy805;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy644;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy809;
				goto yy810;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy811;
					goto yy812;
				} else {
					if (yych <= 0xF3) goto yy813;
					if (yych <= 0xF4) goto yy814;
					goto yy74;
				}
			}
		}
	}
yy947:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy466;
			if (yych >= '\'') goto yy938;
		} else {
			if (yych == '\\') goto yy949;
			if (yych >= 0x80) goto yy466;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy950;
			if (yych <= 0xE0) goto yy951;
			goto yy952;
		} else {
			if (yych <= 0xF0) goto yy953;
			if (yych <= 0xF3) goto yy954;
			if (yych <= 0xF4) goto yy955;
			goto yy466;
		}
	}
yy948:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '&') goto yy948;
			goto yy938;
		} else {
			if (yych == '\\') goto yy949;
			if (yych <= 0x7F) goto yy948;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy950;
			if (yych <= 0xE0) goto yy951;
			goto yy952;
		} else {
			if (yych <= 0xF0) goto yy953;
			if (yych <= 0xF3) goto yy954;
			if (yych <= 0xF4) goto yy955;
			goto yy74;
		}
	}
yy949:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy948;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy951;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy952;
			goto yy953;
		} else {
			if (yych <= 0xF3) goto yy954;
			if (yych <= 0xF4) goto yy955;
			goto yy74;
		}
	}
yy950:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy948;
	goto yy74;
yy951:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy950;
	goto yy74;
yy952:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy950;
	goto yy74;
yy953:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy952;
	goto yy74;
yy954:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy952;
	goto yy74;
yy955:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy952;
	goto yy74;
yy956:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy816;
				goto yy956;
			} else {
				if (yych == '\r') goto yy956;
				if (yych <= 0x1F) goto yy816;
				goto yy956;
			}
		} else {
			if (yych <= '/') {
				if (yych == '-') goto yy956;
				if (yych <= '.') goto yy816;
			} else {
				if (yych <= ':') goto yy956;
				if (yych <= '<') goto yy816;
				if (yych <= '=') goto yy817;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '?') goto yy957;
				if (yych <= '@') goto yy816;
				if (yych <= 'Z') goto yy956;
				goto yy816;
			} else {
				if (yych == '`') goto yy816;
				if (yych <= 'z') goto yy956;
				goto yy816;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy820;
				if (yych <= 0xE0) goto yy821;
				goto yy822;
			} else {
				if (yych <= 0xF0) goto yy823;
				if (yych <= 0xF3) goto yy824;
				if (yych <= 0xF4) goto yy825;
				goto yy74;
			}
		}
	}
yy957:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy816;
				goto yy956;
			} else {
				if (yych <= '.') goto yy816;
				if (yych <= '/') goto yy957;
				goto yy956;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy816;
				if (yych <= '>') goto yy465;
				goto yy957;
			} else {
				if (yych <= '@') goto yy816;
				if (yych <= 'Z') goto yy956;
				goto yy816;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy816;
				goto yy956;
			} else {
				if (yych <= 0x7F) goto yy816;
				if (yych <= 0xC1) goto yy74;
				goto yy820;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy821;
				if (yych <= 0xEF) goto yy822;
				goto yy823;
			} else {
				if (yych <= 0xF3) goto yy824;
				if (yych <= 0xF4) goto yy825;
				goto yy74;
			}
		}
	}
yy958:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy958;
				goto yy818;
			} else {
				if (yych <= '\r') goto yy958;
				if (yych == ' ') goto yy958;
				goto yy818;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy816;
				if (yych == '-') goto yy958;
				goto yy818;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy958;
				} else {
					if (yych <= '<') goto yy818;
					if (yych <= '=') goto yy1088;
					goto yy960;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy959;
				if (yych <= '@') goto yy818;
				if (yych <= 'Z') goto yy958;
				goto yy818;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy961;
					goto yy818;
				} else {
					if (yych == '`') goto yy818;
					goto yy958;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy818;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy962;
				goto yy963;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy964;
					goto yy965;
				} else {
					if (yych <= 0xF3) goto yy966;
					if (yych <= 0xF4) goto yy967;
					goto yy74;
				}
			}
		}
	}
yy959:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '!') goto yy818;
				goto yy816;
			} else {
				if (yych == '-') goto yy958;
				if (yych <= '.') goto yy818;
				goto yy959;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy958;
				if (yych <= '=') goto yy818;
				goto yy1089;
			} else {
				if (yych <= '?') goto yy959;
				if (yych <= '@') goto yy818;
				if (yych <= 'Z') goto yy958;
				goto yy818;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy961;
				if (yych <= '^') goto yy818;
				goto yy958;
			} else {
				if (yych <= '`') goto yy818;
				if (yych <= 'z') goto yy958;
				if (yych <= 0x7F) goto yy818;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy962;
				if (yych <= 0xE0) goto yy963;
				goto yy964;
			} else {
				if (yych <= 0xF0) goto yy965;
				if (yych <= 0xF3) goto yy966;
				if (yych <= 0xF4) goto yy967;
				goto yy74;
			}
		}
	}
yy960:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '!') goto yy960;
			goto yy1090;
		} else {
			if (yych == '\\') goto yy1091;
			if (yych <= 0x7F) goto yy960;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1092;
			if (yych <= 0xE0) goto yy1093;
			goto yy1094;
		} else {
			if (yych <= 0xF0) goto yy1095;
			if (yych <= 0xF3) goto yy1096;
			if (yych <= 0xF4) goto yy1097;
			goto yy74;
		}
	}
yy961:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy818;
				goto yy958;
			} else {
				if (yych <= '.') goto yy818;
				if (yych <= '/') goto yy959;
				goto yy958;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy818;
				if (yych <= '>') goto yy960;
				goto yy959;
			} else {
				if (yych <= '@') goto yy818;
				if (yych <= 'Z') goto yy958;
				goto yy818;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy818;
				goto yy958;
			} else {
				if (yych <= 0x7F) goto yy818;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy963;
				if (yych <= 0xEF) goto yy964;
				goto yy965;
			} else {
				if (yych <= 0xF3) goto yy966;
				if (yych <= 0xF4) goto yy967;
				goto yy74;
			}
		}
	}
yy962:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy818;
	goto yy74;
yy963:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy962;
	goto yy74;
yy964:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy962;
	goto yy74;
yy965:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy964;
	goto yy74;
yy966:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy964;
	goto yy74;
yy967:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy964;
	goto yy74;
yy968:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy968;
				goto yy819;
			} else {
				if (yych <= '\r') goto yy968;
				if (yych == ' ') goto yy968;
				goto yy819;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy816;
				if (yych == '-') goto yy968;
				goto yy819;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy968;
				} else {
					if (yych <= '<') goto yy819;
					if (yych <= '=') goto yy1098;
					goto yy970;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy969;
				if (yych <= '@') goto yy819;
				if (yych <= 'Z') goto yy968;
				goto yy819;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy971;
					goto yy819;
				} else {
					if (yych == '`') goto yy819;
					goto yy968;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy819;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy972;
				goto yy973;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy974;
					goto yy975;
				} else {
					if (yych <= 0xF3) goto yy976;
					if (yych <= 0xF4) goto yy977;
					goto yy74;
				}
			}
		}
	}
yy969:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '&') goto yy819;
				goto yy816;
			} else {
				if (yych == '-') goto yy968;
				if (yych <= '.') goto yy819;
				goto yy969;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy968;
				if (yych <= '=') goto yy819;
				goto yy1099;
			} else {
				if (yych <= '?') goto yy969;
				if (yych <= '@') goto yy819;
				if (yych <= 'Z') goto yy968;
				goto yy819;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy971;
				if (yych <= '^') goto yy819;
				goto yy968;
			} else {
				if (yych <= '`') goto yy819;
				if (yych <= 'z') goto yy968;
				if (yych <= 0x7F) goto yy819;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy972;
				if (yych <= 0xE0) goto yy973;
				goto yy974;
			} else {
				if (yych <= 0xF0) goto yy975;
				if (yych <= 0xF3) goto yy976;
				if (yych <= 0xF4) goto yy977;
				goto yy74;
			}
		}
	}
yy970:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '&') goto yy970;
			goto yy1090;
		} else {
			if (yych == '\\') goto yy1100;
			if (yych <= 0x7F) goto yy970;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1101;
			if (yych <= 0xE0) goto yy1102;
			goto yy1103;
		} else {
			if (yych <= 0xF0) goto yy1104;
			if (yych <= 0xF3) goto yy1105;
			if (yych <= 0xF4) goto yy1106;
			goto yy74;
		}
	}
yy971:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy819;
				goto yy968;
			} else {
				if (yych <= '.') goto yy819;
				if (yych <= '/') goto yy969;
				goto yy968;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy819;
				if (yych <= '>') goto yy970;
				goto yy969;
			} else {
				if (yych <= '@') goto yy819;
				if (yych <= 'Z') goto yy968;
				goto yy819;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy819;
				goto yy968;
			} else {
				if (yych <= 0x7F) goto yy819;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy973;
				if (yych <= 0xEF) goto yy974;
				goto yy975;
			} else {
				if (yych <= 0xF3) goto yy976;
				if (yych <= 0xF4) goto yy977;
				goto yy74;
			}
		}
	}
yy972:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy819;
	goto yy74;
yy973:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy972;
	goto yy74;
yy974:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy972;
	goto yy74;
yy975:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy974;
	goto yy74;
yy976:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy974;
	goto yy74;
yy977:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy974;
	goto yy74;
yy978:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1107;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy978;
	goto yy74;
yy979:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy172;
		if (yych <= '9') goto yy1108;
		goto yy172;
	} else {
		if (yych <= 'F') goto yy1108;
		if (yych <= '`') goto yy172;
		if (yych <= 'f') goto yy1109;
		goto yy172;
	}
yy980:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy203;
		if (yych <= '9') goto yy1109;
		goto yy203;
	} else {
		if (yych <= 'F') goto yy1109;
		if (yych <= '`') goto yy203;
		if (yych <= 'f') goto yy1109;
		goto yy203;
	}
yy981:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1110;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1112;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1112;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy982:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1113;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1112;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1112;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy983:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '4') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1113;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy1114;
				if (yych <= '9') goto yy1110;
				yyt2 = YYCURSOR;
				goto yy1111;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1112;
				goto yy484;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'f') goto yy1112;
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy984:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1112;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1112;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1112;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy985:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy496;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1115;
			if (yych <= '1') goto yy1116;
			if (yych <= '2') goto yy1117;
			goto yy1115;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy74;
			if (yych <= 'Z') goto yy484;
			goto yy74;
		} else {
			if (yych == '`') goto yy74;
			if (yych <= 'z') goto yy484;
			goto yy74;
		}
	}
yy986:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy985;
	goto yy485;
yy987:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy985;
	if (yych <= '/') goto yy485;
	if (yych <= '5') goto yy986;
	goto yy485;
yy988:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy660;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy484;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy663;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy989:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy154;
	if (yych <= '9') goto yy1118;
	goto yy154;
yy990:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1119;
	goto yy74;
yy991:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy156;
				goto yy89;
			} else {
				if (yych == '+') goto yy89;
				goto yy156;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				yyt3 = YYCURSOR;
				goto yy351;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy1109;
				yyt3 = YYCURSOR;
				goto yy355;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy156;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy1109;
				if (yych <= 'Z') goto yy202;
				goto yy156;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy50;
				goto yy156;
			} else {
				if (yych <= 'f') goto yy1120;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy992:
	yych = *++YYCURSOR;
	if (yych == 'T') goto yy1121;
	goto yy154;
yy993:
	yych = *++YYCURSOR;
	if (yych == 'g') goto yy1122;
	goto yy74;
yy994:
	++YYCURSOR;
	yyt1 = yyt2;
yy995:
	YYCURSOR = yyt1;
#line 302 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 24576 "../../lnav/src/data_scanner_re.cc"
yy996:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	yyt1 = yyt2;
	if (yych <= '/') goto yy995;
	if (yych <= '9') goto yy1123;
	goto yy995;
yy997:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy994;
	goto yy74;
yy998:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy997;
	goto yy74;
yy999:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy997;
	goto yy74;
yy1000:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy999;
	goto yy74;
yy1001:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy999;
	goto yy74;
yy1002:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy999;
	goto yy74;
yy1003:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy1124;
	goto yy74;
yy1004:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy1125;
	goto yy115;
yy1005:
	++YYCURSOR;
yy1006:
	YYCURSOR = yyt1;
#line 400 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_IPV4_ADDRESS);
       }
#line 24631 "../../lnav/src/data_scanner_re.cc"
yy1007:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy1006;
		} else {
			if (yych == ',') goto yy1006;
			if (yych <= '.') goto yy90;
			goto yy1006;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy1006;
			if (yych <= 'Z') goto yy90;
			goto yy1006;
		} else {
			if (yych == '`') goto yy1006;
			if (yych <= 'z') goto yy90;
			goto yy1006;
		}
	}
yy1008:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy1006;
			goto yy89;
		} else {
			if (yych <= ',') goto yy1006;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy1006;
			goto yy534;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1006;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy534;
			goto yy1006;
		} else {
			if (yych == '`') goto yy1006;
			if (yych <= 'z') goto yy534;
			goto yy1006;
		}
	}
yy1009:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy1006;
			goto yy89;
		} else {
			if (yych <= ',') goto yy1006;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy1006;
			goto yy700;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1006;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy1006;
		} else {
			if (yych == '`') goto yy1006;
			if (yych <= 'z') goto yy138;
			goto yy1006;
		}
	}
yy1010:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1006;
		if (yych <= '9') goto yy279;
		if (yych <= '@') goto yy1006;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1006;
			goto yy279;
		} else {
			if (yych <= '`') goto yy1006;
			if (yych <= 'z') goto yy279;
			goto yy1006;
		}
	}
yy1011:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1008;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1009;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1010;
				}
				goto yy700;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1012;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1013;
				}
				yyt1 = YYCURSOR;
				goto yy1014;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				goto yy536;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1015;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1016;
				}
				yyt1 = YYCURSOR;
				goto yy1017;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1018;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1019;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1020;
				}
				goto yy536;
			}
		}
	}
yy1012:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy124;
	goto yy1006;
yy1013:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1006;
			goto yy216;
		} else {
			if (yych <= '/') goto yy1006;
			if (yych <= '9') goto yy292;
			goto yy1006;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy292;
			if (yych <= '^') goto yy1006;
			goto yy279;
		} else {
			if (yych <= '`') goto yy1006;
			if (yych <= 'z') goto yy292;
			goto yy1006;
		}
	}
yy1014:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy701;
			goto yy1006;
		} else {
			if (yych == ',') goto yy1006;
			goto yy701;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1006;
			if (yych <= 'Z') goto yy701;
			goto yy1006;
		} else {
			if (yych == '`') goto yy1006;
			if (yych <= 'z') goto yy701;
			goto yy1006;
		}
	}
yy1015:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1005;
	goto yy74;
yy1016:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1015;
	goto yy74;
yy1017:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1015;
	goto yy74;
yy1018:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1017;
	goto yy74;
yy1019:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1017;
	goto yy74;
yy1020:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1017;
	goto yy74;
yy1021:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				yyt1 = YYCURSOR;
				goto yy1005;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1007;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1008;
			}
		} else {
			if (yych <= '5') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1009;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1010;
				}
				goto yy1011;
			} else {
				if (yych <= '9') goto yy700;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1012;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1013;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				yyt1 = YYCURSOR;
				goto yy1014;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1014;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1005;
				}
				goto yy536;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1015;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1016;
				}
				yyt1 = YYCURSOR;
				goto yy1017;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1018;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1019;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1020;
				}
				goto yy536;
			}
		}
	}
yy1022:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy848;
	if (yych == 'b') goto yy848;
	goto yy535;
yy1023:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy848;
		goto yy535;
	} else {
		if (yych <= 'b') goto yy848;
		if (yych == 'n') goto yy848;
		goto yy535;
	}
yy1024:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy718;
	if (yych == ':') goto yy718;
	goto yy548;
yy1025:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1126;
	goto yy280;
yy1026:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1127;
	goto yy74;
yy1027:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych == ' ') goto yy1128;
			if (yych >= '%') goto yy139;
		} else {
			if (yych <= '+') {
				if (yych >= '+') goto yy139;
			} else {
				if (yych <= ',') goto yy1028;
				if (yych <= '/') goto yy139;
				goto yy1129;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy139;
			} else {
				if (yych == 'T') goto yy1130;
				goto yy139;
			}
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy139;
			} else {
				if (yych <= '`') goto yy1028;
				if (yych <= 'z') goto yy139;
			}
		}
	}
yy1028:
#line 316 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE);
       }
#line 25084 "../../lnav/src/data_scanner_re.cc"
yy1029:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= ',') {
			if (yych == ' ') goto yy1128;
			goto yy1028;
		} else {
			if (yych <= '/') goto yy280;
			if (yych <= '9') goto yy1131;
			if (yych <= ':') goto yy280;
			goto yy1028;
		}
	} else {
		if (yych <= '^') {
			if (yych == 'T') goto yy1132;
			if (yych <= 'Z') goto yy280;
			goto yy1028;
		} else {
			if (yych == '`') goto yy1028;
			if (yych <= 'z') goto yy280;
			goto yy1028;
		}
	}
yy1030:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy1133;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1134;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1135;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1031:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		if (yych <= ',') goto yy299;
		goto yy1133;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1135;
		goto yy299;
	}
yy1032:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1133;
	goto yy299;
yy1033:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy308;
		}
		if (yych <= '9') goto yy862;
		yyt2 = yyt1;
		goto yy308;
	} else {
		if (yych <= 'Z') goto yy862;
		if (yych <= '`') {
			yyt2 = yyt1;
			goto yy308;
		}
		if (yych <= 'z') goto yy862;
		yyt2 = yyt1;
		goto yy308;
	}
yy1034:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1136;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1136;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1136;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1035:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy74;
			yyt2 = yyt1;
			goto yy307;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy1137;
			goto yy124;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1137;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1137;
			goto yy74;
		}
	}
yy1036:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy1136;
				yyt1 = YYCURSOR;
				goto yy1035;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1136;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1136;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1037:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1138;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1035;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1136;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1136;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1038:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy1138;
				if (yych <= '9') goto yy1136;
				yyt1 = YYCURSOR;
				goto yy1035;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1136;
				goto yy279;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'f') goto yy1136;
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1039:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '/') {
			if (yych <= ',') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= '-') {
				yyt1 = YYCURSOR;
				goto yy1139;
			}
			if (yych <= '.') {
				yyt2 = YYCURSOR;
				goto yy996;
			}
			yyt2 = YYCURSOR;
			goto yy994;
		} else {
			if (yych <= ':') {
				if (yych <= '9') {
					yyt1 = YYCURSOR;
					goto yy1041;
				}
				goto yy1140;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy994;
				}
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy1041;
				}
				yyt2 = YYCURSOR;
				goto yy994;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy1041;
				}
				yyt2 = YYCURSOR;
				goto yy994;
			} else {
				if (yych <= 0xC1) goto yy548;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy997;
				}
				yyt2 = YYCURSOR;
				goto yy998;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy999;
				}
				yyt2 = YYCURSOR;
				goto yy1000;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1001;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1002;
				}
				goto yy548;
			}
		}
	}
yy1040:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy870;
			goto yy1140;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy870;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy870;
			goto yy548;
		}
	}
yy1041:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy995;
		if (yych <= '9') goto yy1042;
		if (yych <= ':') goto yy723;
		goto yy995;
	} else {
		if (yych <= 'F') goto yy1042;
		if (yych <= '`') goto yy995;
		if (yych >= 'g') goto yy995;
	}
yy1042:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy723;
	goto yy74;
yy1043:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1141;
		if (yych >= ';') goto yy74;
	} else {
		if (yych <= 'F') goto yy1141;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1141;
		goto yy74;
	}
yy1044:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1142;
		if (yych <= ':') goto yy1143;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1142;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1142;
		goto yy74;
	}
yy1045:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1144;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1046:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1147;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1047:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1147;
				goto yy1148;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1144;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1145;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1146;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1146;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1048:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1146;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1145;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1146;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1146;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1049:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1149;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy875;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1149;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1149;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1050:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1149;
				yyt4 = YYCURSOR;
				goto yy875;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1149;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1149;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1051:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1150;
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1151;
	goto yy280;
yy1052:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1150;
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1051;
	goto yy280;
yy1053:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy1150;
		goto yy280;
	} else {
		if (yych <= '4') goto yy1051;
		if (yych <= '5') goto yy1152;
		if (yych <= '9') goto yy1151;
		goto yy280;
	}
yy1054:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1153;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy884;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1153;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1153;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1055:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '-': goto yy1133;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1154;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1056:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy299;
			goto yy24;
		} else {
			if (yych <= '+') goto yy299;
			if (yych <= ',') goto yy24;
			if (yych <= '-') goto yy1133;
			goto yy299;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy299;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy299;
			goto yy24;
		}
	}
yy1057:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy304;
	if (yych == '-') goto yy1157;
	goto yy299;
yy1058:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy74;
		if (yych <= '0') goto yy1158;
		goto yy1159;
	} else {
		if (yych <= '2') goto yy1160;
		if (yych <= '9') goto yy1158;
		goto yy74;
	}
yy1059:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy1058;
	goto yy74;
yy1060:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy1058;
	if (yych <= '/') goto yy74;
	if (yych <= '5') goto yy1059;
	goto yy74;
yy1061:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy748;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1062:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1161;
				yyt2 = YYCURSOR;
				goto yy1162;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1063:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1163;
				yyt2 = YYCURSOR;
				goto yy748;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1064:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1163;
				goto yy1164;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1161;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy748;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy747;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy747;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1065:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1161;
				yyt2 = YYCURSOR;
				goto yy748;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy747;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1066:
	yych = *++YYCURSOR;
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy624;
				goto yy1066;
			} else {
				if (yych == '\r') goto yy1066;
				goto yy624;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1066;
				if (yych <= '!') goto yy624;
				goto yy627;
			} else {
				if (yych == '\'') goto yy628;
				if (yych <= ')') goto yy624;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '=') {
				if (yych == '-') goto yy752;
				goto yy624;
			} else {
				if (yych <= '>') goto yy631;
				if (yych <= 0x7F) goto yy624;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy632;
				if (yych <= 0xE0) goto yy633;
				goto yy634;
			} else {
				if (yych <= 0xF0) goto yy635;
				if (yych <= 0xF3) goto yy636;
				if (yych <= 0xF4) goto yy637;
				goto yy74;
			}
		}
	}
yy1067:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[512+yych] & 128) {
		goto yy755;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy901;
			if (yych <= '"') goto yy897;
			goto yy608;
		} else {
			if (yych <= '-') goto yy756;
			if (yych <= '\\') goto yy757;
			goto yy901;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy758;
			if (yych <= 0xE0) goto yy759;
			goto yy760;
		} else {
			if (yych <= 0xF0) goto yy761;
			if (yych <= 0xF3) goto yy762;
			if (yych <= 0xF4) goto yy763;
			goto yy901;
		}
	}
yy1068:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[256+yych] & 1) {
		goto yy764;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy901;
			if (yych <= '\'') goto yy897;
			goto yy616;
		} else {
			if (yych <= '-') goto yy765;
			if (yych <= '\\') goto yy766;
			goto yy901;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy767;
			if (yych <= 0xE0) goto yy768;
			goto yy769;
		} else {
			if (yych <= 0xF0) goto yy770;
			if (yych <= 0xF3) goto yy771;
			if (yych <= 0xF4) goto yy772;
			goto yy901;
		}
	}
yy1069:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ',') goto yy902;
			goto yy1165;
		} else {
			if (yych == '>') goto yy1070;
			if (yych <= 0x7F) goto yy902;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1071;
			if (yych <= 0xE0) goto yy1072;
			goto yy1073;
		} else {
			if (yych <= 0xF0) goto yy1074;
			if (yych <= 0xF3) goto yy1075;
			if (yych <= 0xF4) goto yy1076;
			goto yy74;
		}
	}
yy1070:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[0+yych] & 32) {
		goto yy1166;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy461;
			if (yych <= '"') goto yy1167;
			goto yy1168;
		} else {
			if (yych <= '*') goto yy1169;
			if (yych <= '-') goto yy1170;
			if (yych <= '\\') goto yy1171;
			goto yy461;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy461;
		}
	}
yy1071:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy902;
	goto yy74;
yy1072:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1071;
	goto yy74;
yy1073:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1071;
	goto yy74;
yy1074:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1073;
	goto yy74;
yy1075:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1073;
	goto yy74;
yy1076:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1073;
	goto yy74;
yy1077:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy902;
			goto yy775;
		} else {
			if (yych <= '-') goto yy1077;
			if (yych == '>') goto yy1067;
			goto yy775;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy74;
		}
	}
yy1078:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy902;
			goto yy776;
		} else {
			if (yych <= '-') goto yy1078;
			if (yych == '>') goto yy1068;
			goto yy776;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy913;
			if (yych <= 0xE0) goto yy914;
			goto yy915;
		} else {
			if (yych <= 0xF0) goto yy916;
			if (yych <= 0xF3) goto yy917;
			if (yych <= 0xF4) goto yy918;
			goto yy74;
		}
	}
yy1079:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[0+yych] & 64) {
		goto yy1169;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy461;
			if (yych <= '"') goto yy1178;
			goto yy1179;
		} else {
			if (yych <= '\\') goto yy1180;
			if (yych <= 0xC1) goto yy461;
			goto yy1181;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1182;
			if (yych <= 0xEF) goto yy1183;
			goto yy1184;
		} else {
			if (yych <= 0xF3) goto yy1185;
			if (yych <= 0xF4) goto yy1186;
			goto yy461;
		}
	}
yy1080:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy920;
	goto yy74;
yy1081:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1080;
	goto yy74;
yy1082:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1080;
	goto yy74;
yy1083:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1082;
	goto yy74;
yy1084:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1082;
	goto yy74;
yy1085:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1082;
	goto yy74;
yy1086:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 16) {
		goto yy1086;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy644;
			if (yych <= '\'') goto yy643;
			goto yy1187;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy1188;
				goto yy1187;
			} else {
				if (yych <= '>') goto yy1189;
				if (yych <= '?') goto yy1188;
				goto yy1187;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy1190;
				goto yy1187;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1191;
				goto yy1192;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy1193;
				goto yy1194;
			} else {
				if (yych <= 0xF3) goto yy1195;
				if (yych <= 0xF4) goto yy1196;
				goto yy74;
			}
		}
	}
yy1087:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy1087;
			goto yy74;
		} else {
			if (yych <= '\r') goto yy1087;
			if (yych <= 0x1F) goto yy74;
			goto yy1087;
		}
	} else {
		if (yych <= '=') {
			if (yych == '/') goto yy326;
			goto yy74;
		} else {
			if (yych <= '>') goto yy327;
			if (yych <= '?') goto yy326;
			goto yy74;
		}
	}
yy1088:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1088;
				goto yy818;
			} else {
				if (yych <= '\r') goto yy1088;
				if (yych == ' ') goto yy1088;
				goto yy818;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy1197;
				if (yych == '-') goto yy958;
				goto yy818;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy959;
					goto yy958;
				} else {
					if (yych <= '=') goto yy818;
					if (yych <= '>') goto yy960;
					goto yy959;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy818;
				if (yych <= 'Z') goto yy958;
				if (yych <= '[') goto yy818;
				goto yy961;
			} else {
				if (yych == '_') goto yy958;
				if (yych <= '`') goto yy818;
				goto yy958;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy818;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy962;
				goto yy963;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy964;
					goto yy965;
				} else {
					if (yych <= 0xF3) goto yy966;
					if (yych <= 0xF4) goto yy967;
					goto yy74;
				}
			}
		}
	}
yy1089:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy466;
			if (yych <= '!') goto yy960;
		} else {
			if (yych == '\\') goto yy1091;
			if (yych <= 0x7F) goto yy960;
			goto yy466;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1092;
			if (yych <= 0xE0) goto yy1093;
			goto yy1094;
		} else {
			if (yych <= 0xF0) goto yy1095;
			if (yych <= 0xF3) goto yy1096;
			if (yych <= 0xF4) goto yy1097;
			goto yy466;
		}
	}
yy1090:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy74;
				goto yy1198;
			} else {
				if (yych == '\r') goto yy1198;
				goto yy74;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy1198;
				goto yy74;
			} else {
				if (yych <= '-') goto yy651;
				if (yych <= '.') goto yy74;
				goto yy326;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy651;
				goto yy74;
			} else {
				if (yych <= '?') goto yy326;
				if (yych <= '@') goto yy74;
				goto yy651;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy74;
				goto yy651;
			} else {
				if (yych <= '`') goto yy74;
				if (yych <= 'z') goto yy651;
				goto yy74;
			}
		}
	}
yy1091:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy960;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy1093;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1094;
			goto yy1095;
		} else {
			if (yych <= 0xF3) goto yy1096;
			if (yych <= 0xF4) goto yy1097;
			goto yy74;
		}
	}
yy1092:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy960;
	goto yy74;
yy1093:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1092;
	goto yy74;
yy1094:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1092;
	goto yy74;
yy1095:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1094;
	goto yy74;
yy1096:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1094;
	goto yy74;
yy1097:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1094;
	goto yy74;
yy1098:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1098;
				goto yy819;
			} else {
				if (yych <= '\r') goto yy1098;
				if (yych == ' ') goto yy1098;
				goto yy819;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy1197;
				if (yych == '-') goto yy968;
				goto yy819;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy969;
					goto yy968;
				} else {
					if (yych <= '=') goto yy819;
					if (yych <= '>') goto yy970;
					goto yy969;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy819;
				if (yych <= 'Z') goto yy968;
				if (yych <= '[') goto yy819;
				goto yy971;
			} else {
				if (yych == '_') goto yy968;
				if (yych <= '`') goto yy819;
				goto yy968;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy819;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy972;
				goto yy973;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy974;
					goto yy975;
				} else {
					if (yych <= 0xF3) goto yy976;
					if (yych <= 0xF4) goto yy977;
					goto yy74;
				}
			}
		}
	}
yy1099:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy466;
			if (yych <= '&') goto yy970;
			goto yy1090;
		} else {
			if (yych == '\\') goto yy1100;
			if (yych <= 0x7F) goto yy970;
			goto yy466;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1101;
			if (yych <= 0xE0) goto yy1102;
			goto yy1103;
		} else {
			if (yych <= 0xF0) goto yy1104;
			if (yych <= 0xF3) goto yy1105;
			if (yych <= 0xF4) goto yy1106;
			goto yy466;
		}
	}
yy1100:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy970;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy1102;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1103;
			goto yy1104;
		} else {
			if (yych <= 0xF3) goto yy1105;
			if (yych <= 0xF4) goto yy1106;
			goto yy74;
		}
	}
yy1101:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy970;
	goto yy74;
yy1102:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1101;
	goto yy74;
yy1103:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1101;
	goto yy74;
yy1104:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1103;
	goto yy74;
yy1105:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1103;
	goto yy74;
yy1106:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1103;
	goto yy74;
yy1107:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy1199;
	goto yy74;
yy1108:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1200;
	goto yy172;
yy1109:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1133;
	goto yy203;
yy1110:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1201;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1201;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1201;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1111:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy74;
			goto yy307;
		} else {
			if (yych <= '/') goto yy74;
			if (yych <= '9') goto yy1137;
			goto yy359;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1137;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1137;
			goto yy74;
		}
	}
yy1112:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1201;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1201;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1201;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1113:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1202;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1201;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1201;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1114:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy1202;
				if (yych <= '9') goto yy1201;
				yyt2 = YYCURSOR;
				goto yy1111;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1201;
				goto yy484;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'f') goto yy1201;
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1115:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1203;
	if (yych <= '/') goto yy485;
	if (yych <= '9') goto yy1204;
	goto yy485;
yy1116:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1203;
	if (yych <= '/') goto yy485;
	if (yych <= '9') goto yy1115;
	goto yy485;
yy1117:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy1203;
		goto yy485;
	} else {
		if (yych <= '4') goto yy1115;
		if (yych <= '5') goto yy1205;
		if (yych <= '9') goto yy1204;
		goto yy485;
	}
yy1118:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy154;
	if (yych <= '9') goto yy1206;
	goto yy154;
yy1119:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1207;
	goto yy74;
yy1120:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy156;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy156;
				goto yy1133;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy351;
				}
				if (yych <= '/') goto yy124;
				goto yy202;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy355;
			}
			if (yych <= '?') goto yy156;
			if (yych <= '@') goto yy130;
			goto yy202;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy156;
				goto yy50;
			} else {
				if (yych <= '`') goto yy156;
				if (yych <= 'z') goto yy492;
				goto yy156;
			}
		}
	}
yy1121:
	yych = *++YYCURSOR;
	if (yych == 'A') goto yy1208;
	goto yy154;
yy1122:
	yych = *++YYCURSOR;
	if (yych == 'i') goto yy1209;
	goto yy74;
yy1123:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1210;
	goto yy74;
yy1124:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1211;
	goto yy74;
yy1125:
	yych = *++YYCURSOR;
	if (yych == 'F') goto yy1212;
	goto yy74;
yy1126:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1213;
	goto yy280;
yy1127:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1214;
	goto yy74;
yy1128:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1215;
	goto yy74;
yy1129:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '%') {
			if (yych == ' ') goto yy1128;
			if (yych <= '$') goto yy1028;
			goto yy139;
		} else {
			if (yych == '+') goto yy139;
			if (yych <= ',') goto yy1028;
			goto yy139;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') goto yy1028;
			if (yych != 'T') goto yy139;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1028;
				goto yy139;
			} else {
				if (yych <= '`') goto yy1028;
				if (yych <= 'z') goto yy139;
				goto yy1028;
			}
		}
	}
yy1130:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy139;
	if (yych <= '9') goto yy1216;
	goto yy139;
yy1131:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych == ' ') goto yy1128;
			goto yy1028;
		} else {
			if (yych <= ':') goto yy280;
			if (yych <= '?') goto yy1028;
			goto yy280;
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'T') goto yy1132;
			if (yych <= 'Z') goto yy280;
			goto yy1028;
		} else {
			if (yych == '`') goto yy1028;
			if (yych <= 'z') goto yy280;
			goto yy1028;
		}
	}
yy1132:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1217;
	goto yy280;
yy1133:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy1218;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy1218;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy1218;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy1134:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1219;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1220;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1135:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1220;
		goto yy299;
	}
yy1136:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '9') goto yy1221;
				yyt1 = YYCURSOR;
				goto yy1035;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1221;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1221;
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1137:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1222;
				yyt2 = YYCURSOR;
				goto yy1223;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1222;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1222;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1138:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy570;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1221;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1035;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1221;
				if (yych <= 'Z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1221;
			} else {
				if (yych <= 'z') goto yy279;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy134;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1139:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy995;
		if (yych <= '9') goto yy852;
		goto yy995;
	} else {
		if (yych <= 'F') goto yy852;
		if (yych <= '`') goto yy995;
		if (yych <= 'f') goto yy852;
		goto yy995;
	}
yy1140:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1224;
		if (yych <= ':') goto yy872;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1224;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1224;
		goto yy74;
	}
yy1141:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1225;
		if (yych <= ':') goto yy1044;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1225;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1225;
		goto yy74;
	}
yy1142:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1226;
		if (yych <= ':') goto yy1227;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1226;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1226;
		goto yy74;
	}
yy1143:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1228;
			} else {
				if (yych <= '1') goto yy1229;
				if (yych <= '2') goto yy1230;
				goto yy1228;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1231;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1231;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1144:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1232;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1145:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1231;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1231;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1231;
			goto yy74;
		}
	}
yy1146:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1232;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1145;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1232;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1232;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1147:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1233;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1148:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1233;
				goto yy1232;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1145;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1232;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1232;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1149:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy875;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1150:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy74;
			if (yych <= '0') goto yy1234;
			goto yy1235;
		} else {
			if (yych <= '2') goto yy1236;
			if (yych <= '9') goto yy1234;
			goto yy74;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy279;
			if (yych <= '^') goto yy74;
			goto yy279;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy279;
			goto yy74;
		}
	}
yy1151:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1150;
	goto yy280;
yy1152:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1150;
	if (yych <= '/') goto yy280;
	if (yych <= '5') goto yy1151;
	goto yy280;
yy1153:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy884;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1154:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1237;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1155:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy299;
			goto yy24;
		} else {
			if (yych == ',') goto yy24;
			goto yy299;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy299;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy299;
			goto yy24;
		}
	}
yy1156:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy304;
	if (yych == '-') goto yy443;
	goto yy299;
yy1157:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy1238;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy1218;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy1218;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy1158:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1239;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1159:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1158;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1160:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '4') goto yy1158;
			if (yych <= '5') goto yy1240;
			goto yy1239;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy312;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1161:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy891;
				yyt2 = YYCURSOR;
				goto yy748;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy891;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1162:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy307;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1241;
			if (yych <= '1') goto yy1242;
			goto yy1243;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy1244;
			if (yych <= '@') goto yy74;
			goto yy732;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy732;
			goto yy74;
		}
	}
yy1163:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1245;
				yyt2 = YYCURSOR;
				goto yy748;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy891;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1164:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1245;
				goto yy891;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy748;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy891;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy891;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1165:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ',') goto yy902;
			goto yy1165;
		} else {
			if (yych == '>') goto yy1246;
			if (yych <= 0x7F) goto yy902;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1071;
			if (yych <= 0xE0) goto yy1072;
			goto yy1073;
		} else {
			if (yych <= 0xF0) goto yy1074;
			if (yych <= 0xF3) goto yy1075;
			if (yych <= 0xF4) goto yy1076;
			goto yy74;
		}
	}
yy1166:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 32) {
		goto yy1166;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych >= '#') goto yy1168;
		} else {
			if (yych <= '*') goto yy1169;
			if (yych <= '-') goto yy1170;
			if (yych <= '\\') goto yy1171;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy74;
		}
	}
yy1167:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1247;
				goto yy764;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1247;
					goto yy764;
				} else {
					if (yych <= ' ') goto yy1247;
					if (yych <= '!') goto yy764;
					goto yy1248;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy1249;
				if (yych <= ')') goto yy764;
				goto yy616;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy764;
					goto yy1250;
				} else {
					if (yych <= '/') goto yy764;
					if (yych <= ':') goto yy1251;
					goto yy764;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy912;
				if (yych <= '@') goto yy764;
				if (yych <= 'Z') goto yy1251;
				goto yy764;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy766;
					goto yy764;
				} else {
					if (yych == '`') goto yy764;
					goto yy1251;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy764;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy767;
				goto yy768;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy769;
					goto yy770;
				} else {
					if (yych <= 0xF3) goto yy771;
					if (yych <= 0xF4) goto yy772;
					goto yy74;
				}
			}
		}
	}
yy1168:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1252;
				goto yy755;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1252;
					goto yy755;
				} else {
					if (yych <= ' ') goto yy1252;
					if (yych <= '!') goto yy755;
					goto yy1253;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy1254;
				if (yych <= ')') goto yy755;
				goto yy608;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy755;
					goto yy1255;
				} else {
					if (yych <= '/') goto yy755;
					if (yych <= ':') goto yy1256;
					goto yy755;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy904;
				if (yych <= '@') goto yy755;
				if (yych <= 'Z') goto yy1256;
				goto yy755;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy757;
					goto yy755;
				} else {
					if (yych == '`') goto yy755;
					goto yy1256;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy755;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy758;
				goto yy759;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy760;
					goto yy761;
				} else {
					if (yych <= 0xF3) goto yy762;
					if (yych <= 0xF4) goto yy763;
					goto yy74;
				}
			}
		}
	}
yy1169:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 64) {
		goto yy1169;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy1178;
			goto yy1179;
		} else {
			if (yych <= '\\') goto yy1180;
			if (yych <= 0xC1) goto yy74;
			goto yy1181;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1182;
			if (yych <= 0xEF) goto yy1183;
			goto yy1184;
		} else {
			if (yych <= 0xF3) goto yy1185;
			if (yych <= 0xF4) goto yy1186;
			goto yy74;
		}
	}
yy1170:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy1167;
			goto yy1166;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy1168;
				goto yy1166;
			} else {
				if (yych <= '-') goto yy1257;
				if (yych <= '[') goto yy1166;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy1166;
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1172;
			goto yy1173;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy1174;
				goto yy1175;
			} else {
				if (yych <= 0xF3) goto yy1176;
				if (yych <= 0xF4) goto yy1177;
				goto yy74;
			}
		}
	}
yy1171:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= ')') goto yy1166;
			goto yy1169;
		} else {
			if (yych == '-') goto yy1170;
			if (yych <= 0x7F) goto yy1166;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy74;
		}
	}
yy1172:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1166;
	goto yy74;
yy1173:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1172;
	goto yy74;
yy1174:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1172;
	goto yy74;
yy1175:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1174;
	goto yy74;
yy1176:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1174;
	goto yy74;
yy1177:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1174;
	goto yy74;
yy1178:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1258;
				goto yy616;
			} else {
				if (yych <= '\r') goto yy1258;
				if (yych == ' ') goto yy1258;
				goto yy616;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1259;
				if (yych == '\'') goto yy1260;
				goto yy616;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1261;
					goto yy616;
				} else {
					if (yych <= ':') goto yy1261;
					if (yych <= '=') goto yy616;
					goto yy928;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy616;
				if (yych <= 'Z') goto yy1261;
				if (yych <= '[') goto yy616;
				goto yy617;
			} else {
				if (yych == '_') goto yy1261;
				if (yych <= '`') goto yy616;
				goto yy1261;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy616;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy618;
				goto yy619;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy620;
					goto yy621;
				} else {
					if (yych <= 0xF3) goto yy622;
					if (yych <= 0xF4) goto yy623;
					goto yy74;
				}
			}
		}
	}
yy1179:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1262;
				goto yy608;
			} else {
				if (yych <= '\r') goto yy1262;
				if (yych == ' ') goto yy1262;
				goto yy608;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1263;
				if (yych == '\'') goto yy1264;
				goto yy608;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1265;
					goto yy608;
				} else {
					if (yych <= ':') goto yy1265;
					if (yych <= '=') goto yy608;
					goto yy921;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy608;
				if (yych <= 'Z') goto yy1265;
				if (yych <= '[') goto yy608;
				goto yy609;
			} else {
				if (yych == '_') goto yy1265;
				if (yych <= '`') goto yy608;
				goto yy1265;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy608;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy610;
				goto yy611;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy612;
					goto yy613;
				} else {
					if (yych <= 0xF3) goto yy614;
					if (yych <= 0xF4) goto yy615;
					goto yy74;
				}
			}
		}
	}
yy1180:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy1169;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy1182;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1183;
			goto yy1184;
		} else {
			if (yych <= 0xF3) goto yy1185;
			if (yych <= 0xF4) goto yy1186;
			goto yy74;
		}
	}
yy1181:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1169;
	goto yy74;
yy1182:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1181;
	goto yy74;
yy1183:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1181;
	goto yy74;
yy1184:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1183;
	goto yy74;
yy1185:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1183;
	goto yy74;
yy1186:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1183;
	goto yy74;
yy1187:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1187;
				goto yy1086;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1187;
					goto yy1086;
				} else {
					if (yych <= ' ') goto yy1187;
					if (yych <= '!') goto yy1086;
					goto yy644;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy643;
				if (yych <= ',') goto yy1086;
				goto yy1187;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy1086;
				} else {
					if (yych <= ':') goto yy1187;
					if (yych <= '<') goto yy1086;
					goto yy1266;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy1189;
				if (yych <= '?') goto yy1188;
				if (yych <= '@') goto yy1086;
				goto yy1187;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy1086;
					goto yy1190;
				} else {
					if (yych == '_') goto yy1187;
					goto yy1086;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy1187;
					goto yy1086;
				} else {
					if (yych <= 0xC1) goto yy74;
					if (yych <= 0xDF) goto yy1191;
					goto yy1192;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1193;
					goto yy1194;
				} else {
					if (yych <= 0xF3) goto yy1195;
					if (yych <= 0xF4) goto yy1196;
					goto yy74;
				}
			}
		}
	}
yy1188:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 16) {
		goto yy1086;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy644;
			if (yych <= '\'') goto yy643;
			goto yy1187;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy1188;
				goto yy1187;
			} else {
				if (yych <= '>') goto yy1267;
				if (yych <= '?') goto yy1188;
				goto yy1187;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy1190;
				goto yy1187;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1191;
				goto yy1192;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy1193;
				goto yy1194;
			} else {
				if (yych <= 0xF3) goto yy1195;
				if (yych <= 0xF4) goto yy1196;
				goto yy74;
			}
		}
	}
yy1189:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy328;
			if (yych == '"') goto yy1269;
			goto yy1268;
		} else {
			if (yych <= '\'') goto yy1270;
			if (yych == '\\') goto yy1271;
			goto yy1268;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy328;
			if (yych <= 0xDF) goto yy1272;
			if (yych <= 0xE0) goto yy1273;
			goto yy1274;
		} else {
			if (yych <= 0xF0) goto yy1275;
			if (yych <= 0xF3) goto yy1276;
			if (yych <= 0xF4) goto yy1277;
			goto yy328;
		}
	}
yy1190:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy1086;
				goto yy1187;
			} else {
				if (yych <= '.') goto yy1086;
				if (yych <= '/') goto yy1188;
				goto yy1187;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy1086;
				if (yych <= '>') goto yy1189;
				goto yy1188;
			} else {
				if (yych <= '@') goto yy1086;
				if (yych <= 'Z') goto yy1187;
				goto yy1086;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy1086;
				goto yy1187;
			} else {
				if (yych <= 0x7F) goto yy1086;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1192;
				if (yych <= 0xEF) goto yy1193;
				goto yy1194;
			} else {
				if (yych <= 0xF3) goto yy1195;
				if (yych <= 0xF4) goto yy1196;
				goto yy74;
			}
		}
	}
yy1191:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1086;
	goto yy74;
yy1192:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1191;
	goto yy74;
yy1193:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1191;
	goto yy74;
yy1194:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1193;
	goto yy74;
yy1195:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1193;
	goto yy74;
yy1196:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1193;
	goto yy74;
yy1197:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '!') goto yy1197;
				goto yy819;
			} else {
				if (yych == '\'') goto yy818;
				if (yych <= ',') goto yy1197;
				goto yy1278;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy1197;
				if (yych <= '/') goto yy1279;
				if (yych <= ':') goto yy1278;
				goto yy1197;
			} else {
				if (yych <= '>') goto yy1280;
				if (yych <= '?') goto yy1279;
				if (yych <= '@') goto yy1197;
				goto yy1278;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy1281;
				goto yy1197;
			} else {
				if (yych == '`') goto yy1197;
				if (yych <= 'z') goto yy1278;
				goto yy1197;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1282;
				if (yych <= 0xE0) goto yy1283;
				goto yy1284;
			} else {
				if (yych <= 0xF0) goto yy1285;
				if (yych <= 0xF3) goto yy1286;
				if (yych <= 0xF4) goto yy1287;
				goto yy74;
			}
		}
	}
yy1198:
	yych = *++YYCURSOR;
	if (yych <= 0x1F) {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy74;
			goto yy1198;
		} else {
			if (yych == '\r') goto yy1198;
			goto yy74;
		}
	} else {
		if (yych <= '/') {
			if (yych <= ' ') goto yy1198;
			if (yych <= '.') goto yy74;
			goto yy326;
		} else {
			if (yych == '?') goto yy326;
			goto yy74;
		}
	}
yy1199:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1288;
	goto yy74;
yy1200:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy154;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy154;
				goto yy337;
			} else {
				if (yych <= '.') goto yy89;
				if (yych <= '/') goto yy154;
				goto yy1289;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy154;
				goto yy91;
			} else {
				if (yych <= 'F') goto yy1289;
				if (yych <= 'Z') goto yy338;
				goto yy154;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy338;
				goto yy154;
			} else {
				if (yych <= 'f') goto yy1218;
				if (yych <= 'z') goto yy138;
				goto yy154;
			}
		}
	}
yy1201:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1290;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1290;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1290;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1202:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy661;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1290;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'F') goto yy1290;
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1290;
			} else {
				if (yych <= 'z') goto yy484;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy39;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				yyt2 = YYCURSOR;
				goto yy317;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy318;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1203:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy496;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1291;
			if (yych <= '1') goto yy1292;
			if (yych <= '2') goto yy1293;
			goto yy1291;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy74;
			if (yych <= 'Z') goto yy484;
			goto yy74;
		} else {
			if (yych == '`') goto yy74;
			if (yych <= 'z') goto yy484;
			goto yy74;
		}
	}
yy1204:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1203;
	goto yy485;
yy1205:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1203;
	if (yych <= '/') goto yy485;
	if (yych <= '5') goto yy1204;
	goto yy485;
yy1206:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1294;
	goto yy154;
yy1207:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1294;
	goto yy74;
yy1208:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy1295;
	goto yy154;
yy1209:
	yych = *++YYCURSOR;
	if (yych == 't') goto yy1296;
	goto yy74;
yy1210:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1297;
	goto yy74;
yy1211:
	yych = *++YYCURSOR;
	if (yych == '\n') goto yy74;
	goto yy1299;
yy1212:
	yych = *++YYCURSOR;
	if (yych == 'i') goto yy1306;
	goto yy74;
yy1213:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1131;
	goto yy280;
yy1214:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1307;
	goto yy74;
yy1215:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1308;
	goto yy74;
yy1216:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy139;
	if (yych <= '9') goto yy1309;
	goto yy139;
yy1217:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy280;
	if (yych <= '9') goto yy1310;
	goto yy280;
yy1218:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1311;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1311;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1311;
		goto yy139;
	}
yy1219:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1312;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1313;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1220:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1313;
		goto yy299;
	}
yy1221:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy279;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1035;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1222:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1314;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1314;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1314;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1223:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1315;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1315;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1315;
			goto yy74;
		}
	}
yy1224:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1316;
		if (yych <= ':') goto yy1044;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1316;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1316;
		goto yy74;
	}
yy1225:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1044;
	goto yy74;
yy1226:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1317;
		if (yych >= ';') goto yy74;
	} else {
		if (yych <= 'F') goto yy1317;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1317;
		goto yy74;
	}
yy1227:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1318;
		if (yych <= ':') goto yy1319;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1318;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1318;
		goto yy74;
	}
yy1228:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1320;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1229:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1323;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1230:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1323;
				goto yy1324;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1320;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1321;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1322;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1322;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1231:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1322;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1321;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1322;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1322;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1232:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1325;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1145;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1325;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1325;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1233:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1325;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1325;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1325;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1234:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy1326;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1327;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy134;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1235:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy1234;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1327;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy134;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1236:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '4') goto yy1234;
				if (yych <= '5') goto yy1328;
				goto yy1326;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1327;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1237:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1329;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1238:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy139;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1330;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy1311;
		case 'B':
		case 'b': goto yy1331;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy594;
		case 'S':
		case 's': goto yy595;
		case 'm': goto yy596;
		case 'p':
		case 'u': goto yy597;
		default: goto yy24;
	}
yy1239:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1240:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '5') goto yy1239;
			if (yych <= '9') goto yy74;
			yyt2 = YYCURSOR;
			goto yy312;
		} else {
			if (yych <= '@') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			goto yy74;
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1241:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1332;
				yyt2 = YYCURSOR;
				goto yy1333;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1242:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1334;
				yyt2 = YYCURSOR;
				goto yy884;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1243:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1334;
				goto yy1335;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1332;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy884;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy883;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy883;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1244:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1332;
				yyt2 = YYCURSOR;
				goto yy884;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1245:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1061;
				yyt2 = YYCURSOR;
				goto yy748;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1061;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1061;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1246:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[0+yych] & 32) {
		goto yy1166;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy901;
			if (yych <= '"') goto yy1167;
			goto yy1168;
		} else {
			if (yych <= '*') goto yy1169;
			if (yych <= '-') goto yy1170;
			if (yych <= '\\') goto yy1171;
			goto yy901;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy901;
		}
	}
yy1247:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy764;
				goto yy1247;
			} else {
				if (yych == '\r') goto yy1247;
				if (yych <= 0x1F) goto yy764;
				goto yy1247;
			}
		} else {
			if (yych <= '&') {
				if (yych != '"') goto yy764;
			} else {
				if (yych <= '\'') goto yy1249;
				if (yych == '*') goto yy616;
				goto yy764;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '>') {
				if (yych <= '-') goto yy765;
				if (yych <= '=') goto yy764;
				goto yy912;
			} else {
				if (yych == '\\') goto yy766;
				if (yych <= 0x7F) goto yy764;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy767;
				if (yych <= 0xE0) goto yy768;
				goto yy769;
			} else {
				if (yych <= 0xF0) goto yy770;
				if (yych <= 0xF3) goto yy771;
				if (yych <= 0xF4) goto yy772;
				goto yy74;
			}
		}
	}
yy1248:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 32) {
		goto yy1166;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy764;
			goto yy1168;
		} else {
			if (yych <= '*') goto yy1169;
			if (yych <= '-') goto yy1170;
			if (yych <= '\\') goto yy1171;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy74;
		}
	}
yy1249:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1247;
				goto yy764;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1247;
					goto yy764;
				} else {
					if (yych <= ' ') goto yy1247;
					if (yych <= '!') goto yy764;
					goto yy1248;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy628;
				if (yych <= ')') goto yy764;
				goto yy616;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy764;
				} else {
					if (yych <= '/') goto yy764;
					if (yych <= ':') goto yy1251;
					goto yy764;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy912;
				if (yych <= '@') goto yy764;
				if (yych <= 'Z') goto yy1251;
				goto yy764;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy766;
					goto yy764;
				} else {
					if (yych == '`') goto yy764;
					goto yy1251;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy764;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy767;
				goto yy768;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy769;
					goto yy770;
				} else {
					if (yych <= 0xF3) goto yy771;
					if (yych <= 0xF4) goto yy772;
					goto yy74;
				}
			}
		}
	}
yy1250:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1336;
				goto yy764;
			} else {
				if (yych <= '\r') goto yy1336;
				if (yych == ' ') goto yy1336;
				goto yy764;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1248;
				if (yych == '\'') goto yy1249;
				goto yy764;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1337;
					goto yy764;
				} else {
					if (yych <= ':') goto yy1251;
					if (yych <= '<') goto yy764;
					goto yy1338;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy912;
				if (yych <= '@') goto yy764;
				if (yych >= '[') goto yy764;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy766;
					goto yy764;
				} else {
					if (yych == '`') goto yy764;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy764;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy767;
				goto yy768;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy769;
					goto yy770;
				} else {
					if (yych <= 0xF3) goto yy771;
					if (yych <= 0xF4) goto yy772;
					goto yy74;
				}
			}
		}
	}
yy1251:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1336;
				goto yy764;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1336;
					goto yy764;
				} else {
					if (yych <= ' ') goto yy1336;
					if (yych <= '!') goto yy764;
					goto yy1248;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy1249;
				if (yych <= ')') goto yy764;
				goto yy616;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy764;
					goto yy1250;
				} else {
					if (yych <= '/') goto yy764;
					if (yych <= ':') goto yy1251;
					goto yy764;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy1338;
				if (yych <= '>') goto yy912;
				if (yych <= '@') goto yy764;
				goto yy1251;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy764;
					goto yy766;
				} else {
					if (yych == '_') goto yy1251;
					goto yy764;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy1251;
					goto yy764;
				} else {
					if (yych <= 0xC1) goto yy74;
					if (yych <= 0xDF) goto yy767;
					goto yy768;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy769;
					goto yy770;
				} else {
					if (yych <= 0xF3) goto yy771;
					if (yych <= 0xF4) goto yy772;
					goto yy74;
				}
			}
		}
	}
yy1252:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy755;
				goto yy1252;
			} else {
				if (yych == '\r') goto yy1252;
				if (yych <= 0x1F) goto yy755;
				goto yy1252;
			}
		} else {
			if (yych <= '&') {
				if (yych != '"') goto yy755;
			} else {
				if (yych <= '\'') goto yy1254;
				if (yych == '*') goto yy608;
				goto yy755;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '>') {
				if (yych <= '-') goto yy756;
				if (yych <= '=') goto yy755;
				goto yy904;
			} else {
				if (yych == '\\') goto yy757;
				if (yych <= 0x7F) goto yy755;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy758;
				if (yych <= 0xE0) goto yy759;
				goto yy760;
			} else {
				if (yych <= 0xF0) goto yy761;
				if (yych <= 0xF3) goto yy762;
				if (yych <= 0xF4) goto yy763;
				goto yy74;
			}
		}
	}
yy1253:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1252;
				goto yy755;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1252;
					goto yy755;
				} else {
					if (yych <= ' ') goto yy1252;
					if (yych <= '!') goto yy755;
					goto yy627;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy1254;
				if (yych <= ')') goto yy755;
				goto yy608;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy755;
					goto yy1255;
				} else {
					if (yych <= '/') goto yy755;
					if (yych <= ':') goto yy1256;
					goto yy755;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy904;
				if (yych <= '@') goto yy755;
				if (yych <= 'Z') goto yy1256;
				goto yy755;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy757;
					goto yy755;
				} else {
					if (yych == '`') goto yy755;
					goto yy1256;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy755;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy758;
				goto yy759;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy760;
					goto yy761;
				} else {
					if (yych <= 0xF3) goto yy762;
					if (yych <= 0xF4) goto yy763;
					goto yy74;
				}
			}
		}
	}
yy1254:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 32) {
		goto yy1166;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy1167;
			goto yy755;
		} else {
			if (yych <= '*') goto yy1169;
			if (yych <= '-') goto yy1170;
			if (yych <= '\\') goto yy1171;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1172;
			if (yych <= 0xE0) goto yy1173;
			goto yy1174;
		} else {
			if (yych <= 0xF0) goto yy1175;
			if (yych <= 0xF3) goto yy1176;
			if (yych <= 0xF4) goto yy1177;
			goto yy74;
		}
	}
yy1255:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1339;
				goto yy755;
			} else {
				if (yych <= '\r') goto yy1339;
				if (yych == ' ') goto yy1339;
				goto yy755;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1253;
				if (yych == '\'') goto yy1254;
				goto yy755;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1340;
					goto yy755;
				} else {
					if (yych <= ':') goto yy1256;
					if (yych <= '<') goto yy755;
					goto yy1341;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy904;
				if (yych <= '@') goto yy755;
				if (yych >= '[') goto yy755;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy757;
					goto yy755;
				} else {
					if (yych == '`') goto yy755;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy755;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy758;
				goto yy759;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy760;
					goto yy761;
				} else {
					if (yych <= 0xF3) goto yy762;
					if (yych <= 0xF4) goto yy763;
					goto yy74;
				}
			}
		}
	}
yy1256:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1339;
				goto yy755;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1339;
					goto yy755;
				} else {
					if (yych <= ' ') goto yy1339;
					if (yych <= '!') goto yy755;
					goto yy1253;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy1254;
				if (yych <= ')') goto yy755;
				goto yy608;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy755;
					goto yy1255;
				} else {
					if (yych <= '/') goto yy755;
					if (yych <= ':') goto yy1256;
					goto yy755;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy1341;
				if (yych <= '>') goto yy904;
				if (yych <= '@') goto yy755;
				goto yy1256;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy755;
					goto yy757;
				} else {
					if (yych == '_') goto yy1256;
					goto yy755;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy1256;
					goto yy755;
				} else {
					if (yych <= 0xC1) goto yy74;
					if (yych <= 0xDF) goto yy758;
					goto yy759;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy760;
					goto yy761;
				} else {
					if (yych <= 0xF3) goto yy762;
					if (yych <= 0xF4) goto yy763;
					goto yy74;
				}
			}
		}
	}
yy1257:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '\'') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy74;
				goto yy1166;
			} else {
				if (yych <= '"') goto yy1167;
				if (yych <= '&') goto yy1166;
				goto yy1168;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1166;
				goto yy1257;
			} else {
				if (yych == '>') goto yy1246;
				goto yy1166;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= '\\') goto yy1171;
				goto yy1166;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1172;
				goto yy1173;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy1174;
				goto yy1175;
			} else {
				if (yych <= 0xF3) goto yy1176;
				if (yych <= 0xF4) goto yy1177;
				goto yy74;
			}
		}
	}
yy1258:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy616;
				goto yy1258;
			} else {
				if (yych == '\r') goto yy1258;
				goto yy616;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1258;
				if (yych <= '!') goto yy616;
			} else {
				if (yych == '\'') goto yy1260;
				goto yy616;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy928;
				if (yych <= '[') goto yy616;
				goto yy617;
			} else {
				if (yych <= 0x7F) goto yy616;
				if (yych <= 0xC1) goto yy74;
				goto yy618;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy619;
				if (yych <= 0xEF) goto yy620;
				goto yy621;
			} else {
				if (yych <= 0xF3) goto yy622;
				if (yych <= 0xF4) goto yy623;
				goto yy74;
			}
		}
	}
yy1259:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 64) {
		goto yy1169;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy616;
			goto yy1179;
		} else {
			if (yych <= '\\') goto yy1180;
			if (yych <= 0xC1) goto yy74;
			goto yy1181;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1182;
			if (yych <= 0xEF) goto yy1183;
			goto yy1184;
		} else {
			if (yych <= 0xF3) goto yy1185;
			if (yych <= 0xF4) goto yy1186;
			goto yy74;
		}
	}
yy1260:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1258;
				goto yy616;
			} else {
				if (yych <= '\r') goto yy1258;
				if (yych == ' ') goto yy1258;
				goto yy616;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1259;
				if (yych == '\'') goto yy457;
				goto yy616;
			} else {
				if (yych <= '/') {
					if (yych >= '.') goto yy616;
				} else {
					if (yych <= ':') goto yy1261;
					if (yych <= '=') goto yy616;
					goto yy928;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy616;
				if (yych <= 'Z') goto yy1261;
				if (yych <= '[') goto yy616;
				goto yy617;
			} else {
				if (yych == '_') goto yy1261;
				if (yych <= '`') goto yy616;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy616;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy618;
				goto yy619;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy620;
					goto yy621;
				} else {
					if (yych <= 0xF3) goto yy622;
					if (yych <= 0xF4) goto yy623;
					goto yy74;
				}
			}
		}
	}
yy1261:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1342;
				goto yy616;
			} else {
				if (yych <= '\r') goto yy1342;
				if (yych == ' ') goto yy1342;
				goto yy616;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1259;
				if (yych == '\'') goto yy1260;
				goto yy616;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1261;
					goto yy616;
				} else {
					if (yych <= ':') goto yy1261;
					if (yych <= '<') goto yy616;
					goto yy1343;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy928;
				if (yych <= '@') goto yy616;
				if (yych <= 'Z') goto yy1261;
				goto yy616;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy617;
					goto yy616;
				} else {
					if (yych == '`') goto yy616;
					goto yy1261;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy616;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy618;
				goto yy619;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy620;
					goto yy621;
				} else {
					if (yych <= 0xF3) goto yy622;
					if (yych <= 0xF4) goto yy623;
					goto yy74;
				}
			}
		}
	}
yy1262:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy608;
				goto yy1262;
			} else {
				if (yych == '\r') goto yy1262;
				goto yy608;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1262;
				if (yych <= '!') goto yy608;
			} else {
				if (yych == '\'') goto yy1264;
				goto yy608;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy921;
				if (yych <= '[') goto yy608;
				goto yy609;
			} else {
				if (yych <= 0x7F) goto yy608;
				if (yych <= 0xC1) goto yy74;
				goto yy610;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy611;
				if (yych <= 0xEF) goto yy612;
				goto yy613;
			} else {
				if (yych <= 0xF3) goto yy614;
				if (yych <= 0xF4) goto yy615;
				goto yy74;
			}
		}
	}
yy1263:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1262;
				goto yy608;
			} else {
				if (yych <= '\r') goto yy1262;
				if (yych == ' ') goto yy1262;
				goto yy608;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy456;
				if (yych != '\'') goto yy608;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1265;
					goto yy608;
				} else {
					if (yych <= ':') goto yy1265;
					if (yych <= '=') goto yy608;
					goto yy921;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy608;
				if (yych <= 'Z') goto yy1265;
				if (yych <= '[') goto yy608;
				goto yy609;
			} else {
				if (yych == '_') goto yy1265;
				if (yych <= '`') goto yy608;
				goto yy1265;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy608;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy610;
				goto yy611;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy612;
					goto yy613;
				} else {
					if (yych <= 0xF3) goto yy614;
					if (yych <= 0xF4) goto yy615;
					goto yy74;
				}
			}
		}
	}
yy1264:
	yych = *++YYCURSOR;
	if (yybm_bol[0+yych] & 64) {
		goto yy1169;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy1178;
			goto yy608;
		} else {
			if (yych <= '\\') goto yy1180;
			if (yych <= 0xC1) goto yy74;
			goto yy1181;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1182;
			if (yych <= 0xEF) goto yy1183;
			goto yy1184;
		} else {
			if (yych <= 0xF3) goto yy1185;
			if (yych <= 0xF4) goto yy1186;
			goto yy74;
		}
	}
yy1265:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1344;
				goto yy608;
			} else {
				if (yych <= '\r') goto yy1344;
				if (yych == ' ') goto yy1344;
				goto yy608;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1263;
				if (yych == '\'') goto yy1264;
				goto yy608;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1265;
					goto yy608;
				} else {
					if (yych <= ':') goto yy1265;
					if (yych <= '<') goto yy608;
					goto yy1345;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy921;
				if (yych <= '@') goto yy608;
				if (yych <= 'Z') goto yy1265;
				goto yy608;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy609;
					goto yy608;
				} else {
					if (yych == '`') goto yy608;
					goto yy1265;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy608;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy610;
				goto yy611;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy612;
					goto yy613;
				} else {
					if (yych <= 0xF3) goto yy614;
					if (yych <= 0xF4) goto yy615;
					goto yy74;
				}
			}
		}
	}
yy1266:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1266;
				goto yy1086;
			} else {
				if (yych <= '\r') goto yy1266;
				if (yych == ' ') goto yy1266;
				goto yy1086;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy1187;
				if (yych <= '.') goto yy1086;
				if (yych <= '/') goto yy1188;
				goto yy1187;
			} else {
				if (yych <= '=') goto yy1086;
				if (yych <= '>') goto yy1189;
				if (yych <= '?') goto yy1188;
				goto yy1086;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy1187;
				if (yych == '\\') goto yy1190;
				goto yy1086;
			} else {
				if (yych == '`') goto yy1086;
				if (yych <= 'z') goto yy1187;
				goto yy1086;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1191;
				if (yych <= 0xE0) goto yy1192;
				goto yy1193;
			} else {
				if (yych <= 0xF0) goto yy1194;
				if (yych <= 0xF3) goto yy1195;
				if (yych <= 0xF4) goto yy1196;
				goto yy74;
			}
		}
	}
yy1267:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy466;
			if (yych == '"') goto yy1269;
		} else {
			if (yych <= '\'') goto yy1270;
			if (yych == '\\') goto yy1271;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy466;
			if (yych <= 0xDF) goto yy1272;
			if (yych <= 0xE0) goto yy1273;
			goto yy1274;
		} else {
			if (yych <= 0xF0) goto yy1275;
			if (yych <= 0xF3) goto yy1276;
			if (yych <= 0xF4) goto yy1277;
			goto yy466;
		}
	}
yy1268:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy74;
			if (yych != '"') goto yy1268;
		} else {
			if (yych <= '\'') goto yy1270;
			if (yych == '\\') goto yy1271;
			goto yy1268;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1272;
			if (yych <= 0xE0) goto yy1273;
			goto yy1274;
		} else {
			if (yych <= 0xF0) goto yy1275;
			if (yych <= 0xF3) goto yy1276;
			if (yych <= 0xF4) goto yy1277;
			goto yy74;
		}
	}
yy1269:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1346;
				goto yy948;
			} else {
				if (yych <= '\r') goto yy1346;
				if (yych == ' ') goto yy1346;
				goto yy948;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy938;
				if (yych == '-') goto yy1347;
				goto yy948;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy1348;
					goto yy1347;
				} else {
					if (yych <= '=') goto yy948;
					if (yych <= '>') goto yy807;
					goto yy1348;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy948;
				if (yych <= 'Z') goto yy1347;
				if (yych <= '[') goto yy948;
				goto yy949;
			} else {
				if (yych == '_') goto yy1347;
				if (yych <= '`') goto yy948;
				goto yy1347;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy948;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy950;
				goto yy951;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy952;
					goto yy953;
				} else {
					if (yych <= 0xF3) goto yy954;
					if (yych <= 0xF4) goto yy955;
					goto yy74;
				}
			}
		}
	}
yy1270:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1349;
				goto yy937;
			} else {
				if (yych <= '\r') goto yy1349;
				if (yych == ' ') goto yy1349;
				goto yy937;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy938;
				if (yych == '-') goto yy1350;
				goto yy937;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy1351;
					goto yy1350;
				} else {
					if (yych <= '=') goto yy937;
					if (yych <= '>') goto yy797;
					goto yy1351;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy937;
				if (yych <= 'Z') goto yy1350;
				if (yych <= '[') goto yy937;
				goto yy939;
			} else {
				if (yych == '_') goto yy1350;
				if (yych <= '`') goto yy937;
				goto yy1350;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy937;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy940;
				goto yy941;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy942;
					goto yy943;
				} else {
					if (yych <= 0xF3) goto yy944;
					if (yych <= 0xF4) goto yy945;
					goto yy74;
				}
			}
		}
	}
yy1271:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy1268;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy1273;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1274;
			goto yy1275;
		} else {
			if (yych <= 0xF3) goto yy1276;
			if (yych <= 0xF4) goto yy1277;
			goto yy74;
		}
	}
yy1272:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1268;
	goto yy74;
yy1273:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1272;
	goto yy74;
yy1274:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1272;
	goto yy74;
yy1275:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1274;
	goto yy74;
yy1276:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1274;
	goto yy74;
yy1277:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1274;
	goto yy74;
yy1278:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1278;
				goto yy1197;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy1278;
					goto yy1197;
				} else {
					if (yych <= ' ') goto yy1278;
					if (yych <= '!') goto yy1197;
					goto yy819;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy818;
				if (yych <= ',') goto yy1197;
				goto yy1278;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy1197;
				} else {
					if (yych <= ':') goto yy1278;
					if (yych <= '<') goto yy1197;
					goto yy1352;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy1280;
				if (yych <= '?') goto yy1279;
				if (yych <= '@') goto yy1197;
				goto yy1278;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy1197;
					goto yy1281;
				} else {
					if (yych == '_') goto yy1278;
					goto yy1197;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy1278;
					goto yy1197;
				} else {
					if (yych <= 0xC1) goto yy74;
					if (yych <= 0xDF) goto yy1282;
					goto yy1283;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1284;
					goto yy1285;
				} else {
					if (yych <= 0xF3) goto yy1286;
					if (yych <= 0xF4) goto yy1287;
					goto yy74;
				}
			}
		}
	}
yy1279:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '!') goto yy1197;
				goto yy819;
			} else {
				if (yych == '\'') goto yy818;
				if (yych <= ',') goto yy1197;
				goto yy1278;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy1197;
				if (yych <= '/') goto yy1279;
				if (yych <= ':') goto yy1278;
				goto yy1197;
			} else {
				if (yych <= '>') goto yy1353;
				if (yych <= '?') goto yy1279;
				if (yych <= '@') goto yy1197;
				goto yy1278;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy1281;
				goto yy1197;
			} else {
				if (yych == '`') goto yy1197;
				if (yych <= 'z') goto yy1278;
				goto yy1197;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1282;
				if (yych <= 0xE0) goto yy1283;
				goto yy1284;
			} else {
				if (yych <= 0xF0) goto yy1285;
				if (yych <= 0xF3) goto yy1286;
				if (yych <= 0xF4) goto yy1287;
				goto yy74;
			}
		}
	}
yy1280:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy1354;
			goto yy1280;
		} else {
			if (yych <= '\'') goto yy1355;
			if (yych == '\\') goto yy1356;
			goto yy1280;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1357;
			if (yych <= 0xE0) goto yy1358;
			goto yy1359;
		} else {
			if (yych <= 0xF0) goto yy1360;
			if (yych <= 0xF3) goto yy1361;
			if (yych <= 0xF4) goto yy1362;
			goto yy74;
		}
	}
yy1281:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy74;
				if (yych <= ',') goto yy1197;
				goto yy1278;
			} else {
				if (yych <= '.') goto yy1197;
				if (yych <= '/') goto yy1279;
				goto yy1278;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy1197;
				if (yych <= '>') goto yy1280;
				goto yy1279;
			} else {
				if (yych <= '@') goto yy1197;
				if (yych <= 'Z') goto yy1278;
				goto yy1197;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy1197;
				goto yy1278;
			} else {
				if (yych <= 0x7F) goto yy1197;
				if (yych <= 0xC1) goto yy74;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1283;
				if (yych <= 0xEF) goto yy1284;
				goto yy1285;
			} else {
				if (yych <= 0xF3) goto yy1286;
				if (yych <= 0xF4) goto yy1287;
				goto yy74;
			}
		}
	}
yy1282:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1197;
	goto yy74;
yy1283:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1282;
	goto yy74;
yy1284:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1282;
	goto yy74;
yy1285:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1284;
	goto yy74;
yy1286:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1284;
	goto yy74;
yy1287:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1284;
	goto yy74;
yy1288:
	yych = *++YYCURSOR;
	if (yych == ',') goto yy1363;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1288;
	goto yy74;
yy1289:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1364;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1364;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1311;
		goto yy339;
	}
yy1290:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1033;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy484;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1111;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1291:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy660;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1365;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1366;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1292:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy660;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1291;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1366;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1293:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy660;
				}
				yyt2 = YYCURSOR;
				goto yy667;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= '4') goto yy1291;
				goto yy1367;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1365;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1366;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1294:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1368;
	goto yy74;
yy1295:
	yych = *++YYCURSOR;
	if (yych == 'L') goto yy1369;
	goto yy154;
yy1296:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1370;
	goto yy74;
yy1297:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy1371;
			}
			goto yy74;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= 0xC1) goto yy74;
			yyt2 = YYCURSOR;
			goto yy997;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy998;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy999;
			}
			yyt2 = YYCURSOR;
			goto yy1000;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1001;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1002;
			}
			goto yy74;
		}
	}
yy1298:
	yych = *++YYCURSOR;
yy1299:
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\t') goto yy1298;
			goto yy1372;
		} else {
			if (yych <= 0x7F) goto yy1298;
			if (yych <= 0xC1) goto yy74;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1301;
			if (yych <= 0xEF) goto yy1302;
			goto yy1303;
		} else {
			if (yych <= 0xF3) goto yy1304;
			if (yych <= 0xF4) goto yy1305;
			goto yy74;
		}
	}
yy1300:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1298;
	goto yy74;
yy1301:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1300;
	goto yy74;
yy1302:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1300;
	goto yy74;
yy1303:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1302;
	goto yy74;
yy1304:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1302;
	goto yy74;
yy1305:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1302;
	goto yy74;
yy1306:
	yych = *++YYCURSOR;
	if (yych == 'l') goto yy1373;
	goto yy74;
yy1307:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1374;
	goto yy74;
yy1308:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1375;
	goto yy74;
yy1309:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1376;
	goto yy139;
yy1310:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1376;
	goto yy280;
yy1311:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1377;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1377;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1377;
		goto yy139;
	}
yy1312:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1378;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1379;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1313:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1379;
		goto yy299;
	}
yy1314:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1380;
				yyt2 = YYCURSOR;
				goto yy1223;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1380;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1380;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1315:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1381;
				yyt2 = YYCURSOR;
				goto yy1382;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1381;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1381;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1316:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy1141;
			goto yy1383;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy1141;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy1141;
			goto yy548;
		}
	}
yy1317:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1384;
		if (yych <= ':') goto yy1227;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1384;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1384;
		goto yy74;
	}
yy1318:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1385;
		if (yych <= ':') goto yy1386;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1385;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1385;
		goto yy74;
	}
yy1319:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1387;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1387;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1387;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1320:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1388;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1321:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1387;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1387;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1387;
			goto yy74;
		}
	}
yy1322:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1388;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1321;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1388;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1388;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1323:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1389;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1324:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1389;
				goto yy1388;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1321;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1388;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1388;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1325:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1145;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1326:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '9') goto yy279;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy279;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy434;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy279;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy134;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1327:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy307;
	if (yych == ':') goto yy124;
	goto yy74;
yy1328:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= '5') goto yy1326;
				if (yych <= '9') goto yy279;
				yyt2 = YYCURSOR;
				goto yy1327;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy279;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy434;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy279;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy134;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy134;
			}
		}
	}
yy1329:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1390;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1330:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy139;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1391;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy1377;
		case 'B':
		case 'b': goto yy1392;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy594;
		case 'S':
		case 's': goto yy595;
		case 'm': goto yy596;
		case 'p':
		case 'u': goto yy597;
		default: goto yy24;
	}
yy1331:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy139;
			if (yych <= '*') goto yy24;
			goto yy139;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy139;
			} else {
				if (yych <= '9') goto yy1377;
				if (yych <= ':') goto yy139;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy139;
			if (yych <= 'F') goto yy1377;
			if (yych <= 'Z') goto yy139;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy139;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy1377;
				if (yych <= 'z') goto yy139;
				goto yy24;
			}
		}
	}
yy1332:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1054;
				yyt2 = YYCURSOR;
				goto yy884;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1054;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1054;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1333:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy307;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1393;
			if (yych <= '1') goto yy1394;
			goto yy1395;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy1396;
			if (yych <= '@') goto yy74;
			goto yy1048;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1048;
			goto yy74;
		}
	}
yy1334:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1397;
				yyt2 = YYCURSOR;
				goto yy884;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1054;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1054;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1335:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1397;
				goto yy1054;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy884;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1054;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1054;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1336:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy764;
				goto yy1336;
			} else {
				if (yych == '\r') goto yy1336;
				if (yych <= 0x1F) goto yy764;
				goto yy1336;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy1248;
				goto yy764;
			} else {
				if (yych <= '\'') goto yy1249;
				if (yych == '*') goto yy616;
				goto yy764;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '=') {
				if (yych <= '-') goto yy765;
				if (yych <= '<') goto yy764;
				goto yy1338;
			} else {
				if (yych <= '>') goto yy912;
				if (yych == '\\') goto yy766;
				goto yy764;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy767;
				if (yych <= 0xE0) goto yy768;
				goto yy769;
			} else {
				if (yych <= 0xF0) goto yy770;
				if (yych <= 0xF3) goto yy771;
				if (yych <= 0xF4) goto yy772;
				goto yy74;
			}
		}
	}
yy1337:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1336;
				goto yy764;
			} else {
				if (yych <= '\r') goto yy1336;
				if (yych == ' ') goto yy1336;
				goto yy764;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1248;
				if (yych == '\'') goto yy1249;
				goto yy764;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1337;
					goto yy764;
				} else {
					if (yych <= ':') goto yy1251;
					if (yych <= '<') goto yy764;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy1068;
				if (yych <= '@') goto yy764;
				if (yych <= 'Z') goto yy1251;
				goto yy764;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy766;
					goto yy764;
				} else {
					if (yych == '`') goto yy764;
					goto yy1251;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy764;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy767;
				goto yy768;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy769;
					goto yy770;
				} else {
					if (yych <= 0xF3) goto yy771;
					if (yych <= 0xF4) goto yy772;
					goto yy74;
				}
			}
		}
	}
yy1338:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 16) {
		goto yy776;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '"') goto yy902;
			goto yy786;
		} else {
			if (yych <= '-') goto yy911;
			if (yych <= '>') goto yy764;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy913;
			if (yych <= 0xE0) goto yy914;
			goto yy915;
		} else {
			if (yych <= 0xF0) goto yy916;
			if (yych <= 0xF3) goto yy917;
			if (yych <= 0xF4) goto yy918;
			goto yy74;
		}
	}
yy1339:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy755;
				goto yy1339;
			} else {
				if (yych == '\r') goto yy1339;
				if (yych <= 0x1F) goto yy755;
				goto yy1339;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy1253;
				goto yy755;
			} else {
				if (yych <= '\'') goto yy1254;
				if (yych == '*') goto yy608;
				goto yy755;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '=') {
				if (yych <= '-') goto yy756;
				if (yych <= '<') goto yy755;
				goto yy1341;
			} else {
				if (yych <= '>') goto yy904;
				if (yych == '\\') goto yy757;
				goto yy755;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy758;
				if (yych <= 0xE0) goto yy759;
				goto yy760;
			} else {
				if (yych <= 0xF0) goto yy761;
				if (yych <= 0xF3) goto yy762;
				if (yych <= 0xF4) goto yy763;
				goto yy74;
			}
		}
	}
yy1340:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1339;
				goto yy755;
			} else {
				if (yych <= '\r') goto yy1339;
				if (yych == ' ') goto yy1339;
				goto yy755;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy1253;
				if (yych == '\'') goto yy1254;
				goto yy755;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy1340;
					goto yy755;
				} else {
					if (yych <= ':') goto yy1256;
					if (yych <= '<') goto yy755;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy1067;
				if (yych <= '@') goto yy755;
				if (yych <= 'Z') goto yy1256;
				goto yy755;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy757;
					goto yy755;
				} else {
					if (yych == '`') goto yy755;
					goto yy1256;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy755;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy758;
				goto yy759;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy760;
					goto yy761;
				} else {
					if (yych <= 0xF3) goto yy762;
					if (yych <= 0xF4) goto yy763;
					goto yy74;
				}
			}
		}
	}
yy1341:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 8) {
		goto yy775;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\'') goto yy902;
			goto yy785;
		} else {
			if (yych <= '-') goto yy903;
			if (yych <= '>') goto yy755;
			goto yy74;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy905;
			if (yych <= 0xE0) goto yy906;
			goto yy907;
		} else {
			if (yych <= 0xF0) goto yy908;
			if (yych <= 0xF3) goto yy909;
			if (yych <= 0xF4) goto yy910;
			goto yy74;
		}
	}
yy1342:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy616;
				goto yy1342;
			} else {
				if (yych == '\r') goto yy1342;
				goto yy616;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1342;
				if (yych <= '!') goto yy616;
				goto yy1259;
			} else {
				if (yych == '\'') goto yy1260;
				goto yy616;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy1343;
				if (yych <= '>') goto yy928;
				goto yy616;
			} else {
				if (yych <= '\\') goto yy617;
				if (yych <= 0x7F) goto yy616;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy618;
				if (yych <= 0xE0) goto yy619;
				goto yy620;
			} else {
				if (yych <= 0xF0) goto yy621;
				if (yych <= 0xF3) goto yy622;
				if (yych <= 0xF4) goto yy623;
				goto yy74;
			}
		}
	}
yy1343:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 128) {
		goto yy786;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy74;
			goto yy920;
		} else {
			if (yych <= '>') goto yy616;
			if (yych <= 0xC1) goto yy74;
			goto yy929;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy930;
			if (yych <= 0xEF) goto yy931;
			goto yy932;
		} else {
			if (yych <= 0xF3) goto yy933;
			if (yych <= 0xF4) goto yy934;
			goto yy74;
		}
	}
yy1344:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy608;
				goto yy1344;
			} else {
				if (yych == '\r') goto yy1344;
				goto yy608;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1344;
				if (yych <= '!') goto yy608;
				goto yy1263;
			} else {
				if (yych == '\'') goto yy1264;
				goto yy608;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy1345;
				if (yych <= '>') goto yy921;
				goto yy608;
			} else {
				if (yych <= '\\') goto yy609;
				if (yych <= 0x7F) goto yy608;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy610;
				if (yych <= 0xE0) goto yy611;
				goto yy612;
			} else {
				if (yych <= 0xF0) goto yy613;
				if (yych <= 0xF3) goto yy614;
				if (yych <= 0xF4) goto yy615;
				goto yy74;
			}
		}
	}
yy1345:
	yych = *++YYCURSOR;
	if (yybm_bol[256+yych] & 64) {
		goto yy785;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy74;
			goto yy920;
		} else {
			if (yych <= '>') goto yy608;
			if (yych <= 0xC1) goto yy74;
			goto yy922;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy923;
			if (yych <= 0xEF) goto yy924;
			goto yy925;
		} else {
			if (yych <= 0xF3) goto yy926;
			if (yych <= 0xF4) goto yy927;
			goto yy74;
		}
	}
yy1346:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy948;
				goto yy1346;
			} else {
				if (yych == '\r') goto yy1346;
				goto yy948;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy1346;
				if (yych <= '&') goto yy948;
				goto yy938;
			} else {
				if (yych == '/') goto yy1348;
				goto yy948;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy807;
				if (yych <= '?') goto yy1348;
				goto yy948;
			} else {
				if (yych <= '\\') goto yy949;
				if (yych <= 0x7F) goto yy948;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy950;
				if (yych <= 0xE0) goto yy951;
				goto yy952;
			} else {
				if (yych <= 0xF0) goto yy953;
				if (yych <= 0xF3) goto yy954;
				if (yych <= 0xF4) goto yy955;
				goto yy74;
			}
		}
	}
yy1347:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1398;
				goto yy948;
			} else {
				if (yych <= '\r') goto yy1398;
				if (yych == ' ') goto yy1398;
				goto yy948;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy938;
				if (yych == '-') goto yy1347;
				goto yy948;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy1347;
				} else {
					if (yych <= '<') goto yy948;
					if (yych <= '=') goto yy1399;
					goto yy807;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy1348;
				if (yych <= '@') goto yy948;
				if (yych <= 'Z') goto yy1347;
				goto yy948;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy949;
					goto yy948;
				} else {
					if (yych == '`') goto yy948;
					goto yy1347;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy948;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy950;
				goto yy951;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy952;
					goto yy953;
				} else {
					if (yych <= 0xF3) goto yy954;
					if (yych <= 0xF4) goto yy955;
					goto yy74;
				}
			}
		}
	}
yy1348:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy938;
			goto yy948;
		} else {
			if (yych <= '>') goto yy947;
			if (yych == '\\') goto yy949;
			goto yy948;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy950;
			if (yych <= 0xE0) goto yy951;
			goto yy952;
		} else {
			if (yych <= 0xF0) goto yy953;
			if (yych <= 0xF3) goto yy954;
			if (yych <= 0xF4) goto yy955;
			goto yy74;
		}
	}
yy1349:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy937;
				goto yy1349;
			} else {
				if (yych == '\r') goto yy1349;
				goto yy937;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1349;
				if (yych <= '!') goto yy937;
				goto yy938;
			} else {
				if (yych == '/') goto yy1351;
				goto yy937;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy797;
				if (yych <= '?') goto yy1351;
				goto yy937;
			} else {
				if (yych <= '\\') goto yy939;
				if (yych <= 0x7F) goto yy937;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy940;
				if (yych <= 0xE0) goto yy941;
				goto yy942;
			} else {
				if (yych <= 0xF0) goto yy943;
				if (yych <= 0xF3) goto yy944;
				if (yych <= 0xF4) goto yy945;
				goto yy74;
			}
		}
	}
yy1350:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1400;
				goto yy937;
			} else {
				if (yych <= '\r') goto yy1400;
				if (yych == ' ') goto yy1400;
				goto yy937;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy938;
				if (yych == '-') goto yy1350;
				goto yy937;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy1350;
				} else {
					if (yych <= '<') goto yy937;
					if (yych <= '=') goto yy1401;
					goto yy797;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy1351;
				if (yych <= '@') goto yy937;
				if (yych <= 'Z') goto yy1350;
				goto yy937;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy939;
					goto yy937;
				} else {
					if (yych == '`') goto yy937;
					goto yy1350;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy937;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy940;
				goto yy941;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy942;
					goto yy943;
				} else {
					if (yych <= 0xF3) goto yy944;
					if (yych <= 0xF4) goto yy945;
					goto yy74;
				}
			}
		}
	}
yy1351:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy938;
			goto yy937;
		} else {
			if (yych <= '>') goto yy936;
			if (yych == '\\') goto yy939;
			goto yy937;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy940;
			if (yych <= 0xE0) goto yy941;
			goto yy942;
		} else {
			if (yych <= 0xF0) goto yy943;
			if (yych <= 0xF3) goto yy944;
			if (yych <= 0xF4) goto yy945;
			goto yy74;
		}
	}
yy1352:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1352;
				goto yy1197;
			} else {
				if (yych <= '\r') goto yy1352;
				if (yych == ' ') goto yy1352;
				goto yy1197;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy1278;
				if (yych <= '.') goto yy1197;
				if (yych <= '/') goto yy1279;
				goto yy1278;
			} else {
				if (yych <= '=') goto yy1197;
				if (yych <= '>') goto yy1280;
				if (yych <= '?') goto yy1279;
				goto yy1197;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy1278;
				if (yych == '\\') goto yy1281;
				goto yy1197;
			} else {
				if (yych == '`') goto yy1197;
				if (yych <= 'z') goto yy1278;
				goto yy1197;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1282;
				if (yych <= 0xE0) goto yy1283;
				goto yy1284;
			} else {
				if (yych <= 0xF0) goto yy1285;
				if (yych <= 0xF3) goto yy1286;
				if (yych <= 0xF4) goto yy1287;
				goto yy74;
			}
		}
	}
yy1353:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy466;
			if (yych != '"') goto yy1280;
		} else {
			if (yych <= '\'') goto yy1355;
			if (yych == '\\') goto yy1356;
			goto yy1280;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy466;
			if (yych <= 0xDF) goto yy1357;
			if (yych <= 0xE0) goto yy1358;
			goto yy1359;
		} else {
			if (yych <= 0xF0) goto yy1360;
			if (yych <= 0xF3) goto yy1361;
			if (yych <= 0xF4) goto yy1362;
			goto yy466;
		}
	}
yy1354:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1402;
				goto yy970;
			} else {
				if (yych <= '\r') goto yy1402;
				if (yych == ' ') goto yy1402;
				goto yy970;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy1090;
				if (yych == '-') goto yy1403;
				goto yy970;
			} else {
				if (yych <= '/') goto yy1404;
				if (yych <= ':') goto yy1403;
				if (yych <= '>') goto yy970;
				goto yy1404;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy970;
				if (yych <= 'Z') goto yy1403;
				if (yych <= '[') goto yy970;
				goto yy1100;
			} else {
				if (yych == '_') goto yy1403;
				if (yych <= '`') goto yy970;
				goto yy1403;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy970;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1101;
				goto yy1102;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1103;
					goto yy1104;
				} else {
					if (yych <= 0xF3) goto yy1105;
					if (yych <= 0xF4) goto yy1106;
					goto yy74;
				}
			}
		}
	}
yy1355:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1405;
				goto yy960;
			} else {
				if (yych <= '\r') goto yy1405;
				if (yych == ' ') goto yy1405;
				goto yy960;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy1090;
				if (yych == '-') goto yy1406;
				goto yy960;
			} else {
				if (yych <= '/') goto yy1407;
				if (yych <= ':') goto yy1406;
				if (yych <= '>') goto yy960;
				goto yy1407;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy960;
				if (yych <= 'Z') goto yy1406;
				if (yych <= '[') goto yy960;
				goto yy1091;
			} else {
				if (yych == '_') goto yy1406;
				if (yych <= '`') goto yy960;
				goto yy1406;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy960;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1092;
				goto yy1093;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1094;
					goto yy1095;
				} else {
					if (yych <= 0xF3) goto yy1096;
					if (yych <= 0xF4) goto yy1097;
					goto yy74;
				}
			}
		}
	}
yy1356:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy74;
			goto yy1280;
		} else {
			if (yych <= 0xC1) goto yy74;
			if (yych >= 0xE0) goto yy1358;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1359;
			goto yy1360;
		} else {
			if (yych <= 0xF3) goto yy1361;
			if (yych <= 0xF4) goto yy1362;
			goto yy74;
		}
	}
yy1357:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1280;
	goto yy74;
yy1358:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1357;
	goto yy74;
yy1359:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1357;
	goto yy74;
yy1360:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1359;
	goto yy74;
yy1361:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1359;
	goto yy74;
yy1362:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1359;
	goto yy74;
yy1363:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1408;
	goto yy74;
yy1364:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1409;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1409;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1377;
		goto yy339;
	}
yy1365:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy660;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy484;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1366;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy569;
			} else {
				if (yych <= 'Z') goto yy484;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy486;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1366:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy307;
	if (yych == ':') goto yy359;
	goto yy74;
yy1367:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy660;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy667;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				goto yy1365;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy484;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1366;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy569;
				}
				if (yych <= 'Z') goto yy484;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy486;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy484;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy39;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy39;
			}
		}
	}
yy1368:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1410;
	goto yy74;
yy1369:
	yych = *++YYCURSOR;
	if (yych == 'E') goto yy1412;
	goto yy154;
yy1370:
	yych = *++YYCURSOR;
	if (yych == '\n') goto yy74;
	goto yy1414;
yy1371:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy1421;
			}
			goto yy995;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= 0xC1) goto yy995;
			yyt2 = YYCURSOR;
			goto yy997;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy998;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy999;
			}
			yyt2 = YYCURSOR;
			goto yy1000;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1001;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1002;
			}
			goto yy995;
		}
	}
yy1372:
	++YYCURSOR;
#line 366 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_DIFF_FILE_HEADER);
           cap_inner.c_end -= 1;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 36392 "../../lnav/src/data_scanner_re.cc"
yy1373:
	yych = *++YYCURSOR;
	if (yych == 'e') goto yy1422;
	goto yy74;
yy1374:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1423;
	goto yy74;
yy1375:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1424;
	goto yy74;
yy1376:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1424;
	if (yych <= ':') goto yy124;
	goto yy74;
yy1377:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1425;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1425;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1425;
		goto yy139;
	}
yy1378:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1426;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1427;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1379:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1427;
		goto yy299;
	}
yy1380:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy1033;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			goto yy74;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1223;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1381:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1428;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1428;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1428;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1382:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1429;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1429;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1429;
			goto yy74;
		}
	}
yy1383:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1430;
		if (yych <= ':') goto yy1143;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1430;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1430;
		goto yy74;
	}
yy1384:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1227;
	goto yy74;
yy1385:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1431;
		if (yych >= ';') goto yy74;
	} else {
		if (yych <= 'F') goto yy1431;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1431;
		goto yy74;
	}
yy1386:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1432;
		if (yych <= ':') goto yy1433;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1432;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1432;
		goto yy74;
	}
yy1387:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1434;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1435;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1434;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1434;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1388:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1436;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1321;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1436;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1436;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1389:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt4 = YYCURSOR;
				goto yy873;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1436;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1436;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1436;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1390:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1437;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1391:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy139;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1438;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy1425;
		case 'B':
		case 'b': goto yy1439;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy594;
		case 'S':
		case 's': goto yy595;
		case 'm': goto yy596;
		case 'p':
		case 'u': goto yy597;
		default: goto yy24;
	}
yy1392:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy139;
			if (yych <= '*') goto yy24;
			goto yy139;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy24;
				goto yy139;
			} else {
				if (yych <= '9') goto yy1425;
				if (yych <= ':') goto yy139;
				goto yy24;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy139;
			if (yych <= 'F') goto yy1425;
			if (yych <= 'Z') goto yy139;
			goto yy24;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy139;
				goto yy24;
			} else {
				if (yych <= 'f') goto yy1425;
				if (yych <= 'z') goto yy139;
				goto yy24;
			}
		}
	}
yy1393:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1440;
				yyt2 = YYCURSOR;
				goto yy1441;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1394:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1442;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1395:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1442;
				goto yy1443;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1440;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1145;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1146;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1146;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1396:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1440;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1146;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1397:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1153;
				yyt2 = YYCURSOR;
				goto yy884;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1153;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1153;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1398:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy948;
				goto yy1398;
			} else {
				if (yych == '\r') goto yy1398;
				goto yy948;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy1398;
				if (yych <= '&') goto yy948;
				goto yy938;
			} else {
				if (yych == '/') goto yy1348;
				if (yych <= '<') goto yy948;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy807;
				if (yych <= '?') goto yy1348;
				goto yy948;
			} else {
				if (yych <= '\\') goto yy949;
				if (yych <= 0x7F) goto yy948;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy950;
				if (yych <= 0xE0) goto yy951;
				goto yy952;
			} else {
				if (yych <= 0xF0) goto yy953;
				if (yych <= 0xF3) goto yy954;
				if (yych <= 0xF4) goto yy955;
				goto yy74;
			}
		}
	}
yy1399:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy644;
			} else {
				if (yych <= '\t') goto yy946;
				if (yych <= '\f') goto yy644;
				goto yy946;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy946;
				goto yy644;
			} else {
				if (yych <= '"') goto yy1086;
				if (yych <= '=') goto yy644;
				goto yy948;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy644;
				goto yy808;
			} else {
				if (yych <= 0x7F) goto yy644;
				if (yych <= 0xC1) goto yy74;
				goto yy809;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy810;
				if (yych <= 0xEF) goto yy811;
				goto yy812;
			} else {
				if (yych <= 0xF3) goto yy813;
				if (yych <= 0xF4) goto yy814;
				goto yy74;
			}
		}
	}
yy1400:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy937;
				goto yy1400;
			} else {
				if (yych == '\r') goto yy1400;
				goto yy937;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1400;
				if (yych <= '!') goto yy937;
				goto yy938;
			} else {
				if (yych == '/') goto yy1351;
				if (yych <= '<') goto yy937;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy797;
				if (yych <= '?') goto yy1351;
				goto yy937;
			} else {
				if (yych <= '\\') goto yy939;
				if (yych <= 0x7F) goto yy937;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy940;
				if (yych <= 0xE0) goto yy941;
				goto yy942;
			} else {
				if (yych <= 0xF0) goto yy943;
				if (yych <= 0xF3) goto yy944;
				if (yych <= 0xF4) goto yy945;
				goto yy74;
			}
		}
	}
yy1401:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy643;
			} else {
				if (yych <= '\t') goto yy935;
				if (yych <= '\f') goto yy643;
				goto yy935;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy935;
				goto yy643;
			} else {
				if (yych <= '\'') goto yy1086;
				if (yych <= '=') goto yy643;
				goto yy937;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy643;
				goto yy798;
			} else {
				if (yych <= 0x7F) goto yy643;
				if (yych <= 0xC1) goto yy74;
				goto yy799;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy800;
				if (yych <= 0xEF) goto yy801;
				goto yy802;
			} else {
				if (yych <= 0xF3) goto yy803;
				if (yych <= 0xF4) goto yy804;
				goto yy74;
			}
		}
	}
yy1402:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy970;
				goto yy1402;
			} else {
				if (yych == '\r') goto yy1402;
				goto yy970;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy1402;
				if (yych <= '&') goto yy970;
				goto yy1090;
			} else {
				if (yych == '/') goto yy1404;
				goto yy970;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy1404;
				if (yych <= '[') goto yy970;
				goto yy1100;
			} else {
				if (yych <= 0x7F) goto yy970;
				if (yych <= 0xC1) goto yy74;
				goto yy1101;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1102;
				if (yych <= 0xEF) goto yy1103;
				goto yy1104;
			} else {
				if (yych <= 0xF3) goto yy1105;
				if (yych <= 0xF4) goto yy1106;
				goto yy74;
			}
		}
	}
yy1403:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1444;
				goto yy970;
			} else {
				if (yych <= '\r') goto yy1444;
				if (yych == ' ') goto yy1444;
				goto yy970;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy1090;
				if (yych == '-') goto yy1403;
				goto yy970;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy1403;
				} else {
					if (yych == '=') goto yy1445;
					goto yy970;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy1404;
				if (yych <= '@') goto yy970;
				if (yych <= 'Z') goto yy1403;
				goto yy970;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy1100;
					goto yy970;
				} else {
					if (yych == '`') goto yy970;
					goto yy1403;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy970;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1101;
				goto yy1102;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1103;
					goto yy1104;
				} else {
					if (yych <= 0xF3) goto yy1105;
					if (yych <= 0xF4) goto yy1106;
					goto yy74;
				}
			}
		}
	}
yy1404:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy74;
			if (yych == '\'') goto yy1090;
			goto yy970;
		} else {
			if (yych <= '>') goto yy1099;
			if (yych == '\\') goto yy1100;
			goto yy970;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1101;
			if (yych <= 0xE0) goto yy1102;
			goto yy1103;
		} else {
			if (yych <= 0xF0) goto yy1104;
			if (yych <= 0xF3) goto yy1105;
			if (yych <= 0xF4) goto yy1106;
			goto yy74;
		}
	}
yy1405:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy960;
				goto yy1405;
			} else {
				if (yych == '\r') goto yy1405;
				goto yy960;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1405;
				if (yych <= '!') goto yy960;
				goto yy1090;
			} else {
				if (yych == '/') goto yy1407;
				goto yy960;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy1407;
				if (yych <= '[') goto yy960;
				goto yy1091;
			} else {
				if (yych <= 0x7F) goto yy960;
				if (yych <= 0xC1) goto yy74;
				goto yy1092;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1093;
				if (yych <= 0xEF) goto yy1094;
				goto yy1095;
			} else {
				if (yych <= 0xF3) goto yy1096;
				if (yych <= 0xF4) goto yy1097;
				goto yy74;
			}
		}
	}
yy1406:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy74;
				if (yych == '\t') goto yy1446;
				goto yy960;
			} else {
				if (yych <= '\r') goto yy1446;
				if (yych == ' ') goto yy1446;
				goto yy960;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy1090;
				if (yych == '-') goto yy1406;
				goto yy960;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy1406;
				} else {
					if (yych == '=') goto yy1447;
					goto yy960;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy1407;
				if (yych <= '@') goto yy960;
				if (yych <= 'Z') goto yy1406;
				goto yy960;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy1091;
					goto yy960;
				} else {
					if (yych == '`') goto yy960;
					goto yy1406;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy960;
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) goto yy1092;
				goto yy1093;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1094;
					goto yy1095;
				} else {
					if (yych <= 0xF3) goto yy1096;
					if (yych <= 0xF4) goto yy1097;
					goto yy74;
				}
			}
		}
	}
yy1407:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy74;
			if (yych == '"') goto yy1090;
			goto yy960;
		} else {
			if (yych <= '>') goto yy1089;
			if (yych == '\\') goto yy1091;
			goto yy960;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) goto yy1092;
			if (yych <= 0xE0) goto yy1093;
			goto yy1094;
		} else {
			if (yych <= 0xF0) goto yy1095;
			if (yych <= 0xF3) goto yy1096;
			if (yych <= 0xF4) goto yy1097;
			goto yy74;
		}
	}
yy1408:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1448;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1408;
	goto yy74;
yy1409:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1449;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1449;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1425;
		goto yy339;
	}
yy1410:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1450;
yy1411:
#line 313 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 37890 "../../lnav/src/data_scanner_re.cc"
yy1412:
	yych = *++YYCURSOR;
	if (yych == '\t') goto yy1451;
	if (yych == ' ') goto yy1453;
	goto yy154;
yy1413:
	yych = *++YYCURSOR;
yy1414:
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\t') goto yy1413;
			goto yy1454;
		} else {
			if (yych <= 0x7F) goto yy1413;
			if (yych <= 0xC1) goto yy74;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1416;
			if (yych <= 0xEF) goto yy1417;
			goto yy1418;
		} else {
			if (yych <= 0xF3) goto yy1419;
			if (yych <= 0xF4) goto yy1420;
			goto yy74;
		}
	}
yy1415:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1413;
	goto yy74;
yy1416:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1415;
	goto yy74;
yy1417:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1415;
	goto yy74;
yy1418:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1417;
	goto yy74;
yy1419:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1417;
	goto yy74;
yy1420:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1417;
	goto yy74;
yy1421:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy1455;
			}
			goto yy995;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= 0xC1) goto yy995;
			yyt2 = YYCURSOR;
			goto yy997;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy998;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy999;
			}
			yyt2 = YYCURSOR;
			goto yy1000;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1001;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1002;
			}
			goto yy995;
		}
	}
yy1422:
	yych = *++YYCURSOR;
	if (yych == 's') goto yy1456;
	goto yy74;
yy1423:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1457;
	goto yy74;
yy1424:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1458;
	goto yy74;
yy1425:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1460;
	goto yy139;
yy1426:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1461;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1462;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1427:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1462;
		goto yy299;
	}
yy1428:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1463;
				yyt2 = YYCURSOR;
				goto yy1382;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1463;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1463;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1429:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1464;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1464;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1464;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1430:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1465;
		if (yych <= ':') goto yy1227;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1465;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1465;
		goto yy74;
	}
yy1431:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1466;
		if (yych <= ':') goto yy1386;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1466;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1466;
		goto yy74;
	}
yy1432:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1467;
		if (yych <= ':') goto yy1468;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1467;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1467;
		goto yy74;
	}
yy1433:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1469;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1469;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1469;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1434:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1470;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1470;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1470;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1435:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy1469;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy1469;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1469;
			goto yy74;
		}
	}
yy1436:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1321;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1437:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1471;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1438:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1460;
	goto yy588;
yy1439:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy139;
			goto yy24;
		} else {
			if (yych <= '+') goto yy139;
			if (yych <= ',') goto yy24;
			if (yych <= '-') goto yy1460;
			goto yy139;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy24;
			if (yych <= 'Z') goto yy139;
			goto yy24;
		} else {
			if (yych == '`') goto yy24;
			if (yych <= 'z') goto yy139;
			goto yy24;
		}
	}
yy1440:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1232;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1441:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy307;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1472;
			if (yych <= '1') goto yy1473;
			goto yy1474;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy1475;
			if (yych <= '@') goto yy74;
			goto yy1231;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1231;
			goto yy74;
		}
	}
yy1442:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1476;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1232;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1443:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1476;
				goto yy1232;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1145;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1232;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1232;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1444:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy970;
				goto yy1444;
			} else {
				if (yych == '\r') goto yy1444;
				goto yy970;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy1444;
				if (yych <= '&') goto yy970;
				goto yy1090;
			} else {
				if (yych == '/') goto yy1404;
				if (yych <= '<') goto yy970;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy1404;
				goto yy970;
			} else {
				if (yych <= '\\') goto yy1100;
				if (yych <= 0x7F) goto yy970;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy1101;
				if (yych <= 0xE0) goto yy1102;
				goto yy1103;
			} else {
				if (yych <= 0xF0) goto yy1104;
				if (yych <= 0xF3) goto yy1105;
				if (yych <= 0xF4) goto yy1106;
				goto yy74;
			}
		}
	}
yy1445:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy819;
			} else {
				if (yych <= '\t') goto yy1098;
				if (yych <= '\f') goto yy819;
				goto yy1098;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy1098;
				goto yy819;
			} else {
				if (yych <= '"') goto yy1197;
				if (yych <= '=') goto yy819;
				goto yy970;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy819;
				goto yy971;
			} else {
				if (yych <= 0x7F) goto yy819;
				if (yych <= 0xC1) goto yy74;
				goto yy972;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy973;
				if (yych <= 0xEF) goto yy974;
				goto yy975;
			} else {
				if (yych <= 0xF3) goto yy976;
				if (yych <= 0xF4) goto yy977;
				goto yy74;
			}
		}
	}
yy1446:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy960;
				goto yy1446;
			} else {
				if (yych == '\r') goto yy1446;
				goto yy960;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy1446;
				if (yych <= '!') goto yy960;
				goto yy1090;
			} else {
				if (yych == '/') goto yy1407;
				if (yych <= '<') goto yy960;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy1407;
				goto yy960;
			} else {
				if (yych <= '\\') goto yy1091;
				if (yych <= 0x7F) goto yy960;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy1092;
				if (yych <= 0xE0) goto yy1093;
				goto yy1094;
			} else {
				if (yych <= 0xF0) goto yy1095;
				if (yych <= 0xF3) goto yy1096;
				if (yych <= 0xF4) goto yy1097;
				goto yy74;
			}
		}
	}
yy1447:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy818;
			} else {
				if (yych <= '\t') goto yy1088;
				if (yych <= '\f') goto yy818;
				goto yy1088;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy1088;
				goto yy818;
			} else {
				if (yych <= '\'') goto yy1197;
				if (yych <= '=') goto yy818;
				goto yy960;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy818;
				goto yy961;
			} else {
				if (yych <= 0x7F) goto yy818;
				if (yych <= 0xC1) goto yy74;
				goto yy962;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy963;
				if (yych <= 0xEF) goto yy964;
				goto yy965;
			} else {
				if (yych <= 0xF3) goto yy966;
				if (yych <= 0xF4) goto yy967;
				goto yy74;
			}
		}
	}
yy1448:
	yych = *++YYCURSOR;
	if (yych == '@') goto yy1477;
	goto yy74;
yy1449:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1478;
	goto yy339;
yy1450:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1479;
	goto yy74;
yy1451:
	yych = *++YYCURSOR;
yy1452:
	if (yybm_bol[0+yych] & 128) {
		goto yy1451;
	}
	if (yych <= '^') {
		if (yych <= '@') goto yy74;
		if (yych <= 'Z') {
			yyt1 = YYCURSOR;
			goto yy1480;
		}
		goto yy74;
	} else {
		if (yych == '`') goto yy74;
		if (yych <= 'z') {
			yyt1 = YYCURSOR;
			goto yy1480;
		}
		goto yy74;
	}
yy1453:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1452;
		} else {
			if (yych <= ' ') goto yy1481;
			if (yych == '-') goto yy153;
			goto yy1452;
		}
	} else {
		if (yych <= 'I') {
			if (yych <= '9') goto yy153;
			if (yych <= '@') goto yy1452;
			if (yych <= 'H') {
				yyt1 = YYCURSOR;
				goto yy1482;
			}
			yyt1 = YYCURSOR;
			goto yy1483;
		} else {
			if (yych <= 'Z') {
				yyt1 = YYCURSOR;
				goto yy1482;
			}
			if (yych == '_') {
				yyt1 = YYCURSOR;
				goto yy1482;
			}
			goto yy1452;
		}
	}
yy1454:
	++YYCURSOR;
#line 351 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_H1);
           cap_inner.c_end = cap_inner.c_begin;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 39046 "../../lnav/src/data_scanner_re.cc"
yy1455:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			goto yy995;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy994;
			}
			if (yych <= 0xC1) goto yy995;
			yyt2 = YYCURSOR;
			goto yy997;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy998;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy999;
			}
			yyt2 = YYCURSOR;
			goto yy1000;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1001;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1002;
			}
			goto yy995;
		}
	}
yy1456:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy1484;
	goto yy115;
yy1457:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1485;
	goto yy74;
yy1458:
	yyaccept = 45;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1486;
yy1459:
#line 310 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 39108 "../../lnav/src/data_scanner_re.cc"
yy1460:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy1487;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy1487;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy1487;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy1461:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1488;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1489;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1462:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1489;
		goto yy299;
	}
yy1463:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy1033;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			goto yy74;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1382;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1464:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1490;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1490;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1490;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1465:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy1317;
			goto yy1491;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy1317;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy1317;
			goto yy548;
		}
	}
yy1466:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1386;
	goto yy74;
yy1467:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1492;
		if (yych >= ';') goto yy74;
	} else {
		if (yych <= 'F') goto yy1492;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1492;
		goto yy74;
	}
yy1468:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1469;
		if (yych <= ':') goto yy1239;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1469;
		if (yych <= '`') goto yy74;
		if (yych >= 'g') goto yy74;
	}
yy1469:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1493;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1493;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1493;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1470:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1494;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1435;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1494;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1494;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1471:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1495;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1472:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1496;
				yyt2 = YYCURSOR;
				goto yy1497;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1473:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1498;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1474:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1498;
				goto yy1499;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1496;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1321;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1322;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1322;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1475:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1496;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1322;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1476:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1325;
				yyt2 = YYCURSOR;
				goto yy1145;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1325;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1325;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1477:
	yych = *++YYCURSOR;
	if (yych == '@') goto yy1500;
	goto yy74;
yy1478:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy154;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy154;
				goto yy337;
			} else {
				if (yych <= '.') goto yy89;
				if (yych <= '/') goto yy154;
				goto yy1501;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy154;
				goto yy91;
			} else {
				if (yych <= 'F') goto yy1501;
				if (yych <= 'Z') goto yy338;
				goto yy154;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy338;
				goto yy154;
			} else {
				if (yych <= 'f') goto yy1487;
				if (yych <= 'z') goto yy138;
				goto yy154;
			}
		}
	}
yy1479:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1502;
	goto yy74;
yy1480:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1503;
		if (yych <= '@') goto yy74;
		goto yy1503;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy74;
			goto yy1503;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'z') goto yy1503;
			goto yy74;
		}
	}
yy1481:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= 0x08) goto yy74;
			if (yych <= '\t') goto yy1451;
			if (yych <= '\n') goto yy333;
			goto yy74;
		} else {
			if (yych <= ' ') goto yy1481;
			if (yych == '-') goto yy153;
			goto yy74;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy153;
			if (yych <= '@') goto yy74;
			if (yych >= '[') goto yy74;
			yyt1 = YYCURSOR;
		} else {
			if (yych <= '_') {
				yyt1 = YYCURSOR;
				goto yy1482;
			}
			if (yych <= '`') goto yy74;
			if (yych <= 'z') {
				yyt1 = YYCURSOR;
				goto yy1480;
			}
			goto yy74;
		}
	}
yy1482:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy154;
		if (yych <= '9') goto yy1504;
		if (yych <= '@') goto yy154;
		goto yy1504;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy154;
			goto yy1504;
		} else {
			if (yych <= '`') goto yy154;
			if (yych <= 'z') goto yy1503;
			goto yy154;
		}
	}
yy1483:
	yych = *++YYCURSOR;
	if (yych <= 'F') {
		if (yych <= '9') {
			if (yych <= '/') goto yy154;
			goto yy1504;
		} else {
			if (yych <= '@') goto yy154;
			if (yych <= 'E') goto yy1504;
			goto yy1506;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1504;
			if (yych <= '^') goto yy154;
			goto yy1504;
		} else {
			if (yych <= '`') goto yy154;
			if (yych <= 'z') goto yy1503;
			goto yy154;
		}
	}
yy1484:
	yych = *++YYCURSOR;
	if (yych == '(') goto yy1507;
	goto yy74;
yy1485:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1508;
	goto yy74;
yy1486:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1509;
	goto yy74;
yy1487:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1510;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1510;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1510;
		goto yy139;
	}
yy1488:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy120;
			if (yych <= '*') goto yy126;
			goto yy299;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy126;
				goto yy299;
			} else {
				if (yych <= '.') goto yy412;
				if (yych <= '/') goto yy299;
				goto yy1511;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1512;
			if (yych <= ':') goto yy299;
			if (yych <= '?') goto yy126;
			goto yy299;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy126;
				goto yy299;
			} else {
				if (yych <= '`') goto yy126;
				if (yych <= 'z') goto yy299;
				goto yy126;
			}
		}
	}
yy1489:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy120;
		goto yy299;
	} else {
		if (yych <= '.') goto yy412;
		if (yych <= '/') goto yy299;
		if (yych <= '9') goto yy1512;
		goto yy299;
	}
yy1490:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy1033;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1513;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1513;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1513;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1491:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1514;
		if (yych <= ':') goto yy1319;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1514;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1514;
		goto yy74;
	}
yy1492:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1515;
		if (yych <= ':') goto yy1468;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1515;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1515;
		goto yy74;
	}
yy1493:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1516;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1516;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1516;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1494:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy74;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1435;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') goto yy74;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 'z') goto yy74;
			yyt2 = YYCURSOR;
			goto yy307;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy316;
			}
			yyt2 = YYCURSOR;
			goto yy317;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy318;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy319;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy320;
			}
			goto yy74;
		}
	}
yy1495:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy299;
		case '.': goto yy446;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1517;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'S':
		case 's': goto yy137;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1496:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1388;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1497:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy307;
			goto yy74;
		} else {
			if (yych <= '0') goto yy1518;
			if (yych <= '1') goto yy1519;
			goto yy1520;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy1518;
			if (yych <= '@') goto yy74;
			goto yy1387;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy1387;
			goto yy74;
		}
	}
yy1498:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1521;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1388;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1499:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1521;
				goto yy1388;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1321;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1388;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1388;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1500:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1522;
	goto yy74;
yy1501:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1523;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1523;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1510;
		goto yy339;
	}
yy1502:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1524;
	goto yy1411;
yy1503:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				yyt2 = YYCURSOR;
				goto yy1525;
			} else {
				if (yych <= '\n') {
					yyt2 = YYCURSOR;
					goto yy1527;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				goto yy1503;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				goto yy1503;
			} else {
				if (yych == '_') goto yy1503;
				yyt2 = YYCURSOR;
				goto yy1525;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1503;
				yyt2 = YYCURSOR;
				goto yy1525;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1528;
				}
				yyt2 = YYCURSOR;
				goto yy1529;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1530;
				}
				yyt2 = YYCURSOR;
				goto yy1531;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1532;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1533;
				}
				goto yy74;
			}
		}
	}
yy1504:
	yych = *++YYCURSOR;
yy1505:
	if (yych <= '^') {
		if (yych <= ',') {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy74;
				if (yych <= '\t') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				goto yy333;
			} else {
				if (yych == ' ') {
					yyt2 = YYCURSOR;
					goto yy1534;
				}
				yyt2 = YYCURSOR;
				goto yy1525;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1534;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				goto yy1504;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				if (yych <= 'Z') goto yy1504;
				yyt2 = YYCURSOR;
				goto yy1525;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') goto yy1504;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				goto yy1503;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1525;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy1528;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1529;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1530;
				}
				yyt2 = YYCURSOR;
				goto yy1531;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1532;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1533;
				}
				goto yy74;
			}
		}
	}
yy1506:
	yych = *++YYCURSOR;
	if (yych == ' ') {
		yyt2 = YYCURSOR;
		goto yy1535;
	}
	goto yy1505;
yy1507:
	yych = *++YYCURSOR;
	if (yych == 'x') goto yy1536;
	goto yy74;
yy1508:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1537;
	goto yy74;
yy1509:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1538;
	goto yy74;
yy1510:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1539;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1539;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1539;
		goto yy139;
	}
yy1511:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= ',') {
			if (yych <= '$') {
				if (yych == ' ') {
					yyt1 = YYCURSOR;
					goto yy1542;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			} else {
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy1543;
				}
				if (yych == '+') {
					yyt1 = YYCURSOR;
					goto yy1543;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy1544;
				}
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1545;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1546;
				}
				goto yy1548;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1550;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1551;
				}
				yyt1 = YYCURSOR;
				goto yy1552;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy1553;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				yyt1 = YYCURSOR;
				goto yy1554;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy1552;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1553;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1548;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1555;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1556;
				}
				yyt1 = YYCURSOR;
				goto yy1557;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1558;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1559;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1560;
				}
				goto yy1548;
			}
		}
	}
yy1512:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= ',') {
			if (yych <= '$') {
				if (yych == ' ') {
					yyt1 = YYCURSOR;
					goto yy1542;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			} else {
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy1543;
				}
				if (yych == '+') {
					yyt1 = YYCURSOR;
					goto yy1543;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy1544;
				}
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1545;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1546;
				}
				goto yy1549;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy1550;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1551;
				}
				yyt1 = YYCURSOR;
				goto yy1552;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy1553;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				yyt1 = YYCURSOR;
				goto yy1554;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy1540;
				}
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy1552;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1553;
				}
				yyt1 = YYCURSOR;
				goto yy1540;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy128;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1555;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1556;
				}
				yyt1 = YYCURSOR;
				goto yy1557;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1558;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1559;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1560;
				}
				goto yy128;
			}
		}
	}
yy1513:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy1033;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			goto yy74;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1321;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1514:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1561;
		if (yych <= ':') goto yy1386;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1561;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1561;
		goto yy74;
	}
yy1515:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1468;
	goto yy74;
yy1516:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1239;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1239;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1239;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1517:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '+') {
				if (yych <= '$') {
					if (yych == ' ') {
						yyt1 = YYCURSOR;
						goto yy1542;
					}
					yyt1 = YYCURSOR;
					goto yy1540;
				} else {
					if (yych <= '%') {
						yyt1 = YYCURSOR;
						goto yy1543;
					}
					if (yych <= '*') {
						yyt1 = YYCURSOR;
						goto yy1540;
					}
					yyt1 = YYCURSOR;
					goto yy1543;
				}
			} else {
				if (yych <= '.') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy1540;
					}
					if (yych <= '-') {
						yyt1 = YYCURSOR;
						goto yy1544;
					}
					yyt1 = YYCURSOR;
					goto yy1562;
				} else {
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy1546;
					}
					if (yych <= '9') goto yy1563;
					yyt1 = YYCURSOR;
					goto yy1550;
				}
			}
		} else {
			if (yych <= 'G') {
				if (yych <= 'D') {
					if (yych <= '?') {
						yyt1 = YYCURSOR;
						goto yy1540;
					}
					if (yych <= '@') {
						yyt1 = YYCURSOR;
						goto yy1551;
					}
					yyt1 = YYCURSOR;
					goto yy1552;
				} else {
					if (yych <= 'E') {
						yyt1 = YYCURSOR;
						goto yy1564;
					}
					if (yych <= 'F') {
						yyt1 = YYCURSOR;
						goto yy1552;
					}
					yyt1 = YYCURSOR;
					goto yy1565;
				}
			} else {
				if (yych <= 'L') {
					if (yych == 'K') {
						yyt1 = YYCURSOR;
						goto yy1565;
					}
					yyt1 = YYCURSOR;
					goto yy1553;
				} else {
					if (yych <= 'M') {
						yyt1 = YYCURSOR;
						goto yy1565;
					}
					if (yych <= 'Z') {
						yyt1 = YYCURSOR;
						goto yy1553;
					}
					yyt1 = YYCURSOR;
					goto yy1540;
				}
			}
		}
	} else {
		if (yych <= 'p') {
			if (yych <= 'g') {
				if (yych <= 'd') {
					if (yych <= '_') {
						yyt1 = YYCURSOR;
						goto yy1554;
					}
					if (yych <= '`') {
						yyt1 = YYCURSOR;
						goto yy1540;
					}
					yyt1 = YYCURSOR;
					goto yy1552;
				} else {
					if (yych <= 'e') {
						yyt1 = YYCURSOR;
						goto yy1564;
					}
					if (yych <= 'f') {
						yyt1 = YYCURSOR;
						goto yy1552;
					}
					yyt1 = YYCURSOR;
					goto yy1565;
				}
			} else {
				if (yych <= 'l') {
					if (yych == 'k') {
						yyt1 = YYCURSOR;
						goto yy1565;
					}
					yyt1 = YYCURSOR;
					goto yy1553;
				} else {
					if (yych <= 'm') {
						yyt1 = YYCURSOR;
						goto yy1566;
					}
					if (yych <= 'o') {
						yyt1 = YYCURSOR;
						goto yy1553;
					}
					yyt1 = YYCURSOR;
					goto yy1567;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= 'z') {
					if (yych == 'u') {
						yyt1 = YYCURSOR;
						goto yy1567;
					}
					yyt1 = YYCURSOR;
					goto yy1553;
				} else {
					if (yych <= 0x7F) {
						yyt1 = YYCURSOR;
						goto yy1540;
					}
					if (yych <= 0xC1) goto yy24;
					yyt1 = YYCURSOR;
					goto yy1555;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) {
						yyt1 = YYCURSOR;
						goto yy1556;
					}
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy1557;
					}
					yyt1 = YYCURSOR;
					goto yy1558;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy1559;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy1560;
					}
					goto yy24;
				}
			}
		}
	}
yy1518:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1568;
				yyt2 = YYCURSOR;
				goto yy1435;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1434;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1434;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1519:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1569;
				yyt2 = YYCURSOR;
				goto yy1435;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1434;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1434;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1520:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '4') goto yy1569;
				goto yy1570;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy1568;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1435;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1434;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1434;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1521:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1436;
				yyt2 = YYCURSOR;
				goto yy1321;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1436;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1436;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1522:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\t') {
				yyt1 = YYCURSOR;
				goto yy1571;
			}
			goto yy74;
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy1571;
			}
			if (yych <= 0xC1) goto yy74;
			yyt1 = YYCURSOR;
			goto yy1572;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy1573;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy1574;
			}
			yyt1 = YYCURSOR;
			goto yy1575;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy1576;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy1577;
			}
			goto yy74;
		}
	}
yy1523:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1578;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1578;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1539;
		goto yy339;
	}
yy1524:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1579;
	goto yy74;
yy1525:
	yych = *++YYCURSOR;
yy1526:
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\t') goto yy1525;
		} else {
			if (yych <= 0x7F) goto yy1525;
			if (yych <= 0xC1) goto yy74;
			goto yy1528;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1529;
			if (yych <= 0xEF) goto yy1530;
			goto yy1531;
		} else {
			if (yych <= 0xF3) goto yy1532;
			if (yych <= 0xF4) goto yy1533;
			goto yy74;
		}
	}
yy1527:
	++YYCURSOR;
	table_start = yyt1;
	table_end = yyt2;
#line 358 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_H1);
           cap_inner.c_begin = table_start.val - this->ds_input.udata();
           cap_inner.c_end = table_end.val - this->ds_input.udata();
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 41833 "../../lnav/src/data_scanner_re.cc"
yy1528:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1525;
	goto yy74;
yy1529:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1528;
	goto yy74;
yy1530:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1528;
	goto yy74;
yy1531:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1530;
	goto yy74;
yy1532:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1530;
	goto yy74;
yy1533:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1530;
	goto yy74;
yy1534:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				goto yy1525;
			} else {
				if (yych <= '\n') goto yy333;
				if (yych <= 0x1F) goto yy1525;
				goto yy1534;
			}
		} else {
			if (yych <= '/') {
				if (yych == '-') goto yy1534;
				goto yy1525;
			} else {
				if (yych <= '9') goto yy1534;
				if (yych <= '@') goto yy1525;
				goto yy1534;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '_') {
				if (yych <= '^') goto yy1525;
				goto yy1534;
			} else {
				if (yych <= 0x7F) goto yy1525;
				if (yych <= 0xC1) goto yy74;
				goto yy1528;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1529;
				if (yych <= 0xEF) goto yy1530;
				goto yy1531;
			} else {
				if (yych <= 0xF3) goto yy1532;
				if (yych <= 0xF4) goto yy1533;
				goto yy74;
			}
		}
	}
yy1535:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'N') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'M') goto yy1534;
			goto yy1580;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1536:
	yych = *++YYCURSOR;
	if (yych == '8') goto yy1581;
	goto yy74;
yy1537:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1582;
	goto yy74;
yy1538:
	yyaccept = 45;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1583;
	goto yy1459;
yy1539:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1584;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1584;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1584;
		goto yy139;
	}
yy1540:
	++YYCURSOR;
yy1541:
	YYCURSOR = yyt1;
#line 406 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_CREDIT_CARD_NUMBER);
           if (!this->is_credit_card(this->to_string_fragment(cap_all))) {
               if (cap_all.length() > 16) {
                   cap_all.c_end = cap_all.c_begin + 4;
                   cap_inner.c_end = cap_inner.c_begin + 4;
               }
               this->ds_next_offset = cap_all.c_end;
               token_out = DT_HEX_DUMP;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 41975 "../../lnav/src/data_scanner_re.cc"
yy1542:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych == '%') goto yy274;
	goto yy1541;
yy1543:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy90;
			goto yy1541;
		} else {
			if (yych == ',') goto yy1541;
			if (yych <= '.') goto yy90;
			goto yy1541;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy90;
			if (yych <= '?') goto yy1541;
			if (yych <= 'Z') goto yy90;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy90;
			goto yy1541;
		}
	}
yy1544:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy1541;
			goto yy89;
		} else {
			if (yych <= ',') goto yy1541;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy1541;
			goto yy138;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy138;
			goto yy1541;
		}
	}
yy1545:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy1541;
			goto yy89;
		} else {
			if (yych <= ',') goto yy1541;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy1541;
			goto yy552;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy138;
			goto yy1541;
		}
	}
yy1546:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1541;
		if (yych <= '9') goto yy279;
		if (yych <= '@') goto yy1541;
		goto yy279;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1541;
			goto yy279;
		} else {
			if (yych <= '`') goto yy1541;
			if (yych <= 'z') goto yy279;
			goto yy1541;
		}
	}
yy1547:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
yy1548:
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy126;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy126;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '7') goto yy1547;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy294;
				goto yy126;
			} else {
				if (yych <= '@') goto yy130;
				if (yych <= 'F') goto yy298;
				goto yy135;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy138;
				goto yy126;
			} else {
				if (yych <= 'f') goto yy298;
				if (yych <= 'z') goto yy135;
				goto yy126;
			}
		}
	}
yy1549:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy128;
				goto yy120;
			} else {
				if (yych == '+') goto yy89;
				goto yy128;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy122;
				goto yy412;
			} else {
				if (yych <= '/') goto yy124;
				if (yych <= '9') goto yy1549;
				goto yy294;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy128;
				goto yy130;
			} else {
				if (yych <= 'F') goto yy298;
				if (yych <= 'Z') goto yy135;
				goto yy128;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy138;
				goto yy128;
			} else {
				if (yych <= 'f') goto yy298;
				if (yych <= 'z') goto yy135;
				goto yy128;
			}
		}
	}
yy1550:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy288;
	if (yych == ':') goto yy124;
	goto yy1541;
yy1551:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1541;
			goto yy216;
		} else {
			if (yych <= '/') goto yy1541;
			if (yych <= '9') goto yy292;
			goto yy1541;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy292;
			if (yych <= '^') goto yy1541;
			goto yy279;
		} else {
			if (yych <= '`') goto yy1541;
			if (yych <= 'z') goto yy292;
			goto yy1541;
		}
	}
yy1552:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy299;
			goto yy1541;
		} else {
			if (yych == ',') goto yy1541;
			goto yy299;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= 'Z') goto yy299;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy299;
			goto yy1541;
		}
	}
yy1553:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy136;
			goto yy1541;
		} else {
			if (yych == ',') goto yy1541;
			goto yy136;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= 'Z') goto yy136;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy136;
			goto yy1541;
		}
	}
yy1554:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy139;
			goto yy1541;
		} else {
			if (yych == ',') goto yy1541;
			goto yy139;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= 'Z') goto yy139;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy139;
			goto yy1541;
		}
	}
yy1555:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1540;
	goto yy74;
yy1556:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1555;
	goto yy74;
yy1557:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1555;
	goto yy74;
yy1558:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1557;
	goto yy74;
yy1559:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1557;
	goto yy74;
yy1560:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1557;
	goto yy74;
yy1561:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy1431;
			goto yy1585;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy1431;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy1431;
			goto yy548;
		}
	}
yy1562:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy1541;
			goto yy89;
		} else {
			if (yych <= ',') goto yy1541;
			if (yych <= '.') goto yy89;
			if (yych <= '/') goto yy1541;
			goto yy598;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= '@') goto yy91;
			if (yych <= 'Z') goto yy138;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy138;
			goto yy1541;
		}
	}
yy1563:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_bol[1280+yych] & 2) {
		goto yy118;
	}
	switch (yych) {
		case '%': goto yy120;
		case '+': goto yy89;
		case '-': goto yy122;
		case '.': goto yy446;
		case '/': goto yy124;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1563;
		case ':': goto yy294;
		case '@': goto yy130;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy298;
		case 'B':
		case 'b': goto yy1155;
		case 'E':
		case 'e': goto yy1156;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy133;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy135;
		case 'S':
		case 's': goto yy137;
		case '_': goto yy138;
		case 'm': goto yy140;
		case 'p':
		case 'u': goto yy141;
		default: goto yy24;
	}
yy1564:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy299;
			goto yy1541;
		} else {
			if (yych <= '+') goto yy304;
			if (yych <= ',') goto yy1541;
			if (yych <= '-') goto yy443;
			goto yy299;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1541;
			if (yych <= 'Z') goto yy299;
			goto yy1541;
		} else {
			if (yych == '`') goto yy1541;
			if (yych <= 'z') goto yy299;
			goto yy1541;
		}
	}
yy1565:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'B') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy1541;
			goto yy136;
		} else {
			if (yych <= ':') {
				if (yych <= ',') goto yy1541;
				goto yy136;
			} else {
				if (yych <= '?') goto yy1541;
				if (yych <= 'A') goto yy136;
				goto yy137;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy136;
				goto yy1541;
			} else {
				if (yych == '`') goto yy1541;
				goto yy136;
			}
		} else {
			if (yych <= 'h') {
				if (yych <= 'b') goto yy137;
				goto yy136;
			} else {
				if (yych <= 'i') goto yy295;
				if (yych <= 'z') goto yy136;
				goto yy1541;
			}
		}
	}
yy1566:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1541;
				goto yy136;
			} else {
				if (yych == '+') goto yy136;
				goto yy1541;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy136;
				if (yych <= '?') goto yy1541;
				goto yy136;
			} else {
				if (yych <= 'B') goto yy137;
				if (yych <= 'R') goto yy136;
				goto yy137;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy136;
				if (yych <= '^') goto yy1541;
				goto yy136;
			} else {
				if (yych <= '`') goto yy1541;
				if (yych <= 'a') goto yy136;
				goto yy137;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy297;
				goto yy136;
			} else {
				if (yych <= 's') goto yy137;
				if (yych <= 'z') goto yy136;
				goto yy1541;
			}
		}
	}
yy1567:
	yyaccept = 46;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '+') {
			if (yych == '%') goto yy136;
			if (yych <= '*') goto yy1541;
			goto yy136;
		} else {
			if (yych <= ',') goto yy1541;
			if (yych <= ':') goto yy136;
			if (yych <= '?') goto yy1541;
			goto yy136;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'S') goto yy137;
			if (yych <= 'Z') goto yy136;
			if (yych <= '^') goto yy1541;
			goto yy136;
		} else {
			if (yych <= 'r') {
				if (yych <= '`') goto yy1541;
				goto yy136;
			} else {
				if (yych <= 's') goto yy137;
				if (yych <= 'z') goto yy136;
				goto yy1541;
			}
		}
	}
yy1568:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1470;
				yyt2 = YYCURSOR;
				goto yy1435;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1470;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1470;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1569:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1586;
				yyt2 = YYCURSOR;
				goto yy1435;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1470;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1470;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1570:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '5') goto yy1586;
				goto yy1470;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1435;
				}
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 'F') goto yy1470;
				if (yych <= 'Z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy1470;
				goto yy74;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 0xC1) goto yy74;
				yyt2 = YYCURSOR;
				goto yy315;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy316;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1571:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy74;
			if (yych <= '\t') goto yy1571;
			goto yy1587;
		} else {
			if (yych <= 0x7F) goto yy1571;
			if (yych <= 0xC1) goto yy74;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1573;
			if (yych <= 0xEF) goto yy1574;
			goto yy1575;
		} else {
			if (yych <= 0xF3) goto yy1576;
			if (yych <= 0xF4) goto yy1577;
			goto yy74;
		}
	}
yy1572:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1571;
	goto yy74;
yy1573:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy74;
	if (yych <= 0xBF) goto yy1572;
	goto yy74;
yy1574:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1572;
	goto yy74;
yy1575:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy74;
	if (yych <= 0xBF) goto yy1574;
	goto yy74;
yy1576:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0xBF) goto yy1574;
	goto yy74;
yy1577:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy74;
	if (yych <= 0x8F) goto yy1574;
	goto yy74;
yy1578:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1588;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1588;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1584;
		goto yy339;
	}
yy1579:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1589;
	goto yy74;
yy1580:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'O') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'N') goto yy1534;
			goto yy1590;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1581:
	yych = *++YYCURSOR;
	if (yych == '6') goto yy1591;
	goto yy74;
yy1582:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1592;
	goto yy74;
yy1583:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1593;
	goto yy74;
yy1584:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1594;
	goto yy139;
yy1585:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1595;
		if (yych <= ':') goto yy1433;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1595;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1595;
		goto yy74;
	}
yy1586:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt1 = YYCURSOR;
				goto yy447;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1494;
				yyt2 = YYCURSOR;
				goto yy1435;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1494;
			} else {
				if (yych <= 'Z') goto yy74;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1494;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy74;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy74;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1587:
	++YYCURSOR;
	hunk_heading = yyt1;
#line 373 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_DIFF_HUNK_HEADING);
           cap_inner.c_begin = hunk_heading.val - this->ds_input.udata();
           cap_inner.c_end -= 1;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 42984 "../../lnav/src/data_scanner_re.cc"
yy1588:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1596;
	goto yy339;
yy1589:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1597;
	goto yy74;
yy1590:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'T') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'S') goto yy1534;
			goto yy1598;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1591:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy114;
	goto yy74;
yy1592:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy1540;
			}
			goto yy74;
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy1540;
			}
			if (yych <= 0xC1) goto yy74;
			yyt1 = YYCURSOR;
			goto yy1555;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy1556;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy1557;
			}
			yyt1 = YYCURSOR;
			goto yy1558;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy1559;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy1560;
			}
			goto yy74;
		}
	}
yy1593:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1599;
	goto yy74;
yy1594:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy1600;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy1600;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy1600;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy1595:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1601;
		if (yych <= ':') goto yy1468;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1601;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1601;
		goto yy74;
	}
yy1596:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy154;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy154;
				goto yy337;
			} else {
				if (yych <= '.') goto yy89;
				if (yych <= '/') goto yy154;
				goto yy1602;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy154;
				goto yy91;
			} else {
				if (yych <= 'F') goto yy1602;
				if (yych <= 'Z') goto yy338;
				goto yy154;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy338;
				goto yy154;
			} else {
				if (yych <= 'f') goto yy1600;
				if (yych <= 'z') goto yy138;
				goto yy154;
			}
		}
	}
yy1597:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1603;
	goto yy1411;
yy1598:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1604;
			if (yych <= ',') goto yy1526;
			goto yy1534;
		}
	} else {
		if (yych <= '@') {
			if (yych <= '/') goto yy1526;
			if (yych <= '9') goto yy1534;
			goto yy1526;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1599:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy74;
	if (yych <= '9') goto yy1605;
	goto yy74;
yy1600:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1606;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1606;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1606;
		goto yy139;
	}
yy1601:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy548;
			goto yy718;
		} else {
			if (yych <= '/') goto yy548;
			if (yych <= '9') goto yy1492;
			goto yy1607;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy548;
			goto yy1492;
		} else {
			if (yych <= '`') goto yy548;
			if (yych <= 'f') goto yy1492;
			goto yy548;
		}
	}
yy1602:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1608;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1608;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1606;
		goto yy339;
	}
yy1603:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1609;
	goto yy1411;
yy1604:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'E') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'D') goto yy1534;
			goto yy1610;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1605:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1459;
	if (yych <= '9') goto yy1611;
	goto yy1459;
yy1606:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1612;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1612;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1612;
		goto yy139;
	}
yy1607:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy74;
		if (yych <= '9') goto yy1613;
		if (yych <= ':') goto yy1239;
		goto yy74;
	} else {
		if (yych <= 'F') goto yy1613;
		if (yych <= '`') goto yy74;
		if (yych <= 'f') goto yy1613;
		goto yy74;
	}
yy1608:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1614;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1614;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1612;
		goto yy339;
	}
yy1609:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1411;
	if (yych <= '9') goto yy1615;
	goto yy1411;
yy1610:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'X') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'W') goto yy1534;
			goto yy1616;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1611:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1459;
	if (yych <= '9') goto yy1617;
	goto yy1459;
yy1612:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1618;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1618;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1618;
		goto yy139;
	}
yy1613:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= '9') goto yy1619;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy312;
			}
			yyt2 = YYCURSOR;
			goto yy307;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy1619;
				goto yy74;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= 'f') goto yy1619;
				goto yy74;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy307;
			}
			if (yych <= 0xC1) goto yy74;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy315;
			}
			yyt2 = YYCURSOR;
			goto yy316;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy74;
			}
		}
	}
yy1614:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1620;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1620;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1618;
		goto yy339;
	}
yy1615:
	++YYCURSOR;
	goto yy1411;
yy1616:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'I') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'H') goto yy1534;
			goto yy1621;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1617:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1459;
	if (yych <= '9') goto yy1622;
	goto yy1459;
yy1618:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1623;
	goto yy139;
yy1619:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				yyt2 = YYCURSOR;
				goto yy1624;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				if (yych <= '9') goto yy1516;
				yyt2 = YYCURSOR;
				goto yy1625;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1516;
			} else {
				if (yych <= 'Z') goto yy548;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy307;
				}
				goto yy1516;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy548;
				yyt2 = YYCURSOR;
				goto yy307;
			} else {
				if (yych <= 0xC1) goto yy548;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy315;
				}
				yyt2 = YYCURSOR;
				goto yy316;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy317;
				}
				yyt2 = YYCURSOR;
				goto yy318;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy319;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy320;
				}
				goto yy548;
			}
		}
	}
yy1620:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy1626;
	goto yy339;
yy1621:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'S') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'R') goto yy1534;
			goto yy1627;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1622:
	++YYCURSOR;
	goto yy1459;
yy1623:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy90;
			goto yy1628;
		} else {
			if (yych <= '@') goto yy90;
			if (yych <= 'F') goto yy1628;
			goto yy138;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy138;
			goto yy90;
		} else {
			if (yych <= 'f') goto yy1628;
			if (yych <= 'z') goto yy138;
			goto yy90;
		}
	}
yy1624:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy308;
		if (yych <= '9') goto yy852;
		goto yy308;
	} else {
		if (yych <= 'F') goto yy852;
		if (yych <= '`') goto yy308;
		if (yych <= 'f') goto yy852;
		goto yy308;
	}
yy1625:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy307;
		if (yych <= '/') goto yy74;
		goto yy852;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy74;
			goto yy852;
		} else {
			if (yych <= '`') goto yy74;
			if (yych <= 'f') goto yy852;
			goto yy74;
		}
	}
yy1626:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy89;
			if (yych <= '*') goto yy154;
			goto yy89;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy154;
				goto yy337;
			} else {
				if (yych <= '.') goto yy89;
				if (yych <= '/') goto yy154;
				goto yy1629;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy154;
				goto yy91;
			} else {
				if (yych <= 'F') goto yy1629;
				if (yych <= 'Z') goto yy338;
				goto yy154;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy338;
				goto yy154;
			} else {
				if (yych <= 'f') goto yy1628;
				if (yych <= 'z') goto yy138;
				goto yy154;
			}
		}
	}
yy1627:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'T') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'S') goto yy1534;
			goto yy1630;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1628:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1631;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1631;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1631;
		goto yy139;
	}
yy1629:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1632;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1632;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1631;
		goto yy339;
	}
yy1630:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= 'S') {
			if (yych <= '9') goto yy1534;
			if (yych <= '@') goto yy1526;
			if (yych <= 'R') goto yy1534;
			goto yy1633;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1631:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1634;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1634;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1634;
		goto yy139;
	}
yy1632:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1635;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1635;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1634;
		goto yy339;
	}
yy1633:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '\n') {
			if (yych <= 0x08) goto yy1526;
			if (yych <= '\t') goto yy1636;
			goto yy333;
		} else {
			if (yych == ' ') goto yy1637;
			if (yych <= ',') goto yy1526;
			goto yy1534;
		}
	} else {
		if (yych <= '@') {
			if (yych <= '/') goto yy1526;
			if (yych <= '9') goto yy1534;
			goto yy1526;
		} else {
			if (yych <= 'Z') goto yy1534;
			if (yych == '_') goto yy1534;
			goto yy1526;
		}
	}
yy1634:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1638;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1638;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1638;
		goto yy139;
	}
yy1635:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1639;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1639;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1638;
		goto yy339;
	}
yy1636:
	yych = *++YYCURSOR;
	if (yych <= '_') {
		if (yych <= 0x1F) {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy74;
				goto yy1525;
			} else {
				if (yych <= '\t') goto yy1636;
				if (yych <= '\n') goto yy1527;
				goto yy1525;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ' ') goto yy1636;
				goto yy1525;
			} else {
				if (yych <= 'Z') {
					yyt3 = YYCURSOR;
					goto yy1640;
				}
				if (yych <= '^') goto yy1525;
				yyt3 = YYCURSOR;
				goto yy1640;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') goto yy1525;
				yyt3 = YYCURSOR;
				goto yy1640;
			} else {
				if (yych <= 0x7F) goto yy1525;
				if (yych <= 0xC1) goto yy74;
				goto yy1528;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy1529;
				if (yych <= 0xEF) goto yy1530;
				goto yy1531;
			} else {
				if (yych <= 0xF3) goto yy1532;
				if (yych <= 0xF4) goto yy1533;
				goto yy74;
			}
		}
	}
yy1637:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy74;
				if (yych <= 0x08) goto yy1525;
				goto yy1636;
			} else {
				if (yych <= '\n') goto yy333;
				if (yych <= 0x1F) goto yy1525;
				goto yy1637;
			}
		} else {
			if (yych <= '/') {
				if (yych == '-') goto yy1534;
				goto yy1525;
			} else {
				if (yych <= '9') goto yy1534;
				if (yych <= '@') goto yy1525;
				yyt3 = YYCURSOR;
				goto yy1641;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt3 = YYCURSOR;
					goto yy1641;
				}
				goto yy1525;
			} else {
				if (yych <= 'z') {
					yyt3 = YYCURSOR;
					goto yy1640;
				}
				if (yych <= 0x7F) goto yy1525;
				goto yy74;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy1528;
				if (yych <= 0xE0) goto yy1529;
				goto yy1530;
			} else {
				if (yych <= 0xF0) goto yy1531;
				if (yych <= 0xF3) goto yy1532;
				if (yych <= 0xF4) goto yy1533;
				goto yy74;
			}
		}
	}
yy1638:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1642;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1642;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1642;
		goto yy139;
	}
yy1639:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1643;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1643;
		if (yych <= '`') goto yy339;
		if (yych <= 'f') goto yy1642;
		goto yy339;
	}
yy1640:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1526;
		if (yych <= '9') {
			yyt1 = yyt3;
			goto yy1503;
		}
		if (yych <= '@') goto yy1526;
		yyt1 = yyt3;
		goto yy1503;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1526;
			yyt1 = yyt3;
			goto yy1503;
		} else {
			if (yych <= '`') goto yy1526;
			if (yych <= 'z') {
				yyt1 = yyt3;
				goto yy1503;
			}
			goto yy1526;
		}
	}
yy1641:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych == '\n') goto yy333;
			goto yy1526;
		} else {
			if (yych <= ' ') goto yy1534;
			if (yych == '-') goto yy1534;
			goto yy1526;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') {
				yyt1 = yyt3;
				goto yy1504;
			}
			if (yych <= '@') goto yy1526;
			if (yych <= 'Z') {
				yyt1 = yyt3;
				goto yy1504;
			}
			goto yy1526;
		} else {
			if (yych <= '_') {
				yyt1 = yyt3;
				goto yy1504;
			}
			if (yych <= '`') goto yy1526;
			if (yych <= 'z') {
				yyt1 = yyt3;
				goto yy1503;
			}
			goto yy1526;
		}
	}
yy1642:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1644;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1644;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1644;
		goto yy139;
	}
yy1643:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1645;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1645;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1644:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1646;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1646;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1646;
		goto yy139;
	}
yy1645:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1647;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1647;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1646:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1648;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1648;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1648;
		goto yy139;
	}
yy1647:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1649;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1649;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1648:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1650;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1650;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1650;
		goto yy139;
	}
yy1649:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1651;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1651;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1650:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1652;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1652;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1652;
		goto yy139;
	}
yy1651:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1653;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1653;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1652:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1654;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1654;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1654;
		goto yy139;
	}
yy1653:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1655;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1655;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1654:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy139;
		if (yych <= '9') goto yy1656;
		goto yy139;
	} else {
		if (yych <= 'F') goto yy1656;
		if (yych <= '`') goto yy139;
		if (yych <= 'f') goto yy1656;
		goto yy139;
	}
yy1655:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy339;
		if (yych <= '9') goto yy1658;
		goto yy339;
	} else {
		if (yych <= 'F') goto yy1658;
		if (yych <= '`') goto yy339;
		if (yych >= 'g') goto yy339;
	}
yy1656:
	yyaccept = 47;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy139;
		} else {
			if (yych != ',') goto yy139;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1657;
			if (yych <= 'Z') goto yy139;
		} else {
			if (yych == '`') goto yy1657;
			if (yych <= 'z') goto yy139;
		}
	}
yy1657:
#line 404 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_UUID); }
#line 44174 "../../lnav/src/data_scanner_re.cc"
yy1658:
	yyaccept = 47;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych == '\n') goto yy339;
			if (yych <= 0x1F) goto yy1657;
			goto yy339;
		} else {
			if (yych == '%') goto yy339;
			if (yych <= '*') goto yy1657;
			goto yy339;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ',') goto yy1657;
			if (yych <= ':') goto yy339;
			if (yych <= '?') goto yy1657;
			goto yy339;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1657;
				goto yy339;
			} else {
				if (yych <= '`') goto yy1657;
				if (yych <= 'z') goto yy339;
				goto yy1657;
			}
		}
	}
/* *********************************** */
yyc_init:
	yych = *YYCURSOR;
	if (yybm_init[1536+yych] & 32) {
		goto yy1667;
	}
	if (yych <= 'A') {
		if (yych <= '(') {
			if (yych <= 0x1F) {
				if (yych <= '\n') {
					if (yych <= 0x00) goto yy1660;
					if (yych <= 0x08) goto yy1661;
					if (yych <= '\t') goto yy1663;
					goto yy1665;
				} else {
					if (yych <= '\r') {
						if (yych <= '\f') goto yy1661;
						goto yy1666;
					} else {
						if (yych == 0x1B) goto yy1668;
						goto yy1661;
					}
				}
			} else {
				if (yych <= '#') {
					if (yych <= ' ') goto yy1663;
					if (yych <= '!') goto yy1661;
					if (yych <= '"') goto yy1669;
					goto yy1670;
				} else {
					if (yych <= '%') {
						if (yych <= '$') goto yy1661;
						goto yy1671;
					} else {
						if (yych <= '&') goto yy1661;
						if (yych <= '\'') goto yy1672;
						goto yy1673;
					}
				}
			}
		} else {
			if (yych <= '1') {
				if (yych <= ',') {
					if (yych <= ')') goto yy1674;
					if (yych <= '*') goto yy1661;
					if (yych <= '+') goto yy1675;
					goto yy1676;
				} else {
					if (yych <= '.') {
						if (yych <= '-') goto yy1677;
						goto yy1678;
					} else {
						if (yych <= '/') goto yy1680;
						if (yych <= '0') goto yy1682;
						goto yy1684;
					}
				}
			} else {
				if (yych <= '<') {
					if (yych <= '9') {
						if (yych <= '2') goto yy1685;
						goto yy1686;
					} else {
						if (yych <= ':') goto yy1687;
						if (yych <= ';') goto yy1689;
						goto yy1690;
					}
				} else {
					if (yych <= '>') {
						if (yych <= '=') goto yy1692;
						goto yy1693;
					} else {
						if (yych <= '?') goto yy1661;
						if (yych <= '@') goto yy1694;
						goto yy1695;
					}
				}
			}
		}
	} else {
		if (yych <= 'm') {
			if (yych <= 'Z') {
				if (yych <= 'I') {
					if (yych <= 'E') goto yy1697;
					if (yych <= 'F') goto yy1699;
					if (yych <= 'H') goto yy1700;
					goto yy1701;
				} else {
					if (yych <= 'N') {
						if (yych <= 'M') goto yy1700;
						goto yy1702;
					} else {
						if (yych == 'T') goto yy1703;
						goto yy1700;
					}
				}
			} else {
				if (yych <= '_') {
					if (yych <= '\\') {
						if (yych <= '[') goto yy1704;
						goto yy1705;
					} else {
						if (yych <= ']') goto yy1706;
						if (yych <= '^') goto yy1661;
						goto yy1707;
					}
				} else {
					if (yych <= 'a') {
						if (yych <= '`') goto yy1661;
						goto yy1709;
					} else {
						if (yych <= 'e') goto yy1710;
						if (yych <= 'f') goto yy1711;
						goto yy1712;
					}
				}
			}
		} else {
			if (yych <= '|') {
				if (yych <= 's') {
					if (yych <= 'n') goto yy1713;
					if (yych == 'r') goto yy1714;
					goto yy1712;
				} else {
					if (yych <= 'u') {
						if (yych <= 't') goto yy1715;
						goto yy1716;
					} else {
						if (yych <= 'z') goto yy1712;
						if (yych <= '{') goto yy1717;
						goto yy1661;
					}
				}
			} else {
				if (yych <= 0xE0) {
					if (yych <= 0x7F) {
						if (yych <= '}') goto yy1718;
						goto yy1661;
					} else {
						if (yych <= 0xC1) goto yy1719;
						if (yych <= 0xDF) goto yy1721;
						goto yy1722;
					}
				} else {
					if (yych <= 0xF0) {
						if (yych <= 0xEF) goto yy1723;
						goto yy1724;
					} else {
						if (yych <= 0xF3) goto yy1725;
						if (yych <= 0xF4) goto yy1726;
						goto yy1719;
					}
				}
			}
		}
	}
yy1660:
	++YYCURSOR;
#line 174 "../../lnav/src/data_scanner_re.re"
	{ return std::nullopt; }
#line 44365 "../../lnav/src/data_scanner_re.cc"
yy1661:
	++YYCURSOR;
yy1662:
#line 491 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_GARBAGE); }
#line 44371 "../../lnav/src/data_scanner_re.cc"
yy1663:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1728;
	if (yych <= '9') goto yy1729;
	if (yych <= ':') goto yy1731;
	goto yy1728;
yy1664:
#line 488 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WHITE); }
#line 44382 "../../lnav/src/data_scanner_re.cc"
yy1665:
	++YYCURSOR;
#line 484 "../../lnav/src/data_scanner_re.re"
	{
           this->ds_bol = true;
           RET(DT_LINE);
       }
#line 44390 "../../lnav/src/data_scanner_re.cc"
yy1666:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '\n') goto yy1665;
		goto yy1728;
	} else {
		if (yych <= '9') goto yy1729;
		if (yych <= ':') goto yy1731;
		goto yy1728;
	}
yy1667:
	yych = *++YYCURSOR;
	if (yybm_init[1536+yych] & 32) {
		goto yy1667;
	}
#line 176 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_ZERO_WIDTH_SPACE);
       }
#line 44411 "../../lnav/src/data_scanner_re.cc"
yy1668:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '[') goto yy1732;
	goto yy1662;
yy1669:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1A) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1662;
			if (yych <= '\t') goto yy1734;
			goto yy1662;
		} else {
			if (yych == 0x16) goto yy1662;
			goto yy1734;
		}
	} else {
		if (yych <= '"') {
			if (yych <= 0x1B) goto yy1662;
			if (yych <= '!') goto yy1734;
			goto yy1735;
		} else {
			if (yych <= 0x7F) goto yy1734;
			if (yych <= 0xC1) goto yy1662;
			if (yych <= 0xF4) goto yy1734;
			goto yy1662;
		}
	}
yy1670:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '[') goto yy1744;
	goto yy1662;
yy1671:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy1662;
		} else {
			if (yych == ',') goto yy1662;
			if (yych <= '.') goto yy1746;
			goto yy1662;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1662;
			if (yych <= 'Z') goto yy1746;
			goto yy1662;
		} else {
			if (yych == '`') goto yy1662;
			if (yych <= 'z') goto yy1746;
			goto yy1662;
		}
	}
yy1672:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x1A) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1662;
			if (yych <= '\t') goto yy1749;
			goto yy1662;
		} else {
			if (yych == 0x16) goto yy1662;
			goto yy1749;
		}
	} else {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy1662;
			if (yych <= '&') goto yy1749;
			goto yy1750;
		} else {
			if (yych <= 0x7F) goto yy1749;
			if (yych <= 0xC1) goto yy1662;
			if (yych <= 0xF4) goto yy1749;
			goto yy1662;
		}
	}
yy1673:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy1758;
#line 395 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LPAREN); }
#line 44499 "../../lnav/src/data_scanner_re.cc"
yy1674:
	++YYCURSOR;
#line 396 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RPAREN); }
#line 44504 "../../lnav/src/data_scanner_re.cc"
yy1675:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '0') {
		if (yych <= '+') {
			if (yych == '%') goto yy1763;
			if (yych <= '*') goto yy1662;
			goto yy1759;
		} else {
			if (yych <= ',') goto yy1662;
			if (yych <= '.') goto yy1763;
			if (yych <= '/') goto yy1662;
			goto yy1760;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1761;
			if (yych <= '?') goto yy1662;
			if (yych <= 'Z') goto yy1763;
			goto yy1662;
		} else {
			if (yych == '`') goto yy1662;
			if (yych <= 'z') goto yy1763;
			goto yy1662;
		}
	}
yy1676:
	++YYCURSOR;
#line 387 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COMMA); }
#line 44535 "../../lnav/src/data_scanner_re.cc"
yy1677:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy1763;
			if (yych <= '*') goto yy1662;
			goto yy1763;
		} else {
			if (yych <= ',') goto yy1662;
			if (yych <= '-') goto yy1764;
			if (yych <= '.') goto yy1763;
			goto yy1662;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '0') goto yy1765;
			if (yych <= '9') goto yy1766;
			if (yych <= '?') goto yy1662;
			goto yy1763;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1662;
				goto yy1763;
			} else {
				if (yych <= '`') goto yy1662;
				if (yych <= 'z') goto yy1763;
				goto yy1662;
			}
		}
	}
yy1678:
	yyaccept = 2;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy1746;
			if (yych >= '+') goto yy1746;
		} else {
			if (yych <= ',') goto yy1679;
			if (yych <= '-') goto yy1746;
			if (yych <= '.') goto yy1767;
			goto yy1768;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1679;
			if (yych <= 'Z') goto yy1746;
		} else {
			if (yych == '`') goto yy1679;
			if (yych <= 'z') goto yy1746;
		}
	}
yy1679:
#line 489 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_DOT); }
#line 44593 "../../lnav/src/data_scanner_re.cc"
yy1680:
	yych = *++YYCURSOR;
	if (yych == '*') goto yy1771;
	if (yych == 'P') goto yy1772;
	goto yy1770;
yy1681:
#line 300 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PATH); }
#line 44602 "../../lnav/src/data_scanner_re.cc"
yy1682:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1778;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7': goto yy1780;
		case '8':
		case '9': goto yy1782;
		case ':': goto yy1784;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy1786;
		case 'B':
		case 'b': goto yy1787;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		case 'x': goto yy1797;
		default: goto yy1683;
	}
yy1683:
#line 444 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_NUMBER);
           auto sf = this->to_string_fragment(cap_all);
           if (isalpha(sf.back())) {
               while (isalpha(sf.back())) {
                   sf.pop_back();
               }
               cap_all.c_end = sf.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = sf.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               this->ds_units = true;
           }
           return tokenize_result{DT_NUMBER, cap_all, cap_inner, this->ds_input.data()};
       }
#line 44698 "../../lnav/src/data_scanner_re.cc"
yy1684:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1778;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1798;
		case ':': goto yy1784;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1786;
		case 'B':
		case 'b': goto yy1787;
		case 'E':
		case 'e': goto yy1799;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1685:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1778;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy1798;
		case '5': goto yy1800;
		case '6':
		case '7':
		case '8':
		case '9': goto yy1801;
		case ':': goto yy1784;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1786;
		case 'B':
		case 'b': goto yy1787;
		case 'E':
		case 'e': goto yy1799;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1686:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1778;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1801;
		case ':': goto yy1784;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1786;
		case 'B':
		case 'b': goto yy1787;
		case 'E':
		case 'e': goto yy1799;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1687:
	yyaccept = 4;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1802;
yy1688:
#line 385 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_COLON); }
#line 44943 "../../lnav/src/data_scanner_re.cc"
yy1689:
	++YYCURSOR;
#line 388 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_SEMI); }
#line 44948 "../../lnav/src/data_scanner_re.cc"
yy1690:
	yyaccept = 5;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '>') {
		if (yych <= '-') {
			if (yych == '!') goto yy1803;
			if (yych >= '-') goto yy1804;
		} else {
			if (yych <= '.') goto yy1691;
			if (yych <= '/') goto yy1805;
			if (yych <= ':') goto yy1804;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1806;
			if (yych <= '@') goto yy1691;
			if (yych <= 'Z') goto yy1804;
		} else {
			if (yych == '`') goto yy1691;
			if (yych <= 'z') goto yy1804;
		}
	}
yy1691:
#line 397 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LANGLE); }
#line 44974 "../../lnav/src/data_scanner_re.cc"
yy1692:
	++YYCURSOR;
#line 386 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EQUALS); }
#line 44979 "../../lnav/src/data_scanner_re.cc"
yy1693:
	++YYCURSOR;
#line 398 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RANGLE); }
#line 44984 "../../lnav/src/data_scanner_re.cc"
yy1694:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '@') goto yy1807;
	goto yy1662;
yy1695:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych >= '%') goto yy1708;
			} else {
				if (yych == '\'') goto yy1808;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy1708;
			} else {
				if (yych <= '/') goto yy1708;
				if (yych <= '9') goto yy1812;
				goto yy1813;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych >= '@') goto yy1708;
			} else {
				if (yych <= 'F') goto yy1814;
				if (yych <= 'Z') goto yy1815;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1708;
			} else {
				if (yych <= 'f') goto yy1816;
				if (yych <= 'z') goto yy1817;
			}
		}
	}
yy1696:
#line 466 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WORD); }
#line 45029 "../../lnav/src/data_scanner_re.cc"
yy1697:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '\'') goto yy1808;
			goto yy1708;
		} else {
			if (yych <= '9') goto yy1812;
			if (yych <= ':') goto yy1813;
			goto yy1708;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy1814;
			if (yych <= 'Z') goto yy1815;
			goto yy1708;
		} else {
			if (yych <= 'f') goto yy1816;
			if (yych <= 'z') goto yy1817;
			goto yy1708;
		}
	}
yy1698:
#line 476 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 45058 "../../lnav/src/data_scanner_re.cc"
yy1699:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'A') {
		if (yych <= '/') {
			if (yych == '\'') goto yy1808;
			goto yy1708;
		} else {
			if (yych <= '9') goto yy1812;
			if (yych <= ':') goto yy1813;
			if (yych <= '@') goto yy1708;
			goto yy1818;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy1814;
			if (yych <= 'Z') goto yy1815;
			goto yy1708;
		} else {
			if (yych <= 'a') goto yy1819;
			if (yych <= 'f') goto yy1816;
			if (yych <= 'z') goto yy1817;
			goto yy1708;
		}
	}
yy1700:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych == '\'') goto yy1808;
		if (yych <= '9') goto yy1821;
		goto yy1822;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1821;
			goto yy1815;
		} else {
			if (yych <= '`') goto yy1821;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1701:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '\'') {
			if (yych == '%') goto yy1821;
			if (yych <= '&') goto yy1696;
			goto yy1808;
		} else {
			if (yych == '+') goto yy1821;
			if (yych <= ',') goto yy1696;
			goto yy1821;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1822;
			if (yych <= '?') goto yy1696;
			if (yych <= '@') goto yy1821;
			goto yy1815;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1696;
				goto yy1821;
			} else {
				if (yych <= '`') goto yy1696;
				if (yych <= 'z') goto yy1817;
				goto yy1696;
			}
		}
	}
yy1702:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'T') {
		if (yych <= '9') {
			if (yych == '\'') goto yy1808;
			goto yy1821;
		} else {
			if (yych <= ':') goto yy1822;
			if (yych <= '@') goto yy1821;
			goto yy1815;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'U') goto yy1823;
			if (yych <= 'Z') goto yy1815;
			goto yy1821;
		} else {
			if (yych == 'o') goto yy1824;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1703:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Q') {
		if (yych <= '9') {
			if (yych == '\'') goto yy1808;
			goto yy1821;
		} else {
			if (yych <= ':') goto yy1822;
			if (yych <= '@') goto yy1821;
			goto yy1815;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'R') goto yy1825;
			if (yych <= 'Z') goto yy1815;
			goto yy1821;
		} else {
			if (yych == 'r') goto yy1826;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1704:
	yych = *++YYCURSOR;
	if (yych == ']') goto yy1758;
#line 393 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LSQUARE); }
#line 45182 "../../lnav/src/data_scanner_re.cc"
yy1705:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '[') {
			if (yych == '\n') goto yy1662;
			goto yy1827;
		} else {
			if (yych <= '\\') goto yy1768;
			if (yych == 'n') goto yy1665;
			goto yy1827;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1662;
			if (yych <= 0xDF) goto yy1828;
			if (yych <= 0xE0) goto yy1829;
			goto yy1830;
		} else {
			if (yych <= 0xF0) goto yy1831;
			if (yych <= 0xF3) goto yy1832;
			if (yych <= 0xF4) goto yy1833;
			goto yy1662;
		}
	}
yy1706:
	++YYCURSOR;
#line 394 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RSQUARE); }
#line 45212 "../../lnav/src/data_scanner_re.cc"
yy1707:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy1708:
	if (yybm_init[1536+yych] & 64) {
		goto yy1707;
	}
	switch (yych) {
		case '%':
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1811;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
		case ':': goto yy1834;
		case '@': goto yy1785;
		default: goto yy1698;
	}
yy1709:
	yyaccept = 6;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1696;
				goto yy1708;
			} else {
				if (yych == '\'') goto yy1808;
				goto yy1696;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy1708;
				goto yy1696;
			} else {
				if (yych <= '/') goto yy1708;
				if (yych <= '9') goto yy1812;
				goto yy1835;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1696;
				goto yy1708;
			} else {
				if (yych <= 'F') goto yy1814;
				if (yych <= 'Z') goto yy1815;
				goto yy1696;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1708;
				goto yy1696;
			} else {
				if (yych <= 'f') goto yy1816;
				if (yych <= 'z') goto yy1817;
				goto yy1696;
			}
		}
	}
yy1710:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '\'') goto yy1808;
			goto yy1708;
		} else {
			if (yych <= '9') goto yy1812;
			if (yych <= ':') goto yy1835;
			goto yy1708;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy1814;
			if (yych <= 'Z') goto yy1815;
			goto yy1708;
		} else {
			if (yych <= 'f') goto yy1816;
			if (yych <= 'z') goto yy1817;
			goto yy1708;
		}
	}
yy1711:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '&') {
			if (yych == '"') goto yy1836;
			goto yy1708;
		} else {
			if (yych <= '\'') goto yy1837;
			if (yych <= '/') goto yy1708;
			if (yych <= '9') goto yy1812;
			goto yy1835;
		}
	} else {
		if (yych <= '`') {
			if (yych <= '@') goto yy1708;
			if (yych <= 'F') goto yy1814;
			if (yych <= 'Z') goto yy1815;
			goto yy1708;
		} else {
			if (yych <= 'a') goto yy1819;
			if (yych <= 'f') goto yy1816;
			if (yych <= 'z') goto yy1817;
			goto yy1708;
		}
	}
yy1712:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych == '\'') goto yy1808;
		goto yy1821;
	} else {
		if (yych <= 'Z') goto yy1815;
		if (yych <= '`') goto yy1821;
		if (yych <= 'z') goto yy1817;
		goto yy1821;
	}
yy1713:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych == '\'') goto yy1808;
		if (yych <= '@') goto yy1821;
		goto yy1815;
	} else {
		if (yych <= 't') {
			if (yych <= '`') goto yy1821;
			goto yy1817;
		} else {
			if (yych <= 'u') goto yy1839;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1714:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '"') {
			if (yych <= '!') goto yy1821;
			goto yy1836;
		} else {
			if (yych == '\'') goto yy1837;
			goto yy1821;
		}
	} else {
		if (yych <= 'd') {
			if (yych <= 'Z') goto yy1815;
			if (yych <= '`') goto yy1821;
			goto yy1817;
		} else {
			if (yych <= 'e') goto yy1840;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1715:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych == '\'') goto yy1808;
		if (yych <= '@') goto yy1821;
		goto yy1815;
	} else {
		if (yych <= 'q') {
			if (yych <= '`') goto yy1821;
			goto yy1817;
		} else {
			if (yych <= 'r') goto yy1826;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1716:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '\'') {
		if (yych == '"') goto yy1836;
		if (yych <= '&') goto yy1821;
		goto yy1837;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1821;
			goto yy1815;
		} else {
			if (yych <= '`') goto yy1821;
			if (yych <= 'z') goto yy1817;
			goto yy1821;
		}
	}
yy1717:
	yych = *++YYCURSOR;
	if (yych == '}') goto yy1758;
#line 391 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_LCURLY); }
#line 45422 "../../lnav/src/data_scanner_re.cc"
yy1718:
	++YYCURSOR;
#line 392 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_RCURLY); }
#line 45427 "../../lnav/src/data_scanner_re.cc"
yy1719:
	++YYCURSOR;
yy1720:
#line 175 "../../lnav/src/data_scanner_re.re"
	{ return std::nullopt; }
#line 45433 "../../lnav/src/data_scanner_re.cc"
yy1721:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1720;
	if (yych <= 0xBF) goto yy1661;
	goto yy1720;
yy1722:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x9F) goto yy1720;
	if (yych <= 0xBF) goto yy1841;
	goto yy1720;
yy1723:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy1720;
	if (yych <= 0xBF) goto yy1841;
	goto yy1720;
yy1724:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x8F) goto yy1720;
	if (yych <= 0xBF) goto yy1842;
	goto yy1720;
yy1725:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy1720;
	if (yych <= 0xBF) goto yy1842;
	goto yy1720;
yy1726:
	yyaccept = 8;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) goto yy1720;
	if (yych <= 0x8F) goto yy1842;
	goto yy1720;
yy1727:
	yych = *++YYCURSOR;
yy1728:
	if (yybm_init[1536+yych] & 128) {
		goto yy1727;
	}
	goto yy1664;
yy1729:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1843;
yy1730:
	YYCURSOR = YYMARKER;
	if (yyaccept <= 22) {
		if (yyaccept <= 11) {
			if (yyaccept <= 5) {
				if (yyaccept <= 2) {
					if (yyaccept <= 1) {
						if (yyaccept == 0) {
							goto yy1664;
						} else {
							goto yy1662;
						}
					} else {
						goto yy1679;
					}
				} else {
					if (yyaccept <= 4) {
						if (yyaccept == 3) {
							goto yy1683;
						} else {
							goto yy1688;
						}
					} else {
						goto yy1691;
					}
				}
			} else {
				if (yyaccept <= 8) {
					if (yyaccept <= 7) {
						if (yyaccept == 6) {
							goto yy1696;
						} else {
							goto yy1698;
						}
					} else {
						goto yy1720;
					}
				} else {
					if (yyaccept <= 10) {
						if (yyaccept == 9) {
							goto yy1681;
						} else {
							goto yy1776;
						}
					} else {
						goto yy1781;
					}
				}
			}
		} else {
			if (yyaccept <= 17) {
				if (yyaccept <= 14) {
					if (yyaccept <= 13) {
						if (yyaccept == 12) {
							goto yy1783;
						} else {
							goto yy1789;
						}
					} else {
						yyt3 = YYCURSOR;
						goto yy1810;
					}
				} else {
					if (yyaccept <= 16) {
						if (yyaccept == 15) {
							goto yy1736;
						} else {
							goto yy1848;
						}
					} else {
						goto yy1856;
					}
				}
			} else {
				if (yyaccept <= 20) {
					if (yyaccept <= 19) {
						if (yyaccept == 18) {
							goto yy1865;
						} else {
							goto yy1877;
						}
					} else {
						goto yy1867;
					}
				} else {
					if (yyaccept == 21) {
						goto yy1909;
					} else {
						goto yy1810;
					}
				}
			}
		}
	} else {
		if (yyaccept <= 34) {
			if (yyaccept <= 28) {
				if (yyaccept <= 25) {
					if (yyaccept <= 24) {
						if (yyaccept == 23) {
							goto yy1854;
						} else {
							goto yy1941;
						}
					} else {
						yyt2 = yyt1;
						goto yy1941;
					}
				} else {
					if (yyaccept <= 27) {
						if (yyaccept == 26) {
							yyt1 = YYCURSOR;
							goto yy2116;
						} else {
							goto yy2128;
						}
					} else {
						goto yy2015;
					}
				}
			} else {
				if (yyaccept <= 31) {
					if (yyaccept <= 30) {
						if (yyaccept == 29) {
							goto yy2143;
						} else {
							goto yy2155;
						}
					} else {
						goto yy2171;
					}
				} else {
					if (yyaccept <= 33) {
						if (yyaccept == 32) {
							goto yy2082;
						} else {
							goto yy2116;
						}
					} else {
						goto yy2285;
					}
				}
			}
		} else {
			if (yyaccept <= 40) {
				if (yyaccept <= 37) {
					if (yyaccept <= 36) {
						if (yyaccept == 35) {
							goto yy1961;
						} else {
							yyt2 = yyt4;
							goto yy1941;
						}
					} else {
						goto yy2496;
					}
				} else {
					if (yyaccept <= 39) {
						if (yyaccept == 38) {
							goto yy2087;
						} else {
							goto yy2586;
						}
					} else {
						goto yy2596;
					}
				}
			} else {
				if (yyaccept <= 43) {
					if (yyaccept <= 42) {
						if (yyaccept == 41) {
							goto yy2618;
						} else {
							goto yy2975;
						}
					} else {
						goto yy3009;
					}
				} else {
					if (yyaccept == 44) {
						goto yy3069;
					} else {
						goto yy3152;
					}
				}
			}
		}
	}
yy1731:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy1844;
	goto yy1730;
yy1732:
	yych = *++YYCURSOR;
	if (yybm_init[1280+yych] & 1) {
		goto yy1732;
	}
	if (yych <= '@') goto yy1730;
	if (yych <= 'Z') goto yy1845;
	if (yych <= '`') goto yy1730;
	if (yych <= 'z') goto yy1845;
	goto yy1730;
yy1733:
	yych = *++YYCURSOR;
yy1734:
	if (yybm_init[1280+yych] & 2) {
		goto yy1733;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x1B) goto yy1730;
			goto yy1846;
		} else {
			if (yych <= '\\') goto yy1737;
			if (yych <= 0xC1) goto yy1730;
			goto yy1738;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1739;
			if (yych <= 0xEF) goto yy1740;
			goto yy1741;
		} else {
			if (yych <= 0xF3) goto yy1742;
			if (yych <= 0xF4) goto yy1743;
			goto yy1730;
		}
	}
yy1735:
	yych = *++YYCURSOR;
	if (yych == '"') goto yy1847;
yy1736:
#line 180 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 45725 "../../lnav/src/data_scanner_re.cc"
yy1737:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy1733;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy1739;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1740;
			goto yy1741;
		} else {
			if (yych <= 0xF3) goto yy1742;
			if (yych <= 0xF4) goto yy1743;
			goto yy1730;
		}
	}
yy1738:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1733;
	goto yy1730;
yy1739:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1738;
	goto yy1730;
yy1740:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1738;
	goto yy1730;
yy1741:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1740;
	goto yy1730;
yy1742:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1740;
	goto yy1730;
yy1743:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1740;
	goto yy1730;
yy1744:
	yych = *++YYCURSOR;
	if (yybm_init[1280+yych] & 4) {
		goto yy1744;
	}
	if (yych == '[') goto yy1849;
	goto yy1730;
yy1745:
	yych = *++YYCURSOR;
yy1746:
	if (yybm_init[1280+yych] & 8) {
		goto yy1745;
	}
	if (yych != '@') goto yy1730;
yy1747:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy1850;
	goto yy1851;
yy1748:
	yych = *++YYCURSOR;
yy1749:
	if (yybm_init[1280+yych] & 16) {
		goto yy1748;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy1730;
			goto yy1852;
		} else {
			if (yych <= '\\') goto yy1751;
			if (yych <= 0xC1) goto yy1730;
			goto yy1752;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1753;
			if (yych <= 0xEF) goto yy1754;
			goto yy1755;
		} else {
			if (yych <= 0xF3) goto yy1756;
			if (yych <= 0xF4) goto yy1757;
			goto yy1730;
		}
	}
yy1750:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= 'R') {
			if (yych == '\'') goto yy1855;
			yyt2 = YYCURSOR;
			goto yy1853;
		} else {
			if (yych <= 'S') goto yy1730;
			if (yych == 's') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1853;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1857;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1858;
			}
			yyt2 = YYCURSOR;
			goto yy1859;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1860;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1861;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1862;
			}
			goto yy1730;
		}
	}
yy1751:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy1748;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy1753;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1754;
			goto yy1755;
		} else {
			if (yych <= 0xF3) goto yy1756;
			if (yych <= 0xF4) goto yy1757;
			goto yy1730;
		}
	}
yy1752:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1748;
	goto yy1730;
yy1753:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1752;
	goto yy1730;
yy1754:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1752;
	goto yy1730;
yy1755:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1754;
	goto yy1730;
yy1756:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1754;
	goto yy1730;
yy1757:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1754;
	goto yy1730;
yy1758:
	++YYCURSOR;
#line 390 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMPTY_CONTAINER); }
#line 45916 "../../lnav/src/data_scanner_re.cc"
yy1759:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1730;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy1730;
			goto yy1863;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1730;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1863;
			goto yy1730;
		} else {
			if (yych <= '_') goto yy1745;
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1863;
			goto yy1730;
		}
	}
yy1760:
	yych = *++YYCURSOR;
	if (yybm_init[1280+yych] & 32) {
		goto yy1762;
	}
	if (yych <= '\r') {
		if (yych <= 0x08) {
			if (yych <= 0x00) goto yy1866;
			goto yy1746;
		} else {
			if (yych <= '\n') goto yy1866;
			if (yych <= '\f') goto yy1746;
			goto yy1866;
		}
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1746;
			goto yy1866;
		} else {
			if (yych == '.') goto yy1868;
			goto yy1746;
		}
	}
yy1761:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1868;
	goto yy1870;
yy1762:
	yych = *++YYCURSOR;
yy1763:
	if (yybm_init[1280+yych] & 32) {
		goto yy1762;
	}
	if (yych <= '%') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy1866;
			if (yych <= 0x08) goto yy1730;
			if (yych <= '\n') goto yy1866;
			goto yy1730;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy1866;
				goto yy1730;
			} else {
				if (yych <= ' ') goto yy1866;
				if (yych <= '$') goto yy1730;
				goto yy1745;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych == '+') goto yy1745;
			if (yych <= ',') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= '@') {
				if (yych <= '?') goto yy1730;
				goto yy1747;
			} else {
				if (yych == '_') goto yy1745;
				goto yy1730;
			}
		}
	}
yy1764:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy1878;
				}
				yyt1 = YYCURSOR;
				goto yy1876;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy1878;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy1876;
				}
				goto yy1745;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy1878;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy1876;
				}
				yyt1 = YYCURSOR;
				goto yy1879;
			} else {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy1876;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy1880;
				}
				yyt1 = YYCURSOR;
				goto yy1879;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy1878;
				}
				yyt1 = YYCURSOR;
				goto yy1876;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy1879;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy1876;
				}
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy1881;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy1882;
				}
				yyt1 = YYCURSOR;
				goto yy1883;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy1884;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy1885;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy1886;
				}
				goto yy1730;
			}
		}
	}
yy1765:
	yych = *++YYCURSOR;
	if (yybm_init[1024+yych] & 32) {
		goto yy1889;
	}
	if (yych <= '.') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1866;
				goto yy1746;
			} else {
				if (yych <= '\n') goto yy1866;
				if (yych <= '\f') goto yy1746;
				goto yy1866;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy1887;
				goto yy1746;
			} else {
				if (yych <= '%') goto yy1775;
				if (yych <= '-') goto yy1746;
				goto yy1888;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') {
				if (yych <= '/') goto yy1746;
				goto yy1890;
			} else {
				if (yych <= '@') goto yy1746;
				if (yych <= 'F') goto yy1891;
				goto yy1762;
			}
		} else {
			if (yych <= 'w') {
				if (yych <= '`') goto yy1746;
				if (yych <= 'f') goto yy1891;
				goto yy1762;
			} else {
				if (yych <= 'x') goto yy1893;
				if (yych <= 'z') goto yy1762;
				goto yy1746;
			}
		}
	}
yy1766:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy1888;
	goto yy1895;
yy1767:
	yych = *++YYCURSOR;
	if (yych != '/') goto yy1746;
yy1768:
	yych = *++YYCURSOR;
	if (yych == 'P') goto yy1772;
	goto yy1770;
yy1769:
	yych = *++YYCURSOR;
yy1770:
	if (yybm_init[1280+yych] & 64) {
		goto yy1769;
	}
	goto yy1681;
yy1771:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1280+yych] & 128) {
		goto yy1771;
	}
	if (yych <= 0xDF) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1681;
			if (yych <= '\'') goto yy1898;
			goto yy1899;
		} else {
			if (yych <= 0x7F) goto yy1898;
			if (yych <= 0xC1) goto yy1681;
			goto yy1900;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1901;
			if (yych <= 0xEF) goto yy1902;
			goto yy1903;
		} else {
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy1681;
		}
	}
yy1772:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy1906;
	goto yy1770;
yy1773:
	yych = *++YYCURSOR;
yy1774:
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych == '%') goto yy1907;
	goto yy1730;
yy1775:
	yyaccept = 10;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
		} else {
			if (yych == ',') goto yy1776;
			if (yych <= '.') goto yy1746;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1776;
			if (yych <= 'Z') goto yy1746;
		} else {
			if (yych == '`') goto yy1776;
			if (yych <= 'z') goto yy1746;
		}
	}
yy1776:
#line 443 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_PERCENTAGE); }
#line 46232 "../../lnav/src/data_scanner_re.cc"
yy1777:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1730;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy1730;
			goto yy1793;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1730;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy1730;
		} else {
			if (yych == '`') goto yy1730;
			if (yych <= 'z') goto yy1793;
			goto yy1730;
		}
	}
yy1778:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1746;
			if (yych <= '0') goto yy1908;
			goto yy1910;
		} else {
			if (yych <= '2') goto yy1911;
			if (yych <= '9') goto yy1908;
			goto yy1746;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1793;
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy1779:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy1912;
		if (yych <= '@') goto yy1730;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy1780:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych >= '%') goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1914;
				goto yy1915;
			} else {
				if (yych <= '/') goto yy1916;
				if (yych <= '7') goto yy1917;
				goto yy1918;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1919;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy1920;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
			} else {
				if (yych <= 'f') goto yy1920;
				if (yych <= 'z') goto yy1790;
			}
		}
	}
yy1781:
#line 442 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_OCTAL_NUMBER); }
#line 46342 "../../lnav/src/data_scanner_re.cc"
yy1782:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych >= '%') goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1914;
				goto yy1915;
			} else {
				if (yych <= '/') goto yy1916;
				if (yych <= '9') goto yy1918;
				goto yy1919;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych >= '@') goto yy1785;
			} else {
				if (yych <= 'F') goto yy1920;
				if (yych <= 'Z') goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
			} else {
				if (yych <= 'f') goto yy1920;
				if (yych <= 'z') goto yy1790;
			}
		}
	}
yy1783:
#line 458 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_HEX_NUMBER); }
#line 46386 "../../lnav/src/data_scanner_re.cc"
yy1784:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy1922;
		goto yy1923;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy1785:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1730;
			goto yy1850;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy1925;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1925;
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1925;
			goto yy1730;
		}
	}
yy1786:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy1914;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy1920;
				goto yy1926;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy1920;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy1920;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy1787:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1683;
				goto yy1914;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy1920;
				goto yy1926;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy1920;
				if (yych <= 'Z') goto yy1791;
				goto yy1683;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy1920;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy1788:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy1792;
		goto yy1791;
	} else {
		if (yych <= 'b') goto yy1792;
		if (yych == 'i') goto yy1928;
		goto yy1791;
	}
yy1789:
#line 480 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_ID);
       }
#line 46520 "../../lnav/src/data_scanner_re.cc"
yy1790:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy1791:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1789;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1789;
			if (yych <= '.') goto yy1777;
			if (yych <= '/') goto yy1779;
			goto yy1790;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1927;
			if (yych <= '?') goto yy1789;
			if (yych <= '@') goto yy1785;
			goto yy1790;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1789;
				goto yy1793;
			} else {
				if (yych <= '`') goto yy1789;
				if (yych <= 'z') goto yy1790;
				goto yy1789;
			}
		}
	}
yy1792:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1791;
			goto yy1683;
		} else {
			if (yych == ',') goto yy1683;
			goto yy1791;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1791;
			goto yy1683;
		}
	}
yy1793:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy1794:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1789;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1789;
			if (yych <= '.') goto yy1777;
			if (yych <= '/') goto yy1779;
			goto yy1793;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy1789;
			if (yych <= '@') goto yy1785;
			goto yy1793;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1789;
				goto yy1793;
			} else {
				if (yych <= '`') goto yy1789;
				if (yych <= 'z') goto yy1793;
				goto yy1789;
			}
		}
	}
yy1795:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1683;
				goto yy1791;
			} else {
				if (yych == '+') goto yy1791;
				goto yy1683;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy1791;
				if (yych <= '?') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= 'B') goto yy1792;
				if (yych <= 'R') goto yy1791;
				goto yy1792;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1791;
				if (yych <= '^') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '`') goto yy1683;
				if (yych <= 'a') goto yy1791;
				goto yy1792;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy1930;
				goto yy1791;
			} else {
				if (yych <= 's') goto yy1792;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy1796:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy1792;
	if (yych == 's') goto yy1792;
	goto yy1791;
yy1797:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1932;
			goto yy1789;
		} else {
			if (yych == ',') goto yy1789;
			goto yy1932;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1789;
			if (yych <= 'Z') goto yy1932;
			goto yy1789;
		} else {
			if (yych == '`') goto yy1789;
			if (yych <= 'z') goto yy1932;
			goto yy1789;
		}
	}
yy1798:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1914;
		case '.': goto yy1933;
		case '/': goto yy1916;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1934;
		case ':': goto yy1919;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1920;
		case 'B':
		case 'b': goto yy1935;
		case 'E':
		case 'e': goto yy1936;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1799:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1937;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy1938;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy1920;
				goto yy1926;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy1920;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy1920;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy1800:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1914;
		case '.': goto yy1933;
		case '/': goto yy1916;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5': goto yy1934;
		case '6':
		case '7':
		case '8':
		case '9': goto yy1939;
		case ':': goto yy1919;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1920;
		case 'B':
		case 'b': goto yy1935;
		case 'E':
		case 'e': goto yy1936;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1801:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1914;
		case '.': goto yy1933;
		case '/': goto yy1916;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1939;
		case ':': goto yy1919;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1920;
		case 'B':
		case 'b': goto yy1935;
		case 'E':
		case 'e': goto yy1936;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1802:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1942;
			} else {
				if (yych <= '1') goto yy1943;
				if (yych <= '2') goto yy1944;
				goto yy1942;
			}
		} else {
			if (yych <= 'E') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1946;
			} else {
				if (yych <= 'F') goto yy1947;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'e') goto yy1946;
				if (yych <= 'f') goto yy1947;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1803:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy1954;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= ':') goto yy1955;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1955;
			if (yych <= '^') goto yy1730;
			goto yy1955;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1955;
			goto yy1730;
		}
	}
yy1804:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 1) {
		goto yy1958;
	}
	goto yy1957;
yy1805:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy1730;
		if (yych <= '\f') goto yy1963;
		goto yy1730;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1963;
			goto yy1730;
		} else {
			if (yych == '>') goto yy1730;
			goto yy1963;
		}
	}
yy1806:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy1964;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= ':') goto yy1964;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1964;
			if (yych <= '^') goto yy1730;
			goto yy1964;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1964;
			goto yy1730;
		}
	}
yy1807:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1965;
	goto yy1730;
yy1808:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
yy1809:
	if (yybm_init[1024+yych] & 2) {
		goto yy1808;
	}
	if (yych <= '&') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				yyt3 = YYCURSOR;
				goto yy1810;
			}
			if (yych <= '\n') {
				yyt3 = YYCURSOR;
				goto yy1966;
			}
			if (yych >= '\r') {
				yyt3 = YYCURSOR;
				goto yy1966;
			}
			yyt3 = YYCURSOR;
		} else {
			if (yych <= ' ') {
				if (yych >= ' ') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych <= '!') {
					yyt3 = YYCURSOR;
					goto yy1967;
				}
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '+') {
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy1967;
				}
				if (yych >= '.') {
					yyt3 = YYCURSOR;
					goto yy1968;
				}
				yyt3 = YYCURSOR;
			}
		} else {
			if (yych <= ';') {
				if (yych >= ':') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
			} else {
				if (yych == '?') {
					yyt3 = YYCURSOR;
					goto yy1967;
				}
				yyt3 = YYCURSOR;
			}
		}
	}
yy1810:
	YYCURSOR = yyt3;
#line 464 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_WORD); }
#line 47178 "../../lnav/src/data_scanner_re.cc"
yy1811:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1730;
			if (yych <= '-') goto yy1969;
			if (yych <= '.') goto yy1745;
			goto yy1730;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1970;
			if (yych <= '?') goto yy1730;
			if (yych <= '@') goto yy1747;
			goto yy1970;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1730;
				goto yy1970;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy1970;
				goto yy1730;
			}
		}
	}
yy1812:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1821;
			goto yy1914;
		} else {
			if (yych <= '/') goto yy1821;
			if (yych <= '9') goto yy1971;
			goto yy1972;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1821;
			goto yy1971;
		} else {
			if (yych <= '`') goto yy1821;
			if (yych <= 'f') goto yy1971;
			goto yy1821;
		}
	}
yy1813:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych <= '.') goto yy1730;
			goto yy1921;
		} else {
			if (yych <= '9') goto yy1924;
			if (yych <= ':') goto yy1973;
			goto yy1730;
		}
	} else {
		if (yych <= '\\') {
			if (yych <= 'F') goto yy1924;
			if (yych <= '[') goto yy1730;
			goto yy1768;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy1814:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych == '-') goto yy1914;
			goto yy1708;
		} else {
			if (yych <= '9') goto yy1971;
			if (yych <= ':') goto yy1972;
			goto yy1708;
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'F') goto yy1974;
			if (yych <= 'Z') goto yy1975;
			goto yy1708;
		} else {
			if (yych <= 'f') goto yy1974;
			if (yych <= 'z') goto yy1975;
			goto yy1708;
		}
	}
yy1815:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy1821;
	if (yych <= 'Z') goto yy1975;
	if (yych <= '`') goto yy1821;
	if (yych <= 'z') goto yy1975;
	goto yy1821;
yy1816:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1914;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1971;
				yyt3 = YYCURSOR;
				goto yy1977;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy1974;
				if (yych <= 'Z') goto yy1975;
				goto yy1809;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy1978;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1817:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1809;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1809;
				goto yy1777;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy1980;
			}
			if (yych <= '?') goto yy1809;
			if (yych <= '@') goto yy1785;
			goto yy1975;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1809;
				goto yy1707;
			} else {
				if (yych <= '`') goto yy1809;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1818:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= '/') {
			if (yych == '-') goto yy1914;
			goto yy1708;
		} else {
			if (yych <= '9') goto yy1971;
			if (yych <= ':') goto yy1972;
			if (yych <= '@') goto yy1708;
			goto yy1974;
		}
	} else {
		if (yych <= 'Z') {
			if (yych == 'L') goto yy1981;
			goto yy1975;
		} else {
			if (yych <= '`') goto yy1708;
			if (yych <= 'f') goto yy1974;
			if (yych <= 'z') goto yy1975;
			goto yy1708;
		}
	}
yy1819:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '/') {
				if (yych <= '-') goto yy1914;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				goto yy1779;
			} else {
				if (yych <= '9') goto yy1971;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy1977;
				}
				goto yy1809;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'F') {
				if (yych <= '@') goto yy1785;
				goto yy1974;
			} else {
				if (yych <= 'Z') goto yy1975;
				if (yych <= '^') goto yy1809;
				goto yy1707;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '`') goto yy1809;
				if (yych <= 'f') goto yy1978;
				goto yy1979;
			} else {
				if (yych <= 'l') goto yy1982;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1820:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy1821:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1698;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1698;
				goto yy1777;
			} else {
				if (yych <= '.') goto yy1811;
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1838;
			if (yych <= '?') goto yy1698;
			if (yych <= '@') goto yy1785;
			goto yy1820;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1698;
				goto yy1707;
			} else {
				if (yych <= '`') goto yy1698;
				if (yych <= 'z') goto yy1820;
				goto yy1698;
			}
		}
	}
yy1822:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == '/') goto yy1921;
		goto yy1730;
	} else {
		if (yych <= ':') goto yy1983;
		if (yych == '\\') goto yy1768;
		goto yy1730;
	}
yy1823:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'L') {
		if (yych <= '@') goto yy1821;
		if (yych <= 'K') goto yy1975;
		goto yy1984;
	} else {
		if (yych <= 'Z') goto yy1975;
		if (yych <= '`') goto yy1821;
		if (yych <= 'z') goto yy1975;
		goto yy1821;
	}
yy1824:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1820;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'Z') goto yy1975;
				if (yych <= '^') goto yy1809;
				goto yy1707;
			}
		} else {
			if (yych <= 'm') {
				if (yych <= '`') goto yy1809;
				goto yy1979;
			} else {
				if (yych <= 'n') goto yy1985;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1825:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'U') {
		if (yych <= '@') goto yy1821;
		if (yych <= 'T') goto yy1975;
		goto yy1986;
	} else {
		if (yych <= 'Z') goto yy1975;
		if (yych <= '`') goto yy1821;
		if (yych <= 'z') goto yy1975;
		goto yy1821;
	}
yy1826:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1820;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'Z') goto yy1975;
				if (yych <= '^') goto yy1809;
				goto yy1707;
			}
		} else {
			if (yych <= 't') {
				if (yych <= '`') goto yy1809;
				goto yy1979;
			} else {
				if (yych <= 'u') goto yy1985;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1827:
	++YYCURSOR;
#line 490 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_ESCAPED_CHAR); }
#line 47611 "../../lnav/src/data_scanner_re.cc"
yy1828:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1827;
	goto yy1730;
yy1829:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1828;
	goto yy1730;
yy1830:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1828;
	goto yy1730;
yy1831:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1830;
	goto yy1730;
yy1832:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1830;
	goto yy1730;
yy1833:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1830;
	goto yy1730;
yy1834:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1983;
	goto yy1730;
yy1835:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy1924;
		goto yy1973;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy1836:
	yych = *++YYCURSOR;
	if (yych == '"') goto yy1735;
	goto yy1734;
yy1837:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 4) {
		goto yy1991;
	}
	if (yych <= '"') {
		if (yych <= 0x15) {
			if (yych <= '\t') {
				if (yych <= 0x00) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0x08) goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1987;
			} else {
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych == '\r') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				goto yy1749;
			}
		} else {
			if (yych <= 0x1B) {
				if (yych <= 0x16) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0x1A) goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= 0x1F) goto yy1749;
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				yyt3 = YYCURSOR;
				goto yy1987;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych <= '&') goto yy1749;
				if (yych <= '\'') goto yy1989;
				yyt3 = YYCURSOR;
				goto yy1987;
			} else {
				if (yych == ',') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				if (yych <= '-') goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1990;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1749;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				if (yych <= '>') goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1988;
			} else {
				if (yych <= 0x7F) goto yy1749;
				if (yych <= 0xC1) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0xF4) goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	}
yy1838:
	yych = *++YYCURSOR;
	if (yych == '/') goto yy1921;
	if (yych == ':') goto yy1983;
	goto yy1730;
yy1839:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1820;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'Z') goto yy1975;
				if (yych <= '^') goto yy1809;
				goto yy1707;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '`') goto yy1809;
				goto yy1979;
			} else {
				if (yych <= 'l') goto yy1992;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1840:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1809;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1809;
				goto yy1993;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy1980;
			}
			if (yych <= '?') goto yy1809;
			if (yych <= '@') goto yy1785;
			goto yy1975;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1809;
				goto yy1707;
			} else {
				if (yych <= '`') goto yy1809;
				if (yych <= 'z') goto yy1979;
				goto yy1809;
			}
		}
	}
yy1841:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1661;
	goto yy1730;
yy1842:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1841;
	goto yy1730;
yy1843:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy1994;
	goto yy1730;
yy1844:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy1995;
	goto yy1730;
yy1845:
	++YYCURSOR;
#line 381 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_CSI);
       }
#line 47868 "../../lnav/src/data_scanner_re.cc"
yy1846:
	yyaccept = 15;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '"') goto yy1733;
	goto yy1736;
yy1847:
	yyaccept = 16;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych >= 0x01) goto yy1734;
		} else {
			if (yych <= '\n') goto yy1848;
			if (yych <= 0x15) goto yy1734;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych != 0x1B) goto yy1734;
		} else {
			if (yych <= 0xC1) goto yy1848;
			if (yych <= 0xF4) goto yy1734;
		}
	}
yy1848:
#line 193 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 3;
           goto yyc_dbldocstring;
       }
#line 47906 "../../lnav/src/data_scanner_re.cc"
yy1849:
	yych = *++YYCURSOR;
	if (yybm_init[1024+yych] & 4) {
		goto yy1849;
	}
	if (yych <= 0xE0) {
		if (yych <= ']') {
			if (yych <= 0x00) goto yy1730;
			goto yy1996;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy1997;
			goto yy1998;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1999;
			goto yy2000;
		} else {
			if (yych <= 0xF3) goto yy2001;
			if (yych <= 0xF4) goto yy2002;
			goto yy1730;
		}
	}
yy1850:
	yych = *++YYCURSOR;
yy1851:
	if (yybm_init[1024+yych] & 8) {
		goto yy1850;
	}
	if (yych <= ',') goto yy1730;
	if (yych <= '.') goto yy2003;
	goto yy1730;
yy1852:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= 'R') {
			if (yych == '\'') {
				yyt1 = YYCURSOR;
				goto yy2004;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'S') goto yy1730;
			if (yych == 's') goto yy1730;
			yyt2 = YYCURSOR;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1857;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1858;
			}
			yyt2 = YYCURSOR;
			goto yy1859;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1860;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1861;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1862;
			}
			goto yy1730;
		}
	}
yy1853:
	++YYCURSOR;
	yyt1 = yyt2;
yy1854:
	YYCURSOR = yyt1;
#line 278 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           if (tf == text_format_t::TF_RUST) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when([](char ch) { return ch != '\'' && !isalnum(ch); });
               cap_all.c_end = split_res.first.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = split_res.first.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               return tokenize_result{DT_SYMBOL, cap_all, cap_inner, this->ds_input.data()};
           }
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 1;
           cap_inner.c_end -= 1;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 48010 "../../lnav/src/data_scanner_re.cc"
yy1855:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych >= 0x01) goto yy1749;
		} else {
			if (yych <= '\n') goto yy1856;
			if (yych <= 0x15) goto yy1749;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych != 0x1B) goto yy1749;
		} else {
			if (yych <= 0xC1) goto yy1856;
			if (yych <= 0xF4) goto yy1749;
		}
	}
yy1856:
#line 221 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           switch (this->ds_input[cap_inner.c_begin]) {
           case 'f':
           case 'u':
           case 'r':
               cap_inner.c_begin += 1;
               break;
           }
           cap_inner.c_begin += 3;
           goto yyc_sdocstring;
       }
#line 48043 "../../lnav/src/data_scanner_re.cc"
yy1857:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1853;
	goto yy1730;
yy1858:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1857;
	goto yy1730;
yy1859:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1857;
	goto yy1730;
yy1860:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1859;
	goto yy1730;
yy1861:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1859;
	goto yy1730;
yy1862:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1859;
	goto yy1730;
yy1863:
	yyaccept = 18;
	yych = *(YYMARKER = ++YYCURSOR);
yy1864:
	if (yych <= '/') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych >= '+') goto yy1745;
		} else {
			if (yych <= ',') goto yy1865;
			if (yych <= '-') goto yy1759;
			if (yych <= '.') goto yy1745;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy1863;
			if (yych <= '?') goto yy1865;
			if (yych <= '@') goto yy1747;
			goto yy1863;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy1745;
			} else {
				if (yych <= '`') goto yy1865;
				if (yych <= 'z') goto yy1863;
			}
		}
	}
yy1865:
#line 468 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 48107 "../../lnav/src/data_scanner_re.cc"
yy1866:
	++YYCURSOR;
yy1867:
	YYCURSOR -= 1;
#line 472 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_SYMBOL);
       }
#line 48116 "../../lnav/src/data_scanner_re.cc"
yy1868:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2005;
	goto yy1746;
yy1869:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy1870:
	if (yybm_init[1024+yych] & 16) {
		goto yy1869;
	}
	switch (yych) {
		case 0x00:
		case '\t':
		case '\n':
		case '\r':
		case ' ': goto yy1866;
		case '%':
		case '+':
		case '-':
		case '_': goto yy1745;
		case '.': goto yy2006;
		case '@': goto yy1747;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1762;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy1871;
		case 'E':
		case 'e': goto yy1872;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1873;
		case 'm': goto yy1874;
		case 'p':
		case 'u': goto yy1875;
		default: goto yy1683;
	}
yy1871:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1763;
				goto yy1683;
			} else {
				if (yych <= '\n') goto yy1763;
				if (yych <= '\f') goto yy1683;
				goto yy1763;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy1763;
				goto yy1683;
			} else {
				if (yych <= '%') goto yy1763;
				if (yych <= '*') goto yy1683;
				goto yy1763;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '.') {
				if (yych <= ',') goto yy1683;
				goto yy1763;
			} else {
				if (yych <= '/') goto yy1683;
				if (yych <= '9') goto yy1763;
				goto yy1683;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1763;
				if (yych <= '^') goto yy1683;
				goto yy1763;
			} else {
				if (yych <= '`') goto yy1683;
				if (yych <= 'z') goto yy1763;
				goto yy1683;
			}
		}
	}
yy1872:
	yych = *++YYCURSOR;
	if (yybm_init[1280+yych] & 32) {
		goto yy1762;
	}
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1866;
			if (yych <= 0x08) goto yy1746;
			goto yy1866;
		} else {
			if (yych == '\r') goto yy1866;
			goto yy1746;
		}
	} else {
		if (yych <= '+') {
			if (yych <= ' ') goto yy1866;
			if (yych <= '*') goto yy1746;
			goto yy1937;
		} else {
			if (yych == '-') goto yy1937;
			goto yy1746;
		}
	}
yy1873:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1866;
				goto yy1746;
			} else {
				if (yych <= '\n') goto yy1866;
				if (yych <= '\f') goto yy1746;
				goto yy1866;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1746;
				goto yy1866;
			} else {
				if (yych <= '/') goto yy1746;
				if (yych <= '9') goto yy1762;
				goto yy1746;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'B') {
				if (yych <= 'A') goto yy1762;
				goto yy1871;
			} else {
				if (yych <= 'Z') goto yy1762;
				if (yych <= '`') goto yy1746;
				goto yy1762;
			}
		} else {
			if (yych <= 'h') {
				if (yych <= 'b') goto yy1871;
				goto yy1762;
			} else {
				if (yych <= 'i') goto yy2007;
				if (yych <= 'z') goto yy1762;
				goto yy1746;
			}
		}
	}
yy1874:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '$') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1763;
				if (yych <= 0x08) goto yy1683;
				if (yych <= '\n') goto yy1763;
				goto yy1683;
			} else {
				if (yych <= '\r') goto yy1763;
				if (yych == ' ') goto yy1763;
				goto yy1683;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '%') goto yy1763;
				if (yych == '+') goto yy1763;
				goto yy1683;
			} else {
				if (yych == '/') goto yy1683;
				if (yych <= '9') goto yy1763;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'S') {
				if (yych == 'B') goto yy1871;
				if (yych <= 'R') goto yy1763;
				goto yy1871;
			} else {
				if (yych <= 'Z') goto yy1763;
				if (yych == '_') goto yy1763;
				goto yy1683;
			}
		} else {
			if (yych <= 'i') {
				if (yych == 'b') goto yy1871;
				if (yych <= 'h') goto yy1763;
				goto yy2008;
			} else {
				if (yych == 's') goto yy1871;
				if (yych <= 'z') goto yy1763;
				goto yy1683;
			}
		}
	}
yy1875:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy1866;
			if (yych <= 0x08) goto yy1746;
			if (yych <= '\n') goto yy1866;
			goto yy1746;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy1866;
				goto yy1746;
			} else {
				if (yych <= ' ') goto yy1866;
				if (yych <= '/') goto yy1746;
				goto yy1762;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1746;
			if (yych == 'S') goto yy1871;
			goto yy1762;
		} else {
			if (yych <= 'r') {
				if (yych <= '`') goto yy1746;
				goto yy1762;
			} else {
				if (yych <= 's') goto yy1871;
				if (yych <= 'z') goto yy1762;
				goto yy1746;
			}
		}
	}
yy1876:
	++YYCURSOR;
yy1877:
	YYCURSOR = yyt1;
#line 389 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMDASH); }
#line 48396 "../../lnav/src/data_scanner_re.cc"
yy1878:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy1877;
		} else {
			if (yych == ',') goto yy1877;
			if (yych <= '.') goto yy1746;
			goto yy1877;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1877;
			if (yych <= 'Z') goto yy1746;
			goto yy1877;
		} else {
			if (yych == '`') goto yy1877;
			if (yych <= 'z') goto yy1746;
			goto yy1877;
		}
	}
yy1879:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1864;
			goto yy1877;
		} else {
			if (yych == ',') goto yy1877;
			if (yych <= '.') goto yy1864;
			goto yy1877;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1864;
			if (yych <= '?') goto yy1877;
			if (yych <= 'Z') goto yy1864;
			goto yy1877;
		} else {
			if (yych == '`') goto yy1877;
			if (yych <= 'z') goto yy1864;
			goto yy1877;
		}
	}
yy1880:
	yyaccept = 19;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= ',') goto yy1877;
		if (yych == '/') goto yy1877;
		goto yy1850;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1877;
			goto yy1850;
		} else {
			if (yych <= '`') goto yy1877;
			if (yych <= 'z') goto yy1850;
			goto yy1877;
		}
	}
yy1881:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1876;
	goto yy1730;
yy1882:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1881;
	goto yy1730;
yy1883:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1881;
	goto yy1730;
yy1884:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1883;
	goto yy1730;
yy1885:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1883;
	goto yy1730;
yy1886:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1883;
	goto yy1730;
yy1887:
	yyaccept = 20;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych == '%') goto yy1907;
	goto yy1867;
yy1888:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2009;
	goto yy1746;
yy1889:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 32) {
		goto yy1889;
	}
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1866;
				if (yych <= 0x08) goto yy1781;
				goto yy1866;
			} else {
				if (yych == '\r') goto yy1866;
				goto yy1781;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy1887;
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				if (yych <= ',') goto yy1781;
				goto yy1745;
			}
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') {
				if (yych <= '.') goto yy2010;
				if (yych <= '/') goto yy1781;
			} else {
				if (yych <= '?') goto yy1781;
				if (yych <= '@') goto yy1747;
				goto yy1891;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1762;
				if (yych <= '^') goto yy1781;
				goto yy1745;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'f') goto yy1891;
				if (yych <= 'z') goto yy1762;
				goto yy1781;
			}
		}
	}
yy1890:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '-') {
		if (yych <= 0x1F) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1866;
				if (yych <= 0x08) goto yy1783;
				goto yy1866;
			} else {
				if (yych == '\r') goto yy1866;
				goto yy1783;
			}
		} else {
			if (yych <= '%') {
				if (yych <= ' ') goto yy1887;
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				if (yych <= ',') goto yy1783;
				goto yy1745;
			}
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') {
				if (yych <= '.') goto yy2010;
				if (yych <= '/') goto yy1783;
				goto yy1890;
			} else {
				if (yych <= '?') goto yy1783;
				if (yych <= '@') goto yy1747;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1762;
				if (yych <= '^') goto yy1783;
				goto yy1745;
			} else {
				if (yych <= '`') goto yy1783;
				if (yych <= 'f') goto yy1891;
				if (yych <= 'z') goto yy1762;
				goto yy1783;
			}
		}
	}
yy1891:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
yy1892:
	if (yybm_init[1024+yych] & 64) {
		goto yy1891;
	}
	if (yych <= '*') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1866;
				goto yy1783;
			} else {
				if (yych <= '\n') goto yy1866;
				if (yych <= '\f') goto yy1783;
				goto yy1866;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1783;
				goto yy1866;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ',') {
				if (yych <= '+') goto yy1745;
				goto yy1783;
			} else {
				if (yych <= '.') goto yy1745;
				if (yych <= '?') goto yy1783;
				goto yy1747;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1762;
				if (yych <= '^') goto yy1783;
				goto yy1745;
			} else {
				if (yych <= '`') goto yy1783;
				if (yych <= 'z') goto yy1762;
				goto yy1783;
			}
		}
	}
yy1893:
	yych = *++YYCURSOR;
	if (yybm_init[1024+yych] & 64) {
		goto yy1891;
	}
	if (yych <= 0x1F) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1866;
			if (yych <= 0x08) goto yy1746;
			goto yy1866;
		} else {
			if (yych == '\r') goto yy1866;
			goto yy1746;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ' ') goto yy1866;
			if (yych <= '@') goto yy1746;
			goto yy1762;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1762;
			goto yy1746;
		}
	}
yy1894:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy1895:
	switch (yych) {
		case 0x00:
		case '\t':
		case '\n':
		case '\r': goto yy1866;
		case ' ': goto yy1887;
		case '%': goto yy1775;
		case '+':
		case '-':
		case '_': goto yy1745;
		case '.': goto yy2011;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1894;
		case '@': goto yy1747;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1891;
		case 'B':
		case 'b': goto yy1896;
		case 'E':
		case 'e': goto yy1897;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1873;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1762;
		case 'S':
		case 's': goto yy1871;
		case 'm': goto yy1874;
		case 'p':
		case 'u': goto yy1875;
		default: goto yy1683;
	}
yy1896:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1892;
				goto yy1683;
			} else {
				if (yych <= '\n') goto yy1892;
				if (yych <= '\f') goto yy1683;
				goto yy1892;
			}
		} else {
			if (yych <= '$') {
				if (yych == ' ') goto yy1892;
				goto yy1683;
			} else {
				if (yych <= '%') goto yy1892;
				if (yych <= '*') goto yy1683;
				goto yy1892;
			}
		}
	} else {
		if (yych <= '?') {
			if (yych <= '.') {
				if (yych <= ',') goto yy1683;
				goto yy1892;
			} else {
				if (yych <= '/') goto yy1683;
				if (yych <= '9') goto yy1892;
				goto yy1683;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1892;
				if (yych <= '^') goto yy1683;
				goto yy1892;
			} else {
				if (yych <= '`') goto yy1683;
				if (yych <= 'z') goto yy1892;
				goto yy1683;
			}
		}
	}
yy1897:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy1937;
	if (yych == '-') goto yy1937;
	goto yy1892;
yy1898:
	yych = *++YYCURSOR;
	if (yybm_init[1024+yych] & 128) {
		goto yy1898;
	}
	if (yych <= 0xE0) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			goto yy2012;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy1900;
			goto yy1901;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1902;
			goto yy1903;
		} else {
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy1730;
		}
	}
yy1899:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 16) {
		goto yy2013;
	}
	if (yych <= '\\') {
		if (yych <= ')') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy1681;
				if (yych <= ' ') goto yy1898;
				goto yy1771;
			} else {
				if (yych <= '"') goto yy1898;
				if (yych == '\'') goto yy1898;
				goto yy1771;
			}
		} else {
			if (yych <= '9') {
				if (yych <= ',') goto yy1898;
				if (yych == '/') goto yy2014;
				goto yy1771;
			} else {
				if (yych <= '?') goto yy1898;
				if (yych == '[') goto yy1898;
				goto yy1771;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '`') {
				if (yych <= ']') goto yy1898;
				if (yych <= '_') goto yy1771;
				goto yy1898;
			} else {
				if (yych <= 'z') goto yy1771;
				if (yych == '~') goto yy1771;
				goto yy1898;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1681;
				if (yych <= 0xDF) goto yy1900;
				if (yych <= 0xE0) goto yy1901;
				goto yy1902;
			} else {
				if (yych <= 0xF0) goto yy1903;
				if (yych <= 0xF3) goto yy1904;
				if (yych <= 0xF4) goto yy1905;
				goto yy1681;
			}
		}
	}
yy1900:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1898;
	goto yy1730;
yy1901:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1900;
	goto yy1730;
yy1902:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1900;
	goto yy1730;
yy1903:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1902;
	goto yy1730;
yy1904:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1902;
	goto yy1730;
yy1905:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1902;
	goto yy1730;
yy1906:
	yych = *++YYCURSOR;
	if (yych == 'o') goto yy2016;
	goto yy1770;
yy1907:
	++YYCURSOR;
	goto yy1776;
yy1908:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2019;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2022;
		case 'E':
		case 'e': goto yy2023;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2024;
		case 'm': goto yy2025;
		case 'p':
		case 'u': goto yy2026;
		default: goto yy1909;
	}
yy1909:
#line 419 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_NUMBER);
           auto sf = this->to_string_fragment(cap_all);
           if (isalpha(sf.back())) {
               while (isalpha(sf.back())) {
                   sf.pop_back();
               }
               cap_all.c_end = sf.sf_end - this->ds_input.sf_begin;
               cap_inner.c_end = sf.sf_end - this->ds_input.sf_begin;
               this->ds_next_offset = cap_all.c_end;
               this->ds_units = true;
           }
           return tokenize_result{DT_NUMBER, cap_all, cap_inner, this->ds_input.data()};
       }
#line 49015 "../../lnav/src/data_scanner_re.cc"
yy1910:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy1908;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2022;
		case 'E':
		case 'e': goto yy2023;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2024;
		case 'm': goto yy2025;
		case 'p':
		case 'u': goto yy2026;
		default: goto yy1909;
	}
yy1911:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy1908;
		case '5': goto yy2027;
		case '6':
		case '7':
		case '8':
		case '9': goto yy2019;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2022;
		case 'E':
		case 'e': goto yy2023;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2024;
		case 'm': goto yy2025;
		case 'p':
		case 'u': goto yy2026;
		default: goto yy1909;
	}
yy1912:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy1913:
	if (yych <= '@') {
		if (yych <= '9') {
			if (yych <= ',') goto yy1789;
			if (yych <= '/') goto yy1779;
			goto yy1912;
		} else {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy1789;
			goto yy1779;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1912;
			if (yych <= '^') goto yy1789;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1789;
			if (yych <= 'z') goto yy1912;
			goto yy1789;
		}
	}
yy1914:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy1730;
				goto yy1745;
			} else {
				if (yych <= '/') goto yy1730;
				if (yych <= '9') goto yy2028;
				goto yy1730;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1747;
			if (yych <= 'F') goto yy2028;
			if (yych <= 'Z') goto yy1793;
			goto yy1730;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1730;
			} else {
				if (yych <= 'f') goto yy2028;
				if (yych <= 'z') goto yy1793;
				goto yy1730;
			}
		}
	}
yy1915:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1746;
			if (yych <= '0') goto yy2029;
			goto yy2030;
		} else {
			if (yych <= '2') goto yy2031;
			if (yych <= '9') goto yy2029;
			goto yy1746;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1793;
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy1916:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy1912;
		if (yych <= '@') goto yy1730;
		goto yy2032;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy2032;
			goto yy1730;
		}
	}
yy1917:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1781;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '7') goto yy2034;
				goto yy2035;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy2036;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2037;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy2037;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy1918:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1783;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2035;
				goto yy2036;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2037;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2037;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy1919:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy2038;
		goto yy1923;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2039;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2039;
			goto yy1730;
		}
	}
yy1920:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2037;
				if (yych <= ':') goto yy2036;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2037;
			if (yych <= 'Z') goto yy1791;
			goto yy1783;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2037;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy1921:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= ',') {
			if (yych <= '$') goto yy2041;
			if (yych <= '&') goto yy2040;
			goto yy2041;
		} else {
			if (yych == '.') goto yy2041;
			if (yych <= '9') goto yy2040;
			goto yy2041;
		}
	} else {
		if (yych <= '@') {
			if (yych == '>') goto yy2041;
			if (yych <= '?') goto yy2040;
			goto yy2041;
		} else {
			if (yych <= 'Z') goto yy2040;
			if (yych <= '`') goto yy2041;
			if (yych <= 'z') goto yy2040;
			goto yy2041;
		}
	}
yy1922:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2048;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2050;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2050;
		goto yy1730;
	}
yy1923:
	yych = *++YYCURSOR;
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2051;
			} else {
				if (yych <= '1') goto yy2052;
				if (yych <= '2') goto yy2053;
				goto yy2051;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2054;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2054;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1924:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2050;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2050;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2050;
		goto yy1730;
	}
yy1925:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1789;
			if (yych <= '-') goto yy2056;
			goto yy2057;
		} else {
			if (yych <= '/') goto yy1779;
			if (yych <= '9') goto yy1925;
			if (yych <= ':') goto yy1929;
			goto yy1789;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1779;
			if (yych <= 'Z') goto yy1925;
			goto yy1789;
		} else {
			if (yych <= '_') goto yy1912;
			if (yych <= '`') goto yy1789;
			if (yych <= 'z') goto yy1925;
			goto yy1789;
		}
	}
yy1926:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy2039;
		goto yy1923;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2039;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2039;
			goto yy1730;
		}
	}
yy1927:
	yych = *++YYCURSOR;
	if (yych == '/') goto yy1921;
	if (yych == ':') goto yy1779;
	goto yy1730;
yy1928:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy1792;
	if (yych == 'b') goto yy1792;
	goto yy1791;
yy1929:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy1779;
	goto yy1730;
yy1930:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy1792;
		goto yy1791;
	} else {
		if (yych <= 'b') goto yy1792;
		if (yych == 'n') goto yy1792;
		goto yy1791;
	}
yy1931:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
yy1932:
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1783;
			goto yy1745;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy1783;
				goto yy1777;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1931;
				goto yy1927;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy1931;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy1931;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy1933:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1746;
			if (yych <= '0') goto yy2058;
			goto yy2059;
		} else {
			if (yych <= '2') goto yy2060;
			if (yych <= '9') goto yy2058;
			goto yy1746;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1793;
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy1934:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy1933;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2061;
		case ':': goto yy2036;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2037;
		case 'B':
		case 'b': goto yy2062;
		case 'E':
		case 'e': goto yy2063;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1935:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2037;
				if (yych <= ':') goto yy2036;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2037;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2037;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy1936:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy2064;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy2037;
				goto yy2036;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy2037;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2037;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy1937:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2065;
	goto yy1746;
yy1938:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy2066;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy2028;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy2028;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy1939:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy2067;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2061;
		case ':': goto yy2036;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2037;
		case 'B':
		case 'b': goto yy2062;
		case 'E':
		case 'e': goto yy2063;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy1940:
	++YYCURSOR;
yy1941:
	YYCURSOR = yyt2;
#line 319 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_IPV6_ADDRESS); }
#line 49941 "../../lnav/src/data_scanner_re.cc"
yy1942:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2069;
				yyt1 = YYCURSOR;
				goto yy2070;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1943:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2072;
				yyt1 = YYCURSOR;
				goto yy2070;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1944:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2072;
				goto yy2073;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2069;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2070;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2071;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2071;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1945:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1940;
	goto yy1730;
yy1946:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2071;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy2070;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2071;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2071;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1947:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2070;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2071;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2074;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'e') goto yy2071;
				goto yy2074;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych >= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1948:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1940;
	goto yy1730;
yy1949:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1948;
	goto yy1730;
yy1950:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1948;
	goto yy1730;
yy1951:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1950;
	goto yy1730;
yy1952:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1950;
	goto yy1730;
yy1953:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1950;
	goto yy1730;
yy1954:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2076;
			goto yy2079;
		} else {
			if (yych <= '/') goto yy2076;
			if (yych <= ':') goto yy2080;
			goto yy2076;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2080;
			if (yych <= '^') goto yy2076;
			goto yy2080;
		} else {
			if (yych <= '`') goto yy2076;
			if (yych <= 'z') goto yy2080;
			goto yy2076;
		}
	}
yy1955:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2076;
			goto yy2080;
		} else {
			if (yych <= '/') goto yy2076;
			if (yych <= ':') goto yy2080;
			goto yy2076;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2080;
			if (yych <= '^') goto yy2076;
			goto yy2080;
		} else {
			if (yych <= '`') goto yy2076;
			if (yych <= 'z') goto yy2080;
			goto yy2076;
		}
	}
yy1956:
	yych = *++YYCURSOR;
yy1957:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy1956;
			} else {
				if (yych == '\r') goto yy1956;
				goto yy1730;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy1956;
				goto yy1730;
			} else {
				if (yych <= '-') goto yy2083;
				if (yych <= '.') goto yy1730;
				goto yy1959;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy2083;
				goto yy1730;
			} else {
				if (yych <= '>') goto yy1960;
				if (yych <= '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2083;
				if (yych <= '^') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2083;
				goto yy1730;
			}
		}
	}
yy1958:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 1) {
		goto yy1958;
	}
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2084;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2084;
			if (yych <= 0x1F) goto yy1730;
			goto yy2084;
		}
	} else {
		if (yych <= '<') {
			if (yych != '/') goto yy1730;
		} else {
			if (yych <= '=') goto yy2085;
			if (yych <= '>') goto yy1960;
			if (yych >= '@') goto yy1730;
		}
	}
yy1959:
	yych = *++YYCURSOR;
	if (yych == '>') goto yy2086;
	goto yy1730;
yy1960:
	++YYCURSOR;
yy1961:
#line 329 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_OPEN_TAG);
       }
#line 50447 "../../lnav/src/data_scanner_re.cc"
yy1962:
	yych = *++YYCURSOR;
yy1963:
	if (yych <= '/') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy2088;
			if (yych <= '\f') goto yy1730;
			goto yy2088;
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1730;
				goto yy2088;
			} else {
				if (yych == '-') goto yy1962;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '=') {
				if (yych <= ':') goto yy1962;
				goto yy1730;
			} else {
				if (yych <= '>') goto yy2089;
				if (yych <= '@') goto yy1730;
				goto yy1962;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1730;
				goto yy1962;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy1962;
				goto yy1730;
			}
		}
	}
yy1964:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2091;
			goto yy2092;
		} else {
			if (yych <= '/') goto yy2091;
			if (yych <= ':') goto yy2092;
			goto yy2091;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2092;
			if (yych <= '^') goto yy2091;
			goto yy2092;
		} else {
			if (yych <= '`') goto yy2091;
			if (yych <= 'z') goto yy2092;
			goto yy2091;
		}
	}
yy1965:
	yych = *++YYCURSOR;
	if (yych == '-') goto yy2093;
	goto yy1730;
yy1966:
	++YYCURSOR;
	goto yy1810;
yy1967:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy1966;
		goto yy1810;
	} else {
		if (yych <= '\r') goto yy1966;
		if (yych == ' ') goto yy1966;
		goto yy1810;
	}
yy1968:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy1966;
		goto yy1730;
	} else {
		if (yych <= '\r') goto yy1966;
		if (yych == ' ') goto yy1966;
		goto yy1730;
	}
yy1969:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 2) {
		goto yy1969;
	}
	if (yych <= ',') {
		if (yych <= '%') {
			if (yych <= '$') goto yy1698;
			goto yy1745;
		} else {
			if (yych == '+') goto yy1745;
			goto yy1698;
		}
	} else {
		if (yych <= ':') {
			if (yych <= '.') goto yy2094;
			if (yych <= '/') goto yy1698;
			goto yy2095;
		} else {
			if (yych == '@') goto yy1747;
			goto yy1698;
		}
	}
yy1970:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1698;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1698;
				goto yy2096;
			} else {
				if (yych <= '.') goto yy1811;
				if (yych <= '/') goto yy1779;
				goto yy1970;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1834;
			if (yych <= '?') goto yy1698;
			if (yych <= '@') goto yy1785;
			goto yy1970;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1698;
				goto yy1970;
			} else {
				if (yych <= '`') goto yy1698;
				if (yych <= 'z') goto yy1970;
				goto yy1698;
			}
		}
	}
yy1971:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1821;
		if (yych <= '9') goto yy2097;
		if (yych <= ':') goto yy1835;
		goto yy1821;
	} else {
		if (yych <= 'F') goto yy2097;
		if (yych <= '`') goto yy1821;
		if (yych <= 'f') goto yy2097;
		goto yy1821;
	}
yy1972:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy2039;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2039;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2039;
			goto yy1730;
		}
	}
yy1973:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= '2') {
			if (yych <= '/') {
				if (yych == '-') {
					yyt2 = YYCURSOR;
					goto yy2098;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '0') goto yy2099;
				if (yych <= '1') goto yy2100;
				goto yy2101;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2099;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2102;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2102;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy1974:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1821;
			goto yy2106;
		} else {
			if (yych <= '/') goto yy1821;
			if (yych <= '9') goto yy2097;
			goto yy1835;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1821;
			goto yy2097;
		} else {
			if (yych <= '`') goto yy1821;
			if (yych <= 'f') goto yy2097;
			goto yy1821;
		}
	}
yy1975:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2106;
	goto yy1821;
yy1976:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy1966;
			if (yych <= '\f') goto yy1746;
			goto yy1966;
		} else {
			if (yych == ' ') goto yy1966;
			if (yych <= ',') goto yy1746;
			goto yy1969;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') goto yy1746;
			if (yych <= '9') goto yy1970;
			if (yych <= '@') goto yy1746;
			goto yy1970;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1746;
				goto yy1970;
			} else {
				if (yych <= '`') goto yy1746;
				if (yych <= 'z') goto yy1970;
				goto yy1746;
			}
		}
	}
yy1977:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') goto yy1810;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy2039;
		goto yy1973;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1810;
			goto yy2039;
		} else {
			if (yych <= '`') goto yy1810;
			if (yych <= 'f') goto yy2039;
			goto yy1810;
		}
	}
yy1978:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1809;
				yyt3 = YYCURSOR;
				goto yy2107;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1809;
				goto yy1777;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy2097;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy2108;
				}
				goto yy1809;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2097;
				goto yy1820;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy2109;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy1979:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1809;
				yyt3 = YYCURSOR;
				goto yy2107;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '+') goto yy1745;
				goto yy1809;
			} else {
				if (yych <= '-') goto yy1777;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				goto yy1779;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				if (yych <= '9') goto yy1820;
				yyt3 = YYCURSOR;
			} else {
				if (yych <= '?') goto yy1809;
				if (yych <= '@') goto yy1785;
				goto yy1820;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1809;
				goto yy1707;
			} else {
				if (yych <= '`') goto yy1809;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy1980:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy1921;
	if (yych == ':') goto yy1983;
	goto yy1810;
yy1981:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2106;
	if (yych == 'S') goto yy2111;
	goto yy1821;
yy1982:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1809;
				yyt3 = YYCURSOR;
				goto yy2107;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1809;
				goto yy1777;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy1980;
				}
				goto yy1809;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= 'r') {
				if (yych <= '_') goto yy1707;
				if (yych <= '`') goto yy1809;
				goto yy2110;
			} else {
				if (yych <= 's') goto yy2112;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy1983:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy2113;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2103;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2103;
			if (yych <= '^') goto yy1730;
			goto yy2103;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy2103;
			goto yy1730;
		}
	}
yy1984:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2106;
	if (yych == 'L') goto yy2115;
	goto yy1821;
yy1985:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1809;
				yyt3 = YYCURSOR;
				goto yy2107;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1809;
				goto yy1777;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy1980;
				}
				goto yy1809;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= 'd') {
				if (yych <= '_') goto yy1707;
				if (yych <= '`') goto yy1809;
				goto yy2110;
			} else {
				if (yych <= 'e') goto yy2117;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy1986:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2106;
	if (yych == 'E') goto yy2115;
	goto yy1821;
yy1987:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\t') {
			if (yych <= 0x00) goto yy1810;
			goto yy1749;
		} else {
			if (yych <= '\n') goto yy1810;
			if (yych <= 0x15) goto yy1749;
			goto yy1810;
		}
	} else {
		if (yych <= 0x7F) {
			if (yych == 0x1B) goto yy1810;
			goto yy1749;
		} else {
			if (yych <= 0xC1) goto yy1810;
			if (yych <= 0xF4) goto yy1749;
			goto yy1810;
		}
	}
yy1988:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x16) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1810;
			if (yych <= 0x08) goto yy1749;
			if (yych <= '\t') goto yy1987;
			goto yy1810;
		} else {
			if (yych == '\r') goto yy1987;
			if (yych <= 0x15) goto yy1749;
			goto yy1810;
		}
	} else {
		if (yych <= ' ') {
			if (yych == 0x1B) goto yy1810;
			if (yych <= 0x1F) goto yy1749;
			goto yy1987;
		} else {
			if (yych <= 0x7F) goto yy1749;
			if (yych <= 0xC1) goto yy1810;
			if (yych <= 0xF4) goto yy1749;
			goto yy1810;
		}
	}
yy1989:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '&') {
				if (yych == '!') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				yyt2 = YYCURSOR;
				goto yy1853;
			} else {
				if (yych <= '\'') goto yy2119;
				if (yych <= '+') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt1 = yyt3 = YYCURSOR;
				goto yy2118;
			}
		} else {
			if (yych <= '>') {
				if (yych == '.') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				yyt2 = YYCURSOR;
				goto yy1853;
			} else {
				if (yych <= '?') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				if (yych <= 'R') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 's') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				if (yych <= 'r') {
					yyt1 = YYCURSOR;
					goto yy2120;
				}
				goto yy1808;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2120;
				}
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1857;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1858;
				}
				yyt2 = YYCURSOR;
				goto yy1859;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1860;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1861;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1862;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	}
yy1990:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy1987;
		goto yy1749;
	} else {
		if (yych <= '\r') goto yy1987;
		if (yych == ' ') goto yy1987;
		goto yy1749;
	}
yy1991:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 4) {
		goto yy1991;
	}
	if (yych <= '+') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0x08) goto yy1748;
				if (yych <= '\t') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				yyt3 = YYCURSOR;
				goto yy1966;
			} else {
				if (yych <= '\r') {
					if (yych <= '\f') goto yy1748;
					yyt3 = YYCURSOR;
					goto yy1987;
				} else {
					if (yych == 0x16) {
						yyt3 = YYCURSOR;
						goto yy1810;
					}
					goto yy1748;
				}
			}
		} else {
			if (yych <= '!') {
				if (yych <= 0x1B) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0x1F) goto yy1748;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				yyt3 = YYCURSOR;
				goto yy1988;
			} else {
				if (yych <= '&') {
					if (yych <= '"') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				} else {
					if (yych <= '\'') goto yy2121;
					if (yych <= '*') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				}
			}
		}
	} else {
		if (yych <= '\\') {
			if (yych <= '9') {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				if (yych == '.') {
					yyt3 = YYCURSOR;
					goto yy1990;
				}
				goto yy1748;
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				} else {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy1988;
					}
					if (yych <= '[') goto yy1748;
					goto yy1751;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy1748;
				if (yych <= 0xC1) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 0xDF) goto yy1752;
				goto yy1753;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1754;
					goto yy1755;
				} else {
					if (yych <= 0xF3) goto yy1756;
					if (yych <= 0xF4) goto yy1757;
					yyt3 = YYCURSOR;
					goto yy1810;
				}
			}
		}
	}
yy1992:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1809;
				yyt3 = YYCURSOR;
				goto yy2107;
			} else {
				if (yych == '%') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1809;
				goto yy1777;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') {
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy1980;
				}
				goto yy1809;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= 'k') {
				if (yych <= '_') goto yy1707;
				if (yych <= '`') goto yy1809;
				goto yy2110;
			} else {
				if (yych <= 'l') goto yy2117;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy1993:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy1793;
		if (yych <= '@') goto yy1746;
		goto yy2122;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy2122;
			goto yy1746;
		}
	}
yy1994:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2123;
	goto yy1730;
yy1995:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2124;
	goto yy1730;
yy1996:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '<') goto yy1849;
			goto yy2125;
		} else {
			if (yych == ']') goto yy2126;
			if (yych <= 0x7F) goto yy1849;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1997;
			if (yych <= 0xE0) goto yy1998;
			goto yy1999;
		} else {
			if (yych <= 0xF0) goto yy2000;
			if (yych <= 0xF3) goto yy2001;
			if (yych <= 0xF4) goto yy2002;
			goto yy1730;
		}
	}
yy1997:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1849;
	goto yy1730;
yy1998:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy1997;
	goto yy1730;
yy1999:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1997;
	goto yy1730;
yy2000:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy1999;
	goto yy1730;
yy2001:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy1999;
	goto yy1730;
yy2002:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy1999;
	goto yy1730;
yy2003:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 8) {
		goto yy2003;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy1850;
		if (yych <= '/') goto yy1730;
		goto yy1850;
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1730;
			goto yy2127;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy2127;
			goto yy1730;
		}
	}
yy2004:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1280+yych] & 16) {
		goto yy1748;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x1B) goto yy1854;
			goto yy1852;
		} else {
			if (yych <= '\\') goto yy1751;
			if (yych <= 0xC1) goto yy1854;
			goto yy1752;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1753;
			if (yych <= 0xEF) goto yy1754;
			goto yy1755;
		} else {
			if (yych <= 0xF3) goto yy1756;
			if (yych <= 0xF4) goto yy1757;
			goto yy1854;
		}
	}
yy2005:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2005;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2129;
		case 'E':
		case 'e': goto yy2130;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2131;
		case 'm': goto yy2132;
		case 'p':
		case 'u': goto yy2133;
		default: goto yy1909;
	}
yy2006:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2134;
	goto yy1746;
yy2007:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '\f') {
			if (yych <= 0x00) goto yy1866;
			if (yych <= 0x08) goto yy1746;
			if (yych <= '\n') goto yy1866;
			goto yy1746;
		} else {
			if (yych <= 0x1F) {
				if (yych <= '\r') goto yy1866;
				goto yy1746;
			} else {
				if (yych <= ' ') goto yy1866;
				if (yych <= '/') goto yy1746;
				goto yy1762;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '@') goto yy1746;
			if (yych == 'B') goto yy1871;
			goto yy1762;
		} else {
			if (yych <= 'a') {
				if (yych <= '`') goto yy1746;
				goto yy1762;
			} else {
				if (yych <= 'b') goto yy1871;
				if (yych <= 'z') goto yy1762;
				goto yy1746;
			}
		}
	}
yy2008:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1866;
				goto yy1746;
			} else {
				if (yych <= '\n') goto yy1866;
				if (yych <= '\f') goto yy1746;
				goto yy1866;
			}
		} else {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1746;
				goto yy1866;
			} else {
				if (yych <= '/') goto yy1746;
				if (yych <= '9') goto yy1762;
				goto yy1746;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= 'B') {
				if (yych <= 'A') goto yy1762;
				goto yy1871;
			} else {
				if (yych <= 'Z') goto yy1762;
				if (yych <= '`') goto yy1746;
				goto yy1762;
			}
		} else {
			if (yych <= 'm') {
				if (yych <= 'b') goto yy1871;
				goto yy1762;
			} else {
				if (yych <= 'n') goto yy1871;
				if (yych <= 'z') goto yy1762;
				goto yy1746;
			}
		}
	}
yy2009:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2009;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2129;
		case 'E':
		case 'e': goto yy2130;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2131;
		case 'm': goto yy2132;
		case 'p':
		case 'u': goto yy2133;
		default: goto yy1909;
	}
yy2010:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2135;
	goto yy1746;
yy2011:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2136;
	goto yy1746;
yy2012:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ')') goto yy1898;
			goto yy2137;
		} else {
			if (yych == '/') goto yy2138;
			if (yych <= 0x7F) goto yy1898;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1900;
			if (yych <= 0xE0) goto yy1901;
			goto yy1902;
		} else {
			if (yych <= 0xF0) goto yy1903;
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy1730;
		}
	}
yy2013:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 16) {
		goto yy2013;
	}
	if (yych <= '\\') {
		if (yych <= ')') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy1681;
				if (yych <= ' ') goto yy1898;
				goto yy1771;
			} else {
				if (yych <= '"') goto yy1898;
				if (yych == '\'') goto yy1898;
				goto yy1771;
			}
		} else {
			if (yych <= '9') {
				if (yych <= ',') goto yy1898;
				if (yych == '/') goto yy2139;
				goto yy1771;
			} else {
				if (yych <= '?') goto yy1898;
				if (yych == '[') goto yy1898;
				goto yy1771;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '`') {
				if (yych <= ']') goto yy1898;
				if (yych <= '_') goto yy1771;
				goto yy1898;
			} else {
				if (yych <= 'z') goto yy1771;
				if (yych == '~') goto yy1771;
				goto yy1898;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1681;
				if (yych <= 0xDF) goto yy1900;
				if (yych <= 0xE0) goto yy1901;
				goto yy1902;
			} else {
				if (yych <= 0xF0) goto yy1903;
				if (yych <= 0xF3) goto yy1904;
				if (yych <= 0xF4) goto yy1905;
				goto yy1681;
			}
		}
	}
yy2014:
	yych = *++YYCURSOR;
	if (yybm_init[1280+yych] & 64) {
		goto yy1769;
	}
yy2015:
#line 244 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 51809 "../../lnav/src/data_scanner_re.cc"
yy2016:
	yych = *++YYCURSOR;
	if (yych == 'g') goto yy2140;
	goto yy1770;
yy2017:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2141;
		if (yych <= '@') goto yy1746;
		goto yy2141;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy2141;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy2141;
			goto yy1746;
		}
	}
yy2018:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1746;
			if (yych <= '0') goto yy2144;
			goto yy2145;
		} else {
			if (yych <= '2') goto yy2146;
			if (yych <= '9') goto yy2144;
			goto yy1746;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1793;
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2019:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	goto yy2148;
yy2020:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy2021:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1789;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1789;
				goto yy2017;
			} else {
				if (yych <= '.') goto yy2149;
				if (yych <= '/') goto yy1779;
				goto yy2020;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy1789;
			if (yych <= '@') goto yy1785;
			goto yy2020;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1789;
				goto yy2020;
			} else {
				if (yych <= '`') goto yy1789;
				if (yych <= 'z') goto yy2020;
				goto yy1789;
			}
		}
	}
yy2022:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy2021;
			goto yy1909;
		} else {
			if (yych == ',') goto yy1909;
			goto yy2021;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1909;
			if (yych <= 'Z') goto yy2021;
			goto yy1909;
		} else {
			if (yych == '`') goto yy1909;
			if (yych <= 'z') goto yy2021;
			goto yy1909;
		}
	}
yy2023:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy2150;
	if (yych == '-') goto yy2151;
	goto yy2021;
yy2024:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2022;
		goto yy2021;
	} else {
		if (yych <= 'b') goto yy2022;
		if (yych == 'i') goto yy2152;
		goto yy2021;
	}
yy2025:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1909;
				goto yy2021;
			} else {
				if (yych == '+') goto yy2021;
				goto yy1909;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy2021;
				if (yych <= '?') goto yy1909;
				goto yy2021;
			} else {
				if (yych <= 'B') goto yy2022;
				if (yych <= 'R') goto yy2021;
				goto yy2022;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2021;
				if (yych <= '^') goto yy1909;
				goto yy2021;
			} else {
				if (yych <= '`') goto yy1909;
				if (yych <= 'a') goto yy2021;
				goto yy2022;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy2153;
				goto yy2021;
			} else {
				if (yych <= 's') goto yy2022;
				if (yych <= 'z') goto yy2021;
				goto yy1909;
			}
		}
	}
yy2026:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy2022;
	if (yych == 's') goto yy2022;
	goto yy2021;
yy2027:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	if (yych <= '/') goto yy2148;
	if (yych <= '5') goto yy2019;
	goto yy2148;
yy2028:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1789;
			goto yy1745;
		} else {
			if (yych <= '.') {
				if (yych <= ',') goto yy1789;
				goto yy1777;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2154;
				goto yy1929;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1789;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2154;
				if (yych <= 'Z') goto yy1793;
				goto yy1789;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1789;
			} else {
				if (yych <= 'f') goto yy2154;
				if (yych <= 'z') goto yy1793;
				goto yy1789;
			}
		}
	}
yy2029:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy2021;
	} else {
		if (yych <= '.') goto yy2018;
		if (yych <= '/') goto yy2021;
		if (yych <= '9') goto yy2156;
		goto yy2021;
	}
yy2030:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy2021;
	} else {
		if (yych <= '.') goto yy2018;
		if (yych <= '/') goto yy2021;
		if (yych <= '9') goto yy2029;
		goto yy2021;
	}
yy2031:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy2029;
		case '5': goto yy2157;
		case '6':
		case '7':
		case '8':
		case '9': goto yy2156;
		default: goto yy2021;
	}
yy2032:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy1913;
	if (yych <= 'Z') goto yy2158;
	if (yych <= '`') goto yy1913;
	if (yych <= 'z') goto yy2158;
	goto yy1913;
yy2033:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2159;
		if (yych <= '@') goto yy1746;
		goto yy1793;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2034:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '7') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1781;
				goto yy2161;
			} else {
				if (yych == '%') goto yy1775;
				goto yy1781;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1781;
				goto yy2162;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy2163;
				goto yy2164;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= '9') goto yy2165;
				if (yych <= ':') goto yy2166;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2167;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy2167;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy2035:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych <= ' ') {
				if (yych <= 0x1F) goto yy1783;
				goto yy2161;
			} else {
				if (yych == '%') goto yy1775;
				goto yy1783;
			}
		} else {
			if (yych <= '-') {
				if (yych <= '+') goto yy1745;
				if (yych <= ',') goto yy1783;
				goto yy2162;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy2163;
				goto yy2165;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy2166;
				goto yy1783;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2167;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2167;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy2036:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '.') goto yy1730;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy1924;
		goto yy1923;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy2037:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2167;
				if (yych <= ':') goto yy2166;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2167;
			if (yych <= 'Z') goto yy1791;
			goto yy1783;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2167;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2038:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2168;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2169;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2169;
		goto yy1730;
	}
yy2039:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2169;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2169;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2169;
		goto yy1730;
	}
yy2040:
	yych = *++YYCURSOR;
yy2041:
	if (yybm_init[768+yych] & 32) {
		goto yy2040;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy1730;
			if (yych <= '&') goto yy2170;
			goto yy1730;
		} else {
			if (yych <= 'Z') goto yy2170;
			if (yych <= ']') goto yy1730;
			if (yych <= 'z') goto yy2170;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2042;
			if (yych <= 0xE0) goto yy2043;
			goto yy2044;
		} else {
			if (yych <= 0xF0) goto yy2045;
			if (yych <= 0xF3) goto yy2046;
			if (yych <= 0xF4) goto yy2047;
			goto yy1730;
		}
	}
yy2042:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2040;
	goto yy1730;
yy2043:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2042;
	goto yy1730;
yy2044:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2042;
	goto yy1730;
yy2045:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2044;
	goto yy1730;
yy2046:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2044;
	goto yy1730;
yy2047:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2044;
	goto yy1730;
yy2048:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2172;
		if (yych <= ':') goto yy2173;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2172;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2172;
		goto yy1730;
	}
yy2049:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2174;
		if (yych <= ':') goto yy2175;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2174;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2174;
		goto yy1730;
	}
yy2050:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2172;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2172;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2172;
		goto yy1730;
	}
yy2051:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2178;
				yyt2 = YYCURSOR;
				goto yy2179;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2180;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2180;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2052:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2181;
				yyt2 = YYCURSOR;
				goto yy2179;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2180;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2180;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2053:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '4') goto yy2181;
				goto yy2182;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2178;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2180;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2180;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2054:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy2180;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2180;
				goto yy1912;
			} else {
				if (yych != '_') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2180;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2055:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= ',') goto yy1941;
		if (yych <= ':') goto yy1913;
		if (yych <= '?') goto yy1941;
		goto yy1913;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1941;
			goto yy1913;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy1913;
			goto yy1941;
		}
	}
yy2056:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 8) {
		goto yy2003;
	}
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy1850;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy1925;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1925;
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1925;
			goto yy1730;
		}
	}
yy2057:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 8) {
		goto yy2003;
	}
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy1850;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy1925;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2183;
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy2183;
			goto yy1730;
		}
	}
yy2058:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2184;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2185;
		case 'E':
		case 'e': goto yy2186;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2187;
		case 'm': goto yy2188;
		case 'p':
		case 'u': goto yy2189;
		default: goto yy1683;
	}
yy2059:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2058;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2185;
		case 'E':
		case 'e': goto yy2186;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2187;
		case 'm': goto yy2188;
		case 'p':
		case 'u': goto yy2189;
		default: goto yy1683;
	}
yy2060:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2021;
		case '.': goto yy2018;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4': goto yy2058;
		case '5': goto yy2190;
		case '6':
		case '7':
		case '8':
		case '9': goto yy2184;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2185;
		case 'E':
		case 'e': goto yy2186;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2187;
		case 'm': goto yy2188;
		case 'p':
		case 'u': goto yy2189;
		default: goto yy1683;
	}
yy2061:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case ' ': goto yy2161;
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy2162;
		case '.': goto yy2067;
		case '/': goto yy2163;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2191;
		case ':': goto yy2166;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2167;
		case 'B':
		case 'b': goto yy2192;
		case 'E':
		case 'e': goto yy2193;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2062:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2167;
				if (yych <= ':') goto yy2166;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2167;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2167;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy2063:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1937;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy2167;
				goto yy2166;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy2167;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2167;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2064:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2194;
		if (yych <= '@') goto yy1746;
		goto yy1793;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2065:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2065;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2196;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2197;
		case 'm': goto yy2198;
		case 'p':
		case 'u': goto yy2199;
		default: goto yy1683;
	}
yy2066:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1794;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2200;
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy2154;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2201;
		case 'S':
		case 's': goto yy2202;
		case 'm': goto yy2203;
		case 'p':
		case 'u': goto yy2204;
		default: goto yy1683;
	}
yy2067:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2205;
		if (yych <= '@') goto yy1746;
		goto yy1793;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2068:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy1941;
		}
		if (yych <= '0') goto yy2207;
		goto yy2208;
	} else {
		if (yych <= '2') goto yy2209;
		if (yych <= '9') goto yy2207;
		yyt2 = yyt1;
		goto yy1941;
	}
yy2069:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2210;
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2070:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') {
			yyt2 = yyt1;
			goto yy1940;
		}
		if (yych <= '/') goto yy1730;
		goto yy2211;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2211;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2211;
			goto yy1730;
		}
	}
yy2071:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2210;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy2070;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2210;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2210;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2072:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2212;
				yyt1 = YYCURSOR;
				goto yy2070;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2073:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2212;
				goto yy2210;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2070;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2210;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2210;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2074:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2070;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2210;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2213;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'e') goto yy2210;
				goto yy2213;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2075:
	yych = *++YYCURSOR;
yy2076:
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2075;
			} else {
				if (yych == '\r') goto yy2075;
				goto yy1730;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2075;
				if (yych <= '!') goto yy1730;
			} else {
				if (yych == '\'') goto yy2078;
				goto yy1730;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy2214;
				if (yych <= '/') goto yy1730;
				goto yy2214;
			} else {
				if (yych == '>') goto yy2081;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2214;
				if (yych <= '^') goto yy1730;
				goto yy2214;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2214;
				goto yy1730;
			}
		}
	}
yy2077:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 2) {
		goto yy2215;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '"') goto yy1730;
			goto yy2216;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2217;
			goto yy2218;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2219;
			goto yy2220;
		} else {
			if (yych <= 0xF3) goto yy2221;
			if (yych <= 0xF4) goto yy2222;
			goto yy1730;
		}
	}
yy2078:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 4) {
		goto yy2223;
	}
	if (yych <= 0xE0) {
		if (yych <= '\\') {
			if (yych <= '\'') goto yy1730;
			goto yy2224;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2225;
			goto yy2226;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2227;
			goto yy2228;
		} else {
			if (yych <= 0xF3) goto yy2229;
			if (yych <= 0xF4) goto yy2230;
			goto yy1730;
		}
	}
yy2079:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 64) {
		goto yy2079;
	}
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2231;
				goto yy2233;
			} else {
				if (yych == '\r') goto yy2233;
				goto yy2231;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2233;
				if (yych <= '!') goto yy2231;
				goto yy2234;
			} else {
				if (yych == '\'') goto yy2235;
				if (yych <= ')') goto yy2231;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '<') {
				if (yych == '-') goto yy2236;
				goto yy2231;
			} else {
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2238;
				if (yych <= 0x7F) goto yy2231;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2239;
				if (yych <= 0xE0) goto yy2240;
				goto yy2241;
			} else {
				if (yych <= 0xF0) goto yy2242;
				if (yych <= 0xF3) goto yy2243;
				if (yych <= 0xF4) goto yy2244;
				goto yy1730;
			}
		}
	}
yy2080:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy2245;
				goto yy1730;
			} else {
				if (yych <= '\r') goto yy2245;
				if (yych <= 0x1F) goto yy1730;
				goto yy2245;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy2077;
				goto yy1730;
			} else {
				if (yych <= '\'') goto yy2078;
				if (yych <= ',') goto yy1730;
				goto yy2080;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1730;
				if (yych <= ':') goto yy2080;
				goto yy1730;
			} else {
				if (yych <= '=') goto yy2246;
				if (yych >= '?') goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2080;
				if (yych <= '^') goto yy1730;
				goto yy2080;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2080;
				goto yy1730;
			}
		}
	}
yy2081:
	++YYCURSOR;
yy2082:
#line 321 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_DECL_TAG);
       }
#line 53933 "../../lnav/src/data_scanner_re.cc"
yy2083:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2247;
			} else {
				if (yych == '\r') goto yy2247;
				goto yy1730;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy2247;
				if (yych <= ',') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '.') goto yy1730;
				if (yych <= '/') goto yy1959;
				goto yy2083;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1730;
				goto yy2085;
			} else {
				if (yych <= '>') goto yy1960;
				if (yych <= '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2083;
				if (yych <= '^') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2083;
				goto yy1730;
			}
		}
	}
yy2084:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2084;
			} else {
				if (yych == '\r') goto yy2084;
				goto yy1730;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy2084;
				if (yych <= ',') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '.') goto yy1730;
				if (yych <= '/') goto yy1959;
				goto yy2083;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1730;
			} else {
				if (yych <= '>') goto yy1960;
				if (yych <= '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2083;
				if (yych <= '^') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2083;
				goto yy1730;
			}
		}
	}
yy2085:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2248;
			} else {
				if (yych <= '\t') goto yy2249;
				if (yych <= '\f') goto yy2248;
				goto yy2249;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy2249;
				goto yy2248;
			} else {
				if (yych <= '"') goto yy2250;
				if (yych <= '&') goto yy2248;
				goto yy2251;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy2248;
				goto yy1730;
			} else {
				if (yych <= 0x7F) goto yy2248;
				if (yych <= 0xC1) goto yy1730;
				goto yy2252;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2253;
				if (yych <= 0xEF) goto yy2254;
				goto yy2255;
			} else {
				if (yych <= 0xF3) goto yy2256;
				if (yych <= 0xF4) goto yy2257;
				goto yy1730;
			}
		}
	}
yy2086:
	++YYCURSOR;
yy2087:
#line 325 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_EMPTY_TAG);
       }
#line 54072 "../../lnav/src/data_scanner_re.cc"
yy2088:
	yych = *++YYCURSOR;
	if (yych <= '\r') {
		if (yych == '\t') goto yy2088;
		if (yych <= '\f') goto yy1730;
		goto yy2088;
	} else {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1730;
			goto yy2088;
		} else {
			if (yych != '>') goto yy1730;
		}
	}
yy2089:
	++YYCURSOR;
#line 333 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_XML_CLOSE_TAG);
       }
#line 54093 "../../lnav/src/data_scanner_re.cc"
yy2090:
	yych = *++YYCURSOR;
yy2091:
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2090;
			} else {
				if (yych == '\r') goto yy2090;
				goto yy1730;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy2090;
				goto yy1730;
			} else {
				if (yych <= '-') goto yy2258;
				if (yych <= '.') goto yy1730;
				goto yy1959;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy2258;
				goto yy1730;
			} else {
				if (yych <= '?') goto yy1959;
				if (yych <= '@') goto yy1730;
				goto yy2258;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2258;
				goto yy1730;
			}
		}
	}
yy2092:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2259;
			} else {
				if (yych == '\r') goto yy2259;
				goto yy1730;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy2259;
				if (yych <= ',') goto yy1730;
				goto yy2092;
			} else {
				if (yych <= '.') goto yy1730;
				if (yych <= '/') goto yy1959;
				goto yy2092;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1730;
				goto yy2260;
			} else {
				if (yych == '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2092;
				if (yych <= '^') goto yy1730;
				goto yy2092;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2092;
				goto yy1730;
			}
		}
	}
yy2093:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2261;
	goto yy1730;
yy2094:
	yych = *++YYCURSOR;
	if (yybm_init[768+yych] & 2) {
		goto yy1969;
	}
	if (yych <= '+') {
		if (yych == '%') goto yy1745;
		if (yych <= '*') goto yy1730;
		goto yy1745;
	} else {
		if (yych <= '.') {
			if (yych <= ',') goto yy1730;
			goto yy1745;
		} else {
			if (yych == '@') goto yy1747;
			goto yy1730;
		}
	}
yy2095:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2262;
	goto yy1730;
yy2096:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1698;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1698;
				goto yy1969;
			} else {
				if (yych <= '.') goto yy2094;
				if (yych <= '/') goto yy1698;
				goto yy1970;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy2095;
			if (yych <= '?') goto yy1698;
			if (yych <= '@') goto yy1747;
			goto yy1970;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1698;
				goto yy1970;
			} else {
				if (yych <= '`') goto yy1698;
				if (yych <= 'z') goto yy1970;
				goto yy1698;
			}
		}
	}
yy2097:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1821;
		if (yych <= '9') goto yy2263;
		if (yych <= ':') goto yy2264;
		goto yy1821;
	} else {
		if (yych <= 'F') goto yy2263;
		if (yych <= '`') goto yy1821;
		if (yych <= 'f') goto yy2263;
		goto yy1821;
	}
yy2098:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1941;
			goto yy2114;
		} else {
			if (yych <= '/') goto yy1941;
			if (yych <= ':') goto yy2114;
			goto yy1941;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2114;
			if (yych <= '^') goto yy1941;
			goto yy2114;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy2114;
			goto yy1941;
		}
	}
yy2099:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2267;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2269;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2269;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2100:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2270;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2269;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2269;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2101:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '4') goto yy2270;
				goto yy2271;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2267;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2268;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2269;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2269;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2102:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2272;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2269;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2269;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2269;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2103:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy2104:
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1698;
			if (yych <= '-') goto yy2273;
			goto yy1983;
		} else {
			if (yych <= '/') goto yy1779;
			if (yych <= '9') goto yy2103;
			if (yych <= ':') goto yy1834;
			goto yy1698;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1779;
			if (yych <= 'Z') goto yy2103;
			goto yy1698;
		} else {
			if (yych == '`') goto yy1698;
			if (yych <= 'z') goto yy2103;
			goto yy1698;
		}
	}
yy2105:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= ',') goto yy1941;
		if (yych <= ':') goto yy2104;
		if (yych <= '?') goto yy1941;
		goto yy2104;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1941;
			goto yy2104;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy2104;
			goto yy1941;
		}
	}
yy2106:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy2274;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2274;
	goto yy1730;
yy2107:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2274;
	if (yych <= '/') goto yy1810;
	if (yych <= '9') goto yy2274;
	goto yy1810;
yy2108:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') goto yy1810;
		if (yych <= '/') goto yy1921;
		if (yych <= '9') goto yy1924;
		goto yy1973;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1810;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1810;
			if (yych <= 'f') goto yy1924;
			goto yy1810;
		}
	}
yy2109:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2263;
				yyt3 = YYCURSOR;
				goto yy2275;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2263;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy2276;
				if (yych >= '{') goto yy1809;
			}
		}
	}
yy2110:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych <= 0x1F) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1967;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych == '%') goto yy1745;
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '\'') goto yy1808;
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych <= '+') goto yy1745;
				yyt3 = YYCURSOR;
				goto yy1967;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= '-') goto yy1777;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				goto yy1779;
			} else {
				if (yych <= '9') goto yy1820;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy1980;
				}
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt3 = YYCURSOR;
					goto yy1967;
				}
				if (yych <= '@') goto yy1785;
				if (yych <= 'Z') goto yy1820;
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '_') goto yy1707;
				if (yych <= '`') {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 'z') goto yy2110;
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	}
yy2111:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'E') goto yy2115;
	goto yy1821;
yy2112:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy1820;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '_') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'Z') goto yy1820;
				if (yych <= '^') goto yy1809;
				goto yy1707;
			}
		} else {
			if (yych <= 'd') {
				if (yych <= '`') goto yy1809;
				goto yy2110;
			} else {
				if (yych <= 'e') goto yy2117;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy2113:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
yy2114:
	if (yych <= ':') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1698;
			if (yych <= '-') goto yy2113;
			goto yy2262;
		} else {
			if (yych <= '/') goto yy1698;
			if (yych <= '9') goto yy2113;
			goto yy2095;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1698;
			if (yych <= 'Z') goto yy2113;
			goto yy1698;
		} else {
			if (yych == '`') goto yy1698;
			if (yych <= 'z') goto yy2113;
			goto yy1698;
		}
	}
yy2115:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
			} else {
				if (yych <= '\r') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				if (yych >= ' ') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= '$') {
				if (yych <= '!') {
					yyt1 = YYCURSOR;
					goto yy2278;
				}
				if (yych <= '"') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
			} else {
				if (yych <= '%') goto yy1821;
				if (yych <= '&') {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				goto yy1821;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2278;
				}
				if (yych <= '-') goto yy1821;
				yyt1 = YYCURSOR;
				goto yy2279;
			} else {
				if (yych <= '9') goto yy1821;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2280;
				}
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2278;
				}
				if (yych <= 'Z') goto yy1821;
				yyt1 = YYCURSOR;
			} else {
				if (yych == '`') {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= 'z') goto yy1821;
				yyt1 = YYCURSOR;
			}
		}
	}
yy2116:
	YYCURSOR = yyt1;
#line 462 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_CONSTANT); }
#line 55022 "../../lnav/src/data_scanner_re.cc"
yy2117:
	yyaccept = 26;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= '\n') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
				goto yy2116;
			} else {
				if (yych <= '\r') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				if (yych <= 0x1F) {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= ' ') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
				goto yy2278;
			}
		} else {
			if (yych <= '%') {
				if (yych <= '"') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				if (yych <= '$') {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				goto yy1821;
			} else {
				if (yych <= '&') {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy2281;
				}
				if (yych <= '*') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				goto yy1821;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2278;
				}
				if (yych <= '-') goto yy1821;
				yyt1 = YYCURSOR;
				goto yy2279;
			} else {
				if (yych <= '9') goto yy1821;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2280;
				}
				if (yych <= ';') {
					yyt1 = YYCURSOR;
					goto yy2277;
				}
				yyt1 = YYCURSOR;
				goto yy2116;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2278;
				}
				if (yych <= 'Z') goto yy1821;
				yyt1 = YYCURSOR;
				goto yy2116;
			} else {
				if (yych <= '_') goto yy1821;
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy2116;
				}
				if (yych <= 'z') goto yy2110;
				yyt1 = YYCURSOR;
				goto yy2116;
			}
		}
	}
yy2118:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy1966;
		goto yy1854;
	} else {
		if (yych <= '\r') goto yy1966;
		if (yych == ' ') goto yy1966;
		goto yy1854;
	}
yy2119:
	yyaccept = 17;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 4) {
		goto yy1991;
	}
	if (yych <= '"') {
		if (yych <= 0x15) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1856;
				if (yych <= 0x08) goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1987;
			} else {
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych == '\r') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				goto yy1749;
			}
		} else {
			if (yych <= 0x1B) {
				if (yych <= 0x16) goto yy1856;
				if (yych <= 0x1A) goto yy1749;
				goto yy1856;
			} else {
				if (yych <= 0x1F) goto yy1749;
				if (yych == '!') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				yyt3 = YYCURSOR;
				goto yy1987;
			}
		}
	} else {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych <= '&') goto yy1749;
				if (yych <= '\'') goto yy2121;
				yyt3 = YYCURSOR;
				goto yy1987;
			} else {
				if (yych == ',') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				if (yych <= '-') goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1990;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy1749;
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				if (yych <= '>') goto yy1749;
				yyt3 = YYCURSOR;
				goto yy1988;
			} else {
				if (yych <= 0x7F) goto yy1749;
				if (yych <= 0xC1) goto yy1856;
				if (yych <= 0xF4) goto yy1749;
				goto yy1856;
			}
		}
	}
yy2120:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) goto yy1854;
			if (yych <= '\n') goto yy1809;
			if (yych <= '\f') goto yy1854;
			goto yy1809;
		} else {
			if (yych <= '"') {
				if (yych <= 0x1F) goto yy1854;
				goto yy1809;
			} else {
				if (yych <= '&') goto yy1854;
				if (yych <= '*') goto yy1809;
				goto yy1854;
			}
		}
	} else {
		if (yych <= ';') {
			if (yych <= '-') {
				if (yych <= ',') goto yy1809;
				goto yy1854;
			} else {
				if (yych <= '.') goto yy1809;
				if (yych <= '9') goto yy1854;
				goto yy1809;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '>') goto yy1854;
				goto yy1809;
			} else {
				if (yych <= '`') goto yy1854;
				if (yych <= 'z') goto yy1809;
				goto yy1854;
			}
		}
	}
yy2121:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '&') {
				if (yych == '!') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				yyt2 = YYCURSOR;
				goto yy1853;
			} else {
				if (yych <= '\'') {
					yyt1 = YYCURSOR;
					goto yy2282;
				}
				if (yych <= '+') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt1 = yyt3 = YYCURSOR;
				goto yy2118;
			}
		} else {
			if (yych <= '>') {
				if (yych == '.') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				yyt2 = YYCURSOR;
				goto yy1853;
			} else {
				if (yych <= '?') {
					yyt1 = yyt3 = YYCURSOR;
					goto yy2118;
				}
				if (yych <= 'R') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 's') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				if (yych <= 'r') {
					yyt1 = YYCURSOR;
					goto yy2120;
				}
				goto yy1808;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2120;
				}
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1853;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1857;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1858;
				}
				yyt2 = YYCURSOR;
				goto yy1859;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1860;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1861;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1862;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	}
yy2122:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '\'') goto yy1808;
	if (yych <= '`') goto yy1794;
	if (yych <= 'z') goto yy2283;
	goto yy1794;
yy2123:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy2284;
			}
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy2284;
			}
			if (yych <= 0xC1) goto yy1730;
			yyt1 = YYCURSOR;
			goto yy2286;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy2287;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy2288;
			}
			yyt1 = YYCURSOR;
			goto yy2289;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy2290;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy2291;
			}
			goto yy1730;
		}
	}
yy2124:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2292;
	goto yy1730;
yy2125:
	yych = *++YYCURSOR;
	if (yych == '=') goto yy2125;
	if (yych != ']') goto yy1730;
yy2126:
	++YYCURSOR;
#line 266 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 55419 "../../lnav/src/data_scanner_re.cc"
yy2127:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 8) {
		goto yy2003;
	}
	if (yych <= '9') {
		if (yych == '-') goto yy1850;
		if (yych >= '0') goto yy1850;
	} else {
		if (yych <= 'Z') {
			if (yych >= 'A') goto yy2127;
		} else {
			if (yych <= '`') goto yy2128;
			if (yych <= 'z') goto yy2127;
		}
	}
yy2128:
#line 460 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_EMAIL); }
#line 55440 "../../lnav/src/data_scanner_re.cc"
yy2129:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy1909;
		} else {
			if (yych == ',') goto yy1909;
			if (yych <= '.') goto yy1746;
			goto yy1909;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1909;
			if (yych <= 'Z') goto yy1746;
			goto yy1909;
		} else {
			if (yych == '`') goto yy1909;
			if (yych <= 'z') goto yy1746;
			goto yy1909;
		}
	}
yy2130:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy2150;
	if (yych == '-') goto yy2150;
	goto yy1746;
yy2131:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy2129;
		goto yy1746;
	} else {
		if (yych <= 'b') goto yy2129;
		if (yych == 'i') goto yy2293;
		goto yy1746;
	}
yy2132:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') goto yy1746;
				goto yy1909;
			} else {
				if (yych == ',') goto yy1909;
				goto yy1746;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '/') goto yy1909;
				if (yych <= '9') goto yy1746;
				goto yy1909;
			} else {
				if (yych == 'B') goto yy2129;
				goto yy1746;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'S') goto yy2129;
				if (yych <= 'Z') goto yy1746;
				goto yy1909;
			} else {
				if (yych == '`') goto yy1909;
				goto yy1746;
			}
		} else {
			if (yych <= 'i') {
				if (yych <= 'b') goto yy2129;
				if (yych <= 'h') goto yy1746;
				goto yy2294;
			} else {
				if (yych == 's') goto yy2129;
				if (yych <= 'z') goto yy1746;
				goto yy1909;
			}
		}
	}
yy2133:
	yych = *++YYCURSOR;
	if (yych == 'S') goto yy2129;
	if (yych == 's') goto yy2129;
	goto yy1746;
yy2134:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2134;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2196;
		case 'E':
		case 'e': goto yy2295;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2197;
		case 'm': goto yy2198;
		case 'p':
		case 'u': goto yy2199;
		default: goto yy1683;
	}
yy2135:
	yych = *++YYCURSOR;
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1730;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy1730;
			goto yy2135;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1730;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1745;
			goto yy1730;
		} else {
			if (yych == '`') goto yy1730;
			if (yych <= 'z') goto yy1745;
			goto yy1730;
		}
	}
yy2136:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2136;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2196;
		case 'E':
		case 'e': goto yy2295;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2197;
		case 'm': goto yy2198;
		case 'p':
		case 'u': goto yy2199;
		default: goto yy1683;
	}
yy2137:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ')') goto yy1898;
			goto yy2137;
		} else {
			if (yych == '/') goto yy2296;
			if (yych <= 0x7F) goto yy1898;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy1900;
			if (yych <= 0xE0) goto yy1901;
			goto yy1902;
		} else {
			if (yych <= 0xF0) goto yy1903;
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy1730;
		}
	}
yy2138:
	++YYCURSOR;
	goto yy2015;
yy2139:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1280+yych] & 128) {
		goto yy1771;
	}
	if (yych <= 0xDF) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2015;
			if (yych <= '\'') goto yy1898;
			goto yy1899;
		} else {
			if (yych <= 0x7F) goto yy1898;
			if (yych <= 0xC1) goto yy2015;
			goto yy1900;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy1901;
			if (yych <= 0xEF) goto yy1902;
			goto yy1903;
		} else {
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy2015;
		}
	}
yy2140:
	yych = *++YYCURSOR;
	if (yych == 'r') goto yy2297;
	goto yy1770;
yy2141:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
yy2142:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych >= '+') goto yy1745;
		} else {
			if (yych <= ',') goto yy2143;
			if (yych <= '.') goto yy1777;
			if (yych <= '/') goto yy1779;
			goto yy2141;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy2143;
			if (yych <= '@') goto yy1785;
			goto yy2141;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy2141;
			} else {
				if (yych <= '`') goto yy2143;
				if (yych <= 'z') goto yy2141;
			}
		}
	}
yy2143:
#line 434 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_VERSION_NUMBER);
       }
#line 55800 "../../lnav/src/data_scanner_re.cc"
yy2144:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2298;
	if (yych <= '/') goto yy2301;
	if (yych <= '9') goto yy2299;
	goto yy2301;
yy2145:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2298;
	if (yych <= '/') goto yy2301;
	if (yych <= '9') goto yy2144;
	goto yy2301;
yy2146:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy2298;
		goto yy2301;
	} else {
		if (yych <= '4') goto yy2144;
		if (yych <= '5') goto yy2302;
		if (yych <= '9') goto yy2299;
		goto yy2301;
	}
yy2147:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
yy2148:
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy2017;
		case '.': goto yy2149;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2147;
		case ':': goto yy1929;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2020;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2022;
		case 'E':
		case 'e': goto yy2023;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2024;
		case 'm': goto yy2025;
		case 'p':
		case 'u': goto yy2026;
		default: goto yy1909;
	}
yy2149:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1730;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy1730;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy1730;
			goto yy2300;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1730;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy1730;
		} else {
			if (yych == '`') goto yy1730;
			if (yych <= 'z') goto yy1793;
			goto yy1730;
		}
	}
yy2150:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1746;
	if (yych <= '9') goto yy2303;
	goto yy1746;
yy2151:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2304;
		if (yych <= '@') goto yy1746;
		goto yy2141;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy2141;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy2141;
			goto yy1746;
		}
	}
yy2152:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy2022;
	if (yych == 'b') goto yy2022;
	goto yy2021;
yy2153:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2022;
		goto yy2021;
	} else {
		if (yych <= 'b') goto yy2022;
		if (yych == 'n') goto yy2022;
		goto yy2021;
	}
yy2154:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych >= '+') goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych >= '-') goto yy1914;
			} else {
				if (yych <= '.') goto yy1777;
				if (yych <= '/') goto yy1779;
				goto yy1793;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy2305;
			if (yych <= '?') goto yy2155;
			if (yych <= '@') goto yy1785;
			goto yy1793;
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy1793;
			} else {
				if (yych <= '`') goto yy2155;
				if (yych <= 'z') goto yy1793;
			}
		}
	}
yy2155:
#line 303 "../../lnav/src/data_scanner_re.re"
	{
           if ((YYCURSOR.val - (this->ds_input.udata() + this->ds_next_offset)) == 17) {
               RET(DT_MAC_ADDRESS);
           } else {
               RET(DT_HEX_DUMP);
           }
       }
#line 56011 "../../lnav/src/data_scanner_re.cc"
yy2156:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	goto yy2160;
yy2157:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	if (yych <= '/') goto yy2160;
	if (yych <= '5') goto yy2156;
	goto yy2160;
yy2158:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') goto yy1913;
	if (yych <= 'Z') goto yy2306;
	if (yych <= '`') goto yy1913;
	if (yych <= 'z') goto yy2306;
	goto yy1913;
yy2159:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
yy2160:
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1789;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1789;
				goto yy2017;
			} else {
				if (yych <= '.') goto yy2149;
				if (yych <= '/') goto yy1779;
				goto yy2159;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy1789;
			if (yych <= '@') goto yy1785;
			goto yy2020;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1789;
				goto yy2020;
			} else {
				if (yych <= '`') goto yy1789;
				if (yych <= 'z') goto yy2020;
				goto yy1789;
			}
		}
	}
yy2161:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1774;
	if (yych <= '9') goto yy2307;
	goto yy1774;
yy2162:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2308;
		if (yych <= '@') goto yy1746;
		goto yy1793;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2163:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2309;
		if (yych <= '@') goto yy1730;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy2164:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1781;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '7') goto yy2310;
				goto yy2311;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1927;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2312;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy2312;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy2165:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1783;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2311;
				goto yy1927;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2312;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2312;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy2166:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy1730;
			goto yy2313;
		} else {
			if (yych <= '.') goto yy1730;
			if (yych <= '/') goto yy1921;
			goto yy1924;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy2314;
			if (yych <= '@') goto yy1730;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy2167:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2312;
				if (yych <= ':') goto yy1791;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2312;
			if (yych <= 'Z') goto yy1791;
			goto yy1783;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2312;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2168:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '-') {
				yyt1 = YYCURSOR;
				goto yy2315;
			}
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy2284;
			}
			yyt1 = YYCURSOR;
			goto yy2316;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy2317;
				yyt1 = YYCURSOR;
				goto yy2284;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy2316;
				}
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy2284;
				}
				yyt1 = YYCURSOR;
				goto yy2316;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy2284;
			}
			if (yych <= 0xC1) goto yy2155;
			if (yych <= 0xDF) {
				yyt1 = YYCURSOR;
				goto yy2286;
			}
			yyt1 = YYCURSOR;
			goto yy2287;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt1 = YYCURSOR;
					goto yy2288;
				}
				yyt1 = YYCURSOR;
				goto yy2289;
			} else {
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2290;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2291;
				}
				goto yy2155;
			}
		}
	}
yy2169:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy2172;
			goto yy2319;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy2172;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy2172;
			goto yy2155;
		}
	}
yy2170:
	yyaccept = 31;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 32) {
		goto yy2040;
	}
	if (yych <= 0xC1) {
		if (yych <= ')') {
			if (yych <= '"') goto yy2171;
			if (yych <= '&') goto yy2170;
		} else {
			if (yych <= 'Z') goto yy2170;
			if (yych <= ']') goto yy2171;
			if (yych <= 'z') goto yy2170;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2042;
			if (yych <= 0xE0) goto yy2043;
			goto yy2044;
		} else {
			if (yych <= 0xF0) goto yy2045;
			if (yych <= 0xF3) goto yy2046;
			if (yych <= 0xF4) goto yy2047;
		}
	}
yy2171:
#line 299 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_URL); }
#line 56377 "../../lnav/src/data_scanner_re.cc"
yy2172:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2320;
		if (yych <= ':') goto yy2049;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2320;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2320;
		goto yy1730;
	}
yy2173:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2321;
		if (yych <= ':') goto yy2175;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2174;
		if (yych <= '`') goto yy1730;
		if (yych >= 'g') goto yy1730;
	}
yy2174:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2322;
		if (yych <= ':') goto yy2323;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2322;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2322;
		goto yy1730;
	}
yy2175:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2324;
			} else {
				if (yych <= '1') goto yy2325;
				if (yych <= '2') goto yy2326;
				goto yy2324;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2327;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2327;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2176:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1941;
		if (yych <= '9') goto yy1912;
		if (yych <= '@') goto yy1941;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1941;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy1912;
			goto yy1941;
		}
	}
yy2177:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1941;
			if (yych <= '0') goto yy2328;
			goto yy2329;
		} else {
			if (yych <= '2') goto yy2330;
			if (yych <= '9') goto yy2328;
			goto yy1941;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1912;
			if (yych <= '^') goto yy1941;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy1912;
			goto yy1941;
		}
	}
yy2178:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2331;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2331;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2331;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2179:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1730;
			goto yy1940;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2332;
			goto yy1779;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2332;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2332;
			goto yy1730;
		}
	}
yy2180:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy2331;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2331;
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2331;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2181:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2333;
				yyt2 = YYCURSOR;
				goto yy2179;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2331;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2331;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2182:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '5') goto yy2333;
				goto yy2331;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'F') goto yy2331;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2331;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2183:
	yyaccept = 27;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '.') {
			if (yych <= ',') goto yy2128;
			if (yych <= '-') goto yy2056;
			goto yy2057;
		} else {
			if (yych <= '/') goto yy1779;
			if (yych <= '9') goto yy1925;
			if (yych <= ':') goto yy1929;
			goto yy2128;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1779;
			if (yych <= 'Z') goto yy2183;
			goto yy2128;
		} else {
			if (yych <= '_') goto yy1912;
			if (yych <= '`') goto yy2128;
			if (yych <= 'z') goto yy2183;
			goto yy2128;
		}
	}
yy2184:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	goto yy2206;
yy2185:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy2021;
			goto yy1683;
		} else {
			if (yych == ',') goto yy1683;
			goto yy2021;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy2021;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy2021;
			goto yy1683;
		}
	}
yy2186:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy1937;
	goto yy2021;
yy2187:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2185;
		goto yy2021;
	} else {
		if (yych <= 'b') goto yy2185;
		if (yych == 'i') goto yy2334;
		goto yy2021;
	}
yy2188:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1683;
				goto yy2021;
			} else {
				if (yych == '+') goto yy2021;
				goto yy1683;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy2021;
				if (yych <= '?') goto yy1683;
				goto yy2021;
			} else {
				if (yych <= 'B') goto yy2185;
				if (yych <= 'R') goto yy2021;
				goto yy2185;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2021;
				if (yych <= '^') goto yy1683;
				goto yy2021;
			} else {
				if (yych <= '`') goto yy1683;
				if (yych <= 'a') goto yy2021;
				goto yy2185;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy2335;
				goto yy2021;
			} else {
				if (yych <= 's') goto yy2185;
				if (yych <= 'z') goto yy2021;
				goto yy1683;
			}
		}
	}
yy2189:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy2185;
	if (yych == 's') goto yy2185;
	goto yy2021;
yy2190:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2018;
	if (yych <= '/') goto yy2206;
	if (yych <= '5') goto yy2184;
	goto yy2206;
yy2191:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy2067;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2336;
		case ':': goto yy1927;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2312;
		case 'B':
		case 'b': goto yy2337;
		case 'E':
		case 'e': goto yy2338;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2192:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2312;
				if (yych <= ':') goto yy1791;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2312;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2312;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy2193:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1937;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy2064;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy2312;
				goto yy1791;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy2312;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2312;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2194:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy2195:
	switch (yych) {
		case '%':
		case '+': goto yy1745;
		case '-':
		case '.': goto yy1777;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2194;
		case ':': goto yy1929;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1793;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2202;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2201;
		case 'm': goto yy2203;
		case 'p':
		case 'u': goto yy2204;
		default: goto yy1683;
	}
yy2196:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy1683;
		} else {
			if (yych == ',') goto yy1683;
			if (yych <= '.') goto yy1746;
			goto yy1683;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1746;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1746;
			goto yy1683;
		}
	}
yy2197:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy2196;
		goto yy1746;
	} else {
		if (yych <= 'b') goto yy2196;
		if (yych == 'i') goto yy2339;
		goto yy1746;
	}
yy2198:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') goto yy1746;
				goto yy1683;
			} else {
				if (yych == ',') goto yy1683;
				goto yy1746;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '/') goto yy1683;
				if (yych <= '9') goto yy1746;
				goto yy1683;
			} else {
				if (yych == 'B') goto yy2196;
				goto yy1746;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'S') goto yy2196;
				if (yych <= 'Z') goto yy1746;
				goto yy1683;
			} else {
				if (yych == '`') goto yy1683;
				goto yy1746;
			}
		} else {
			if (yych <= 'i') {
				if (yych <= 'b') goto yy2196;
				if (yych <= 'h') goto yy1746;
				goto yy2340;
			} else {
				if (yych == 's') goto yy2196;
				if (yych <= 'z') goto yy1746;
				goto yy1683;
			}
		}
	}
yy2199:
	yych = *++YYCURSOR;
	if (yych == 'S') goto yy2196;
	if (yych == 's') goto yy2196;
	goto yy1746;
yy2200:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '*') {
			if (yych == '%') goto yy2195;
			goto yy2155;
		} else {
			if (yych <= '+') goto yy2195;
			if (yych <= ',') goto yy2155;
			if (yych <= '-') goto yy1914;
			goto yy2195;
		}
	} else {
		if (yych <= '^') {
			if (yych <= ':') goto yy2305;
			if (yych <= '?') goto yy2155;
			if (yych <= 'Z') goto yy2195;
			goto yy2155;
		} else {
			if (yych == '`') goto yy2155;
			if (yych <= 'z') goto yy2195;
			goto yy2155;
		}
	}
yy2201:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych != 'B') goto yy1794;
	} else {
		if (yych <= 'b') goto yy2202;
		if (yych == 'i') goto yy2341;
		goto yy1794;
	}
yy2202:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1794;
			goto yy1683;
		} else {
			if (yych == ',') goto yy1683;
			goto yy1794;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1794;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1794;
			goto yy1683;
		}
	}
yy2203:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1683;
				goto yy1794;
			} else {
				if (yych == '+') goto yy1794;
				goto yy1683;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy1794;
				if (yych <= '?') goto yy1683;
				goto yy1794;
			} else {
				if (yych <= 'B') goto yy2202;
				if (yych <= 'R') goto yy1794;
				goto yy2202;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1794;
				if (yych <= '^') goto yy1683;
				goto yy1794;
			} else {
				if (yych <= '`') goto yy1683;
				if (yych <= 'a') goto yy1794;
				goto yy2202;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy2342;
				goto yy1794;
			} else {
				if (yych <= 's') goto yy2202;
				if (yych <= 'z') goto yy1794;
				goto yy1683;
			}
		}
	}
yy2204:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy2202;
	if (yych == 's') goto yy2202;
	goto yy1794;
yy2205:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
yy2206:
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy2017;
		case '.': goto yy2149;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2205;
		case ':': goto yy1929;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2020;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2185;
		case 'E':
		case 'e': goto yy2186;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2187;
		case 'm': goto yy2188;
		case 'p':
		case 'u': goto yy2189;
		default: goto yy1683;
	}
yy2207:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2343;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2344;
	goto yy1730;
yy2208:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2343;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2207;
	goto yy1730;
yy2209:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy2343;
		goto yy1730;
	} else {
		if (yych <= '4') goto yy2207;
		if (yych <= '5') goto yy2345;
		if (yych <= '9') goto yy2344;
		goto yy1730;
	}
yy2210:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2346;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy2070;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2346;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2346;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2211:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2347;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2348;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2347;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2347;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2212:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2346;
				yyt1 = YYCURSOR;
				goto yy2070;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2346;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2346;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2213:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= 'E') {
			if (yych <= '9') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2346;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2070;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2346;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2349;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'e') goto yy2346;
				goto yy2349;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2214:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy2350;
				goto yy1730;
			} else {
				if (yych <= '\r') goto yy2350;
				if (yych <= 0x1F) goto yy1730;
				goto yy2350;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy2077;
				goto yy1730;
			} else {
				if (yych <= '\'') goto yy2078;
				if (yych <= ',') goto yy1730;
				goto yy2214;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1730;
				if (yych <= ':') goto yy2214;
				goto yy1730;
			} else {
				if (yych <= '=') goto yy2246;
				if (yych <= '>') goto yy2081;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2214;
				if (yych <= '^') goto yy1730;
				goto yy2214;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2214;
				goto yy1730;
			}
		}
	}
yy2215:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 2) {
		goto yy2215;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			goto yy2351;
		} else {
			if (yych <= '\\') goto yy2216;
			if (yych <= 0xC1) goto yy1730;
			goto yy2217;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2218;
			if (yych <= 0xEF) goto yy2219;
			goto yy2220;
		} else {
			if (yych <= 0xF3) goto yy2221;
			if (yych <= 0xF4) goto yy2222;
			goto yy1730;
		}
	}
yy2216:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2215;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2218;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2219;
			goto yy2220;
		} else {
			if (yych <= 0xF3) goto yy2221;
			if (yych <= 0xF4) goto yy2222;
			goto yy1730;
		}
	}
yy2217:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2215;
	goto yy1730;
yy2218:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2217;
	goto yy1730;
yy2219:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2217;
	goto yy1730;
yy2220:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2219;
	goto yy1730;
yy2221:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2219;
	goto yy1730;
yy2222:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2219;
	goto yy1730;
yy2223:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 4) {
		goto yy2223;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			goto yy2351;
		} else {
			if (yych <= '\\') goto yy2224;
			if (yych <= 0xC1) goto yy1730;
			goto yy2225;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2226;
			if (yych <= 0xEF) goto yy2227;
			goto yy2228;
		} else {
			if (yych <= 0xF3) goto yy2229;
			if (yych <= 0xF4) goto yy2230;
			goto yy1730;
		}
	}
yy2224:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2223;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2226;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2227;
			goto yy2228;
		} else {
			if (yych <= 0xF3) goto yy2229;
			if (yych <= 0xF4) goto yy2230;
			goto yy1730;
		}
	}
yy2225:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2223;
	goto yy1730;
yy2226:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2225;
	goto yy1730;
yy2227:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2225;
	goto yy1730;
yy2228:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2227;
	goto yy1730;
yy2229:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2227;
	goto yy1730;
yy2230:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2227;
	goto yy1730;
yy2231:
	yych = *++YYCURSOR;
yy2232:
	if (yybm_init[512+yych] & 8) {
		goto yy2231;
	}
	if (yych <= 0xE0) {
		if (yych <= '-') {
			if (yych <= '*') goto yy1730;
			goto yy2352;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2239;
			goto yy2240;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2241;
			goto yy2242;
		} else {
			if (yych <= 0xF3) goto yy2243;
			if (yych <= 0xF4) goto yy2244;
			goto yy1730;
		}
	}
yy2233:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2233;
				goto yy2231;
			} else {
				if (yych <= '\r') goto yy2233;
				if (yych == ' ') goto yy2233;
				goto yy2231;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy2234;
				if (yych == '\'') goto yy2235;
				goto yy2231;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy1730;
					goto yy2231;
				} else {
					if (yych <= '-') goto yy2353;
					if (yych <= '/') goto yy2231;
					goto yy2354;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '<') goto yy2231;
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2238;
				goto yy2231;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy2354;
					goto yy2231;
				} else {
					if (yych == '`') goto yy2231;
					goto yy2354;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2231;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				goto yy2240;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2241;
					goto yy2242;
				} else {
					if (yych <= 0xF3) goto yy2243;
					if (yych <= 0xF4) goto yy2244;
					goto yy1730;
				}
			}
		}
	}
yy2234:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 128) {
		goto yy2355;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2231;
			goto yy2215;
		} else {
			if (yych <= '-') goto yy2356;
			if (yych <= '\\') goto yy2357;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy1730;
		}
	}
yy2235:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 1) {
		goto yy2364;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2231;
			goto yy2223;
		} else {
			if (yych <= '-') goto yy2365;
			if (yych <= '\\') goto yy2366;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy1730;
		}
	}
yy2236:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 2) {
		goto yy2373;
	}
	if (yych <= '=') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2231;
				goto yy2233;
			} else {
				if (yych == '\r') goto yy2233;
				if (yych <= 0x1F) goto yy2231;
				goto yy2233;
			}
		} else {
			if (yych <= '\'') {
				if (yych == '"') goto yy2234;
				if (yych <= '&') goto yy2231;
				goto yy2235;
			} else {
				if (yych <= '/') goto yy2231;
				if (yych <= ':') goto yy2079;
				if (yych <= '<') goto yy2231;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy2238;
				if (yych <= '@') goto yy2231;
				if (yych <= 'Z') goto yy2079;
				goto yy2231;
			} else {
				if (yych == '`') goto yy2231;
				if (yych <= 'z') goto yy2079;
				goto yy2231;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				if (yych <= 0xE0) goto yy2240;
				goto yy2241;
			} else {
				if (yych <= 0xF0) goto yy2242;
				if (yych <= 0xF3) goto yy2243;
				if (yych <= 0xF4) goto yy2244;
				goto yy1730;
			}
		}
	}
yy2237:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 4) {
		goto yy2374;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2375;
			goto yy2376;
		} else {
			if (yych <= '*') goto yy2377;
			if (yych <= '-') goto yy2378;
			if (yych <= '>') goto yy2231;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2379;
			if (yych <= 0xE0) goto yy2380;
			goto yy2381;
		} else {
			if (yych <= 0xF0) goto yy2382;
			if (yych <= 0xF3) goto yy2383;
			if (yych <= 0xF4) goto yy2384;
			goto yy1730;
		}
	}
yy2238:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '*') {
		if (yych <= 0x00) goto yy2082;
		if (yych <= ')') goto yy2232;
		goto yy2082;
	} else {
		if (yych <= 0x7F) goto yy2232;
		if (yych <= 0xC1) goto yy2082;
		if (yych <= 0xF4) goto yy2232;
		goto yy2082;
	}
yy2239:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2231;
	goto yy1730;
yy2240:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2239;
	goto yy1730;
yy2241:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2239;
	goto yy1730;
yy2242:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2241;
	goto yy1730;
yy2243:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2241;
	goto yy1730;
yy2244:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2241;
	goto yy1730;
yy2245:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= ' ') {
			if (yych <= '\f') {
				if (yych == '\t') goto yy2245;
				goto yy1730;
			} else {
				if (yych <= '\r') goto yy2245;
				if (yych <= 0x1F) goto yy1730;
				goto yy2245;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy2077;
				goto yy1730;
			} else {
				if (yych <= '\'') goto yy2078;
				if (yych <= ',') goto yy1730;
				goto yy2214;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '<') {
				if (yych <= '/') goto yy1730;
				if (yych <= ':') goto yy2214;
				goto yy1730;
			} else {
				if (yych <= '=') goto yy2246;
				if (yych <= '>') goto yy2081;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2214;
				if (yych <= '^') goto yy1730;
				goto yy2214;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2214;
				goto yy1730;
			}
		}
	}
yy2246:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 32) {
		goto yy2377;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			goto yy2385;
		} else {
			if (yych <= '\'') goto yy2386;
			if (yych <= 0xC1) goto yy1730;
			goto yy2387;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2388;
			if (yych <= 0xEF) goto yy2389;
			goto yy2390;
		} else {
			if (yych <= 0xF3) goto yy2391;
			if (yych <= 0xF4) goto yy2392;
			goto yy1730;
		}
	}
yy2247:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2247;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2247;
			if (yych <= 0x1F) goto yy1730;
			goto yy2247;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy1959;
			goto yy1730;
		} else {
			if (yych <= '=') goto yy2085;
			if (yych <= '>') goto yy1960;
			if (yych <= '?') goto yy1959;
			goto yy1730;
		}
	}
yy2248:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 16) {
		goto yy2248;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '-') goto yy2393;
			goto yy2394;
		} else {
			if (yych <= ':') goto yy2393;
			if (yych <= '>') goto yy1960;
			if (yych <= '?') goto yy2394;
			goto yy2393;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2252;
			if (yych <= 0xE0) goto yy2253;
			goto yy2254;
		} else {
			if (yych <= 0xF0) goto yy2255;
			if (yych <= 0xF3) goto yy2256;
			if (yych <= 0xF4) goto yy2257;
			goto yy1730;
		}
	}
yy2249:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 2) {
		goto yy2394;
	}
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2249;
				goto yy2248;
			} else {
				if (yych <= '\r') goto yy2249;
				if (yych == ' ') goto yy2249;
				goto yy2248;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2250;
				if (yych == '\'') goto yy2251;
				goto yy2248;
			} else {
				if (yych == '.') goto yy2248;
				if (yych <= ':') goto yy2393;
				goto yy2248;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy1960;
				if (yych <= '@') goto yy2248;
				if (yych <= 'Z') goto yy2393;
				goto yy2248;
			} else {
				if (yych == '`') goto yy2248;
				if (yych <= 'z') goto yy2393;
				goto yy2248;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2252;
				if (yych <= 0xE0) goto yy2253;
				goto yy2254;
			} else {
				if (yych <= 0xF0) goto yy2255;
				if (yych <= 0xF3) goto yy2256;
				if (yych <= 0xF4) goto yy2257;
				goto yy1730;
			}
		}
	}
yy2250:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 32) {
		goto yy2250;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2248;
			if (yych <= '-') goto yy2395;
			goto yy2396;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2395;
				goto yy2397;
			} else {
				if (yych <= '?') goto yy2396;
				if (yych <= 'Z') goto yy2395;
				goto yy2398;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy2395;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2399;
			goto yy2400;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2401;
				goto yy2402;
			} else {
				if (yych <= 0xF3) goto yy2403;
				if (yych <= 0xF4) goto yy2404;
				goto yy1730;
			}
		}
	}
yy2251:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 64) {
		goto yy2251;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2248;
			if (yych <= '-') goto yy2405;
			goto yy2406;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2405;
				goto yy2407;
			} else {
				if (yych <= '?') goto yy2406;
				if (yych <= 'Z') goto yy2405;
				goto yy2408;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy2405;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2409;
			goto yy2410;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2411;
				goto yy2412;
			} else {
				if (yych <= 0xF3) goto yy2413;
				if (yych <= 0xF4) goto yy2414;
				goto yy1730;
			}
		}
	}
yy2252:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2248;
	goto yy1730;
yy2253:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2252;
	goto yy1730;
yy2254:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2252;
	goto yy1730;
yy2255:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2254;
	goto yy1730;
yy2256:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2254;
	goto yy1730;
yy2257:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2254;
	goto yy1730;
yy2258:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2415;
			} else {
				if (yych == '\r') goto yy2415;
				goto yy1730;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy2415;
				if (yych <= ',') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '.') goto yy1730;
				if (yych <= '/') goto yy1959;
				goto yy2258;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1730;
				goto yy2260;
			} else {
				if (yych == '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2258;
				if (yych <= '^') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2258;
				goto yy1730;
			}
		}
	}
yy2259:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2259;
			} else {
				if (yych == '\r') goto yy2259;
				goto yy1730;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ' ') goto yy2259;
				if (yych <= ',') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '.') goto yy1730;
				if (yych <= '/') goto yy1959;
				goto yy2258;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= '<') goto yy1730;
			} else {
				if (yych == '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2258;
				if (yych <= '^') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2258;
				goto yy1730;
			}
		}
	}
yy2260:
	yych = *++YYCURSOR;
	if (yych <= '\'') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2416;
			} else {
				if (yych <= '\t') goto yy2417;
				if (yych <= '\f') goto yy2416;
				goto yy2417;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy2417;
				goto yy2416;
			} else {
				if (yych <= '"') goto yy2418;
				if (yych <= '&') goto yy2416;
				goto yy2419;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '>') {
				if (yych <= '=') goto yy2416;
				goto yy1730;
			} else {
				if (yych <= 0x7F) goto yy2416;
				if (yych <= 0xC1) goto yy1730;
				goto yy2420;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2421;
				if (yych <= 0xEF) goto yy2422;
				goto yy2423;
			} else {
				if (yych <= 0xF3) goto yy2424;
				if (yych <= 0xF4) goto yy2425;
				goto yy1730;
			}
		}
	}
yy2261:
	yych = *++YYCURSOR;
	if (yych == ',') goto yy2426;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2261;
	goto yy1730;
yy2262:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1730;
			goto yy2113;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2113;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2113;
			if (yych <= '^') goto yy1730;
			goto yy2113;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy2113;
			goto yy1730;
		}
	}
yy2263:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1821;
		if (yych <= '9') goto yy2427;
		goto yy1821;
	} else {
		if (yych <= 'F') goto yy2427;
		if (yych <= '`') goto yy1821;
		if (yych <= 'f') goto yy2427;
		goto yy1821;
	}
yy2264:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy1730;
			goto yy2313;
		} else {
			if (yych <= '.') goto yy1730;
			if (yych <= '/') goto yy1921;
			goto yy1924;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy2428;
			if (yych <= '@') goto yy1730;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy1924;
			goto yy1730;
		}
	}
yy2265:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			if (yych <= ',') goto yy1941;
			if (yych <= '.') goto yy2114;
			goto yy1941;
		} else {
			if (yych <= '9') goto yy2103;
			if (yych <= ':') goto yy2114;
			goto yy1941;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2103;
			if (yych <= '^') goto yy1941;
			goto yy2103;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy2103;
			goto yy1941;
		}
	}
yy2266:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy2113;
			goto yy1941;
		} else {
			if (yych <= '0') goto yy2429;
			if (yych <= '1') goto yy2430;
			if (yych <= '2') goto yy2431;
			goto yy2429;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1941;
			if (yych <= 'Z') goto yy2103;
			goto yy1941;
		} else {
			if (yych == '`') goto yy1941;
			if (yych <= 'z') goto yy2103;
			goto yy1941;
		}
	}
yy2267:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2432;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2432;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2432;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2268:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1730;
			goto yy1940;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2332;
			goto yy1983;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2332;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2332;
			goto yy1730;
		}
	}
yy2269:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2272;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2432;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2432;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2432;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2270:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2433;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2432;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2432;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2271:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '5') goto yy2433;
				goto yy2432;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2268;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'F') goto yy2432;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2432;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2272:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= ',') goto yy1941;
			goto yy2113;
		} else {
			if (yych <= '/') goto yy1941;
			if (yych <= '9') goto yy2103;
			goto yy1941;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy2103;
			if (yych <= '^') goto yy1941;
			goto yy2103;
		} else {
			if (yych <= '`') goto yy1941;
			if (yych <= 'z') goto yy2103;
			goto yy1941;
		}
	}
yy2273:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '.') {
			if (yych <= ',') goto yy1698;
			if (yych <= '-') goto yy2113;
			goto yy2262;
		} else {
			if (yych <= '/') goto yy1698;
			if (yych <= '9') goto yy2103;
			goto yy2095;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1698;
			if (yych <= 'Z') goto yy2103;
			goto yy1698;
		} else {
			if (yych == '`') goto yy1698;
			if (yych <= 'z') goto yy2103;
			goto yy1698;
		}
	}
yy2274:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2434;
	goto yy1730;
yy2275:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych <= '$') goto yy1810;
			goto yy2313;
		} else {
			if (yych <= '.') goto yy1810;
			if (yych <= '/') goto yy1921;
			goto yy1924;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= ':') goto yy2428;
			if (yych <= '@') goto yy1810;
			goto yy1924;
		} else {
			if (yych <= '`') goto yy1810;
			if (yych <= 'f') goto yy1924;
			goto yy1810;
		}
	}
yy2276:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2427;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2427;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy2435;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy2277:
	++YYCURSOR;
	goto yy2116;
yy2278:
	yych = *++YYCURSOR;
	if (yych <= '\f') {
		if (yych == '\t') goto yy2277;
		goto yy2116;
	} else {
		if (yych <= '\r') goto yy2277;
		if (yych == ' ') goto yy2277;
		goto yy2116;
	}
yy2279:
	yych = *++YYCURSOR;
	if (yych <= '-') {
		if (yych <= '\r') {
			if (yych == '\t') goto yy2277;
			if (yych <= '\f') goto yy1746;
			goto yy2277;
		} else {
			if (yych == ' ') goto yy2277;
			if (yych <= ',') goto yy1746;
			goto yy1969;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '/') goto yy1746;
			if (yych <= '9') goto yy1970;
			if (yych <= '@') goto yy1746;
			goto yy1970;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1746;
				goto yy1970;
			} else {
				if (yych <= '`') goto yy1746;
				if (yych <= 'z') goto yy1970;
				goto yy1746;
			}
		}
	}
yy2280:
	yyaccept = 33;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy1921;
	if (yych == ':') goto yy1983;
	goto yy2116;
yy2281:
	yyaccept = 33;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '+') {
		if (yych <= '\r') {
			if (yych <= 0x08) goto yy2116;
			if (yych <= '\n') goto yy1809;
			if (yych <= '\f') goto yy2116;
			goto yy1809;
		} else {
			if (yych <= '"') {
				if (yych <= 0x1F) goto yy2116;
				goto yy1809;
			} else {
				if (yych <= '&') goto yy2116;
				if (yych <= '*') goto yy1809;
				goto yy2116;
			}
		}
	} else {
		if (yych <= ';') {
			if (yych <= '-') {
				if (yych <= ',') goto yy1809;
				goto yy2116;
			} else {
				if (yych <= '.') goto yy1809;
				if (yych <= '9') goto yy2116;
				goto yy1809;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '>') goto yy2116;
				goto yy1809;
			} else {
				if (yych <= '`') goto yy2116;
				if (yych <= 'z') goto yy1809;
				goto yy2116;
			}
		}
	}
yy2282:
	yyaccept = 23;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[768+yych] & 4) {
		goto yy1991;
	}
	if (yych <= '+') {
		if (yych <= 0x1A) {
			if (yych <= '\n') {
				if (yych <= 0x00) goto yy1854;
				if (yych <= 0x08) goto yy1748;
				if (yych <= '\t') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				yyt3 = YYCURSOR;
				goto yy1966;
			} else {
				if (yych <= '\r') {
					if (yych <= '\f') goto yy1748;
					yyt3 = YYCURSOR;
					goto yy1987;
				} else {
					if (yych == 0x16) goto yy1854;
					goto yy1748;
				}
			}
		} else {
			if (yych <= '!') {
				if (yych <= 0x1B) goto yy1854;
				if (yych <= 0x1F) goto yy1748;
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy1987;
				}
				yyt3 = YYCURSOR;
				goto yy1988;
			} else {
				if (yych <= '&') {
					if (yych <= '"') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				} else {
					if (yych <= '\'') goto yy2121;
					if (yych <= '*') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				}
			}
		}
	} else {
		if (yych <= '\\') {
			if (yych <= '9') {
				if (yych <= ',') {
					yyt3 = YYCURSOR;
					goto yy1988;
				}
				if (yych == '.') {
					yyt3 = YYCURSOR;
					goto yy1990;
				}
				goto yy1748;
			} else {
				if (yych <= '>') {
					if (yych <= ';') {
						yyt3 = YYCURSOR;
						goto yy1987;
					}
					goto yy1748;
				} else {
					if (yych <= '?') {
						yyt3 = YYCURSOR;
						goto yy1988;
					}
					if (yych <= '[') goto yy1748;
					goto yy1751;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy1748;
				if (yych <= 0xC1) goto yy1854;
				if (yych <= 0xDF) goto yy1752;
				goto yy1753;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy1754;
					goto yy1755;
				} else {
					if (yych <= 0xF3) goto yy1756;
					if (yych <= 0xF4) goto yy1757;
					goto yy1854;
				}
			}
		}
	}
yy2283:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ',') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x08) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= '\n') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '\r') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych <= 0x1F) {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= ' ') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1967;
			}
		} else {
			if (yych <= '&') {
				if (yych <= '"') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych == '%') goto yy1745;
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '\'') goto yy1808;
				if (yych <= '*') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				if (yych <= '+') goto yy1745;
				yyt3 = YYCURSOR;
				goto yy1967;
			}
		}
	} else {
		if (yych <= '>') {
			if (yych <= '/') {
				if (yych <= '-') goto yy1777;
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy2436;
				}
				goto yy1779;
			} else {
				if (yych <= '9') goto yy1793;
				if (yych <= ':') {
					yyt3 = YYCURSOR;
					goto yy2437;
				}
				if (yych <= ';') {
					yyt3 = YYCURSOR;
					goto yy1966;
				}
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		} else {
			if (yych <= '^') {
				if (yych <= '?') {
					yyt3 = YYCURSOR;
					goto yy1967;
				}
				if (yych <= '@') goto yy1785;
				if (yych <= 'Z') goto yy1793;
				yyt3 = YYCURSOR;
				goto yy1810;
			} else {
				if (yych <= '_') goto yy1793;
				if (yych <= '`') {
					yyt3 = YYCURSOR;
					goto yy1810;
				}
				if (yych <= 'z') goto yy2283;
				yyt3 = YYCURSOR;
				goto yy1810;
			}
		}
	}
yy2284:
	++YYCURSOR;
yy2285:
	YYCURSOR = yyt1;
#line 301 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 59577 "../../lnav/src/data_scanner_re.cc"
yy2286:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2284;
	goto yy1730;
yy2287:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2286;
	goto yy1730;
yy2288:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2286;
	goto yy1730;
yy2289:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2288;
	goto yy1730;
yy2290:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2288;
	goto yy1730;
yy2291:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2288;
	goto yy1730;
yy2292:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2438;
	goto yy1730;
yy2293:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy2129;
	if (yych == 'b') goto yy2129;
	goto yy1746;
yy2294:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy2129;
		goto yy1746;
	} else {
		if (yych <= 'b') goto yy2129;
		if (yych == 'n') goto yy2129;
		goto yy1746;
	}
yy2295:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy1937;
	if (yych == '-') goto yy1937;
	goto yy1746;
yy2296:
	yyaccept = 28;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 128) {
		goto yy1898;
	}
	if (yych <= 0xE0) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2015;
			goto yy2012;
		} else {
			if (yych <= 0xC1) goto yy2015;
			if (yych <= 0xDF) goto yy1900;
			goto yy1901;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy1902;
			goto yy1903;
		} else {
			if (yych <= 0xF3) goto yy1904;
			if (yych <= 0xF4) goto yy1905;
			goto yy2015;
		}
	}
yy2297:
	yych = *++YYCURSOR;
	if (yych == 'a') goto yy2439;
	goto yy1770;
yy2298:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1746;
			if (yych <= '0') goto yy2440;
			goto yy2441;
		} else {
			if (yych <= '2') goto yy2442;
			if (yych <= '9') goto yy2440;
			goto yy1746;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1793;
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2299:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2298;
	goto yy2301;
yy2300:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
yy2301:
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy2143;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy2143;
				goto yy2017;
			} else {
				if (yych <= '.') goto yy2149;
				if (yych <= '/') goto yy1779;
				goto yy2300;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') goto yy1929;
			if (yych <= '?') goto yy2143;
			if (yych <= '@') goto yy1785;
			goto yy2300;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy2143;
				goto yy2300;
			} else {
				if (yych <= '`') goto yy2143;
				if (yych <= 'z') goto yy2300;
				goto yy2143;
			}
		}
	}
yy2302:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2298;
	if (yych <= '/') goto yy2301;
	if (yych <= '5') goto yy2299;
	goto yy2301;
yy2303:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1745;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2303;
		case '@': goto yy1747;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2129;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2131;
		case 'm': goto yy2132;
		case 'p':
		case 'u': goto yy2133;
		default: goto yy1909;
	}
yy2304:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+': goto yy1745;
		case '-':
		case '.': goto yy1777;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2304;
		case ':': goto yy1929;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy2141;
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy2443;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2444;
		case 'm': goto yy2445;
		case 'p':
		case 'u': goto yy2446;
		default: goto yy1909;
	}
yy2305:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2447;
		if (yych <= ':') goto yy1779;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2447;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2447;
		goto yy1730;
	}
yy2306:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy2448;
	goto yy1913;
yy2307:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2449;
	goto yy1730;
yy2308:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy2450;
	if (yych <= '/') goto yy1794;
	if (yych <= '9') goto yy2451;
	goto yy1794;
yy2309:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '.') goto yy1913;
	if (yych <= '/') goto yy2452;
	if (yych <= '9') goto yy2453;
	goto yy1913;
yy2310:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1781;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '7') goto yy2454;
				goto yy2455;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1927;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2456;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy2456;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy2311:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1783;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2455;
				goto yy1927;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2456;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2456;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy2312:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2456;
				if (yych <= ':') goto yy1791;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2456;
			if (yych <= 'Z') goto yy1791;
			goto yy1783;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2456;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2313:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2457;
		goto yy1730;
	} else {
		if (yych <= 'Z') goto yy2457;
		if (yych <= '`') goto yy1730;
		if (yych <= 'z') goto yy2457;
		goto yy1730;
	}
yy2314:
	yych = *++YYCURSOR;
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2458;
			} else {
				if (yych <= '1') goto yy2459;
				if (yych <= '2') goto yy2460;
				goto yy2458;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2461;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2461;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2315:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy2285;
		if (yych <= '9') goto yy2447;
		goto yy2285;
	} else {
		if (yych <= 'F') goto yy2447;
		if (yych <= '`') goto yy2285;
		if (yych <= 'f') goto yy2447;
		goto yy2285;
	}
yy2316:
	yyaccept = 34;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy2285;
		if (yych <= '9') goto yy2320;
		if (yych <= ':') goto yy2049;
		goto yy2285;
	} else {
		if (yych <= 'F') goto yy2320;
		if (yych <= '`') goto yy2285;
		if (yych <= 'f') goto yy2320;
		goto yy2285;
	}
yy2317:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2462;
		if (yych <= ':') goto yy2175;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2463;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2463;
		goto yy1730;
	}
yy2318:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2447;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2447;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2447;
		goto yy1730;
	}
yy2319:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2463;
		if (yych <= ':') goto yy2175;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2463;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2463;
		goto yy1730;
	}
yy2320:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2049;
	goto yy1730;
yy2321:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2464;
		if (yych <= ':') goto yy2323;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2322;
		if (yych <= '`') goto yy1730;
		if (yych >= 'g') goto yy1730;
	}
yy2322:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2465;
		if (yych >= ';') goto yy1730;
	} else {
		if (yych <= 'F') goto yy2465;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2465;
		goto yy1730;
	}
yy2323:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2466;
		if (yych <= ':') goto yy2467;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2466;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2466;
		goto yy1730;
	}
yy2324:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2469;
				yyt4 = YYCURSOR;
				goto yy2470;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2471;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2471;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2325:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2472;
				yyt4 = YYCURSOR;
				goto yy2470;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2471;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2471;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2326:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2472;
				goto yy2473;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2469;
				if (yych <= ':') {
					yyt4 = YYCURSOR;
					goto yy2470;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2471;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2471;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2327:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2471;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy2470;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2471;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2471;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2328:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2474;
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2475;
	goto yy1913;
yy2329:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2474;
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2328;
	goto yy1913;
yy2330:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy2474;
		goto yy1913;
	} else {
		if (yych <= '4') goto yy2328;
		if (yych <= '5') goto yy2476;
		if (yych <= '9') goto yy2475;
		goto yy1913;
	}
yy2331:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy2477;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2477;
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2477;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2332:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2478;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2479;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2478;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2478;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2333:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				yyt2 = YYCURSOR;
				goto yy2177;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2477;
				yyt2 = YYCURSOR;
				goto yy2179;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2477;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2477;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2334:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy2185;
	if (yych == 'b') goto yy2185;
	goto yy2021;
yy2335:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2185;
		goto yy2021;
	} else {
		if (yych <= 'b') goto yy2185;
		if (yych == 'n') goto yy2185;
		goto yy2021;
	}
yy2336:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy2067;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2480;
		case ':': goto yy1927;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2456;
		case 'B':
		case 'b': goto yy2481;
		case 'E':
		case 'e': goto yy2482;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2337:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2456;
				if (yych <= ':') goto yy1791;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2456;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2456;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy2338:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1937;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy2064;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy2456;
				goto yy1791;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy2456;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2456;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2339:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy2196;
	if (yych == 'b') goto yy2196;
	goto yy1746;
yy2340:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy2196;
		goto yy1746;
	} else {
		if (yych <= 'b') goto yy2196;
		if (yych == 'n') goto yy2196;
		goto yy1746;
	}
yy2341:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy2202;
	if (yych == 'b') goto yy2202;
	goto yy1794;
yy2342:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2202;
		goto yy1794;
	} else {
		if (yych <= 'b') goto yy2202;
		if (yych == 'n') goto yy2202;
		goto yy1794;
	}
yy2343:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy1730;
		if (yych <= '0') goto yy2483;
		goto yy2484;
	} else {
		if (yych <= '2') goto yy2485;
		if (yych <= '9') goto yy2483;
		goto yy1730;
	}
yy2344:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2343;
	goto yy1730;
yy2345:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2343;
	if (yych <= '/') goto yy1730;
	if (yych <= '5') goto yy2344;
	goto yy1730;
yy2346:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt1 = YYCURSOR;
				goto yy2070;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2347:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2486;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2486;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2486;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2348:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2332;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2332;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2332;
			goto yy1730;
		}
	}
yy2349:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2487;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2350:
	yych = *++YYCURSOR;
	if (yych <= '!') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2350;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2350;
			if (yych == ' ') goto yy2350;
			goto yy1730;
		}
	} else {
		if (yych <= '\'') {
			if (yych <= '"') goto yy2077;
			if (yych <= '&') goto yy1730;
			goto yy2078;
		} else {
			if (yych <= '<') goto yy1730;
			if (yych <= '=') goto yy2246;
			if (yych <= '>') goto yy2081;
			goto yy1730;
		}
	}
yy2351:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2488;
			} else {
				if (yych == '\r') goto yy2488;
				goto yy1730;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2488;
				if (yych <= '!') goto yy1730;
				goto yy2077;
			} else {
				if (yych == '\'') goto yy2078;
				goto yy1730;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= ':') {
				if (yych <= '-') goto yy2214;
				if (yych <= '/') goto yy1730;
				goto yy2214;
			} else {
				if (yych == '>') goto yy2081;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2214;
				if (yych <= '^') goto yy1730;
				goto yy2214;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2214;
				goto yy1730;
			}
		}
	}
yy2352:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ',') goto yy2231;
			goto yy2489;
		} else {
			if (yych <= 0x7F) goto yy2231;
			if (yych <= 0xC1) goto yy1730;
			goto yy2239;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2240;
			if (yych <= 0xEF) goto yy2241;
			goto yy2242;
		} else {
			if (yych <= 0xF3) goto yy2243;
			if (yych <= 0xF4) goto yy2244;
			goto yy1730;
		}
	}
yy2353:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2490;
				goto yy2231;
			} else {
				if (yych <= '\r') goto yy2490;
				if (yych == ' ') goto yy2490;
				goto yy2231;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2234;
				if (yych == '\'') goto yy2235;
				goto yy2231;
			} else {
				if (yych <= '-') goto yy2491;
				if (yych <= '/') goto yy2231;
				if (yych >= ';') goto yy2231;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2238;
				if (yych <= '@') goto yy2231;
			} else {
				if (yych == '_') goto yy2354;
				if (yych <= '`') goto yy2231;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2231;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				goto yy2240;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2241;
					goto yy2242;
				} else {
					if (yych <= 0xF3) goto yy2243;
					if (yych <= 0xF4) goto yy2244;
					goto yy1730;
				}
			}
		}
	}
yy2354:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2490;
				goto yy2231;
			} else {
				if (yych <= '\r') goto yy2490;
				if (yych == ' ') goto yy2490;
				goto yy2231;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy2234;
				if (yych == '\'') goto yy2235;
				goto yy2231;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy1730;
					goto yy2231;
				} else {
					if (yych <= '-') goto yy2353;
					if (yych <= '/') goto yy2231;
					goto yy2354;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '@') {
				if (yych <= '<') goto yy2231;
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2238;
				goto yy2231;
			} else {
				if (yych <= '^') {
					if (yych <= 'Z') goto yy2354;
					goto yy2231;
				} else {
					if (yych == '`') goto yy2231;
					goto yy2354;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2231;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				goto yy2240;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2241;
					goto yy2242;
				} else {
					if (yych <= 0xF3) goto yy2243;
					if (yych <= 0xF4) goto yy2244;
					goto yy1730;
				}
			}
		}
	}
yy2355:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 128) {
		goto yy2355;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2492;
			goto yy2215;
		} else {
			if (yych <= '-') goto yy2356;
			if (yych <= '\\') goto yy2357;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy1730;
		}
	}
yy2356:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2492;
			goto yy2355;
		} else {
			if (yych <= '-') goto yy2493;
			if (yych != '\\') goto yy2355;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy1730;
		}
	}
yy2357:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ')') goto yy2355;
			goto yy2215;
		} else {
			if (yych == '-') goto yy2356;
			if (yych <= 0x7F) goto yy2355;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy1730;
		}
	}
yy2358:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2355;
	goto yy1730;
yy2359:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2358;
	goto yy1730;
yy2360:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2358;
	goto yy1730;
yy2361:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2360;
	goto yy1730;
yy2362:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2360;
	goto yy1730;
yy2363:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2360;
	goto yy1730;
yy2364:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 1) {
		goto yy2364;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2492;
			goto yy2223;
		} else {
			if (yych <= '-') goto yy2365;
			if (yych <= '\\') goto yy2366;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy1730;
		}
	}
yy2365:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2492;
			goto yy2364;
		} else {
			if (yych <= '-') goto yy2494;
			if (yych != '\\') goto yy2364;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy1730;
		}
	}
yy2366:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ')') goto yy2364;
			goto yy2223;
		} else {
			if (yych == '-') goto yy2365;
			if (yych <= 0x7F) goto yy2364;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy1730;
		}
	}
yy2367:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2364;
	goto yy1730;
yy2368:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2367;
	goto yy1730;
yy2369:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2367;
	goto yy1730;
yy2370:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2369;
	goto yy1730;
yy2371:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2369;
	goto yy1730;
yy2372:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2369;
	goto yy1730;
yy2373:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 2) {
		goto yy2373;
	}
	if (yych <= '=') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2231;
				goto yy2233;
			} else {
				if (yych == '\r') goto yy2233;
				if (yych <= 0x1F) goto yy2231;
				goto yy2233;
			}
		} else {
			if (yych <= '\'') {
				if (yych == '"') goto yy2234;
				if (yych <= '&') goto yy2231;
				goto yy2235;
			} else {
				if (yych <= '/') goto yy2231;
				if (yych <= ':') goto yy2079;
				if (yych <= '<') goto yy2231;
				goto yy2237;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '>') goto yy2495;
				if (yych <= '@') goto yy2231;
				if (yych <= 'Z') goto yy2079;
				goto yy2231;
			} else {
				if (yych == '`') goto yy2231;
				if (yych <= 'z') goto yy2079;
				goto yy2231;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				if (yych <= 0xE0) goto yy2240;
				goto yy2241;
			} else {
				if (yych <= 0xF0) goto yy2242;
				if (yych <= 0xF3) goto yy2243;
				if (yych <= 0xF4) goto yy2244;
				goto yy1730;
			}
		}
	}
yy2374:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 4) {
		goto yy2374;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych >= '#') goto yy2376;
		} else {
			if (yych <= '*') goto yy2377;
			if (yych <= '-') goto yy2378;
			if (yych <= '>') goto yy2238;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2379;
			if (yych <= 0xE0) goto yy2380;
			goto yy2381;
		} else {
			if (yych <= 0xF0) goto yy2382;
			if (yych <= 0xF3) goto yy2383;
			if (yych <= 0xF4) goto yy2384;
			goto yy1730;
		}
	}
yy2375:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 8) {
		goto yy2375;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2497;
			goto yy2385;
		} else {
			if (yych <= '-') goto yy2498;
			if (yych <= '>') goto yy2499;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2500;
			if (yych <= 0xE0) goto yy2501;
			goto yy2502;
		} else {
			if (yych <= 0xF0) goto yy2503;
			if (yych <= 0xF3) goto yy2504;
			if (yych <= 0xF4) goto yy2505;
			goto yy1730;
		}
	}
yy2376:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 16) {
		goto yy2376;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2497;
			goto yy2386;
		} else {
			if (yych <= '-') goto yy2506;
			if (yych <= '>') goto yy2507;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2508;
			if (yych <= 0xE0) goto yy2509;
			goto yy2510;
		} else {
			if (yych <= 0xF0) goto yy2511;
			if (yych <= 0xF3) goto yy2512;
			if (yych <= 0xF4) goto yy2513;
			goto yy1730;
		}
	}
yy2377:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 32) {
		goto yy2377;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2385;
			goto yy2386;
		} else {
			if (yych <= '>') goto yy2081;
			if (yych <= 0xC1) goto yy1730;
			goto yy2387;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2388;
			if (yych <= 0xEF) goto yy2389;
			goto yy2390;
		} else {
			if (yych <= 0xF3) goto yy2391;
			if (yych <= 0xF4) goto yy2392;
			goto yy1730;
		}
	}
yy2378:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2375;
			goto yy2374;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy2376;
				goto yy2374;
			} else {
				if (yych <= '-') goto yy2514;
				if (yych <= '=') goto yy2374;
				goto yy2238;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy2374;
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2380;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2381;
				goto yy2382;
			} else {
				if (yych <= 0xF3) goto yy2383;
				if (yych <= 0xF4) goto yy2384;
				goto yy1730;
			}
		}
	}
yy2379:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2374;
	goto yy1730;
yy2380:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2379;
	goto yy1730;
yy2381:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2379;
	goto yy1730;
yy2382:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2381;
	goto yy1730;
yy2383:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2381;
	goto yy1730;
yy2384:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2381;
	goto yy1730;
yy2385:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 64) {
		goto yy2385;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			goto yy2515;
		} else {
			if (yych <= '>') goto yy2516;
			if (yych <= 0xC1) goto yy1730;
			goto yy2517;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2518;
			if (yych <= 0xEF) goto yy2519;
			goto yy2520;
		} else {
			if (yych <= 0xF3) goto yy2521;
			if (yych <= 0xF4) goto yy2522;
			goto yy1730;
		}
	}
yy2386:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 128) {
		goto yy2386;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			goto yy2515;
		} else {
			if (yych <= '>') goto yy2523;
			if (yych <= 0xC1) goto yy1730;
			goto yy2524;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2525;
			if (yych <= 0xEF) goto yy2526;
			goto yy2527;
		} else {
			if (yych <= 0xF3) goto yy2528;
			if (yych <= 0xF4) goto yy2529;
			goto yy1730;
		}
	}
yy2387:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2377;
	goto yy1730;
yy2388:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2387;
	goto yy1730;
yy2389:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2387;
	goto yy1730;
yy2390:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2389;
	goto yy1730;
yy2391:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2389;
	goto yy1730;
yy2392:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2389;
	goto yy1730;
yy2393:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 1) {
		goto yy2393;
	}
	if (yych <= 0x7F) {
		if (yych <= '<') {
			if (yych <= 0x00) goto yy1730;
			if (yych != '/') goto yy2248;
		} else {
			if (yych <= '=') goto yy2249;
			if (yych <= '>') goto yy1960;
			if (yych >= '@') goto yy2248;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2252;
			if (yych <= 0xE0) goto yy2253;
			goto yy2254;
		} else {
			if (yych <= 0xF0) goto yy2255;
			if (yych <= 0xF3) goto yy2256;
			if (yych <= 0xF4) goto yy2257;
			goto yy1730;
		}
	}
yy2394:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 16) {
		goto yy2248;
	}
	if (yych <= 'z') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '-') goto yy2393;
			goto yy2394;
		} else {
			if (yych <= ':') goto yy2393;
			if (yych <= '>') goto yy2086;
			if (yych <= '?') goto yy2394;
			goto yy2393;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2252;
			if (yych <= 0xE0) goto yy2253;
			goto yy2254;
		} else {
			if (yych <= 0xF0) goto yy2255;
			if (yych <= 0xF3) goto yy2256;
			if (yych <= 0xF4) goto yy2257;
			goto yy1730;
		}
	}
yy2395:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2395;
				goto yy2250;
			} else {
				if (yych <= '\r') goto yy2395;
				if (yych == ' ') goto yy2395;
				goto yy2250;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2248;
				if (yych == '-') goto yy2395;
				goto yy2250;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2395;
				} else {
					if (yych <= '<') goto yy2250;
					if (yych <= '=') goto yy2530;
					goto yy2397;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2396;
				if (yych <= '@') goto yy2250;
				if (yych <= 'Z') goto yy2395;
				goto yy2250;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2398;
					goto yy2250;
				} else {
					if (yych == '`') goto yy2250;
					goto yy2395;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2250;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2399;
				goto yy2400;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2401;
					goto yy2402;
				} else {
					if (yych <= 0xF3) goto yy2403;
					if (yych <= 0xF4) goto yy2404;
					goto yy1730;
				}
			}
		}
	}
yy2396:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 32) {
		goto yy2250;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2248;
			if (yych <= '-') goto yy2395;
			goto yy2396;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2395;
				goto yy2531;
			} else {
				if (yych <= '?') goto yy2396;
				if (yych <= 'Z') goto yy2395;
				goto yy2398;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy2395;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2399;
			goto yy2400;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2401;
				goto yy2402;
			} else {
				if (yych <= 0xF3) goto yy2403;
				if (yych <= 0xF4) goto yy2404;
				goto yy1730;
			}
		}
	}
yy2397:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1961;
			if (yych <= '!') goto yy2532;
			goto yy2533;
		} else {
			if (yych == '\\') goto yy2534;
			if (yych <= 0x7F) goto yy2532;
			goto yy1961;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2535;
			if (yych <= 0xE0) goto yy2536;
			goto yy2537;
		} else {
			if (yych <= 0xF0) goto yy2538;
			if (yych <= 0xF3) goto yy2539;
			if (yych <= 0xF4) goto yy2540;
			goto yy1961;
		}
	}
yy2398:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2250;
				goto yy2395;
			} else {
				if (yych <= '.') goto yy2250;
				if (yych <= '/') goto yy2396;
				goto yy2395;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2250;
				if (yych <= '>') goto yy2397;
				goto yy2396;
			} else {
				if (yych <= '@') goto yy2250;
				if (yych <= 'Z') goto yy2395;
				goto yy2250;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2250;
				goto yy2395;
			} else {
				if (yych <= 0x7F) goto yy2250;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2400;
				if (yych <= 0xEF) goto yy2401;
				goto yy2402;
			} else {
				if (yych <= 0xF3) goto yy2403;
				if (yych <= 0xF4) goto yy2404;
				goto yy1730;
			}
		}
	}
yy2399:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2250;
	goto yy1730;
yy2400:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2399;
	goto yy1730;
yy2401:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2399;
	goto yy1730;
yy2402:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2401;
	goto yy1730;
yy2403:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2401;
	goto yy1730;
yy2404:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2401;
	goto yy1730;
yy2405:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2405;
				goto yy2251;
			} else {
				if (yych <= '\r') goto yy2405;
				if (yych == ' ') goto yy2405;
				goto yy2251;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2248;
				if (yych == '-') goto yy2405;
				goto yy2251;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2405;
				} else {
					if (yych <= '<') goto yy2251;
					if (yych <= '=') goto yy2541;
					goto yy2407;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2406;
				if (yych <= '@') goto yy2251;
				if (yych <= 'Z') goto yy2405;
				goto yy2251;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2408;
					goto yy2251;
				} else {
					if (yych == '`') goto yy2251;
					goto yy2405;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2251;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2409;
				goto yy2410;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2411;
					goto yy2412;
				} else {
					if (yych <= 0xF3) goto yy2413;
					if (yych <= 0xF4) goto yy2414;
					goto yy1730;
				}
			}
		}
	}
yy2406:
	yych = *++YYCURSOR;
	if (yybm_init[512+yych] & 64) {
		goto yy2251;
	}
	if (yych <= '\\') {
		if (yych <= '/') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2248;
			if (yych <= '-') goto yy2405;
			goto yy2406;
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2405;
				goto yy2542;
			} else {
				if (yych <= '?') goto yy2406;
				if (yych <= 'Z') goto yy2405;
				goto yy2408;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') goto yy2405;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2409;
			goto yy2410;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2411;
				goto yy2412;
			} else {
				if (yych <= 0xF3) goto yy2413;
				if (yych <= 0xF4) goto yy2414;
				goto yy1730;
			}
		}
	}
yy2407:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1961;
			if (yych <= '&') goto yy2543;
			goto yy2533;
		} else {
			if (yych == '\\') goto yy2544;
			if (yych <= 0x7F) goto yy2543;
			goto yy1961;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2545;
			if (yych <= 0xE0) goto yy2546;
			goto yy2547;
		} else {
			if (yych <= 0xF0) goto yy2548;
			if (yych <= 0xF3) goto yy2549;
			if (yych <= 0xF4) goto yy2550;
			goto yy1961;
		}
	}
yy2408:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2251;
				goto yy2405;
			} else {
				if (yych <= '.') goto yy2251;
				if (yych <= '/') goto yy2406;
				goto yy2405;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2251;
				if (yych <= '>') goto yy2407;
				goto yy2406;
			} else {
				if (yych <= '@') goto yy2251;
				if (yych <= 'Z') goto yy2405;
				goto yy2251;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2251;
				goto yy2405;
			} else {
				if (yych <= 0x7F) goto yy2251;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2410;
				if (yych <= 0xEF) goto yy2411;
				goto yy2412;
			} else {
				if (yych <= 0xF3) goto yy2413;
				if (yych <= 0xF4) goto yy2414;
				goto yy1730;
			}
		}
	}
yy2409:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2251;
	goto yy1730;
yy2410:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2409;
	goto yy1730;
yy2411:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2409;
	goto yy1730;
yy2412:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2411;
	goto yy1730;
yy2413:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2411;
	goto yy1730;
yy2414:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2411;
	goto yy1730;
yy2415:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2415;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2415;
			if (yych <= 0x1F) goto yy1730;
			goto yy2415;
		}
	} else {
		if (yych <= '<') {
			if (yych == '/') goto yy1959;
			goto yy1730;
		} else {
			if (yych <= '=') goto yy2260;
			if (yych == '?') goto yy1959;
			goto yy1730;
		}
	}
yy2416:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2416;
				goto yy2551;
			} else {
				if (yych <= '.') goto yy2416;
				if (yych <= '/') goto yy2552;
				goto yy2551;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2416;
				if (yych <= '>') goto yy1730;
				goto yy2552;
			} else {
				if (yych <= '@') goto yy2416;
				if (yych <= 'Z') goto yy2551;
				goto yy2416;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2416;
				goto yy2551;
			} else {
				if (yych <= 0x7F) goto yy2416;
				if (yych <= 0xC1) goto yy1730;
				goto yy2420;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2421;
				if (yych <= 0xEF) goto yy2422;
				goto yy2423;
			} else {
				if (yych <= 0xF3) goto yy2424;
				if (yych <= 0xF4) goto yy2425;
				goto yy1730;
			}
		}
	}
yy2417:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2417;
				goto yy2416;
			} else {
				if (yych <= '\r') goto yy2417;
				if (yych == ' ') goto yy2417;
				goto yy2416;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2418;
				if (yych == '\'') goto yy2419;
				goto yy2416;
			} else {
				if (yych <= '.') {
					if (yych <= '-') goto yy2551;
					goto yy2416;
				} else {
					if (yych <= '/') goto yy2552;
					if (yych <= ':') goto yy2551;
					goto yy2416;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy1730;
				if (yych <= '?') goto yy2552;
				if (yych <= '@') goto yy2416;
				goto yy2551;
			} else {
				if (yych == '_') goto yy2551;
				if (yych <= '`') goto yy2416;
				goto yy2551;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2416;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2420;
				goto yy2421;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2422;
					goto yy2423;
				} else {
					if (yych <= 0xF3) goto yy2424;
					if (yych <= 0xF4) goto yy2425;
					goto yy1730;
				}
			}
		}
	}
yy2418:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '!') goto yy2418;
				goto yy2416;
			} else {
				if (yych == '-') goto yy2553;
				if (yych <= '.') goto yy2418;
				goto yy2554;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2553;
				if (yych <= '=') goto yy2418;
				goto yy2555;
			} else {
				if (yych <= '?') goto yy2554;
				if (yych <= '@') goto yy2418;
				if (yych <= 'Z') goto yy2553;
				goto yy2418;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy2556;
				if (yych <= '^') goto yy2418;
				goto yy2553;
			} else {
				if (yych <= '`') goto yy2418;
				if (yych <= 'z') goto yy2553;
				if (yych <= 0x7F) goto yy2418;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2557;
				if (yych <= 0xE0) goto yy2558;
				goto yy2559;
			} else {
				if (yych <= 0xF0) goto yy2560;
				if (yych <= 0xF3) goto yy2561;
				if (yych <= 0xF4) goto yy2562;
				goto yy1730;
			}
		}
	}
yy2419:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '&') goto yy2419;
				goto yy2416;
			} else {
				if (yych == '-') goto yy2563;
				if (yych <= '.') goto yy2419;
				goto yy2564;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2563;
				if (yych <= '=') goto yy2419;
				goto yy2565;
			} else {
				if (yych <= '?') goto yy2564;
				if (yych <= '@') goto yy2419;
				if (yych <= 'Z') goto yy2563;
				goto yy2419;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy2566;
				if (yych <= '^') goto yy2419;
				goto yy2563;
			} else {
				if (yych <= '`') goto yy2419;
				if (yych <= 'z') goto yy2563;
				if (yych <= 0x7F) goto yy2419;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2567;
				if (yych <= 0xE0) goto yy2568;
				goto yy2569;
			} else {
				if (yych <= 0xF0) goto yy2570;
				if (yych <= 0xF3) goto yy2571;
				if (yych <= 0xF4) goto yy2572;
				goto yy1730;
			}
		}
	}
yy2420:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2416;
	goto yy1730;
yy2421:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2420;
	goto yy1730;
yy2422:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2420;
	goto yy1730;
yy2423:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2422;
	goto yy1730;
yy2424:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2422;
	goto yy1730;
yy2425:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2422;
	goto yy1730;
yy2426:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2573;
	goto yy1730;
yy2427:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1821;
		if (yych <= '9') goto yy2574;
		goto yy1821;
	} else {
		if (yych <= 'F') goto yy2574;
		if (yych <= '`') goto yy1821;
		if (yych <= 'f') goto yy2574;
		goto yy1821;
	}
yy2428:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= '2') {
			if (yych <= '/') {
				if (yych == '-') {
					yyt2 = YYCURSOR;
					goto yy2098;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '0') goto yy2575;
				if (yych <= '1') goto yy2576;
				goto yy2577;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2575;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2578;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2578;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2429:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2579;
	if (yych <= '/') goto yy2104;
	if (yych <= '9') goto yy2580;
	goto yy2104;
yy2430:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2579;
	if (yych <= '/') goto yy2104;
	if (yych <= '9') goto yy2429;
	goto yy2104;
yy2431:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy2579;
		goto yy2104;
	} else {
		if (yych <= '4') goto yy2429;
		if (yych <= '5') goto yy2581;
		if (yych <= '9') goto yy2580;
		goto yy2104;
	}
yy2432:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2272;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2582;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2582;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2582;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2433:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2266;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2582;
				yyt2 = YYCURSOR;
				goto yy2268;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2582;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2582;
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2434:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy2583;
	goto yy1730;
yy2435:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2574;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2574;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy2584;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy2436:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy1966;
			goto yy1746;
		} else {
			if (yych <= '\r') goto yy1966;
			if (yych == ' ') goto yy1966;
			goto yy1746;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1793;
			if (yych <= '@') goto yy1746;
			if (yych <= 'Z') goto yy1793;
			goto yy1746;
		} else {
			if (yych == '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2437:
	yyaccept = 22;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1779;
	goto yy1810;
yy2438:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy2587;
			}
			yyt2 = YYCURSOR;
			goto yy2585;
		} else {
			if (yych <= ':') goto yy1730;
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy2588;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy2589;
			}
			yyt2 = YYCURSOR;
			goto yy2590;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy2591;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy2592;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy2593;
			}
			goto yy1730;
		}
	}
yy2439:
	yych = *++YYCURSOR;
	if (yych == 'm') goto yy2594;
	goto yy1770;
yy2440:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2598;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy2599;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy2600;
				}
				goto yy2601;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2602;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy2603;
				}
				yyt1 = YYCURSOR;
				goto yy2604;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				goto yy2143;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy2605;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy2606;
				}
				yyt1 = YYCURSOR;
				goto yy2607;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy2608;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2609;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2610;
				}
				goto yy2143;
			}
		}
	}
yy2441:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2598;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy2599;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy2600;
				}
				goto yy2440;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2602;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy2603;
				}
				yyt1 = YYCURSOR;
				goto yy2604;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				goto yy2143;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy2605;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy2606;
				}
				yyt1 = YYCURSOR;
				goto yy2607;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy2608;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2609;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2610;
				}
				goto yy2143;
			}
		}
	}
yy2442:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy2598;
				}
				yyt1 = YYCURSOR;
				goto yy2599;
			}
		} else {
			if (yych <= '5') {
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy2600;
				}
				if (yych <= '4') goto yy2440;
				goto yy2611;
			} else {
				if (yych <= '9') goto yy2601;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2602;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2603;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2604;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				goto yy2143;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy2605;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy2606;
				}
				yyt1 = YYCURSOR;
				goto yy2607;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy2608;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2609;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2610;
				}
				goto yy2143;
			}
		}
	}
yy2443:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy2142;
			goto yy1909;
		} else {
			if (yych == ',') goto yy1909;
			goto yy2142;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1909;
			if (yych <= 'Z') goto yy2142;
			goto yy1909;
		} else {
			if (yych == '`') goto yy1909;
			if (yych <= 'z') goto yy2142;
			goto yy1909;
		}
	}
yy2444:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2443;
		goto yy2142;
	} else {
		if (yych <= 'b') goto yy2443;
		if (yych == 'i') goto yy2612;
		goto yy2142;
	}
yy2445:
	yyaccept = 21;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1909;
				goto yy2142;
			} else {
				if (yych == '+') goto yy2142;
				goto yy1909;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy2142;
				if (yych <= '?') goto yy1909;
				goto yy2142;
			} else {
				if (yych <= 'B') goto yy2443;
				if (yych <= 'R') goto yy2142;
				goto yy2443;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2142;
				if (yych <= '^') goto yy1909;
				goto yy2142;
			} else {
				if (yych <= '`') goto yy1909;
				if (yych <= 'a') goto yy2142;
				goto yy2443;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy2613;
				goto yy2142;
			} else {
				if (yych <= 's') goto yy2443;
				if (yych <= 'z') goto yy2142;
				goto yy1909;
			}
		}
	}
yy2446:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'S') goto yy2443;
	if (yych == 's') goto yy2443;
	goto yy2142;
yy2447:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2614;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2614;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2614;
		goto yy1730;
	}
yy2448:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2615;
		if (yych <= '@') goto yy1730;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy2449:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2616;
	goto yy1730;
yy2450:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1746;
		if (yych <= '9') goto yy2617;
		if (yych <= '@') goto yy1746;
		goto yy1793;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1746;
			goto yy1793;
		} else {
			if (yych <= '`') goto yy1746;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2451:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy2450;
	goto yy1794;
yy2452:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2619;
		if (yych <= '@') goto yy1730;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy2453:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy2452;
	goto yy1913;
yy2454:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1781;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '7') goto yy2620;
				goto yy2621;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1927;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy2622;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy2622;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy2455:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1783;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2621;
				goto yy1927;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2622;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2622;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy2456:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2622;
				if (yych <= ':') goto yy1791;
				goto yy1783;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2622;
			if (yych <= 'Z') goto yy1791;
			goto yy1783;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2622;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2457:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2457;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy2457;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy2457;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2458:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2624;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2625;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2626;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2626;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2459:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2627;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2625;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2626;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2626;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2460:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '4') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2627;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy2628;
				if (yych <= '9') goto yy2624;
				yyt1 = YYCURSOR;
				goto yy2625;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2626;
				goto yy1912;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'f') goto yy2626;
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2461:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2626;
				yyt1 = YYCURSOR;
				goto yy2625;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2626;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2626;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2462:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2629;
		if (yych <= ':') goto yy2323;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2630;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2630;
		goto yy1730;
	}
yy2463:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2630;
		if (yych <= ':') goto yy2323;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2630;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2630;
		goto yy1730;
	}
yy2464:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= '9') {
			if (yych == '.') {
				yyt2 = YYCURSOR;
				goto yy2587;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			yyt1 = YYCURSOR;
			goto yy2631;
		} else {
			if (yych <= '@') {
				if (yych <= ':') goto yy2323;
				yyt2 = YYCURSOR;
				goto yy2585;
			} else {
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy2631;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy2585;
				}
				yyt1 = YYCURSOR;
				goto yy2631;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy2588;
			}
			yyt2 = YYCURSOR;
			goto yy2589;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy2590;
				}
				yyt2 = YYCURSOR;
				goto yy2591;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy2592;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy2593;
				}
				goto yy1730;
			}
		}
	}
yy2465:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2632;
		if (yych <= ':') goto yy2323;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2632;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2632;
		goto yy1730;
	}
yy2466:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2633;
		if (yych <= ':') goto yy2634;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2633;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2633;
		goto yy1730;
	}
yy2467:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2635;
			} else {
				if (yych <= '1') goto yy2636;
				if (yych <= '2') goto yy2637;
				goto yy2635;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2638;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2638;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2468:
	yyaccept = 36;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '1') {
		if (yych <= '/') {
			yyt2 = yyt4;
			goto yy1941;
		}
		if (yych <= '0') goto yy2207;
		goto yy2208;
	} else {
		if (yych <= '2') goto yy2209;
		if (yych <= '9') goto yy2207;
		yyt2 = yyt4;
		goto yy1941;
	}
yy2469:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2639;
				yyt4 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2639;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2639;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2470:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') {
			yyt2 = yyt4;
			goto yy1940;
		}
		if (yych <= '/') goto yy1730;
		goto yy2638;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2638;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2638;
			goto yy1730;
		}
	}
yy2471:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2639;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy2470;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2639;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2639;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2472:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2640;
				yyt4 = YYCURSOR;
				goto yy2470;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2639;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2639;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2473:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2640;
				goto yy2639;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt4 = YYCURSOR;
					goto yy2470;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2639;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2639;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2474:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1730;
			if (yych <= '0') goto yy2641;
			goto yy2642;
		} else {
			if (yych <= '2') goto yy2643;
			if (yych <= '9') goto yy2641;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1912;
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy2475:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2474;
	goto yy1913;
yy2476:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2474;
	if (yych <= '/') goto yy1913;
	if (yych <= '5') goto yy2475;
	goto yy1913;
yy2477:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy1912;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2179;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1789;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2478:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2644;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2644;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2644;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2479:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2638;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2638;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2638;
			goto yy1730;
		}
	}
yy2480:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy2067;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2645;
		case ':': goto yy1927;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy2622;
		case 'B':
		case 'b': goto yy2646;
		case 'E':
		case 'e': goto yy2647;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2481:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1683;
			goto yy1791;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1791;
			} else {
				if (yych <= '9') goto yy2622;
				if (yych <= ':') goto yy1791;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1791;
			if (yych <= 'F') goto yy2622;
			if (yych <= 'Z') goto yy1791;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2622;
				if (yych <= 'z') goto yy1791;
				goto yy1683;
			}
		}
	}
yy2482:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy1783;
			goto yy1937;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1783;
				goto yy2064;
			} else {
				if (yych <= '/') goto yy1791;
				if (yych <= '9') goto yy2622;
				goto yy1791;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1791;
			} else {
				if (yych <= 'F') goto yy2622;
				if (yych <= 'Z') goto yy1791;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1791;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy2622;
				if (yych <= 'z') goto yy1791;
				goto yy1783;
			}
		}
	}
yy2483:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2648;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2649;
	goto yy1730;
yy2484:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2648;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2483;
	goto yy1730;
yy2485:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych == '.') goto yy2648;
		goto yy1730;
	} else {
		if (yych <= '4') goto yy2483;
		if (yych <= '5') goto yy2650;
		if (yych <= '9') goto yy2649;
		goto yy1730;
	}
yy2486:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2651;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2348;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2651;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2651;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2487:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy1940;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy2652;
			if (yych <= '1') goto yy2653;
			goto yy2654;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy2655;
			if (yych <= '@') goto yy1730;
			goto yy2211;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2211;
			goto yy1730;
		}
	}
yy2488:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2488;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2488;
			if (yych <= 0x1F) goto yy1730;
			goto yy2488;
		}
	} else {
		if (yych <= '&') {
			if (yych == '"') goto yy2077;
			goto yy1730;
		} else {
			if (yych <= '\'') goto yy2078;
			if (yych == '>') goto yy2081;
			goto yy1730;
		}
	}
yy2489:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ',') goto yy2231;
			goto yy2489;
		} else {
			if (yych == '>') goto yy2495;
			if (yych <= 0x7F) goto yy2231;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2239;
			if (yych <= 0xE0) goto yy2240;
			goto yy2241;
		} else {
			if (yych <= 0xF0) goto yy2242;
			if (yych <= 0xF3) goto yy2243;
			if (yych <= 0xF4) goto yy2244;
			goto yy1730;
		}
	}
yy2490:
	yych = *++YYCURSOR;
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2231;
				goto yy2490;
			} else {
				if (yych == '\r') goto yy2490;
				goto yy2231;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2490;
				if (yych <= '!') goto yy2231;
				goto yy2234;
			} else {
				if (yych == '\'') goto yy2235;
				if (yych <= ')') goto yy2231;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '<') {
				if (yych == '-') goto yy2352;
				goto yy2231;
			} else {
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2238;
				if (yych <= 0x7F) goto yy2231;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2239;
				if (yych <= 0xE0) goto yy2240;
				goto yy2241;
			} else {
				if (yych <= 0xF0) goto yy2242;
				if (yych <= 0xF3) goto yy2243;
				if (yych <= 0xF4) goto yy2244;
				goto yy1730;
			}
		}
	}
yy2491:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2490;
				goto yy2231;
			} else {
				if (yych <= '\r') goto yy2490;
				if (yych == ' ') goto yy2490;
				goto yy2231;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2234;
				if (yych == '\'') goto yy2235;
				goto yy2231;
			} else {
				if (yych <= '-') goto yy2491;
				if (yych <= '/') goto yy2231;
				if (yych <= ':') goto yy2354;
				goto yy2231;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy2237;
				if (yych <= '>') goto yy2495;
				if (yych <= '@') goto yy2231;
				goto yy2354;
			} else {
				if (yych == '_') goto yy2354;
				if (yych <= '`') goto yy2231;
				goto yy2354;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2231;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				goto yy2240;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2241;
					goto yy2242;
				} else {
					if (yych <= 0xF3) goto yy2243;
					if (yych <= 0xF4) goto yy2244;
					goto yy1730;
				}
			}
		}
	}
yy2492:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2656;
				goto yy2231;
			} else {
				if (yych <= '\r') goto yy2656;
				if (yych == ' ') goto yy2656;
				goto yy2231;
			}
		} else {
			if (yych <= ')') {
				if (yych <= '"') goto yy2234;
				if (yych == '\'') goto yy2235;
				goto yy2231;
			} else {
				if (yych <= ',') {
					if (yych <= '*') goto yy1730;
					goto yy2231;
				} else {
					if (yych <= '-') goto yy2353;
					if (yych <= '/') goto yy2231;
					goto yy2354;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= 'Z') {
				if (yych == '>') goto yy2238;
				if (yych <= '@') goto yy2231;
				goto yy2354;
			} else {
				if (yych == '_') goto yy2354;
				if (yych <= '`') goto yy2231;
				goto yy2354;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2231;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2239;
				goto yy2240;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2241;
					goto yy2242;
				} else {
					if (yych <= 0xF3) goto yy2243;
					if (yych <= 0xF4) goto yy2244;
					goto yy1730;
				}
			}
		}
	}
yy2493:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2492;
			goto yy2355;
		} else {
			if (yych <= '=') {
				if (yych <= '-') goto yy2493;
				goto yy2355;
			} else {
				if (yych <= '>') goto yy2657;
				if (yych <= '[') goto yy2355;
				goto yy2357;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy2355;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2358;
			goto yy2359;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2360;
				goto yy2361;
			} else {
				if (yych <= 0xF3) goto yy2362;
				if (yych <= 0xF4) goto yy2363;
				goto yy1730;
			}
		}
	}
yy2494:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2492;
			goto yy2364;
		} else {
			if (yych <= '=') {
				if (yych <= '-') goto yy2494;
				goto yy2364;
			} else {
				if (yych <= '>') goto yy2658;
				if (yych <= '[') goto yy2364;
				goto yy2366;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy2364;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2367;
			goto yy2368;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2369;
				goto yy2370;
			} else {
				if (yych <= 0xF3) goto yy2371;
				if (yych <= 0xF4) goto yy2372;
				goto yy1730;
			}
		}
	}
yy2495:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[512+yych] & 8) {
		goto yy2231;
	}
	if (yych <= 0xE0) {
		if (yych <= '-') {
			if (yych >= '+') goto yy2352;
		} else {
			if (yych <= 0xC1) goto yy2496;
			if (yych <= 0xDF) goto yy2239;
			goto yy2240;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2241;
			goto yy2242;
		} else {
			if (yych <= 0xF3) goto yy2243;
			if (yych <= 0xF4) goto yy2244;
		}
	}
yy2496:
#line 255 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_COMMENT);
           if (tf == text_format_t::TF_DIFF) {
               auto sf = this->to_string_fragment(cap_all);
               auto split_res = sf.split_when(string_fragment::tag1{'\n'});
               cap_all.c_end = split_res.first.sf_end;
               cap_inner.c_end = split_res.first.sf_end;
               this->ds_next_offset = cap_all.c_end;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 65304 "../../lnav/src/data_scanner_re.cc"
yy2497:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 4) {
		goto yy2497;
	}
	if (yych <= 0xDF) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '*') goto yy2515;
			goto yy2659;
		} else {
			if (yych <= '>') goto yy2660;
			if (yych <= 0xC1) goto yy1730;
			goto yy2661;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2662;
			if (yych <= 0xEF) goto yy2663;
			goto yy2664;
		} else {
			if (yych <= 0xF3) goto yy2665;
			if (yych <= 0xF4) goto yy2666;
			goto yy1730;
		}
	}
yy2498:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2497;
			goto yy2375;
		} else {
			if (yych <= '-') goto yy2667;
			if (yych != '>') goto yy2375;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2500;
			if (yych <= 0xE0) goto yy2501;
			goto yy2502;
		} else {
			if (yych <= 0xF0) goto yy2503;
			if (yych <= 0xF3) goto yy2504;
			if (yych <= 0xF4) goto yy2505;
			goto yy1730;
		}
	}
yy2499:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[512+yych] & 128) {
		goto yy2355;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2082;
			if (yych <= '"') goto yy2492;
			goto yy2215;
		} else {
			if (yych <= '-') goto yy2356;
			if (yych <= '\\') goto yy2357;
			goto yy2082;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy2082;
		}
	}
yy2500:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2375;
	goto yy1730;
yy2501:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2500;
	goto yy1730;
yy2502:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2500;
	goto yy1730;
yy2503:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2502;
	goto yy1730;
yy2504:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2502;
	goto yy1730;
yy2505:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2502;
	goto yy1730;
yy2506:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2497;
			goto yy2376;
		} else {
			if (yych <= '-') goto yy2668;
			if (yych != '>') goto yy2376;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2508;
			if (yych <= 0xE0) goto yy2509;
			goto yy2510;
		} else {
			if (yych <= 0xF0) goto yy2511;
			if (yych <= 0xF3) goto yy2512;
			if (yych <= 0xF4) goto yy2513;
			goto yy1730;
		}
	}
yy2507:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[256+yych] & 1) {
		goto yy2364;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2082;
			if (yych <= '\'') goto yy2492;
			goto yy2223;
		} else {
			if (yych <= '-') goto yy2365;
			if (yych <= '\\') goto yy2366;
			goto yy2082;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy2082;
		}
	}
yy2508:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2376;
	goto yy1730;
yy2509:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2508;
	goto yy1730;
yy2510:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2508;
	goto yy1730;
yy2511:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2510;
	goto yy1730;
yy2512:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2510;
	goto yy1730;
yy2513:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2510;
	goto yy1730;
yy2514:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2375;
			goto yy2374;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy2376;
				goto yy2374;
			} else {
				if (yych <= '-') goto yy2514;
				if (yych <= '=') goto yy2374;
				goto yy2495;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy2374;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2379;
			goto yy2380;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2381;
				goto yy2382;
			} else {
				if (yych <= 0xF3) goto yy2383;
				if (yych <= 0xF4) goto yy2384;
				goto yy1730;
			}
		}
	}
yy2515:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 8) {
		goto yy2515;
	}
	if (yych <= 0xE0) {
		if (yych <= '>') {
			if (yych <= 0x00) goto yy1730;
			goto yy2669;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2670;
			goto yy2671;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2672;
			goto yy2673;
		} else {
			if (yych <= 0xF3) goto yy2674;
			if (yych <= 0xF4) goto yy2675;
			goto yy1730;
		}
	}
yy2516:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[512+yych] & 2) {
		goto yy2215;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy2082;
			goto yy2351;
		} else {
			if (yych <= '\\') goto yy2216;
			if (yych <= 0xC1) goto yy2082;
			goto yy2217;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2218;
			if (yych <= 0xEF) goto yy2219;
			goto yy2220;
		} else {
			if (yych <= 0xF3) goto yy2221;
			if (yych <= 0xF4) goto yy2222;
			goto yy2082;
		}
	}
yy2517:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2385;
	goto yy1730;
yy2518:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2517;
	goto yy1730;
yy2519:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2517;
	goto yy1730;
yy2520:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2519;
	goto yy1730;
yy2521:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2519;
	goto yy1730;
yy2522:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2519;
	goto yy1730;
yy2523:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[512+yych] & 4) {
		goto yy2223;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2082;
			goto yy2351;
		} else {
			if (yych <= '\\') goto yy2224;
			if (yych <= 0xC1) goto yy2082;
			goto yy2225;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2226;
			if (yych <= 0xEF) goto yy2227;
			goto yy2228;
		} else {
			if (yych <= 0xF3) goto yy2229;
			if (yych <= 0xF4) goto yy2230;
			goto yy2082;
		}
	}
yy2524:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2386;
	goto yy1730;
yy2525:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2524;
	goto yy1730;
yy2526:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2524;
	goto yy1730;
yy2527:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2526;
	goto yy1730;
yy2528:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2526;
	goto yy1730;
yy2529:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2526;
	goto yy1730;
yy2530:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2530;
				goto yy2250;
			} else {
				if (yych <= '\r') goto yy2530;
				if (yych == ' ') goto yy2530;
				goto yy2250;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2676;
				if (yych == '-') goto yy2395;
				goto yy2250;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2396;
					goto yy2395;
				} else {
					if (yych <= '=') goto yy2250;
					if (yych <= '>') goto yy2397;
					goto yy2396;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2250;
				if (yych <= 'Z') goto yy2395;
				if (yych <= '[') goto yy2250;
				goto yy2398;
			} else {
				if (yych == '_') goto yy2395;
				if (yych <= '`') goto yy2250;
				goto yy2395;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2250;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2399;
				goto yy2400;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2401;
					goto yy2402;
				} else {
					if (yych <= 0xF3) goto yy2403;
					if (yych <= 0xF4) goto yy2404;
					goto yy1730;
				}
			}
		}
	}
yy2531:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy2087;
			if (yych >= '"') goto yy2533;
		} else {
			if (yych == '\\') goto yy2534;
			if (yych >= 0x80) goto yy2087;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2535;
			if (yych <= 0xE0) goto yy2536;
			goto yy2537;
		} else {
			if (yych <= 0xF0) goto yy2538;
			if (yych <= 0xF3) goto yy2539;
			if (yych <= 0xF4) goto yy2540;
			goto yy2087;
		}
	}
yy2532:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '!') goto yy2532;
		} else {
			if (yych == '\\') goto yy2534;
			if (yych <= 0x7F) goto yy2532;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2535;
			if (yych <= 0xE0) goto yy2536;
			goto yy2537;
		} else {
			if (yych <= 0xF0) goto yy2538;
			if (yych <= 0xF3) goto yy2539;
			if (yych <= 0xF4) goto yy2540;
			goto yy1730;
		}
	}
yy2533:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2677;
			} else {
				if (yych == '\r') goto yy2677;
				goto yy1730;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy2677;
				goto yy1730;
			} else {
				if (yych <= '-') goto yy2083;
				if (yych <= '.') goto yy1730;
				goto yy1959;
			}
		}
	} else {
		if (yych <= '@') {
			if (yych <= '=') {
				if (yych <= ':') goto yy2083;
				goto yy1730;
			} else {
				if (yych <= '>') goto yy1960;
				if (yych <= '?') goto yy1959;
				goto yy1730;
			}
		} else {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy2083;
				if (yych <= '^') goto yy1730;
				goto yy2083;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2083;
				goto yy1730;
			}
		}
	}
yy2534:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2532;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2536;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2537;
			goto yy2538;
		} else {
			if (yych <= 0xF3) goto yy2539;
			if (yych <= 0xF4) goto yy2540;
			goto yy1730;
		}
	}
yy2535:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2532;
	goto yy1730;
yy2536:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2535;
	goto yy1730;
yy2537:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2535;
	goto yy1730;
yy2538:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2537;
	goto yy1730;
yy2539:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2537;
	goto yy1730;
yy2540:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2537;
	goto yy1730;
yy2541:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2541;
				goto yy2251;
			} else {
				if (yych <= '\r') goto yy2541;
				if (yych == ' ') goto yy2541;
				goto yy2251;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2676;
				if (yych == '-') goto yy2405;
				goto yy2251;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2406;
					goto yy2405;
				} else {
					if (yych <= '=') goto yy2251;
					if (yych <= '>') goto yy2407;
					goto yy2406;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2251;
				if (yych <= 'Z') goto yy2405;
				if (yych <= '[') goto yy2251;
				goto yy2408;
			} else {
				if (yych == '_') goto yy2405;
				if (yych <= '`') goto yy2251;
				goto yy2405;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2251;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2409;
				goto yy2410;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2411;
					goto yy2412;
				} else {
					if (yych <= 0xF3) goto yy2413;
					if (yych <= 0xF4) goto yy2414;
					goto yy1730;
				}
			}
		}
	}
yy2542:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2087;
			if (yych >= '\'') goto yy2533;
		} else {
			if (yych == '\\') goto yy2544;
			if (yych >= 0x80) goto yy2087;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2545;
			if (yych <= 0xE0) goto yy2546;
			goto yy2547;
		} else {
			if (yych <= 0xF0) goto yy2548;
			if (yych <= 0xF3) goto yy2549;
			if (yych <= 0xF4) goto yy2550;
			goto yy2087;
		}
	}
yy2543:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '&') goto yy2543;
			goto yy2533;
		} else {
			if (yych == '\\') goto yy2544;
			if (yych <= 0x7F) goto yy2543;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2545;
			if (yych <= 0xE0) goto yy2546;
			goto yy2547;
		} else {
			if (yych <= 0xF0) goto yy2548;
			if (yych <= 0xF3) goto yy2549;
			if (yych <= 0xF4) goto yy2550;
			goto yy1730;
		}
	}
yy2544:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2543;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2546;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2547;
			goto yy2548;
		} else {
			if (yych <= 0xF3) goto yy2549;
			if (yych <= 0xF4) goto yy2550;
			goto yy1730;
		}
	}
yy2545:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2543;
	goto yy1730;
yy2546:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2545;
	goto yy1730;
yy2547:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2545;
	goto yy1730;
yy2548:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2547;
	goto yy1730;
yy2549:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2547;
	goto yy1730;
yy2550:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2547;
	goto yy1730;
yy2551:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2416;
				goto yy2551;
			} else {
				if (yych == '\r') goto yy2551;
				if (yych <= 0x1F) goto yy2416;
				goto yy2551;
			}
		} else {
			if (yych <= '/') {
				if (yych == '-') goto yy2551;
				if (yych <= '.') goto yy2416;
			} else {
				if (yych <= ':') goto yy2551;
				if (yych <= '<') goto yy2416;
				if (yych <= '=') goto yy2417;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= '?') goto yy2552;
				if (yych <= '@') goto yy2416;
				if (yych <= 'Z') goto yy2551;
				goto yy2416;
			} else {
				if (yych == '`') goto yy2416;
				if (yych <= 'z') goto yy2551;
				goto yy2416;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2420;
				if (yych <= 0xE0) goto yy2421;
				goto yy2422;
			} else {
				if (yych <= 0xF0) goto yy2423;
				if (yych <= 0xF3) goto yy2424;
				if (yych <= 0xF4) goto yy2425;
				goto yy1730;
			}
		}
	}
yy2552:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2416;
				goto yy2551;
			} else {
				if (yych <= '.') goto yy2416;
				if (yych <= '/') goto yy2552;
				goto yy2551;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2416;
				if (yych <= '>') goto yy2086;
				goto yy2552;
			} else {
				if (yych <= '@') goto yy2416;
				if (yych <= 'Z') goto yy2551;
				goto yy2416;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2416;
				goto yy2551;
			} else {
				if (yych <= 0x7F) goto yy2416;
				if (yych <= 0xC1) goto yy1730;
				goto yy2420;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2421;
				if (yych <= 0xEF) goto yy2422;
				goto yy2423;
			} else {
				if (yych <= 0xF3) goto yy2424;
				if (yych <= 0xF4) goto yy2425;
				goto yy1730;
			}
		}
	}
yy2553:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2553;
				goto yy2418;
			} else {
				if (yych <= '\r') goto yy2553;
				if (yych == ' ') goto yy2553;
				goto yy2418;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2416;
				if (yych == '-') goto yy2553;
				goto yy2418;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2553;
				} else {
					if (yych <= '<') goto yy2418;
					if (yych <= '=') goto yy2678;
					goto yy2555;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2554;
				if (yych <= '@') goto yy2418;
				if (yych <= 'Z') goto yy2553;
				goto yy2418;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2556;
					goto yy2418;
				} else {
					if (yych == '`') goto yy2418;
					goto yy2553;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2418;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2557;
				goto yy2558;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2559;
					goto yy2560;
				} else {
					if (yych <= 0xF3) goto yy2561;
					if (yych <= 0xF4) goto yy2562;
					goto yy1730;
				}
			}
		}
	}
yy2554:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '!') goto yy2418;
				goto yy2416;
			} else {
				if (yych == '-') goto yy2553;
				if (yych <= '.') goto yy2418;
				goto yy2554;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2553;
				if (yych <= '=') goto yy2418;
				goto yy2679;
			} else {
				if (yych <= '?') goto yy2554;
				if (yych <= '@') goto yy2418;
				if (yych <= 'Z') goto yy2553;
				goto yy2418;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy2556;
				if (yych <= '^') goto yy2418;
				goto yy2553;
			} else {
				if (yych <= '`') goto yy2418;
				if (yych <= 'z') goto yy2553;
				if (yych <= 0x7F) goto yy2418;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2557;
				if (yych <= 0xE0) goto yy2558;
				goto yy2559;
			} else {
				if (yych <= 0xF0) goto yy2560;
				if (yych <= 0xF3) goto yy2561;
				if (yych <= 0xF4) goto yy2562;
				goto yy1730;
			}
		}
	}
yy2555:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '!') goto yy2555;
			goto yy2680;
		} else {
			if (yych == '\\') goto yy2681;
			if (yych <= 0x7F) goto yy2555;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2682;
			if (yych <= 0xE0) goto yy2683;
			goto yy2684;
		} else {
			if (yych <= 0xF0) goto yy2685;
			if (yych <= 0xF3) goto yy2686;
			if (yych <= 0xF4) goto yy2687;
			goto yy1730;
		}
	}
yy2556:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2418;
				goto yy2553;
			} else {
				if (yych <= '.') goto yy2418;
				if (yych <= '/') goto yy2554;
				goto yy2553;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2418;
				if (yych <= '>') goto yy2555;
				goto yy2554;
			} else {
				if (yych <= '@') goto yy2418;
				if (yych <= 'Z') goto yy2553;
				goto yy2418;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2418;
				goto yy2553;
			} else {
				if (yych <= 0x7F) goto yy2418;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2558;
				if (yych <= 0xEF) goto yy2559;
				goto yy2560;
			} else {
				if (yych <= 0xF3) goto yy2561;
				if (yych <= 0xF4) goto yy2562;
				goto yy1730;
			}
		}
	}
yy2557:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2418;
	goto yy1730;
yy2558:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2557;
	goto yy1730;
yy2559:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2557;
	goto yy1730;
yy2560:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2559;
	goto yy1730;
yy2561:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2559;
	goto yy1730;
yy2562:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2559;
	goto yy1730;
yy2563:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2563;
				goto yy2419;
			} else {
				if (yych <= '\r') goto yy2563;
				if (yych == ' ') goto yy2563;
				goto yy2419;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2416;
				if (yych == '-') goto yy2563;
				goto yy2419;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2563;
				} else {
					if (yych <= '<') goto yy2419;
					if (yych <= '=') goto yy2688;
					goto yy2565;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2564;
				if (yych <= '@') goto yy2419;
				if (yych <= 'Z') goto yy2563;
				goto yy2419;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2566;
					goto yy2419;
				} else {
					if (yych == '`') goto yy2419;
					goto yy2563;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2419;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2567;
				goto yy2568;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2569;
					goto yy2570;
				} else {
					if (yych <= 0xF3) goto yy2571;
					if (yych <= 0xF4) goto yy2572;
					goto yy1730;
				}
			}
		}
	}
yy2564:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '/') {
			if (yych <= '\'') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '&') goto yy2419;
				goto yy2416;
			} else {
				if (yych == '-') goto yy2563;
				if (yych <= '.') goto yy2419;
				goto yy2564;
			}
		} else {
			if (yych <= '>') {
				if (yych <= ':') goto yy2563;
				if (yych <= '=') goto yy2419;
				goto yy2689;
			} else {
				if (yych <= '?') goto yy2564;
				if (yych <= '@') goto yy2419;
				if (yych <= 'Z') goto yy2563;
				goto yy2419;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= '\\') goto yy2566;
				if (yych <= '^') goto yy2419;
				goto yy2563;
			} else {
				if (yych <= '`') goto yy2419;
				if (yych <= 'z') goto yy2563;
				if (yych <= 0x7F) goto yy2419;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2567;
				if (yych <= 0xE0) goto yy2568;
				goto yy2569;
			} else {
				if (yych <= 0xF0) goto yy2570;
				if (yych <= 0xF3) goto yy2571;
				if (yych <= 0xF4) goto yy2572;
				goto yy1730;
			}
		}
	}
yy2565:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '&') goto yy2565;
			goto yy2680;
		} else {
			if (yych == '\\') goto yy2690;
			if (yych <= 0x7F) goto yy2565;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2691;
			if (yych <= 0xE0) goto yy2692;
			goto yy2693;
		} else {
			if (yych <= 0xF0) goto yy2694;
			if (yych <= 0xF3) goto yy2695;
			if (yych <= 0xF4) goto yy2696;
			goto yy1730;
		}
	}
yy2566:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2419;
				goto yy2563;
			} else {
				if (yych <= '.') goto yy2419;
				if (yych <= '/') goto yy2564;
				goto yy2563;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2419;
				if (yych <= '>') goto yy2565;
				goto yy2564;
			} else {
				if (yych <= '@') goto yy2419;
				if (yych <= 'Z') goto yy2563;
				goto yy2419;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2419;
				goto yy2563;
			} else {
				if (yych <= 0x7F) goto yy2419;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2568;
				if (yych <= 0xEF) goto yy2569;
				goto yy2570;
			} else {
				if (yych <= 0xF3) goto yy2571;
				if (yych <= 0xF4) goto yy2572;
				goto yy1730;
			}
		}
	}
yy2567:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2419;
	goto yy1730;
yy2568:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2567;
	goto yy1730;
yy2569:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2567;
	goto yy1730;
yy2570:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2569;
	goto yy1730;
yy2571:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2569;
	goto yy1730;
yy2572:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2569;
	goto yy1730;
yy2573:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy2697;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2573;
	goto yy1730;
yy2574:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1821;
		if (yych <= '9') goto yy2698;
		goto yy1821;
	} else {
		if (yych <= 'F') goto yy2698;
		if (yych <= '`') goto yy1821;
		if (yych <= 'f') goto yy2698;
		goto yy1821;
	}
yy2575:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2699;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2701;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2701;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2576:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2702;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2701;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2701;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2577:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '4') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2702;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy2703;
				if (yych <= '9') goto yy2699;
				yyt2 = YYCURSOR;
				goto yy2700;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2701;
				goto yy2103;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'f') goto yy2701;
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2578:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2701;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2701;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2701;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2579:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy2113;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy2704;
			if (yych <= '1') goto yy2705;
			if (yych <= '2') goto yy2706;
			goto yy2704;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1730;
			if (yych <= 'Z') goto yy2103;
			goto yy1730;
		} else {
			if (yych == '`') goto yy1730;
			if (yych <= 'z') goto yy2103;
			goto yy1730;
		}
	}
yy2580:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2579;
	goto yy2104;
yy2581:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2579;
	if (yych <= '/') goto yy2104;
	if (yych <= '5') goto yy2580;
	goto yy2104;
yy2582:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2265;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2103;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2268;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2583:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2707;
	goto yy1730;
yy2584:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1809;
				goto yy1745;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1809;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				yyt3 = YYCURSOR;
				goto yy1976;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy2698;
				yyt3 = YYCURSOR;
				goto yy1980;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1809;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy2698;
				if (yych <= 'Z') goto yy1820;
				goto yy1809;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1707;
				goto yy1809;
			} else {
				if (yych <= 'f') goto yy2708;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy2585:
	++YYCURSOR;
	yyt1 = yyt2;
yy2586:
	YYCURSOR = yyt1;
#line 302 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_TIME); }
#line 67107 "../../lnav/src/data_scanner_re.cc"
yy2587:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	yyt1 = yyt2;
	if (yych <= '/') goto yy2586;
	if (yych <= '9') goto yy2709;
	goto yy2586;
yy2588:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2585;
	goto yy1730;
yy2589:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2588;
	goto yy1730;
yy2590:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2588;
	goto yy1730;
yy2591:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2590;
	goto yy1730;
yy2592:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2590;
	goto yy1730;
yy2593:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2590;
	goto yy1730;
yy2594:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy2710;
	goto yy1770;
yy2595:
	++YYCURSOR;
yy2596:
	YYCURSOR = yyt1;
#line 400 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_IPV4_ADDRESS);
       }
#line 67158 "../../lnav/src/data_scanner_re.cc"
yy2597:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy2596;
		} else {
			if (yych == ',') goto yy2596;
			if (yych <= '.') goto yy1746;
			goto yy2596;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy2596;
			if (yych <= 'Z') goto yy1746;
			goto yy2596;
		} else {
			if (yych == '`') goto yy2596;
			if (yych <= 'z') goto yy1746;
			goto yy2596;
		}
	}
yy2598:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy2596;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy2596;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy2596;
			goto yy2141;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy2596;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy2141;
			goto yy2596;
		} else {
			if (yych == '`') goto yy2596;
			if (yych <= 'z') goto yy2141;
			goto yy2596;
		}
	}
yy2599:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy2596;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy2596;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy2596;
			goto yy2300;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy2596;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy2596;
		} else {
			if (yych == '`') goto yy2596;
			if (yych <= 'z') goto yy1793;
			goto yy2596;
		}
	}
yy2600:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy2596;
		if (yych <= '9') goto yy1912;
		if (yych <= '@') goto yy2596;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy2596;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy2596;
			if (yych <= 'z') goto yy1912;
			goto yy2596;
		}
	}
yy2601:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2598;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy2599;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy2600;
				}
				goto yy2300;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2602;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy2603;
				}
				yyt1 = YYCURSOR;
				goto yy2604;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				goto yy2143;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy2605;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy2606;
				}
				yyt1 = YYCURSOR;
				goto yy2607;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy2608;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2609;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2610;
				}
				goto yy2143;
			}
		}
	}
yy2602:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy1779;
	goto yy2596;
yy2603:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy2596;
			goto yy1850;
		} else {
			if (yych <= '/') goto yy2596;
			if (yych <= '9') goto yy1925;
			goto yy2596;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1925;
			if (yych <= '^') goto yy2596;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy2596;
			if (yych <= 'z') goto yy1925;
			goto yy2596;
		}
	}
yy2604:
	yyaccept = 40;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy2301;
			goto yy2596;
		} else {
			if (yych == ',') goto yy2596;
			goto yy2301;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy2596;
			if (yych <= 'Z') goto yy2301;
			goto yy2596;
		} else {
			if (yych == '`') goto yy2596;
			if (yych <= 'z') goto yy2301;
			goto yy2596;
		}
	}
yy2605:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2595;
	goto yy1730;
yy2606:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2605;
	goto yy1730;
yy2607:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2605;
	goto yy1730;
yy2608:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2607;
	goto yy1730;
yy2609:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2607;
	goto yy1730;
yy2610:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2607;
	goto yy1730;
yy2611:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '-') {
			if (yych <= '*') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				yyt1 = YYCURSOR;
				goto yy2595;
			} else {
				if (yych <= '+') {
					yyt1 = YYCURSOR;
					goto yy2597;
				}
				if (yych <= ',') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2598;
			}
		} else {
			if (yych <= '5') {
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy2599;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy2600;
				}
				goto yy2601;
			} else {
				if (yych <= '9') goto yy2300;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2602;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2603;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				yyt1 = YYCURSOR;
				goto yy2604;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy2604;
				}
				if (yych <= 0x7F) {
					yyt1 = YYCURSOR;
					goto yy2595;
				}
				goto yy2143;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy2605;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy2606;
				}
				yyt1 = YYCURSOR;
				goto yy2607;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy2608;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy2609;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy2610;
				}
				goto yy2143;
			}
		}
	}
yy2612:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == 'B') goto yy2443;
	if (yych == 'b') goto yy2443;
	goto yy2142;
yy2613:
	yyaccept = 29;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy2443;
		goto yy2142;
	} else {
		if (yych <= 'b') goto yy2443;
		if (yych == 'n') goto yy2443;
		goto yy2142;
	}
yy2614:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy2318;
	if (yych == ':') goto yy2318;
	goto yy2155;
yy2615:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2711;
	goto yy1913;
yy2616:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2712;
	goto yy1730;
yy2617:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '%') {
			if (yych == ' ') goto yy2713;
			if (yych >= '%') goto yy1794;
		} else {
			if (yych <= '+') {
				if (yych >= '+') goto yy1794;
			} else {
				if (yych <= ',') goto yy2618;
				if (yych <= '/') goto yy1794;
				goto yy2714;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1794;
			} else {
				if (yych == 'T') goto yy2715;
				goto yy1794;
			}
		} else {
			if (yych <= '_') {
				if (yych >= '_') goto yy1794;
			} else {
				if (yych <= '`') goto yy2618;
				if (yych <= 'z') goto yy1794;
			}
		}
	}
yy2618:
#line 316 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE);
       }
#line 67611 "../../lnav/src/data_scanner_re.cc"
yy2619:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= ',') {
			if (yych == ' ') goto yy2713;
			goto yy2618;
		} else {
			if (yych <= '/') goto yy1913;
			if (yych <= '9') goto yy2716;
			if (yych <= ':') goto yy1913;
			goto yy2618;
		}
	} else {
		if (yych <= '^') {
			if (yych == 'T') goto yy2717;
			if (yych <= 'Z') goto yy1913;
			goto yy2618;
		} else {
			if (yych == '`') goto yy2618;
			if (yych <= 'z') goto yy1913;
			goto yy2618;
		}
	}
yy2620:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy2718;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy2719;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy2720;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2621:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		if (yych <= ',') goto yy1932;
		goto yy2718;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy2720;
		goto yy1932;
	}
yy2622:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy2718;
	goto yy1932;
yy2623:
	yyaccept = 25;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') {
			yyt2 = yyt1;
			goto yy1941;
		}
		if (yych <= '9') goto yy2457;
		yyt2 = yyt1;
		goto yy1941;
	} else {
		if (yych <= 'Z') goto yy2457;
		if (yych <= '`') {
			yyt2 = yyt1;
			goto yy1941;
		}
		if (yych <= 'z') goto yy2457;
		yyt2 = yyt1;
		goto yy1941;
	}
yy2624:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2721;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2721;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2721;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2625:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1730;
			yyt2 = yyt1;
			goto yy1940;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2722;
			goto yy1779;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2722;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2722;
			goto yy1730;
		}
	}
yy2626:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2721;
				yyt1 = YYCURSOR;
				goto yy2625;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2721;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2721;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2627:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2723;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2625;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2721;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2721;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2628:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy2723;
				if (yych <= '9') goto yy2721;
				yyt1 = YYCURSOR;
				goto yy2625;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2721;
				goto yy1912;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'f') goto yy2721;
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2629:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '/') {
			if (yych <= ',') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= '-') {
				yyt1 = YYCURSOR;
				goto yy2724;
			}
			if (yych <= '.') {
				yyt2 = YYCURSOR;
				goto yy2587;
			}
			yyt2 = YYCURSOR;
			goto yy2585;
		} else {
			if (yych <= ':') {
				if (yych <= '9') {
					yyt1 = YYCURSOR;
					goto yy2631;
				}
				goto yy2725;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2585;
				}
				if (yych <= 'F') {
					yyt1 = YYCURSOR;
					goto yy2631;
				}
				yyt2 = YYCURSOR;
				goto yy2585;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy2631;
				}
				yyt2 = YYCURSOR;
				goto yy2585;
			} else {
				if (yych <= 0xC1) goto yy2155;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy2588;
				}
				yyt2 = YYCURSOR;
				goto yy2589;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy2590;
				}
				yyt2 = YYCURSOR;
				goto yy2591;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy2592;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy2593;
				}
				goto yy2155;
			}
		}
	}
yy2630:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy2465;
			goto yy2725;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy2465;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy2465;
			goto yy2155;
		}
	}
yy2631:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy2586;
		if (yych <= '9') goto yy2632;
		if (yych <= ':') goto yy2323;
		goto yy2586;
	} else {
		if (yych <= 'F') goto yy2632;
		if (yych <= '`') goto yy2586;
		if (yych >= 'g') goto yy2586;
	}
yy2632:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2323;
	goto yy1730;
yy2633:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2726;
		if (yych >= ';') goto yy1730;
	} else {
		if (yych <= 'F') goto yy2726;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2726;
		goto yy1730;
	}
yy2634:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2727;
		if (yych <= ':') goto yy2728;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2727;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2727;
		goto yy1730;
	}
yy2635:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2729;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2636:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2732;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2637:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2732;
				goto yy2733;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2729;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2730;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2731;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2731;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2638:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2731;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2730;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2731;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2731;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2639:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2734;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy2470;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2734;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2734;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2640:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2734;
				yyt4 = YYCURSOR;
				goto yy2470;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2734;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2734;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2641:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2735;
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2736;
	goto yy1913;
yy2642:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2735;
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2641;
	goto yy1913;
yy2643:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy2735;
		goto yy1913;
	} else {
		if (yych <= '4') goto yy2641;
		if (yych <= '5') goto yy2737;
		if (yych <= '9') goto yy2736;
		goto yy1913;
	}
yy2644:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2738;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2479;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2738;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2738;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2645:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '-': goto yy2718;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2739;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2646:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1932;
			goto yy1683;
		} else {
			if (yych <= '+') goto yy1932;
			if (yych <= ',') goto yy1683;
			if (yych <= '-') goto yy2718;
			goto yy1932;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1932;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1932;
			goto yy1683;
		}
	}
yy2647:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy1937;
	if (yych == '-') goto yy2742;
	goto yy1932;
yy2648:
	yych = *++YYCURSOR;
	if (yych <= '1') {
		if (yych <= '/') goto yy1730;
		if (yych <= '0') goto yy2743;
		goto yy2744;
	} else {
		if (yych <= '2') goto yy2745;
		if (yych <= '9') goto yy2743;
		goto yy1730;
	}
yy2649:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2648;
	goto yy1730;
yy2650:
	yych = *++YYCURSOR;
	if (yych == '.') goto yy2648;
	if (yych <= '/') goto yy1730;
	if (yych <= '5') goto yy2649;
	goto yy1730;
yy2651:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2348;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2652:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2746;
				yyt2 = YYCURSOR;
				goto yy2747;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2653:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2748;
				yyt2 = YYCURSOR;
				goto yy2348;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2654:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2748;
				goto yy2749;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2746;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2348;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2347;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2347;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2655:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2746;
				yyt2 = YYCURSOR;
				goto yy2348;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2347;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2656:
	yych = *++YYCURSOR;
	if (yych <= '*') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2231;
				goto yy2656;
			} else {
				if (yych == '\r') goto yy2656;
				goto yy2231;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2656;
				if (yych <= '!') goto yy2231;
				goto yy2234;
			} else {
				if (yych == '\'') goto yy2235;
				if (yych <= ')') goto yy2231;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '=') {
				if (yych == '-') goto yy2352;
				goto yy2231;
			} else {
				if (yych <= '>') goto yy2238;
				if (yych <= 0x7F) goto yy2231;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2239;
				if (yych <= 0xE0) goto yy2240;
				goto yy2241;
			} else {
				if (yych <= 0xF0) goto yy2242;
				if (yych <= 0xF3) goto yy2243;
				if (yych <= 0xF4) goto yy2244;
				goto yy1730;
			}
		}
	}
yy2657:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[512+yych] & 128) {
		goto yy2355;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2496;
			if (yych <= '"') goto yy2492;
			goto yy2215;
		} else {
			if (yych <= '-') goto yy2356;
			if (yych <= '\\') goto yy2357;
			goto yy2496;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2358;
			if (yych <= 0xE0) goto yy2359;
			goto yy2360;
		} else {
			if (yych <= 0xF0) goto yy2361;
			if (yych <= 0xF3) goto yy2362;
			if (yych <= 0xF4) goto yy2363;
			goto yy2496;
		}
	}
yy2658:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[256+yych] & 1) {
		goto yy2364;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy2496;
			if (yych <= '\'') goto yy2492;
			goto yy2223;
		} else {
			if (yych <= '-') goto yy2365;
			if (yych <= '\\') goto yy2366;
			goto yy2496;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2367;
			if (yych <= 0xE0) goto yy2368;
			goto yy2369;
		} else {
			if (yych <= 0xF0) goto yy2370;
			if (yych <= 0xF3) goto yy2371;
			if (yych <= 0xF4) goto yy2372;
			goto yy2496;
		}
	}
yy2659:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ',') goto yy2497;
			goto yy2750;
		} else {
			if (yych == '>') goto yy2660;
			if (yych <= 0x7F) goto yy2497;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2661;
			if (yych <= 0xE0) goto yy2662;
			goto yy2663;
		} else {
			if (yych <= 0xF0) goto yy2664;
			if (yych <= 0xF3) goto yy2665;
			if (yych <= 0xF4) goto yy2666;
			goto yy1730;
		}
	}
yy2660:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[0+yych] & 32) {
		goto yy2751;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2082;
			if (yych <= '"') goto yy2752;
			goto yy2753;
		} else {
			if (yych <= '*') goto yy2754;
			if (yych <= '-') goto yy2755;
			if (yych <= '\\') goto yy2756;
			goto yy2082;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy2082;
		}
	}
yy2661:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2497;
	goto yy1730;
yy2662:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2661;
	goto yy1730;
yy2663:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2661;
	goto yy1730;
yy2664:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2663;
	goto yy1730;
yy2665:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2663;
	goto yy1730;
yy2666:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2663;
	goto yy1730;
yy2667:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2497;
			goto yy2375;
		} else {
			if (yych <= '-') goto yy2667;
			if (yych == '>') goto yy2657;
			goto yy2375;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2500;
			if (yych <= 0xE0) goto yy2501;
			goto yy2502;
		} else {
			if (yych <= 0xF0) goto yy2503;
			if (yych <= 0xF3) goto yy2504;
			if (yych <= 0xF4) goto yy2505;
			goto yy1730;
		}
	}
yy2668:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= ',') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2497;
			goto yy2376;
		} else {
			if (yych <= '-') goto yy2668;
			if (yych == '>') goto yy2658;
			goto yy2376;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2508;
			if (yych <= 0xE0) goto yy2509;
			goto yy2510;
		} else {
			if (yych <= 0xF0) goto yy2511;
			if (yych <= 0xF3) goto yy2512;
			if (yych <= 0xF4) goto yy2513;
			goto yy1730;
		}
	}
yy2669:
	yyaccept = 32;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[0+yych] & 64) {
		goto yy2754;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2082;
			if (yych <= '"') goto yy2763;
			goto yy2764;
		} else {
			if (yych <= '\\') goto yy2765;
			if (yych <= 0xC1) goto yy2082;
			goto yy2766;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2767;
			if (yych <= 0xEF) goto yy2768;
			goto yy2769;
		} else {
			if (yych <= 0xF3) goto yy2770;
			if (yych <= 0xF4) goto yy2771;
			goto yy2082;
		}
	}
yy2670:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2515;
	goto yy1730;
yy2671:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2670;
	goto yy1730;
yy2672:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2670;
	goto yy1730;
yy2673:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2672;
	goto yy1730;
yy2674:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2672;
	goto yy1730;
yy2675:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2672;
	goto yy1730;
yy2676:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 16) {
		goto yy2676;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2251;
			if (yych <= '\'') goto yy2250;
			goto yy2772;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy2773;
				goto yy2772;
			} else {
				if (yych <= '>') goto yy2774;
				if (yych <= '?') goto yy2773;
				goto yy2772;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy2775;
				goto yy2772;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2776;
				goto yy2777;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2778;
				goto yy2779;
			} else {
				if (yych <= 0xF3) goto yy2780;
				if (yych <= 0xF4) goto yy2781;
				goto yy1730;
			}
		}
	}
yy2677:
	yych = *++YYCURSOR;
	if (yych <= ' ') {
		if (yych <= '\f') {
			if (yych == '\t') goto yy2677;
			goto yy1730;
		} else {
			if (yych <= '\r') goto yy2677;
			if (yych <= 0x1F) goto yy1730;
			goto yy2677;
		}
	} else {
		if (yych <= '=') {
			if (yych == '/') goto yy1959;
			goto yy1730;
		} else {
			if (yych <= '>') goto yy1960;
			if (yych <= '?') goto yy1959;
			goto yy1730;
		}
	}
yy2678:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2678;
				goto yy2418;
			} else {
				if (yych <= '\r') goto yy2678;
				if (yych == ' ') goto yy2678;
				goto yy2418;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2782;
				if (yych == '-') goto yy2553;
				goto yy2418;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2554;
					goto yy2553;
				} else {
					if (yych <= '=') goto yy2418;
					if (yych <= '>') goto yy2555;
					goto yy2554;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2418;
				if (yych <= 'Z') goto yy2553;
				if (yych <= '[') goto yy2418;
				goto yy2556;
			} else {
				if (yych == '_') goto yy2553;
				if (yych <= '`') goto yy2418;
				goto yy2553;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2418;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2557;
				goto yy2558;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2559;
					goto yy2560;
				} else {
					if (yych <= 0xF3) goto yy2561;
					if (yych <= 0xF4) goto yy2562;
					goto yy1730;
				}
			}
		}
	}
yy2679:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy2087;
			if (yych <= '!') goto yy2555;
		} else {
			if (yych == '\\') goto yy2681;
			if (yych <= 0x7F) goto yy2555;
			goto yy2087;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2682;
			if (yych <= 0xE0) goto yy2683;
			goto yy2684;
		} else {
			if (yych <= 0xF0) goto yy2685;
			if (yych <= 0xF3) goto yy2686;
			if (yych <= 0xF4) goto yy2687;
			goto yy2087;
		}
	}
yy2680:
	yych = *++YYCURSOR;
	if (yych <= '/') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x08) goto yy1730;
				goto yy2783;
			} else {
				if (yych == '\r') goto yy2783;
				goto yy1730;
			}
		} else {
			if (yych <= ',') {
				if (yych <= ' ') goto yy2783;
				goto yy1730;
			} else {
				if (yych <= '-') goto yy2258;
				if (yych <= '.') goto yy1730;
				goto yy1959;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '>') {
				if (yych <= ':') goto yy2258;
				goto yy1730;
			} else {
				if (yych <= '?') goto yy1959;
				if (yych <= '@') goto yy1730;
				goto yy2258;
			}
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1730;
				goto yy2258;
			} else {
				if (yych <= '`') goto yy1730;
				if (yych <= 'z') goto yy2258;
				goto yy1730;
			}
		}
	}
yy2681:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2555;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2683;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2684;
			goto yy2685;
		} else {
			if (yych <= 0xF3) goto yy2686;
			if (yych <= 0xF4) goto yy2687;
			goto yy1730;
		}
	}
yy2682:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2555;
	goto yy1730;
yy2683:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2682;
	goto yy1730;
yy2684:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2682;
	goto yy1730;
yy2685:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2684;
	goto yy1730;
yy2686:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2684;
	goto yy1730;
yy2687:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2684;
	goto yy1730;
yy2688:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2688;
				goto yy2419;
			} else {
				if (yych <= '\r') goto yy2688;
				if (yych == ' ') goto yy2688;
				goto yy2419;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2782;
				if (yych == '-') goto yy2563;
				goto yy2419;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2564;
					goto yy2563;
				} else {
					if (yych <= '=') goto yy2419;
					if (yych <= '>') goto yy2565;
					goto yy2564;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2419;
				if (yych <= 'Z') goto yy2563;
				if (yych <= '[') goto yy2419;
				goto yy2566;
			} else {
				if (yych == '_') goto yy2563;
				if (yych <= '`') goto yy2419;
				goto yy2563;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2419;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2567;
				goto yy2568;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2569;
					goto yy2570;
				} else {
					if (yych <= 0xF3) goto yy2571;
					if (yych <= 0xF4) goto yy2572;
					goto yy1730;
				}
			}
		}
	}
yy2689:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2087;
			if (yych <= '&') goto yy2565;
			goto yy2680;
		} else {
			if (yych == '\\') goto yy2690;
			if (yych <= 0x7F) goto yy2565;
			goto yy2087;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2691;
			if (yych <= 0xE0) goto yy2692;
			goto yy2693;
		} else {
			if (yych <= 0xF0) goto yy2694;
			if (yych <= 0xF3) goto yy2695;
			if (yych <= 0xF4) goto yy2696;
			goto yy2087;
		}
	}
yy2690:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2565;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2692;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2693;
			goto yy2694;
		} else {
			if (yych <= 0xF3) goto yy2695;
			if (yych <= 0xF4) goto yy2696;
			goto yy1730;
		}
	}
yy2691:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2565;
	goto yy1730;
yy2692:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2691;
	goto yy1730;
yy2693:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2691;
	goto yy1730;
yy2694:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2693;
	goto yy1730;
yy2695:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2693;
	goto yy1730;
yy2696:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2693;
	goto yy1730;
yy2697:
	yych = *++YYCURSOR;
	if (yych == '+') goto yy2784;
	goto yy1730;
yy2698:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy2718;
	goto yy1821;
yy2699:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2785;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2785;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2785;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2700:
	yych = *++YYCURSOR;
	if (yych <= ':') {
		if (yych <= ' ') {
			if (yych <= 0x1F) goto yy1730;
			goto yy1940;
		} else {
			if (yych <= '/') goto yy1730;
			if (yych <= '9') goto yy2722;
			goto yy1983;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2722;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2722;
			goto yy1730;
		}
	}
yy2701:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2785;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2785;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2785;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2702:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2786;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2785;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2785;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2703:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'Z') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '5') goto yy2786;
				if (yych <= '9') goto yy2785;
				yyt2 = YYCURSOR;
				goto yy2700;
			} else {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2785;
				goto yy2103;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '`') {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'f') goto yy2785;
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2704:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2787;
	if (yych <= '/') goto yy2104;
	if (yych <= '9') goto yy2788;
	goto yy2104;
yy2705:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2787;
	if (yych <= '/') goto yy2104;
	if (yych <= '9') goto yy2704;
	goto yy2104;
yy2706:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych == '.') goto yy2787;
		goto yy2104;
	} else {
		if (yych <= '4') goto yy2704;
		if (yych <= '5') goto yy2789;
		if (yych <= '9') goto yy2788;
		goto yy2104;
	}
yy2707:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2790;
	goto yy1730;
yy2708:
	yyaccept = 14;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy1809;
			goto yy1745;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1809;
				goto yy2718;
			} else {
				if (yych <= '.') {
					yyt3 = YYCURSOR;
					goto yy1976;
				}
				if (yych <= '/') goto yy1779;
				goto yy1820;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= ':') {
				yyt3 = YYCURSOR;
				goto yy1980;
			}
			if (yych <= '?') goto yy1809;
			if (yych <= '@') goto yy1785;
			goto yy1820;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1809;
				goto yy1707;
			} else {
				if (yych <= '`') goto yy1809;
				if (yych <= 'z') goto yy2110;
				goto yy1809;
			}
		}
	}
yy2709:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2791;
	goto yy1730;
yy2710:
	yych = *++YYCURSOR;
	if (yych == 'F') goto yy2792;
	goto yy1730;
yy2711:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2793;
	goto yy1913;
yy2712:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy2794;
	goto yy1730;
yy2713:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2795;
	goto yy1730;
yy2714:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '%') {
			if (yych == ' ') goto yy2713;
			if (yych <= '$') goto yy2618;
			goto yy1794;
		} else {
			if (yych == '+') goto yy1794;
			if (yych <= ',') goto yy2618;
			goto yy1794;
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') goto yy2618;
			if (yych != 'T') goto yy1794;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy2618;
				goto yy1794;
			} else {
				if (yych <= '`') goto yy2618;
				if (yych <= 'z') goto yy1794;
				goto yy2618;
			}
		}
	}
yy2715:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1794;
	if (yych <= '9') goto yy2796;
	goto yy1794;
yy2716:
	yyaccept = 41;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych == ' ') goto yy2713;
			goto yy2618;
		} else {
			if (yych <= ':') goto yy1913;
			if (yych <= '?') goto yy2618;
			goto yy1913;
		}
	} else {
		if (yych <= '^') {
			if (yych <= 'T') goto yy2717;
			if (yych <= 'Z') goto yy1913;
			goto yy2618;
		} else {
			if (yych == '`') goto yy2618;
			if (yych <= 'z') goto yy1913;
			goto yy2618;
		}
	}
yy2717:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2797;
	goto yy1913;
yy2718:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy2798;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy2798;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy2798;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2719:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy2799;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy2800;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2720:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy2800;
		goto yy1932;
	}
yy2721:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '9') goto yy2801;
				yyt1 = YYCURSOR;
				goto yy2625;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2801;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2801;
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2722:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2802;
				yyt2 = YYCURSOR;
				goto yy2803;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2802;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2802;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2723:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych == '.') {
					yyt2 = YYCURSOR;
					goto yy2177;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2801;
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2625;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2801;
				if (yych <= 'Z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2801;
			} else {
				if (yych <= 'z') goto yy1912;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1789;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2724:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy2586;
		if (yych <= '9') goto yy2447;
		goto yy2586;
	} else {
		if (yych <= 'F') goto yy2447;
		if (yych <= '`') goto yy2586;
		if (yych <= 'f') goto yy2447;
		goto yy2586;
	}
yy2725:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2804;
		if (yych <= ':') goto yy2467;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2804;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2804;
		goto yy1730;
	}
yy2726:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2805;
		if (yych <= ':') goto yy2634;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2805;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2805;
		goto yy1730;
	}
yy2727:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2806;
		if (yych <= ':') goto yy2807;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2806;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2806;
		goto yy1730;
	}
yy2728:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '0') {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2808;
			} else {
				if (yych <= '1') goto yy2809;
				if (yych <= '2') goto yy2810;
				goto yy2808;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2811;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2811;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2729:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2812;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2730:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2811;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2811;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2811;
			goto yy1730;
		}
	}
yy2731:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2812;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2730;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2812;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2812;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2732:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2813;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2733:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2813;
				goto yy2812;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2730;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2812;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2812;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2734:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt4 = YYCURSOR;
				goto yy2470;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2735:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '1') {
			if (yych <= '/') goto yy1730;
			if (yych <= '0') goto yy2814;
			goto yy2815;
		} else {
			if (yych <= '2') goto yy2816;
			if (yych <= '9') goto yy2814;
			goto yy1730;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1912;
			if (yych <= '^') goto yy1730;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'z') goto yy1912;
			goto yy1730;
		}
	}
yy2736:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2735;
	goto yy1913;
yy2737:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2735;
	if (yych <= '/') goto yy1913;
	if (yych <= '5') goto yy2736;
	goto yy1913;
yy2738:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2479;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2739:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2817;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2740:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1932;
			goto yy1683;
		} else {
			if (yych == ',') goto yy1683;
			goto yy1932;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1932;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1932;
			goto yy1683;
		}
	}
yy2741:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '+') goto yy1937;
	if (yych == '-') goto yy2064;
	goto yy1932;
yy2742:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy2818;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy2798;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy2798;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy2743:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2819;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2744:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2743;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2745:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '4') goto yy2743;
			if (yych <= '5') goto yy2820;
			goto yy2819;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy1945;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2746:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2486;
				yyt2 = YYCURSOR;
				goto yy2348;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2486;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2486;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2747:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy1940;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy2821;
			if (yych <= '1') goto yy2822;
			goto yy2823;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy2824;
			if (yych <= '@') goto yy1730;
			goto yy2332;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2332;
			goto yy1730;
		}
	}
yy2748:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2825;
				yyt2 = YYCURSOR;
				goto yy2348;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2486;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2486;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2749:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2825;
				goto yy2486;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2348;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2486;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2486;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2750:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ',') goto yy2497;
			goto yy2750;
		} else {
			if (yych == '>') goto yy2826;
			if (yych <= 0x7F) goto yy2497;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2661;
			if (yych <= 0xE0) goto yy2662;
			goto yy2663;
		} else {
			if (yych <= 0xF0) goto yy2664;
			if (yych <= 0xF3) goto yy2665;
			if (yych <= 0xF4) goto yy2666;
			goto yy1730;
		}
	}
yy2751:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 32) {
		goto yy2751;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych >= '#') goto yy2753;
		} else {
			if (yych <= '*') goto yy2754;
			if (yych <= '-') goto yy2755;
			if (yych <= '\\') goto yy2756;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy1730;
		}
	}
yy2752:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2827;
				goto yy2364;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2827;
					goto yy2364;
				} else {
					if (yych <= ' ') goto yy2827;
					if (yych <= '!') goto yy2364;
					goto yy2828;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2829;
				if (yych <= ')') goto yy2364;
				goto yy2223;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2364;
					goto yy2830;
				} else {
					if (yych <= '/') goto yy2364;
					if (yych <= ':') goto yy2831;
					goto yy2364;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2507;
				if (yych <= '@') goto yy2364;
				if (yych <= 'Z') goto yy2831;
				goto yy2364;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2366;
					goto yy2364;
				} else {
					if (yych == '`') goto yy2364;
					goto yy2831;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2364;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2367;
				goto yy2368;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2369;
					goto yy2370;
				} else {
					if (yych <= 0xF3) goto yy2371;
					if (yych <= 0xF4) goto yy2372;
					goto yy1730;
				}
			}
		}
	}
yy2753:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2832;
				goto yy2355;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2832;
					goto yy2355;
				} else {
					if (yych <= ' ') goto yy2832;
					if (yych <= '!') goto yy2355;
					goto yy2833;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2834;
				if (yych <= ')') goto yy2355;
				goto yy2215;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2355;
					goto yy2835;
				} else {
					if (yych <= '/') goto yy2355;
					if (yych <= ':') goto yy2836;
					goto yy2355;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2499;
				if (yych <= '@') goto yy2355;
				if (yych <= 'Z') goto yy2836;
				goto yy2355;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2357;
					goto yy2355;
				} else {
					if (yych == '`') goto yy2355;
					goto yy2836;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2355;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2358;
				goto yy2359;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2360;
					goto yy2361;
				} else {
					if (yych <= 0xF3) goto yy2362;
					if (yych <= 0xF4) goto yy2363;
					goto yy1730;
				}
			}
		}
	}
yy2754:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 64) {
		goto yy2754;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2763;
			goto yy2764;
		} else {
			if (yych <= '\\') goto yy2765;
			if (yych <= 0xC1) goto yy1730;
			goto yy2766;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2767;
			if (yych <= 0xEF) goto yy2768;
			goto yy2769;
		} else {
			if (yych <= 0xF3) goto yy2770;
			if (yych <= 0xF4) goto yy2771;
			goto yy1730;
		}
	}
yy2755:
	yych = *++YYCURSOR;
	if (yych <= '\\') {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2752;
			goto yy2751;
		} else {
			if (yych <= ',') {
				if (yych <= '\'') goto yy2753;
				goto yy2751;
			} else {
				if (yych <= '-') goto yy2837;
				if (yych <= '[') goto yy2751;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) goto yy2751;
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2757;
			goto yy2758;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2759;
				goto yy2760;
			} else {
				if (yych <= 0xF3) goto yy2761;
				if (yych <= 0xF4) goto yy2762;
				goto yy1730;
			}
		}
	}
yy2756:
	yych = *++YYCURSOR;
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= ')') goto yy2751;
			goto yy2754;
		} else {
			if (yych == '-') goto yy2755;
			if (yych <= 0x7F) goto yy2751;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy1730;
		}
	}
yy2757:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2751;
	goto yy1730;
yy2758:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2757;
	goto yy1730;
yy2759:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2757;
	goto yy1730;
yy2760:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2759;
	goto yy1730;
yy2761:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2759;
	goto yy1730;
yy2762:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2759;
	goto yy1730;
yy2763:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2838;
				goto yy2223;
			} else {
				if (yych <= '\r') goto yy2838;
				if (yych == ' ') goto yy2838;
				goto yy2223;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2839;
				if (yych == '\'') goto yy2840;
				goto yy2223;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2841;
					goto yy2223;
				} else {
					if (yych <= ':') goto yy2841;
					if (yych <= '=') goto yy2223;
					goto yy2523;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2223;
				if (yych <= 'Z') goto yy2841;
				if (yych <= '[') goto yy2223;
				goto yy2224;
			} else {
				if (yych == '_') goto yy2841;
				if (yych <= '`') goto yy2223;
				goto yy2841;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2223;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2225;
				goto yy2226;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2227;
					goto yy2228;
				} else {
					if (yych <= 0xF3) goto yy2229;
					if (yych <= 0xF4) goto yy2230;
					goto yy1730;
				}
			}
		}
	}
yy2764:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2842;
				goto yy2215;
			} else {
				if (yych <= '\r') goto yy2842;
				if (yych == ' ') goto yy2842;
				goto yy2215;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2843;
				if (yych == '\'') goto yy2844;
				goto yy2215;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2845;
					goto yy2215;
				} else {
					if (yych <= ':') goto yy2845;
					if (yych <= '=') goto yy2215;
					goto yy2516;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2215;
				if (yych <= 'Z') goto yy2845;
				if (yych <= '[') goto yy2215;
				goto yy2216;
			} else {
				if (yych == '_') goto yy2845;
				if (yych <= '`') goto yy2215;
				goto yy2845;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2215;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2217;
				goto yy2218;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2219;
					goto yy2220;
				} else {
					if (yych <= 0xF3) goto yy2221;
					if (yych <= 0xF4) goto yy2222;
					goto yy1730;
				}
			}
		}
	}
yy2765:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2754;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2767;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2768;
			goto yy2769;
		} else {
			if (yych <= 0xF3) goto yy2770;
			if (yych <= 0xF4) goto yy2771;
			goto yy1730;
		}
	}
yy2766:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2754;
	goto yy1730;
yy2767:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2766;
	goto yy1730;
yy2768:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2766;
	goto yy1730;
yy2769:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2768;
	goto yy1730;
yy2770:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2768;
	goto yy1730;
yy2771:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2768;
	goto yy1730;
yy2772:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2772;
				goto yy2676;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2772;
					goto yy2676;
				} else {
					if (yych <= ' ') goto yy2772;
					if (yych <= '!') goto yy2676;
					goto yy2251;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy2250;
				if (yych <= ',') goto yy2676;
				goto yy2772;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy2676;
				} else {
					if (yych <= ':') goto yy2772;
					if (yych <= '<') goto yy2676;
					goto yy2846;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy2774;
				if (yych <= '?') goto yy2773;
				if (yych <= '@') goto yy2676;
				goto yy2772;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy2676;
					goto yy2775;
				} else {
					if (yych == '_') goto yy2772;
					goto yy2676;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy2772;
					goto yy2676;
				} else {
					if (yych <= 0xC1) goto yy1730;
					if (yych <= 0xDF) goto yy2776;
					goto yy2777;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2778;
					goto yy2779;
				} else {
					if (yych <= 0xF3) goto yy2780;
					if (yych <= 0xF4) goto yy2781;
					goto yy1730;
				}
			}
		}
	}
yy2773:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 16) {
		goto yy2676;
	}
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2251;
			if (yych <= '\'') goto yy2250;
			goto yy2772;
		} else {
			if (yych <= ':') {
				if (yych <= '/') goto yy2773;
				goto yy2772;
			} else {
				if (yych <= '>') goto yy2847;
				if (yych <= '?') goto yy2773;
				goto yy2772;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 'z') {
				if (yych <= '\\') goto yy2775;
				goto yy2772;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2776;
				goto yy2777;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2778;
				goto yy2779;
			} else {
				if (yych <= 0xF3) goto yy2780;
				if (yych <= 0xF4) goto yy2781;
				goto yy1730;
			}
		}
	}
yy2774:
	yyaccept = 35;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1961;
			if (yych == '"') goto yy2849;
			goto yy2848;
		} else {
			if (yych <= '\'') goto yy2850;
			if (yych == '\\') goto yy2851;
			goto yy2848;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1961;
			if (yych <= 0xDF) goto yy2852;
			if (yych <= 0xE0) goto yy2853;
			goto yy2854;
		} else {
			if (yych <= 0xF0) goto yy2855;
			if (yych <= 0xF3) goto yy2856;
			if (yych <= 0xF4) goto yy2857;
			goto yy1961;
		}
	}
yy2775:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2676;
				goto yy2772;
			} else {
				if (yych <= '.') goto yy2676;
				if (yych <= '/') goto yy2773;
				goto yy2772;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2676;
				if (yych <= '>') goto yy2774;
				goto yy2773;
			} else {
				if (yych <= '@') goto yy2676;
				if (yych <= 'Z') goto yy2772;
				goto yy2676;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2676;
				goto yy2772;
			} else {
				if (yych <= 0x7F) goto yy2676;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2777;
				if (yych <= 0xEF) goto yy2778;
				goto yy2779;
			} else {
				if (yych <= 0xF3) goto yy2780;
				if (yych <= 0xF4) goto yy2781;
				goto yy1730;
			}
		}
	}
yy2776:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2676;
	goto yy1730;
yy2777:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2776;
	goto yy1730;
yy2778:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2776;
	goto yy1730;
yy2779:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2778;
	goto yy1730;
yy2780:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2778;
	goto yy1730;
yy2781:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2778;
	goto yy1730;
yy2782:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '!') goto yy2782;
				goto yy2419;
			} else {
				if (yych == '\'') goto yy2418;
				if (yych <= ',') goto yy2782;
				goto yy2858;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy2782;
				if (yych <= '/') goto yy2859;
				if (yych <= ':') goto yy2858;
				goto yy2782;
			} else {
				if (yych <= '>') goto yy2860;
				if (yych <= '?') goto yy2859;
				if (yych <= '@') goto yy2782;
				goto yy2858;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy2861;
				goto yy2782;
			} else {
				if (yych == '`') goto yy2782;
				if (yych <= 'z') goto yy2858;
				goto yy2782;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2862;
				if (yych <= 0xE0) goto yy2863;
				goto yy2864;
			} else {
				if (yych <= 0xF0) goto yy2865;
				if (yych <= 0xF3) goto yy2866;
				if (yych <= 0xF4) goto yy2867;
				goto yy1730;
			}
		}
	}
yy2783:
	yych = *++YYCURSOR;
	if (yych <= 0x1F) {
		if (yych <= '\t') {
			if (yych <= 0x08) goto yy1730;
			goto yy2783;
		} else {
			if (yych == '\r') goto yy2783;
			goto yy1730;
		}
	} else {
		if (yych <= '/') {
			if (yych <= ' ') goto yy2783;
			if (yych <= '.') goto yy1730;
			goto yy1959;
		} else {
			if (yych == '?') goto yy1959;
			goto yy1730;
		}
	}
yy2784:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2868;
	goto yy1730;
yy2785:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2869;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2869;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2869;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2786:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2266;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2869;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'F') goto yy2869;
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= 'f') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2869;
			} else {
				if (yych <= 'z') goto yy2103;
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1698;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				yyt2 = YYCURSOR;
				goto yy1950;
			} else {
				if (yych <= 0xF0) {
					yyt2 = YYCURSOR;
					goto yy1951;
				}
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2787:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych <= '/') {
			if (yych == '-') goto yy2113;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy2870;
			if (yych <= '1') goto yy2871;
			if (yych <= '2') goto yy2872;
			goto yy2870;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1730;
			if (yych <= 'Z') goto yy2103;
			goto yy1730;
		} else {
			if (yych == '`') goto yy1730;
			if (yych <= 'z') goto yy2103;
			goto yy1730;
		}
	}
yy2788:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2787;
	goto yy2104;
yy2789:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy2787;
	if (yych <= '/') goto yy2104;
	if (yych <= '5') goto yy2788;
	goto yy2104;
yy2790:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2873;
	goto yy1730;
yy2791:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2874;
	goto yy1730;
yy2792:
	yych = *++YYCURSOR;
	if (yych == 'i') goto yy2875;
	goto yy1730;
yy2793:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2716;
	goto yy1913;
yy2794:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2876;
	goto yy1730;
yy2795:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2877;
	goto yy1730;
yy2796:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1794;
	if (yych <= '9') goto yy2878;
	goto yy1794;
yy2797:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') goto yy1913;
	if (yych <= '9') goto yy2879;
	goto yy1913;
yy2798:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy2880;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy2880;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy2880;
		goto yy1794;
	}
yy2799:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy2881;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy2882;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2800:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy2882;
		goto yy1932;
	}
yy2801:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy1912;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy2625;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2802:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2883;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2883;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2883;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2803:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2884;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2884;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2884;
			goto yy1730;
		}
	}
yy2804:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2885;
		if (yych <= ':') goto yy2634;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2885;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2885;
		goto yy1730;
	}
yy2805:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2634;
	goto yy1730;
yy2806:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2886;
		if (yych >= ';') goto yy1730;
	} else {
		if (yych <= 'F') goto yy2886;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2886;
		goto yy1730;
	}
yy2807:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2887;
		if (yych <= ':') goto yy2888;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2887;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2887;
		goto yy1730;
	}
yy2808:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2889;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2809:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2892;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2810:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2892;
				goto yy2893;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2889;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2890;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2891;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2891;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2811:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2891;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2890;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2891;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2891;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2812:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2894;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2730;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2894;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2894;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2813:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2894;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2894;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2894;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2814:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy2895;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2896;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1789;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2815:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy2814;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2896;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1789;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2816:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '4') goto yy2814;
				if (yych <= '5') goto yy2897;
				goto yy2895;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2896;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2817:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2898;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2818:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1794;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2899;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy2880;
		case 'B':
		case 'b': goto yy2900;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2201;
		case 'S':
		case 's': goto yy2202;
		case 'm': goto yy2203;
		case 'p':
		case 'u': goto yy2204;
		default: goto yy1683;
	}
yy2819:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2820:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '5') goto yy2819;
			if (yych <= '9') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1945;
		} else {
			if (yych <= '@') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			goto yy1730;
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2821:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2901;
				yyt2 = YYCURSOR;
				goto yy2902;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2822:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2903;
				yyt2 = YYCURSOR;
				goto yy2479;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2823:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2903;
				goto yy2904;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2901;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2479;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2478;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2478;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2824:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2901;
				yyt2 = YYCURSOR;
				goto yy2479;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2478;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2825:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2651;
				yyt2 = YYCURSOR;
				goto yy2348;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2651;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2651;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2826:
	yyaccept = 37;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[0+yych] & 32) {
		goto yy2751;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy2496;
			if (yych <= '"') goto yy2752;
			goto yy2753;
		} else {
			if (yych <= '*') goto yy2754;
			if (yych <= '-') goto yy2755;
			if (yych <= '\\') goto yy2756;
			goto yy2496;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy2496;
		}
	}
yy2827:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2364;
				goto yy2827;
			} else {
				if (yych == '\r') goto yy2827;
				if (yych <= 0x1F) goto yy2364;
				goto yy2827;
			}
		} else {
			if (yych <= '&') {
				if (yych != '"') goto yy2364;
			} else {
				if (yych <= '\'') goto yy2829;
				if (yych == '*') goto yy2223;
				goto yy2364;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '>') {
				if (yych <= '-') goto yy2365;
				if (yych <= '=') goto yy2364;
				goto yy2507;
			} else {
				if (yych == '\\') goto yy2366;
				if (yych <= 0x7F) goto yy2364;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2367;
				if (yych <= 0xE0) goto yy2368;
				goto yy2369;
			} else {
				if (yych <= 0xF0) goto yy2370;
				if (yych <= 0xF3) goto yy2371;
				if (yych <= 0xF4) goto yy2372;
				goto yy1730;
			}
		}
	}
yy2828:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 32) {
		goto yy2751;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2364;
			goto yy2753;
		} else {
			if (yych <= '*') goto yy2754;
			if (yych <= '-') goto yy2755;
			if (yych <= '\\') goto yy2756;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy1730;
		}
	}
yy2829:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2827;
				goto yy2364;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2827;
					goto yy2364;
				} else {
					if (yych <= ' ') goto yy2827;
					if (yych <= '!') goto yy2364;
					goto yy2828;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2235;
				if (yych <= ')') goto yy2364;
				goto yy2223;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2364;
				} else {
					if (yych <= '/') goto yy2364;
					if (yych <= ':') goto yy2831;
					goto yy2364;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2507;
				if (yych <= '@') goto yy2364;
				if (yych <= 'Z') goto yy2831;
				goto yy2364;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2366;
					goto yy2364;
				} else {
					if (yych == '`') goto yy2364;
					goto yy2831;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2364;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2367;
				goto yy2368;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2369;
					goto yy2370;
				} else {
					if (yych <= 0xF3) goto yy2371;
					if (yych <= 0xF4) goto yy2372;
					goto yy1730;
				}
			}
		}
	}
yy2830:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2905;
				goto yy2364;
			} else {
				if (yych <= '\r') goto yy2905;
				if (yych == ' ') goto yy2905;
				goto yy2364;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2828;
				if (yych == '\'') goto yy2829;
				goto yy2364;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2906;
					goto yy2364;
				} else {
					if (yych <= ':') goto yy2831;
					if (yych <= '<') goto yy2364;
					goto yy2907;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2507;
				if (yych <= '@') goto yy2364;
				if (yych >= '[') goto yy2364;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2366;
					goto yy2364;
				} else {
					if (yych == '`') goto yy2364;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2364;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2367;
				goto yy2368;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2369;
					goto yy2370;
				} else {
					if (yych <= 0xF3) goto yy2371;
					if (yych <= 0xF4) goto yy2372;
					goto yy1730;
				}
			}
		}
	}
yy2831:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2905;
				goto yy2364;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2905;
					goto yy2364;
				} else {
					if (yych <= ' ') goto yy2905;
					if (yych <= '!') goto yy2364;
					goto yy2828;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2829;
				if (yych <= ')') goto yy2364;
				goto yy2223;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2364;
					goto yy2830;
				} else {
					if (yych <= '/') goto yy2364;
					if (yych <= ':') goto yy2831;
					goto yy2364;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy2907;
				if (yych <= '>') goto yy2507;
				if (yych <= '@') goto yy2364;
				goto yy2831;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy2364;
					goto yy2366;
				} else {
					if (yych == '_') goto yy2831;
					goto yy2364;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy2831;
					goto yy2364;
				} else {
					if (yych <= 0xC1) goto yy1730;
					if (yych <= 0xDF) goto yy2367;
					goto yy2368;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2369;
					goto yy2370;
				} else {
					if (yych <= 0xF3) goto yy2371;
					if (yych <= 0xF4) goto yy2372;
					goto yy1730;
				}
			}
		}
	}
yy2832:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2355;
				goto yy2832;
			} else {
				if (yych == '\r') goto yy2832;
				if (yych <= 0x1F) goto yy2355;
				goto yy2832;
			}
		} else {
			if (yych <= '&') {
				if (yych != '"') goto yy2355;
			} else {
				if (yych <= '\'') goto yy2834;
				if (yych == '*') goto yy2215;
				goto yy2355;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '>') {
				if (yych <= '-') goto yy2356;
				if (yych <= '=') goto yy2355;
				goto yy2499;
			} else {
				if (yych == '\\') goto yy2357;
				if (yych <= 0x7F) goto yy2355;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2358;
				if (yych <= 0xE0) goto yy2359;
				goto yy2360;
			} else {
				if (yych <= 0xF0) goto yy2361;
				if (yych <= 0xF3) goto yy2362;
				if (yych <= 0xF4) goto yy2363;
				goto yy1730;
			}
		}
	}
yy2833:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2832;
				goto yy2355;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2832;
					goto yy2355;
				} else {
					if (yych <= ' ') goto yy2832;
					if (yych <= '!') goto yy2355;
					goto yy2234;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2834;
				if (yych <= ')') goto yy2355;
				goto yy2215;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2355;
					goto yy2835;
				} else {
					if (yych <= '/') goto yy2355;
					if (yych <= ':') goto yy2836;
					goto yy2355;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2499;
				if (yych <= '@') goto yy2355;
				if (yych <= 'Z') goto yy2836;
				goto yy2355;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2357;
					goto yy2355;
				} else {
					if (yych == '`') goto yy2355;
					goto yy2836;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2355;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2358;
				goto yy2359;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2360;
					goto yy2361;
				} else {
					if (yych <= 0xF3) goto yy2362;
					if (yych <= 0xF4) goto yy2363;
					goto yy1730;
				}
			}
		}
	}
yy2834:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 32) {
		goto yy2751;
	}
	if (yych <= 0xC1) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2752;
			goto yy2355;
		} else {
			if (yych <= '*') goto yy2754;
			if (yych <= '-') goto yy2755;
			if (yych <= '\\') goto yy2756;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2757;
			if (yych <= 0xE0) goto yy2758;
			goto yy2759;
		} else {
			if (yych <= 0xF0) goto yy2760;
			if (yych <= 0xF3) goto yy2761;
			if (yych <= 0xF4) goto yy2762;
			goto yy1730;
		}
	}
yy2835:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2908;
				goto yy2355;
			} else {
				if (yych <= '\r') goto yy2908;
				if (yych == ' ') goto yy2908;
				goto yy2355;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2833;
				if (yych == '\'') goto yy2834;
				goto yy2355;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2909;
					goto yy2355;
				} else {
					if (yych <= ':') goto yy2836;
					if (yych <= '<') goto yy2355;
					goto yy2910;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2499;
				if (yych <= '@') goto yy2355;
				if (yych >= '[') goto yy2355;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2357;
					goto yy2355;
				} else {
					if (yych == '`') goto yy2355;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2355;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2358;
				goto yy2359;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2360;
					goto yy2361;
				} else {
					if (yych <= 0xF3) goto yy2362;
					if (yych <= 0xF4) goto yy2363;
					goto yy1730;
				}
			}
		}
	}
yy2836:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2908;
				goto yy2355;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2908;
					goto yy2355;
				} else {
					if (yych <= ' ') goto yy2908;
					if (yych <= '!') goto yy2355;
					goto yy2833;
				}
			}
		} else {
			if (yych <= '*') {
				if (yych == '\'') goto yy2834;
				if (yych <= ')') goto yy2355;
				goto yy2215;
			} else {
				if (yych <= '-') {
					if (yych <= ',') goto yy2355;
					goto yy2835;
				} else {
					if (yych <= '/') goto yy2355;
					if (yych <= ':') goto yy2836;
					goto yy2355;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '=') goto yy2910;
				if (yych <= '>') goto yy2499;
				if (yych <= '@') goto yy2355;
				goto yy2836;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy2355;
					goto yy2357;
				} else {
					if (yych == '_') goto yy2836;
					goto yy2355;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy2836;
					goto yy2355;
				} else {
					if (yych <= 0xC1) goto yy1730;
					if (yych <= 0xDF) goto yy2358;
					goto yy2359;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2360;
					goto yy2361;
				} else {
					if (yych <= 0xF3) goto yy2362;
					if (yych <= 0xF4) goto yy2363;
					goto yy1730;
				}
			}
		}
	}
yy2837:
	yych = *++YYCURSOR;
	if (yych <= '[') {
		if (yych <= '\'') {
			if (yych <= '!') {
				if (yych <= 0x00) goto yy1730;
				goto yy2751;
			} else {
				if (yych <= '"') goto yy2752;
				if (yych <= '&') goto yy2751;
				goto yy2753;
			}
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy2751;
				goto yy2837;
			} else {
				if (yych == '>') goto yy2826;
				goto yy2751;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= '\\') goto yy2756;
				goto yy2751;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2757;
				goto yy2758;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) goto yy2759;
				goto yy2760;
			} else {
				if (yych <= 0xF3) goto yy2761;
				if (yych <= 0xF4) goto yy2762;
				goto yy1730;
			}
		}
	}
yy2838:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2223;
				goto yy2838;
			} else {
				if (yych == '\r') goto yy2838;
				goto yy2223;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2838;
				if (yych <= '!') goto yy2223;
			} else {
				if (yych == '\'') goto yy2840;
				goto yy2223;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy2523;
				if (yych <= '[') goto yy2223;
				goto yy2224;
			} else {
				if (yych <= 0x7F) goto yy2223;
				if (yych <= 0xC1) goto yy1730;
				goto yy2225;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2226;
				if (yych <= 0xEF) goto yy2227;
				goto yy2228;
			} else {
				if (yych <= 0xF3) goto yy2229;
				if (yych <= 0xF4) goto yy2230;
				goto yy1730;
			}
		}
	}
yy2839:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 64) {
		goto yy2754;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2223;
			goto yy2764;
		} else {
			if (yych <= '\\') goto yy2765;
			if (yych <= 0xC1) goto yy1730;
			goto yy2766;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2767;
			if (yych <= 0xEF) goto yy2768;
			goto yy2769;
		} else {
			if (yych <= 0xF3) goto yy2770;
			if (yych <= 0xF4) goto yy2771;
			goto yy1730;
		}
	}
yy2840:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2838;
				goto yy2223;
			} else {
				if (yych <= '\r') goto yy2838;
				if (yych == ' ') goto yy2838;
				goto yy2223;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2839;
				if (yych == '\'') goto yy2078;
				goto yy2223;
			} else {
				if (yych <= '/') {
					if (yych >= '.') goto yy2223;
				} else {
					if (yych <= ':') goto yy2841;
					if (yych <= '=') goto yy2223;
					goto yy2523;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2223;
				if (yych <= 'Z') goto yy2841;
				if (yych <= '[') goto yy2223;
				goto yy2224;
			} else {
				if (yych == '_') goto yy2841;
				if (yych <= '`') goto yy2223;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2223;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2225;
				goto yy2226;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2227;
					goto yy2228;
				} else {
					if (yych <= 0xF3) goto yy2229;
					if (yych <= 0xF4) goto yy2230;
					goto yy1730;
				}
			}
		}
	}
yy2841:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2911;
				goto yy2223;
			} else {
				if (yych <= '\r') goto yy2911;
				if (yych == ' ') goto yy2911;
				goto yy2223;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2839;
				if (yych == '\'') goto yy2840;
				goto yy2223;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2841;
					goto yy2223;
				} else {
					if (yych <= ':') goto yy2841;
					if (yych <= '<') goto yy2223;
					goto yy2912;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2523;
				if (yych <= '@') goto yy2223;
				if (yych <= 'Z') goto yy2841;
				goto yy2223;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2224;
					goto yy2223;
				} else {
					if (yych == '`') goto yy2223;
					goto yy2841;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2223;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2225;
				goto yy2226;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2227;
					goto yy2228;
				} else {
					if (yych <= 0xF3) goto yy2229;
					if (yych <= 0xF4) goto yy2230;
					goto yy1730;
				}
			}
		}
	}
yy2842:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2215;
				goto yy2842;
			} else {
				if (yych == '\r') goto yy2842;
				goto yy2215;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2842;
				if (yych <= '!') goto yy2215;
			} else {
				if (yych == '\'') goto yy2844;
				goto yy2215;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '>') goto yy2516;
				if (yych <= '[') goto yy2215;
				goto yy2216;
			} else {
				if (yych <= 0x7F) goto yy2215;
				if (yych <= 0xC1) goto yy1730;
				goto yy2217;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2218;
				if (yych <= 0xEF) goto yy2219;
				goto yy2220;
			} else {
				if (yych <= 0xF3) goto yy2221;
				if (yych <= 0xF4) goto yy2222;
				goto yy1730;
			}
		}
	}
yy2843:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2842;
				goto yy2215;
			} else {
				if (yych <= '\r') goto yy2842;
				if (yych == ' ') goto yy2842;
				goto yy2215;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2077;
				if (yych != '\'') goto yy2215;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2845;
					goto yy2215;
				} else {
					if (yych <= ':') goto yy2845;
					if (yych <= '=') goto yy2215;
					goto yy2516;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2215;
				if (yych <= 'Z') goto yy2845;
				if (yych <= '[') goto yy2215;
				goto yy2216;
			} else {
				if (yych == '_') goto yy2845;
				if (yych <= '`') goto yy2215;
				goto yy2845;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2215;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2217;
				goto yy2218;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2219;
					goto yy2220;
				} else {
					if (yych <= 0xF3) goto yy2221;
					if (yych <= 0xF4) goto yy2222;
					goto yy1730;
				}
			}
		}
	}
yy2844:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 64) {
		goto yy2754;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2763;
			goto yy2215;
		} else {
			if (yych <= '\\') goto yy2765;
			if (yych <= 0xC1) goto yy1730;
			goto yy2766;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2767;
			if (yych <= 0xEF) goto yy2768;
			goto yy2769;
		} else {
			if (yych <= 0xF3) goto yy2770;
			if (yych <= 0xF4) goto yy2771;
			goto yy1730;
		}
	}
yy2845:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2913;
				goto yy2215;
			} else {
				if (yych <= '\r') goto yy2913;
				if (yych == ' ') goto yy2913;
				goto yy2215;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2843;
				if (yych == '\'') goto yy2844;
				goto yy2215;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2845;
					goto yy2215;
				} else {
					if (yych <= ':') goto yy2845;
					if (yych <= '<') goto yy2215;
					goto yy2914;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2516;
				if (yych <= '@') goto yy2215;
				if (yych <= 'Z') goto yy2845;
				goto yy2215;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2216;
					goto yy2215;
				} else {
					if (yych == '`') goto yy2215;
					goto yy2845;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2215;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2217;
				goto yy2218;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2219;
					goto yy2220;
				} else {
					if (yych <= 0xF3) goto yy2221;
					if (yych <= 0xF4) goto yy2222;
					goto yy1730;
				}
			}
		}
	}
yy2846:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2846;
				goto yy2676;
			} else {
				if (yych <= '\r') goto yy2846;
				if (yych == ' ') goto yy2846;
				goto yy2676;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy2772;
				if (yych <= '.') goto yy2676;
				if (yych <= '/') goto yy2773;
				goto yy2772;
			} else {
				if (yych <= '=') goto yy2676;
				if (yych <= '>') goto yy2774;
				if (yych <= '?') goto yy2773;
				goto yy2676;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy2772;
				if (yych == '\\') goto yy2775;
				goto yy2676;
			} else {
				if (yych == '`') goto yy2676;
				if (yych <= 'z') goto yy2772;
				goto yy2676;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2776;
				if (yych <= 0xE0) goto yy2777;
				goto yy2778;
			} else {
				if (yych <= 0xF0) goto yy2779;
				if (yych <= 0xF3) goto yy2780;
				if (yych <= 0xF4) goto yy2781;
				goto yy1730;
			}
		}
	}
yy2847:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy2087;
			if (yych == '"') goto yy2849;
		} else {
			if (yych <= '\'') goto yy2850;
			if (yych == '\\') goto yy2851;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy2087;
			if (yych <= 0xDF) goto yy2852;
			if (yych <= 0xE0) goto yy2853;
			goto yy2854;
		} else {
			if (yych <= 0xF0) goto yy2855;
			if (yych <= 0xF3) goto yy2856;
			if (yych <= 0xF4) goto yy2857;
			goto yy2087;
		}
	}
yy2848:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1730;
			if (yych != '"') goto yy2848;
		} else {
			if (yych <= '\'') goto yy2850;
			if (yych == '\\') goto yy2851;
			goto yy2848;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2852;
			if (yych <= 0xE0) goto yy2853;
			goto yy2854;
		} else {
			if (yych <= 0xF0) goto yy2855;
			if (yych <= 0xF3) goto yy2856;
			if (yych <= 0xF4) goto yy2857;
			goto yy1730;
		}
	}
yy2849:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2915;
				goto yy2543;
			} else {
				if (yych <= '\r') goto yy2915;
				if (yych == ' ') goto yy2915;
				goto yy2543;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2533;
				if (yych == '-') goto yy2916;
				goto yy2543;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2917;
					goto yy2916;
				} else {
					if (yych <= '=') goto yy2543;
					if (yych <= '>') goto yy2407;
					goto yy2917;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2543;
				if (yych <= 'Z') goto yy2916;
				if (yych <= '[') goto yy2543;
				goto yy2544;
			} else {
				if (yych == '_') goto yy2916;
				if (yych <= '`') goto yy2543;
				goto yy2916;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2543;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2545;
				goto yy2546;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2547;
					goto yy2548;
				} else {
					if (yych <= 0xF3) goto yy2549;
					if (yych <= 0xF4) goto yy2550;
					goto yy1730;
				}
			}
		}
	}
yy2850:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2918;
				goto yy2532;
			} else {
				if (yych <= '\r') goto yy2918;
				if (yych == ' ') goto yy2918;
				goto yy2532;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2533;
				if (yych == '-') goto yy2919;
				goto yy2532;
			} else {
				if (yych <= ':') {
					if (yych <= '/') goto yy2920;
					goto yy2919;
				} else {
					if (yych <= '=') goto yy2532;
					if (yych <= '>') goto yy2397;
					goto yy2920;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2532;
				if (yych <= 'Z') goto yy2919;
				if (yych <= '[') goto yy2532;
				goto yy2534;
			} else {
				if (yych == '_') goto yy2919;
				if (yych <= '`') goto yy2532;
				goto yy2919;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2532;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2535;
				goto yy2536;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2537;
					goto yy2538;
				} else {
					if (yych <= 0xF3) goto yy2539;
					if (yych <= 0xF4) goto yy2540;
					goto yy1730;
				}
			}
		}
	}
yy2851:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2848;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2853;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2854;
			goto yy2855;
		} else {
			if (yych <= 0xF3) goto yy2856;
			if (yych <= 0xF4) goto yy2857;
			goto yy1730;
		}
	}
yy2852:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2848;
	goto yy1730;
yy2853:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2852;
	goto yy1730;
yy2854:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2852;
	goto yy1730;
yy2855:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2854;
	goto yy1730;
yy2856:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2854;
	goto yy1730;
yy2857:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2854;
	goto yy1730;
yy2858:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '"') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2858;
				goto yy2782;
			} else {
				if (yych <= 0x1F) {
					if (yych <= '\r') goto yy2858;
					goto yy2782;
				} else {
					if (yych <= ' ') goto yy2858;
					if (yych <= '!') goto yy2782;
					goto yy2419;
				}
			}
		} else {
			if (yych <= '-') {
				if (yych == '\'') goto yy2418;
				if (yych <= ',') goto yy2782;
				goto yy2858;
			} else {
				if (yych <= '/') {
					if (yych <= '.') goto yy2782;
				} else {
					if (yych <= ':') goto yy2858;
					if (yych <= '<') goto yy2782;
					goto yy2921;
				}
			}
		}
	} else {
		if (yych <= '`') {
			if (yych <= 'Z') {
				if (yych <= '>') goto yy2860;
				if (yych <= '?') goto yy2859;
				if (yych <= '@') goto yy2782;
				goto yy2858;
			} else {
				if (yych <= '\\') {
					if (yych <= '[') goto yy2782;
					goto yy2861;
				} else {
					if (yych == '_') goto yy2858;
					goto yy2782;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) {
					if (yych <= 'z') goto yy2858;
					goto yy2782;
				} else {
					if (yych <= 0xC1) goto yy1730;
					if (yych <= 0xDF) goto yy2862;
					goto yy2863;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2864;
					goto yy2865;
				} else {
					if (yych <= 0xF3) goto yy2866;
					if (yych <= 0xF4) goto yy2867;
					goto yy1730;
				}
			}
		}
	}
yy2859:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '-') {
			if (yych <= '"') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= '!') goto yy2782;
				goto yy2419;
			} else {
				if (yych == '\'') goto yy2418;
				if (yych <= ',') goto yy2782;
				goto yy2858;
			}
		} else {
			if (yych <= '=') {
				if (yych <= '.') goto yy2782;
				if (yych <= '/') goto yy2859;
				if (yych <= ':') goto yy2858;
				goto yy2782;
			} else {
				if (yych <= '>') goto yy2922;
				if (yych <= '?') goto yy2859;
				if (yych <= '@') goto yy2782;
				goto yy2858;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych == '\\') goto yy2861;
				goto yy2782;
			} else {
				if (yych == '`') goto yy2782;
				if (yych <= 'z') goto yy2858;
				goto yy2782;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2862;
				if (yych <= 0xE0) goto yy2863;
				goto yy2864;
			} else {
				if (yych <= 0xF0) goto yy2865;
				if (yych <= 0xF3) goto yy2866;
				if (yych <= 0xF4) goto yy2867;
				goto yy1730;
			}
		}
	}
yy2860:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2923;
			goto yy2860;
		} else {
			if (yych <= '\'') goto yy2924;
			if (yych == '\\') goto yy2925;
			goto yy2860;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2926;
			if (yych <= 0xE0) goto yy2927;
			goto yy2928;
		} else {
			if (yych <= 0xF0) goto yy2929;
			if (yych <= 0xF3) goto yy2930;
			if (yych <= 0xF4) goto yy2931;
			goto yy1730;
		}
	}
yy2861:
	yych = *++YYCURSOR;
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= ',') goto yy2782;
				goto yy2858;
			} else {
				if (yych <= '.') goto yy2782;
				if (yych <= '/') goto yy2859;
				goto yy2858;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '=') goto yy2782;
				if (yych <= '>') goto yy2860;
				goto yy2859;
			} else {
				if (yych <= '@') goto yy2782;
				if (yych <= 'Z') goto yy2858;
				goto yy2782;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych == '`') goto yy2782;
				goto yy2858;
			} else {
				if (yych <= 0x7F) goto yy2782;
				if (yych <= 0xC1) goto yy1730;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2863;
				if (yych <= 0xEF) goto yy2864;
				goto yy2865;
			} else {
				if (yych <= 0xF3) goto yy2866;
				if (yych <= 0xF4) goto yy2867;
				goto yy1730;
			}
		}
	}
yy2862:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2782;
	goto yy1730;
yy2863:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2862;
	goto yy1730;
yy2864:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2862;
	goto yy1730;
yy2865:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2864;
	goto yy1730;
yy2866:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2864;
	goto yy1730;
yy2867:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2864;
	goto yy1730;
yy2868:
	yych = *++YYCURSOR;
	if (yych == ',') goto yy2932;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2868;
	goto yy1730;
yy2869:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '/') {
			if (yych <= ',') {
				if (yych == '%') {
					yyt1 = YYCURSOR;
					goto yy2623;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2103;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2700;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2870:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2265;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2933;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2934;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2871:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2265;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2870;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2934;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2872:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy2265;
				}
				yyt2 = YYCURSOR;
				goto yy2272;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= '4') goto yy2870;
				goto yy2935;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2933;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2934;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2873:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2936;
	goto yy1730;
yy2874:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy2937;
			}
			goto yy1730;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= 0xC1) goto yy1730;
			yyt2 = YYCURSOR;
			goto yy2588;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy2589;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy2590;
			}
			yyt2 = YYCURSOR;
			goto yy2591;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy2592;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy2593;
			}
			goto yy1730;
		}
	}
yy2875:
	yych = *++YYCURSOR;
	if (yych == 'l') goto yy2938;
	goto yy1730;
yy2876:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2939;
	goto yy1730;
yy2877:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2940;
	goto yy1730;
yy2878:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy2941;
	goto yy1794;
yy2879:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy2941;
	goto yy1913;
yy2880:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy2942;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy2942;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy2942;
		goto yy1794;
	}
yy2881:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy2943;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy2944;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2882:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy2944;
		goto yy1932;
	}
yy2883:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2945;
				yyt2 = YYCURSOR;
				goto yy2803;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2945;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2945;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2884:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2946;
				yyt2 = YYCURSOR;
				goto yy2947;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2946;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2946;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2885:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy2726;
			goto yy2948;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy2726;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy2726;
			goto yy2155;
		}
	}
yy2886:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2949;
		if (yych <= ':') goto yy2807;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2949;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2949;
		goto yy1730;
	}
yy2887:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2950;
		if (yych <= ':') goto yy2951;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2950;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2950;
		goto yy1730;
	}
yy2888:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2952;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2952;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2952;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2889:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2953;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2890:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2952;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2952;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2952;
			goto yy1730;
		}
	}
yy2891:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2953;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2890;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2953;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2953;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2892:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2954;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2893:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2954;
				goto yy2953;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2890;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2953;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2953;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2894:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2730;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2895:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '`') {
		if (yych <= '?') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '9') goto yy1912;
				if (yych >= ';') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'Z') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy1912;
			} else {
				if (yych == '_') {
					yyt2 = YYCURSOR;
					goto yy2055;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1789;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2896:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1940;
	if (yych == ':') goto yy1779;
	goto yy1730;
yy2897:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= ':') {
			if (yych <= '/') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= '5') goto yy2895;
				if (yych <= '9') goto yy1912;
				yyt2 = YYCURSOR;
				goto yy2896;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy1912;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2055;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1912;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1789;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1789;
			}
		}
	}
yy2898:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2955;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2899:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1794;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2956;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy2942;
		case 'B':
		case 'b': goto yy2957;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2201;
		case 'S':
		case 's': goto yy2202;
		case 'm': goto yy2203;
		case 'p':
		case 'u': goto yy2204;
		default: goto yy1683;
	}
yy2900:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1794;
			if (yych <= '*') goto yy1683;
			goto yy1794;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1794;
			} else {
				if (yych <= '9') goto yy2942;
				if (yych <= ':') goto yy1794;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1794;
			if (yych <= 'F') goto yy2942;
			if (yych <= 'Z') goto yy1794;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1794;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2942;
				if (yych <= 'z') goto yy1794;
				goto yy1683;
			}
		}
	}
yy2901:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2644;
				yyt2 = YYCURSOR;
				goto yy2479;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2644;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2644;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2902:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy1940;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy2958;
			if (yych <= '1') goto yy2959;
			goto yy2960;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy2961;
			if (yych <= '@') goto yy1730;
			goto yy2638;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2638;
			goto yy1730;
		}
	}
yy2903:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2962;
				yyt2 = YYCURSOR;
				goto yy2479;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2644;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2644;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2904:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy2962;
				goto yy2644;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2479;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2644;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2644;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2905:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2364;
				goto yy2905;
			} else {
				if (yych == '\r') goto yy2905;
				if (yych <= 0x1F) goto yy2364;
				goto yy2905;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy2828;
				goto yy2364;
			} else {
				if (yych <= '\'') goto yy2829;
				if (yych == '*') goto yy2223;
				goto yy2364;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '=') {
				if (yych <= '-') goto yy2365;
				if (yych <= '<') goto yy2364;
				goto yy2907;
			} else {
				if (yych <= '>') goto yy2507;
				if (yych == '\\') goto yy2366;
				goto yy2364;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2367;
				if (yych <= 0xE0) goto yy2368;
				goto yy2369;
			} else {
				if (yych <= 0xF0) goto yy2370;
				if (yych <= 0xF3) goto yy2371;
				if (yych <= 0xF4) goto yy2372;
				goto yy1730;
			}
		}
	}
yy2906:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2905;
				goto yy2364;
			} else {
				if (yych <= '\r') goto yy2905;
				if (yych == ' ') goto yy2905;
				goto yy2364;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2828;
				if (yych == '\'') goto yy2829;
				goto yy2364;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2906;
					goto yy2364;
				} else {
					if (yych <= ':') goto yy2831;
					if (yych <= '<') goto yy2364;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2658;
				if (yych <= '@') goto yy2364;
				if (yych <= 'Z') goto yy2831;
				goto yy2364;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2366;
					goto yy2364;
				} else {
					if (yych == '`') goto yy2364;
					goto yy2831;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2364;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2367;
				goto yy2368;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2369;
					goto yy2370;
				} else {
					if (yych <= 0xF3) goto yy2371;
					if (yych <= 0xF4) goto yy2372;
					goto yy1730;
				}
			}
		}
	}
yy2907:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 16) {
		goto yy2376;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '"') goto yy2497;
			goto yy2386;
		} else {
			if (yych <= '-') goto yy2506;
			if (yych <= '>') goto yy2364;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2508;
			if (yych <= 0xE0) goto yy2509;
			goto yy2510;
		} else {
			if (yych <= 0xF0) goto yy2511;
			if (yych <= 0xF3) goto yy2512;
			if (yych <= 0xF4) goto yy2513;
			goto yy1730;
		}
	}
yy2908:
	yych = *++YYCURSOR;
	if (yych <= ',') {
		if (yych <= ' ') {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2355;
				goto yy2908;
			} else {
				if (yych == '\r') goto yy2908;
				if (yych <= 0x1F) goto yy2355;
				goto yy2908;
			}
		} else {
			if (yych <= '&') {
				if (yych == '"') goto yy2833;
				goto yy2355;
			} else {
				if (yych <= '\'') goto yy2834;
				if (yych == '*') goto yy2215;
				goto yy2355;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '=') {
				if (yych <= '-') goto yy2356;
				if (yych <= '<') goto yy2355;
				goto yy2910;
			} else {
				if (yych <= '>') goto yy2499;
				if (yych == '\\') goto yy2357;
				goto yy2355;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2358;
				if (yych <= 0xE0) goto yy2359;
				goto yy2360;
			} else {
				if (yych <= 0xF0) goto yy2361;
				if (yych <= 0xF3) goto yy2362;
				if (yych <= 0xF4) goto yy2363;
				goto yy1730;
			}
		}
	}
yy2909:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2908;
				goto yy2355;
			} else {
				if (yych <= '\r') goto yy2908;
				if (yych == ' ') goto yy2908;
				goto yy2355;
			}
		} else {
			if (yych <= ',') {
				if (yych <= '"') goto yy2833;
				if (yych == '\'') goto yy2834;
				goto yy2355;
			} else {
				if (yych <= '/') {
					if (yych <= '-') goto yy2909;
					goto yy2355;
				} else {
					if (yych <= ':') goto yy2836;
					if (yych <= '<') goto yy2355;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '>') goto yy2657;
				if (yych <= '@') goto yy2355;
				if (yych <= 'Z') goto yy2836;
				goto yy2355;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2357;
					goto yy2355;
				} else {
					if (yych == '`') goto yy2355;
					goto yy2836;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2355;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2358;
				goto yy2359;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2360;
					goto yy2361;
				} else {
					if (yych <= 0xF3) goto yy2362;
					if (yych <= 0xF4) goto yy2363;
					goto yy1730;
				}
			}
		}
	}
yy2910:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 8) {
		goto yy2375;
	}
	if (yych <= 0xC1) {
		if (yych <= '*') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\'') goto yy2497;
			goto yy2385;
		} else {
			if (yych <= '-') goto yy2498;
			if (yych <= '>') goto yy2355;
			goto yy1730;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xDF) goto yy2500;
			if (yych <= 0xE0) goto yy2501;
			goto yy2502;
		} else {
			if (yych <= 0xF0) goto yy2503;
			if (yych <= 0xF3) goto yy2504;
			if (yych <= 0xF4) goto yy2505;
			goto yy1730;
		}
	}
yy2911:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2223;
				goto yy2911;
			} else {
				if (yych == '\r') goto yy2911;
				goto yy2223;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2911;
				if (yych <= '!') goto yy2223;
				goto yy2839;
			} else {
				if (yych == '\'') goto yy2840;
				goto yy2223;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy2912;
				if (yych <= '>') goto yy2523;
				goto yy2223;
			} else {
				if (yych <= '\\') goto yy2224;
				if (yych <= 0x7F) goto yy2223;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2225;
				if (yych <= 0xE0) goto yy2226;
				goto yy2227;
			} else {
				if (yych <= 0xF0) goto yy2228;
				if (yych <= 0xF3) goto yy2229;
				if (yych <= 0xF4) goto yy2230;
				goto yy1730;
			}
		}
	}
yy2912:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 128) {
		goto yy2386;
	}
	if (yych <= 0xDF) {
		if (yych <= '"') {
			if (yych <= 0x00) goto yy1730;
			goto yy2515;
		} else {
			if (yych <= '>') goto yy2223;
			if (yych <= 0xC1) goto yy1730;
			goto yy2524;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2525;
			if (yych <= 0xEF) goto yy2526;
			goto yy2527;
		} else {
			if (yych <= 0xF3) goto yy2528;
			if (yych <= 0xF4) goto yy2529;
			goto yy1730;
		}
	}
yy2913:
	yych = *++YYCURSOR;
	if (yych <= '<') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2215;
				goto yy2913;
			} else {
				if (yych == '\r') goto yy2913;
				goto yy2215;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2913;
				if (yych <= '!') goto yy2215;
				goto yy2843;
			} else {
				if (yych == '\'') goto yy2844;
				goto yy2215;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '=') goto yy2914;
				if (yych <= '>') goto yy2516;
				goto yy2215;
			} else {
				if (yych <= '\\') goto yy2216;
				if (yych <= 0x7F) goto yy2215;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2217;
				if (yych <= 0xE0) goto yy2218;
				goto yy2219;
			} else {
				if (yych <= 0xF0) goto yy2220;
				if (yych <= 0xF3) goto yy2221;
				if (yych <= 0xF4) goto yy2222;
				goto yy1730;
			}
		}
	}
yy2914:
	yych = *++YYCURSOR;
	if (yybm_init[256+yych] & 64) {
		goto yy2385;
	}
	if (yych <= 0xDF) {
		if (yych <= '\'') {
			if (yych <= 0x00) goto yy1730;
			goto yy2515;
		} else {
			if (yych <= '>') goto yy2215;
			if (yych <= 0xC1) goto yy1730;
			goto yy2517;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) goto yy2518;
			if (yych <= 0xEF) goto yy2519;
			goto yy2520;
		} else {
			if (yych <= 0xF3) goto yy2521;
			if (yych <= 0xF4) goto yy2522;
			goto yy1730;
		}
	}
yy2915:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2543;
				goto yy2915;
			} else {
				if (yych == '\r') goto yy2915;
				goto yy2543;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy2915;
				if (yych <= '&') goto yy2543;
				goto yy2533;
			} else {
				if (yych == '/') goto yy2917;
				goto yy2543;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy2407;
				if (yych <= '?') goto yy2917;
				goto yy2543;
			} else {
				if (yych <= '\\') goto yy2544;
				if (yych <= 0x7F) goto yy2543;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2545;
				if (yych <= 0xE0) goto yy2546;
				goto yy2547;
			} else {
				if (yych <= 0xF0) goto yy2548;
				if (yych <= 0xF3) goto yy2549;
				if (yych <= 0xF4) goto yy2550;
				goto yy1730;
			}
		}
	}
yy2916:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2963;
				goto yy2543;
			} else {
				if (yych <= '\r') goto yy2963;
				if (yych == ' ') goto yy2963;
				goto yy2543;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2533;
				if (yych == '-') goto yy2916;
				goto yy2543;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2916;
				} else {
					if (yych <= '<') goto yy2543;
					if (yych <= '=') goto yy2964;
					goto yy2407;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2917;
				if (yych <= '@') goto yy2543;
				if (yych <= 'Z') goto yy2916;
				goto yy2543;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2544;
					goto yy2543;
				} else {
					if (yych == '`') goto yy2543;
					goto yy2916;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2543;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2545;
				goto yy2546;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2547;
					goto yy2548;
				} else {
					if (yych <= 0xF3) goto yy2549;
					if (yych <= 0xF4) goto yy2550;
					goto yy1730;
				}
			}
		}
	}
yy2917:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2533;
			goto yy2543;
		} else {
			if (yych <= '>') goto yy2542;
			if (yych == '\\') goto yy2544;
			goto yy2543;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2545;
			if (yych <= 0xE0) goto yy2546;
			goto yy2547;
		} else {
			if (yych <= 0xF0) goto yy2548;
			if (yych <= 0xF3) goto yy2549;
			if (yych <= 0xF4) goto yy2550;
			goto yy1730;
		}
	}
yy2918:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2532;
				goto yy2918;
			} else {
				if (yych == '\r') goto yy2918;
				goto yy2532;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2918;
				if (yych <= '!') goto yy2532;
				goto yy2533;
			} else {
				if (yych == '/') goto yy2920;
				goto yy2532;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy2397;
				if (yych <= '?') goto yy2920;
				goto yy2532;
			} else {
				if (yych <= '\\') goto yy2534;
				if (yych <= 0x7F) goto yy2532;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2535;
				if (yych <= 0xE0) goto yy2536;
				goto yy2537;
			} else {
				if (yych <= 0xF0) goto yy2538;
				if (yych <= 0xF3) goto yy2539;
				if (yych <= 0xF4) goto yy2540;
				goto yy1730;
			}
		}
	}
yy2919:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2965;
				goto yy2532;
			} else {
				if (yych <= '\r') goto yy2965;
				if (yych == ' ') goto yy2965;
				goto yy2532;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2533;
				if (yych == '-') goto yy2919;
				goto yy2532;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2919;
				} else {
					if (yych <= '<') goto yy2532;
					if (yych <= '=') goto yy2966;
					goto yy2397;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2920;
				if (yych <= '@') goto yy2532;
				if (yych <= 'Z') goto yy2919;
				goto yy2532;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2534;
					goto yy2532;
				} else {
					if (yych == '`') goto yy2532;
					goto yy2919;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2532;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2535;
				goto yy2536;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2537;
					goto yy2538;
				} else {
					if (yych <= 0xF3) goto yy2539;
					if (yych <= 0xF4) goto yy2540;
					goto yy1730;
				}
			}
		}
	}
yy2920:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2533;
			goto yy2532;
		} else {
			if (yych <= '>') goto yy2531;
			if (yych == '\\') goto yy2534;
			goto yy2532;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2535;
			if (yych <= 0xE0) goto yy2536;
			goto yy2537;
		} else {
			if (yych <= 0xF0) goto yy2538;
			if (yych <= 0xF3) goto yy2539;
			if (yych <= 0xF4) goto yy2540;
			goto yy1730;
		}
	}
yy2921:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= ',') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2921;
				goto yy2782;
			} else {
				if (yych <= '\r') goto yy2921;
				if (yych == ' ') goto yy2921;
				goto yy2782;
			}
		} else {
			if (yych <= ':') {
				if (yych <= '-') goto yy2858;
				if (yych <= '.') goto yy2782;
				if (yych <= '/') goto yy2859;
				goto yy2858;
			} else {
				if (yych <= '=') goto yy2782;
				if (yych <= '>') goto yy2860;
				if (yych <= '?') goto yy2859;
				goto yy2782;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy2858;
				if (yych == '\\') goto yy2861;
				goto yy2782;
			} else {
				if (yych == '`') goto yy2782;
				if (yych <= 'z') goto yy2858;
				goto yy2782;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2862;
				if (yych <= 0xE0) goto yy2863;
				goto yy2864;
			} else {
				if (yych <= 0xF0) goto yy2865;
				if (yych <= 0xF3) goto yy2866;
				if (yych <= 0xF4) goto yy2867;
				goto yy1730;
			}
		}
	}
yy2922:
	yyaccept = 38;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0x7F) {
		if (yych <= '&') {
			if (yych <= 0x00) goto yy2087;
			if (yych != '"') goto yy2860;
		} else {
			if (yych <= '\'') goto yy2924;
			if (yych == '\\') goto yy2925;
			goto yy2860;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy2087;
			if (yych <= 0xDF) goto yy2926;
			if (yych <= 0xE0) goto yy2927;
			goto yy2928;
		} else {
			if (yych <= 0xF0) goto yy2929;
			if (yych <= 0xF3) goto yy2930;
			if (yych <= 0xF4) goto yy2931;
			goto yy2087;
		}
	}
yy2923:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2967;
				goto yy2565;
			} else {
				if (yych <= '\r') goto yy2967;
				if (yych == ' ') goto yy2967;
				goto yy2565;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2680;
				if (yych == '-') goto yy2968;
				goto yy2565;
			} else {
				if (yych <= '/') goto yy2969;
				if (yych <= ':') goto yy2968;
				if (yych <= '>') goto yy2565;
				goto yy2969;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2565;
				if (yych <= 'Z') goto yy2968;
				if (yych <= '[') goto yy2565;
				goto yy2690;
			} else {
				if (yych == '_') goto yy2968;
				if (yych <= '`') goto yy2565;
				goto yy2968;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2565;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2691;
				goto yy2692;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2693;
					goto yy2694;
				} else {
					if (yych <= 0xF3) goto yy2695;
					if (yych <= 0xF4) goto yy2696;
					goto yy1730;
				}
			}
		}
	}
yy2924:
	yych = *++YYCURSOR;
	if (yych <= '?') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2970;
				goto yy2555;
			} else {
				if (yych <= '\r') goto yy2970;
				if (yych == ' ') goto yy2970;
				goto yy2555;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2680;
				if (yych == '-') goto yy2971;
				goto yy2555;
			} else {
				if (yych <= '/') goto yy2972;
				if (yych <= ':') goto yy2971;
				if (yych <= '>') goto yy2555;
				goto yy2972;
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '\\') {
				if (yych <= '@') goto yy2555;
				if (yych <= 'Z') goto yy2971;
				if (yych <= '[') goto yy2555;
				goto yy2681;
			} else {
				if (yych == '_') goto yy2971;
				if (yych <= '`') goto yy2555;
				goto yy2971;
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2555;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2682;
				goto yy2683;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2684;
					goto yy2685;
				} else {
					if (yych <= 0xF3) goto yy2686;
					if (yych <= 0xF4) goto yy2687;
					goto yy1730;
				}
			}
		}
	}
yy2925:
	yych = *++YYCURSOR;
	if (yych <= 0xE0) {
		if (yych <= 0x7F) {
			if (yych <= 0x00) goto yy1730;
			goto yy2860;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy2927;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy2928;
			goto yy2929;
		} else {
			if (yych <= 0xF3) goto yy2930;
			if (yych <= 0xF4) goto yy2931;
			goto yy1730;
		}
	}
yy2926:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2860;
	goto yy1730;
yy2927:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy2926;
	goto yy1730;
yy2928:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2926;
	goto yy1730;
yy2929:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy2928;
	goto yy1730;
yy2930:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy2928;
	goto yy1730;
yy2931:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy2928;
	goto yy1730;
yy2932:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2973;
	goto yy1730;
yy2933:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '_') {
		if (yych <= '9') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2265;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2103;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2934;
				}
				if (yych <= '?') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2176;
			} else {
				if (yych <= 'Z') goto yy2103;
				if (yych <= '^') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2105;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2934:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy1940;
	if (yych == ':') goto yy1983;
	goto yy1730;
yy2935:
	yyaccept = 7;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= '5') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy2265;
			} else {
				if (yych <= '.') {
					yyt2 = YYCURSOR;
					goto yy2272;
				}
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				goto yy2933;
			}
		} else {
			if (yych <= '?') {
				if (yych <= '9') goto yy2103;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2934;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy2176;
				}
				if (yych <= 'Z') goto yy2103;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= '_') {
					yyt2 = YYCURSOR;
					goto yy2105;
				}
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2103;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1698;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1698;
			}
		}
	}
yy2936:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2974;
	goto yy1730;
yy2937:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy2976;
			}
			goto yy2586;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= 0xC1) goto yy2586;
			yyt2 = YYCURSOR;
			goto yy2588;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy2589;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy2590;
			}
			yyt2 = YYCURSOR;
			goto yy2591;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy2592;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy2593;
			}
			goto yy2586;
		}
	}
yy2938:
	yych = *++YYCURSOR;
	if (yych == 'e') goto yy2977;
	goto yy1730;
yy2939:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2978;
	goto yy1730;
yy2940:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2979;
	goto yy1730;
yy2941:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2979;
	if (yych <= ':') goto yy1779;
	goto yy1730;
yy2942:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy2980;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy2980;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy2980;
		goto yy1794;
	}
yy2943:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy2981;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy2982;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2944:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy2982;
		goto yy1932;
	}
yy2945:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy2623;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			goto yy1730;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2803;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2946:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2983;
				yyt2 = YYCURSOR;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2983;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2983;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2947:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2984;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2984;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2984;
			goto yy1730;
		}
	}
yy2948:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2985;
		if (yych <= ':') goto yy2728;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2985;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2985;
		goto yy1730;
	}
yy2949:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2807;
	goto yy1730;
yy2950:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2986;
		if (yych >= ';') goto yy1730;
	} else {
		if (yych <= 'F') goto yy2986;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2986;
		goto yy1730;
	}
yy2951:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy2987;
		if (yych <= ':') goto yy2988;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy2987;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy2987;
		goto yy1730;
	}
yy2952:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2989;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2990;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2989;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2989;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2953:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2991;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2890;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2991;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2991;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2954:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt4 = YYCURSOR;
				goto yy2468;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2991;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2991;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2991;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2955:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2992;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2956:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	switch (yych) {
		case '%':
		case '+':
		case '-':
		case '.':
		case '/':
		case ':':
		case '@':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1794;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy2993;
		case 'A':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'e':
		case 'f': goto yy2980;
		case 'B':
		case 'b': goto yy2994;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy2201;
		case 'S':
		case 's': goto yy2202;
		case 'm': goto yy2203;
		case 'p':
		case 'u': goto yy2204;
		default: goto yy1683;
	}
yy2957:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '?') {
		if (yych <= '+') {
			if (yych == '%') goto yy1794;
			if (yych <= '*') goto yy1683;
			goto yy1794;
		} else {
			if (yych <= '/') {
				if (yych <= ',') goto yy1683;
				goto yy1794;
			} else {
				if (yych <= '9') goto yy2980;
				if (yych <= ':') goto yy1794;
				goto yy1683;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') goto yy1794;
			if (yych <= 'F') goto yy2980;
			if (yych <= 'Z') goto yy1794;
			goto yy1683;
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1794;
				goto yy1683;
			} else {
				if (yych <= 'f') goto yy2980;
				if (yych <= 'z') goto yy1794;
				goto yy1683;
			}
		}
	}
yy2958:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2995;
				yyt2 = YYCURSOR;
				goto yy2996;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2959:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2997;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2960:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy2997;
				goto yy2998;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy2995;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2730;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2731;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2731;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2961:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2995;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2731;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2962:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2738;
				yyt2 = YYCURSOR;
				goto yy2479;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2738;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2738;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2963:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2543;
				goto yy2963;
			} else {
				if (yych == '\r') goto yy2963;
				goto yy2543;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy2963;
				if (yych <= '&') goto yy2543;
				goto yy2533;
			} else {
				if (yych == '/') goto yy2917;
				if (yych <= '<') goto yy2543;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy2407;
				if (yych <= '?') goto yy2917;
				goto yy2543;
			} else {
				if (yych <= '\\') goto yy2544;
				if (yych <= 0x7F) goto yy2543;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2545;
				if (yych <= 0xE0) goto yy2546;
				goto yy2547;
			} else {
				if (yych <= 0xF0) goto yy2548;
				if (yych <= 0xF3) goto yy2549;
				if (yych <= 0xF4) goto yy2550;
				goto yy1730;
			}
		}
	}
yy2964:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2251;
			} else {
				if (yych <= '\t') goto yy2541;
				if (yych <= '\f') goto yy2251;
				goto yy2541;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy2541;
				goto yy2251;
			} else {
				if (yych <= '"') goto yy2676;
				if (yych <= '=') goto yy2251;
				goto yy2543;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy2251;
				goto yy2408;
			} else {
				if (yych <= 0x7F) goto yy2251;
				if (yych <= 0xC1) goto yy1730;
				goto yy2409;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2410;
				if (yych <= 0xEF) goto yy2411;
				goto yy2412;
			} else {
				if (yych <= 0xF3) goto yy2413;
				if (yych <= 0xF4) goto yy2414;
				goto yy1730;
			}
		}
	}
yy2965:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2532;
				goto yy2965;
			} else {
				if (yych == '\r') goto yy2965;
				goto yy2532;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2965;
				if (yych <= '!') goto yy2532;
				goto yy2533;
			} else {
				if (yych == '/') goto yy2920;
				if (yych <= '<') goto yy2532;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych <= '>') goto yy2397;
				if (yych <= '?') goto yy2920;
				goto yy2532;
			} else {
				if (yych <= '\\') goto yy2534;
				if (yych <= 0x7F) goto yy2532;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2535;
				if (yych <= 0xE0) goto yy2536;
				goto yy2537;
			} else {
				if (yych <= 0xF0) goto yy2538;
				if (yych <= 0xF3) goto yy2539;
				if (yych <= 0xF4) goto yy2540;
				goto yy1730;
			}
		}
	}
yy2966:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2250;
			} else {
				if (yych <= '\t') goto yy2530;
				if (yych <= '\f') goto yy2250;
				goto yy2530;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy2530;
				goto yy2250;
			} else {
				if (yych <= '\'') goto yy2676;
				if (yych <= '=') goto yy2250;
				goto yy2532;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy2250;
				goto yy2398;
			} else {
				if (yych <= 0x7F) goto yy2250;
				if (yych <= 0xC1) goto yy1730;
				goto yy2399;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2400;
				if (yych <= 0xEF) goto yy2401;
				goto yy2402;
			} else {
				if (yych <= 0xF3) goto yy2403;
				if (yych <= 0xF4) goto yy2404;
				goto yy1730;
			}
		}
	}
yy2967:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2565;
				goto yy2967;
			} else {
				if (yych == '\r') goto yy2967;
				goto yy2565;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy2967;
				if (yych <= '&') goto yy2565;
				goto yy2680;
			} else {
				if (yych == '/') goto yy2969;
				goto yy2565;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy2969;
				if (yych <= '[') goto yy2565;
				goto yy2690;
			} else {
				if (yych <= 0x7F) goto yy2565;
				if (yych <= 0xC1) goto yy1730;
				goto yy2691;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2692;
				if (yych <= 0xEF) goto yy2693;
				goto yy2694;
			} else {
				if (yych <= 0xF3) goto yy2695;
				if (yych <= 0xF4) goto yy2696;
				goto yy1730;
			}
		}
	}
yy2968:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '&') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy2999;
				goto yy2565;
			} else {
				if (yych <= '\r') goto yy2999;
				if (yych == ' ') goto yy2999;
				goto yy2565;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '\'') goto yy2680;
				if (yych == '-') goto yy2968;
				goto yy2565;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2968;
				} else {
					if (yych == '=') goto yy3000;
					goto yy2565;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2969;
				if (yych <= '@') goto yy2565;
				if (yych <= 'Z') goto yy2968;
				goto yy2565;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2690;
					goto yy2565;
				} else {
					if (yych == '`') goto yy2565;
					goto yy2968;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2565;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2691;
				goto yy2692;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2693;
					goto yy2694;
				} else {
					if (yych <= 0xF3) goto yy2695;
					if (yych <= 0xF4) goto yy2696;
					goto yy1730;
				}
			}
		}
	}
yy2969:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '\'') goto yy2680;
			goto yy2565;
		} else {
			if (yych <= '>') goto yy2689;
			if (yych == '\\') goto yy2690;
			goto yy2565;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2691;
			if (yych <= 0xE0) goto yy2692;
			goto yy2693;
		} else {
			if (yych <= 0xF0) goto yy2694;
			if (yych <= 0xF3) goto yy2695;
			if (yych <= 0xF4) goto yy2696;
			goto yy1730;
		}
	}
yy2970:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2555;
				goto yy2970;
			} else {
				if (yych == '\r') goto yy2970;
				goto yy2555;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy2970;
				if (yych <= '!') goto yy2555;
				goto yy2680;
			} else {
				if (yych == '/') goto yy2972;
				goto yy2555;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '?') goto yy2972;
				if (yych <= '[') goto yy2555;
				goto yy2681;
			} else {
				if (yych <= 0x7F) goto yy2555;
				if (yych <= 0xC1) goto yy1730;
				goto yy2682;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2683;
				if (yych <= 0xEF) goto yy2684;
				goto yy2685;
			} else {
				if (yych <= 0xF3) goto yy2686;
				if (yych <= 0xF4) goto yy2687;
				goto yy1730;
			}
		}
	}
yy2971:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '!') {
			if (yych <= '\f') {
				if (yych <= 0x00) goto yy1730;
				if (yych == '\t') goto yy3001;
				goto yy2555;
			} else {
				if (yych <= '\r') goto yy3001;
				if (yych == ' ') goto yy3001;
				goto yy2555;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '"') goto yy2680;
				if (yych == '-') goto yy2971;
				goto yy2555;
			} else {
				if (yych <= ':') {
					if (yych >= '0') goto yy2971;
				} else {
					if (yych == '=') goto yy3002;
					goto yy2555;
				}
			}
		}
	} else {
		if (yych <= 'z') {
			if (yych <= '[') {
				if (yych <= '?') goto yy2972;
				if (yych <= '@') goto yy2555;
				if (yych <= 'Z') goto yy2971;
				goto yy2555;
			} else {
				if (yych <= '^') {
					if (yych <= '\\') goto yy2681;
					goto yy2555;
				} else {
					if (yych == '`') goto yy2555;
					goto yy2971;
				}
			}
		} else {
			if (yych <= 0xE0) {
				if (yych <= 0x7F) goto yy2555;
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) goto yy2682;
				goto yy2683;
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xEF) goto yy2684;
					goto yy2685;
				} else {
					if (yych <= 0xF3) goto yy2686;
					if (yych <= 0xF4) goto yy2687;
					goto yy1730;
				}
			}
		}
	}
yy2972:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '=') {
			if (yych <= 0x00) goto yy1730;
			if (yych == '"') goto yy2680;
			goto yy2555;
		} else {
			if (yych <= '>') goto yy2679;
			if (yych == '\\') goto yy2681;
			goto yy2555;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) goto yy2682;
			if (yych <= 0xE0) goto yy2683;
			goto yy2684;
		} else {
			if (yych <= 0xF0) goto yy2685;
			if (yych <= 0xF3) goto yy2686;
			if (yych <= 0xF4) goto yy2687;
			goto yy1730;
		}
	}
yy2973:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy3003;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy2973;
	goto yy1730;
yy2974:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy3004;
yy2975:
#line 313 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 80223 "../../lnav/src/data_scanner_re.cc"
yy2976:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= '9') {
				yyt1 = YYCURSOR;
				goto yy3005;
			}
			goto yy2586;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= 0xC1) goto yy2586;
			yyt2 = YYCURSOR;
			goto yy2588;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy2589;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy2590;
			}
			yyt2 = YYCURSOR;
			goto yy2591;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy2592;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy2593;
			}
			goto yy2586;
		}
	}
yy2977:
	yych = *++YYCURSOR;
	if (yych == 's') goto yy3006;
	goto yy1730;
yy2978:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3007;
	goto yy1730;
yy2979:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3008;
	goto yy1730;
yy2980:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy3010;
	goto yy1794;
yy2981:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy3011;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy3012;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy2982:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy3012;
		goto yy1932;
	}
yy2983:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3013;
				yyt2 = YYCURSOR;
				goto yy2947;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3013;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3013;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2984:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3014;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3014;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3014;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2985:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3015;
		if (yych <= ':') goto yy2807;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3015;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3015;
		goto yy1730;
	}
yy2986:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3016;
		if (yych <= ':') goto yy2951;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3016;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3016;
		goto yy1730;
	}
yy2987:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3017;
		if (yych <= ':') goto yy3018;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3017;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3017;
		goto yy1730;
	}
yy2988:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3019;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3019;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3019;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2989:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3020;
			if (yych >= ';') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			yyt2 = YYCURSOR;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3020;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3020;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2990:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy3019;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy3019;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy3019;
			goto yy1730;
		}
	}
yy2991:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2890;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy2992:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy3021;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy2993:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy3010;
	goto yy2195;
yy2994:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1794;
			goto yy1683;
		} else {
			if (yych <= '+') goto yy1794;
			if (yych <= ',') goto yy1683;
			if (yych <= '-') goto yy3010;
			goto yy1794;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy1683;
			if (yych <= 'Z') goto yy1794;
			goto yy1683;
		} else {
			if (yych == '`') goto yy1683;
			if (yych <= 'z') goto yy1794;
			goto yy1683;
		}
	}
yy2995:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2812;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2996:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy1940;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy3022;
			if (yych <= '1') goto yy3023;
			goto yy3024;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy3025;
			if (yych <= '@') goto yy1730;
			goto yy2811;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2811;
			goto yy1730;
		}
	}
yy2997:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3026;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2812;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2998:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy3026;
				goto yy2812;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2730;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2812;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2812;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy2999:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2565;
				goto yy2999;
			} else {
				if (yych == '\r') goto yy2999;
				goto yy2565;
			}
		} else {
			if (yych <= '\'') {
				if (yych <= ' ') goto yy2999;
				if (yych <= '&') goto yy2565;
				goto yy2680;
			} else {
				if (yych == '/') goto yy2969;
				if (yych <= '<') goto yy2565;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy2969;
				goto yy2565;
			} else {
				if (yych <= '\\') goto yy2690;
				if (yych <= 0x7F) goto yy2565;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2691;
				if (yych <= 0xE0) goto yy2692;
				goto yy2693;
			} else {
				if (yych <= 0xF0) goto yy2694;
				if (yych <= 0xF3) goto yy2695;
				if (yych <= 0xF4) goto yy2696;
				goto yy1730;
			}
		}
	}
yy3000:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2419;
			} else {
				if (yych <= '\t') goto yy2688;
				if (yych <= '\f') goto yy2419;
				goto yy2688;
			}
		} else {
			if (yych <= '!') {
				if (yych == ' ') goto yy2688;
				goto yy2419;
			} else {
				if (yych <= '"') goto yy2782;
				if (yych <= '=') goto yy2419;
				goto yy2565;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy2419;
				goto yy2566;
			} else {
				if (yych <= 0x7F) goto yy2419;
				if (yych <= 0xC1) goto yy1730;
				goto yy2567;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2568;
				if (yych <= 0xEF) goto yy2569;
				goto yy2570;
			} else {
				if (yych <= 0xF3) goto yy2571;
				if (yych <= 0xF4) goto yy2572;
				goto yy1730;
			}
		}
	}
yy3001:
	yych = *++YYCURSOR;
	if (yych <= '=') {
		if (yych <= 0x1F) {
			if (yych <= '\t') {
				if (yych <= 0x00) goto yy1730;
				if (yych <= 0x08) goto yy2555;
				goto yy3001;
			} else {
				if (yych == '\r') goto yy3001;
				goto yy2555;
			}
		} else {
			if (yych <= '"') {
				if (yych <= ' ') goto yy3001;
				if (yych <= '!') goto yy2555;
				goto yy2680;
			} else {
				if (yych == '/') goto yy2972;
				if (yych <= '<') goto yy2555;
			}
		}
	} else {
		if (yych <= 0xC1) {
			if (yych <= '[') {
				if (yych == '?') goto yy2972;
				goto yy2555;
			} else {
				if (yych <= '\\') goto yy2681;
				if (yych <= 0x7F) goto yy2555;
				goto yy1730;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xDF) goto yy2682;
				if (yych <= 0xE0) goto yy2683;
				goto yy2684;
			} else {
				if (yych <= 0xF0) goto yy2685;
				if (yych <= 0xF3) goto yy2686;
				if (yych <= 0xF4) goto yy2687;
				goto yy1730;
			}
		}
	}
yy3002:
	yych = *++YYCURSOR;
	if (yych <= '>') {
		if (yych <= '\r') {
			if (yych <= 0x08) {
				if (yych <= 0x00) goto yy1730;
				goto yy2418;
			} else {
				if (yych <= '\t') goto yy2678;
				if (yych <= '\f') goto yy2418;
				goto yy2678;
			}
		} else {
			if (yych <= '&') {
				if (yych == ' ') goto yy2678;
				goto yy2418;
			} else {
				if (yych <= '\'') goto yy2782;
				if (yych <= '=') goto yy2418;
				goto yy2555;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= '\\') {
				if (yych <= '[') goto yy2418;
				goto yy2556;
			} else {
				if (yych <= 0x7F) goto yy2418;
				if (yych <= 0xC1) goto yy1730;
				goto yy2557;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) goto yy2558;
				if (yych <= 0xEF) goto yy2559;
				goto yy2560;
			} else {
				if (yych <= 0xF3) goto yy2561;
				if (yych <= 0xF4) goto yy2562;
				goto yy1730;
			}
		}
	}
yy3003:
	yych = *++YYCURSOR;
	if (yych == '@') goto yy3027;
	goto yy1730;
yy3004:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3028;
	goto yy1730;
yy3005:
	yyaccept = 39;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 0xDF) {
		if (yych <= ':') {
			if (yych <= '9') {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			goto yy2586;
		} else {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy2585;
			}
			if (yych <= 0xC1) goto yy2586;
			yyt2 = YYCURSOR;
			goto yy2588;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy2589;
			}
			if (yych <= 0xEF) {
				yyt2 = YYCURSOR;
				goto yy2590;
			}
			yyt2 = YYCURSOR;
			goto yy2591;
		} else {
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy2592;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy2593;
			}
			goto yy2586;
		}
	}
yy3006:
	yyaccept = 9;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ' ') goto yy3029;
	goto yy1770;
yy3007:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy3030;
	goto yy1730;
yy3008:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == ':') goto yy3031;
yy3009:
#line 310 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_DATE_TIME);
       }
#line 81314 "../../lnav/src/data_scanner_re.cc"
yy3010:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy3032;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy3032;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy3032;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy3011:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy3033;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy3034;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy3012:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy3034;
		goto yy1932;
	}
yy3013:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy2623;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			goto yy1730;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2947;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3014:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3035;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3035;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3035;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3015:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy2886;
			goto yy3036;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy2886;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy2886;
			goto yy2155;
		}
	}
yy3016:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy2951;
	goto yy1730;
yy3017:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3037;
		if (yych >= ';') goto yy1730;
	} else {
		if (yych <= 'F') goto yy3037;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3037;
		goto yy1730;
	}
yy3018:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3019;
		if (yych <= ':') goto yy2819;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3019;
		if (yych <= '`') goto yy1730;
		if (yych >= 'g') goto yy1730;
	}
yy3019:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3038;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3038;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3038;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3020:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3039;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2990;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3039;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3039;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3021:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy3040;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy3022:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3041;
				yyt2 = YYCURSOR;
				goto yy3042;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3023:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3043;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3024:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy3043;
				goto yy3044;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy3041;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2890;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2891;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2891;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3025:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3041;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2891;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3026:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2894;
				yyt2 = YYCURSOR;
				goto yy2730;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2894;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2894;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3027:
	yych = *++YYCURSOR;
	if (yych == '@') goto yy3045;
	goto yy1730;
yy3028:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3046;
	goto yy1730;
yy3029:
	yych = *++YYCURSOR;
	if (yych == '(') goto yy3047;
	goto yy1730;
yy3030:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3048;
	goto yy1730;
yy3031:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3049;
	goto yy1730;
yy3032:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3050;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3050;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3050;
		goto yy1794;
	}
yy3033:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '7') {
		if (yych <= '+') {
			if (yych == '%') goto yy1775;
			if (yych <= '*') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '-') {
				if (yych <= ',') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '.') goto yy2033;
				if (yych <= '/') goto yy1932;
				goto yy3051;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '9') goto yy3052;
			if (yych <= ':') goto yy1932;
			if (yych <= '?') goto yy1781;
			goto yy1932;
		} else {
			if (yych <= '_') {
				if (yych <= '^') goto yy1781;
				goto yy1932;
			} else {
				if (yych <= '`') goto yy1781;
				if (yych <= 'z') goto yy1932;
				goto yy1781;
			}
		}
	}
yy3034:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '-') {
		if (yych == '%') goto yy1775;
		goto yy1932;
	} else {
		if (yych <= '.') goto yy2033;
		if (yych <= '/') goto yy1932;
		if (yych <= '9') goto yy3052;
		goto yy1932;
	}
yy3035:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '%') {
				if (yych <= '$') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2623;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3053;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3053;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3053;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3036:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3054;
		if (yych <= ':') goto yy2888;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3054;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3054;
		goto yy1730;
	}
yy3037:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3055;
		if (yych <= ':') goto yy3018;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3055;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3055;
		goto yy1730;
	}
yy3038:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3056;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3056;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3056;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3039:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy1730;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy2990;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') goto yy1730;
			if (yych <= '`') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 'z') goto yy1730;
			yyt2 = YYCURSOR;
			goto yy1940;
		}
	} else {
		if (yych <= 0xEF) {
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			if (yych <= 0xE0) {
				yyt2 = YYCURSOR;
				goto yy1949;
			}
			yyt2 = YYCURSOR;
			goto yy1950;
		} else {
			if (yych <= 0xF0) {
				yyt2 = YYCURSOR;
				goto yy1951;
			}
			if (yych <= 0xF3) {
				yyt2 = YYCURSOR;
				goto yy1952;
			}
			if (yych <= 0xF4) {
				yyt2 = YYCURSOR;
				goto yy1953;
			}
			goto yy1730;
		}
	}
yy3040:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+':
		case '-':
		case '/':
		case ':':
		case '@':
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case '_':
		case 'a':
		case 'c':
		case 'd':
		case 'f':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1932;
		case '.': goto yy2067;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy3057;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'S':
		case 's': goto yy1792;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy3041:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2953;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3042:
	yych = *++YYCURSOR;
	if (yych <= '2') {
		if (yych <= '/') {
			if (yych == ' ') goto yy1940;
			goto yy1730;
		} else {
			if (yych <= '0') goto yy3058;
			if (yych <= '1') goto yy3059;
			goto yy3060;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '9') goto yy3058;
			if (yych <= '@') goto yy1730;
			goto yy2952;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2952;
			goto yy1730;
		}
	}
yy3043:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3061;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2953;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3044:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy3061;
				goto yy2953;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2890;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2953;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2953;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3045:
	yych = *++YYCURSOR;
	if (yych == ' ') goto yy3062;
	goto yy1730;
yy3046:
	yyaccept = 42;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy3063;
	goto yy2975;
yy3047:
	yych = *++YYCURSOR;
	if (yych == 'x') goto yy3064;
	goto yy1730;
yy3048:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3065;
	goto yy1730;
yy3049:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3066;
	goto yy1730;
yy3050:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3067;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3067;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3067;
		goto yy1794;
	}
yy3051:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= ',') {
			if (yych <= '$') {
				if (yych == ' ') {
					yyt1 = YYCURSOR;
					goto yy3070;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			} else {
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy3071;
				}
				if (yych == '+') {
					yyt1 = YYCURSOR;
					goto yy3071;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy3072;
				}
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy3073;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy3074;
				}
				goto yy3076;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy3078;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy3079;
				}
				yyt1 = YYCURSOR;
				goto yy3080;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy3081;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				yyt1 = YYCURSOR;
				goto yy3082;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy3080;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy3081;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy3076;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy3083;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy3084;
				}
				yyt1 = YYCURSOR;
				goto yy3085;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy3086;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy3087;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy3088;
				}
				goto yy3076;
			}
		}
	}
yy3052:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'F') {
		if (yych <= ',') {
			if (yych <= '$') {
				if (yych == ' ') {
					yyt1 = YYCURSOR;
					goto yy3070;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			} else {
				if (yych <= '%') {
					yyt1 = YYCURSOR;
					goto yy3071;
				}
				if (yych == '+') {
					yyt1 = YYCURSOR;
					goto yy3071;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			}
		} else {
			if (yych <= '9') {
				if (yych <= '-') {
					yyt1 = YYCURSOR;
					goto yy3072;
				}
				if (yych <= '.') {
					yyt1 = YYCURSOR;
					goto yy3073;
				}
				if (yych <= '/') {
					yyt1 = YYCURSOR;
					goto yy3074;
				}
				goto yy3077;
			} else {
				if (yych <= ':') {
					yyt1 = YYCURSOR;
					goto yy3078;
				}
				if (yych <= '?') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				if (yych <= '@') {
					yyt1 = YYCURSOR;
					goto yy3079;
				}
				yyt1 = YYCURSOR;
				goto yy3080;
			}
		}
	} else {
		if (yych <= 0x7F) {
			if (yych <= '_') {
				if (yych <= 'Z') {
					yyt1 = YYCURSOR;
					goto yy3081;
				}
				if (yych <= '^') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				yyt1 = YYCURSOR;
				goto yy3082;
			} else {
				if (yych <= '`') {
					yyt1 = YYCURSOR;
					goto yy3068;
				}
				if (yych <= 'f') {
					yyt1 = YYCURSOR;
					goto yy3080;
				}
				if (yych <= 'z') {
					yyt1 = YYCURSOR;
					goto yy3081;
				}
				yyt1 = YYCURSOR;
				goto yy3068;
			}
		} else {
			if (yych <= 0xEF) {
				if (yych <= 0xC1) goto yy1783;
				if (yych <= 0xDF) {
					yyt1 = YYCURSOR;
					goto yy3083;
				}
				if (yych <= 0xE0) {
					yyt1 = YYCURSOR;
					goto yy3084;
				}
				yyt1 = YYCURSOR;
				goto yy3085;
			} else {
				if (yych <= 0xF0) {
					yyt1 = YYCURSOR;
					goto yy3086;
				}
				if (yych <= 0xF3) {
					yyt1 = YYCURSOR;
					goto yy3087;
				}
				if (yych <= 0xF4) {
					yyt1 = YYCURSOR;
					goto yy3088;
				}
				goto yy1783;
			}
		}
	}
yy3053:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '9') {
			if (yych == '%') {
				yyt1 = YYCURSOR;
				goto yy2623;
			}
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			goto yy1730;
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2890;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3054:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3089;
		if (yych <= ':') goto yy2951;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3089;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3089;
		goto yy1730;
	}
yy3055:
	yych = *++YYCURSOR;
	if (yych == ':') goto yy3018;
	goto yy1730;
yy3056:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy2819;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy2819;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy2819;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3057:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '^') {
		if (yych <= ':') {
			if (yych <= '+') {
				if (yych <= '$') {
					if (yych == ' ') {
						yyt1 = YYCURSOR;
						goto yy3070;
					}
					yyt1 = YYCURSOR;
					goto yy3068;
				} else {
					if (yych <= '%') {
						yyt1 = YYCURSOR;
						goto yy3071;
					}
					if (yych <= '*') {
						yyt1 = YYCURSOR;
						goto yy3068;
					}
					yyt1 = YYCURSOR;
					goto yy3071;
				}
			} else {
				if (yych <= '.') {
					if (yych <= ',') {
						yyt1 = YYCURSOR;
						goto yy3068;
					}
					if (yych <= '-') {
						yyt1 = YYCURSOR;
						goto yy3072;
					}
					yyt1 = YYCURSOR;
					goto yy3090;
				} else {
					if (yych <= '/') {
						yyt1 = YYCURSOR;
						goto yy3074;
					}
					if (yych <= '9') goto yy3091;
					yyt1 = YYCURSOR;
					goto yy3078;
				}
			}
		} else {
			if (yych <= 'G') {
				if (yych <= 'D') {
					if (yych <= '?') {
						yyt1 = YYCURSOR;
						goto yy3068;
					}
					if (yych <= '@') {
						yyt1 = YYCURSOR;
						goto yy3079;
					}
					yyt1 = YYCURSOR;
					goto yy3080;
				} else {
					if (yych <= 'E') {
						yyt1 = YYCURSOR;
						goto yy3092;
					}
					if (yych <= 'F') {
						yyt1 = YYCURSOR;
						goto yy3080;
					}
					yyt1 = YYCURSOR;
					goto yy3093;
				}
			} else {
				if (yych <= 'L') {
					if (yych == 'K') {
						yyt1 = YYCURSOR;
						goto yy3093;
					}
					yyt1 = YYCURSOR;
					goto yy3081;
				} else {
					if (yych <= 'M') {
						yyt1 = YYCURSOR;
						goto yy3093;
					}
					if (yych <= 'Z') {
						yyt1 = YYCURSOR;
						goto yy3081;
					}
					yyt1 = YYCURSOR;
					goto yy3068;
				}
			}
		}
	} else {
		if (yych <= 'p') {
			if (yych <= 'g') {
				if (yych <= 'd') {
					if (yych <= '_') {
						yyt1 = YYCURSOR;
						goto yy3082;
					}
					if (yych <= '`') {
						yyt1 = YYCURSOR;
						goto yy3068;
					}
					yyt1 = YYCURSOR;
					goto yy3080;
				} else {
					if (yych <= 'e') {
						yyt1 = YYCURSOR;
						goto yy3092;
					}
					if (yych <= 'f') {
						yyt1 = YYCURSOR;
						goto yy3080;
					}
					yyt1 = YYCURSOR;
					goto yy3093;
				}
			} else {
				if (yych <= 'l') {
					if (yych == 'k') {
						yyt1 = YYCURSOR;
						goto yy3093;
					}
					yyt1 = YYCURSOR;
					goto yy3081;
				} else {
					if (yych <= 'm') {
						yyt1 = YYCURSOR;
						goto yy3094;
					}
					if (yych <= 'o') {
						yyt1 = YYCURSOR;
						goto yy3081;
					}
					yyt1 = YYCURSOR;
					goto yy3095;
				}
			}
		} else {
			if (yych <= 0xDF) {
				if (yych <= 'z') {
					if (yych == 'u') {
						yyt1 = YYCURSOR;
						goto yy3095;
					}
					yyt1 = YYCURSOR;
					goto yy3081;
				} else {
					if (yych <= 0x7F) {
						yyt1 = YYCURSOR;
						goto yy3068;
					}
					if (yych <= 0xC1) goto yy1683;
					yyt1 = YYCURSOR;
					goto yy3083;
				}
			} else {
				if (yych <= 0xF0) {
					if (yych <= 0xE0) {
						yyt1 = YYCURSOR;
						goto yy3084;
					}
					if (yych <= 0xEF) {
						yyt1 = YYCURSOR;
						goto yy3085;
					}
					yyt1 = YYCURSOR;
					goto yy3086;
				} else {
					if (yych <= 0xF3) {
						yyt1 = YYCURSOR;
						goto yy3087;
					}
					if (yych <= 0xF4) {
						yyt1 = YYCURSOR;
						goto yy3088;
					}
					goto yy1683;
				}
			}
		}
	}
yy3058:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3096;
				yyt2 = YYCURSOR;
				goto yy2990;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2989;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2989;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3059:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3097;
				yyt2 = YYCURSOR;
				goto yy2990;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2989;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2989;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3060:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '5') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '4') goto yy3097;
				goto yy3098;
			}
		} else {
			if (yych <= '@') {
				if (yych <= '9') goto yy3096;
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2990;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy2989;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy2989;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3061:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy2991;
				yyt2 = YYCURSOR;
				goto yy2890;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2991;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy2991;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3062:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1730;
			if (yych <= '\t') {
				yyt1 = YYCURSOR;
				goto yy3099;
			}
			goto yy1730;
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy3099;
			}
			if (yych <= 0xC1) goto yy1730;
			yyt1 = YYCURSOR;
			goto yy3100;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy3101;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy3102;
			}
			yyt1 = YYCURSOR;
			goto yy3103;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy3104;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy3105;
			}
			goto yy1730;
		}
	}
yy3063:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3106;
	goto yy1730;
yy3064:
	yych = *++YYCURSOR;
	if (yych == '8') goto yy3107;
	goto yy1730;
yy3065:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3108;
	goto yy1730;
yy3066:
	yyaccept = 43;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '.') goto yy3109;
	goto yy3009;
yy3067:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3110;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3110;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3110;
		goto yy1794;
	}
yy3068:
	++YYCURSOR;
yy3069:
	YYCURSOR = yyt1;
#line 406 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_CREDIT_CARD_NUMBER);
           if (!this->is_credit_card(this->to_string_fragment(cap_all))) {
               if (cap_all.length() > 16) {
                   cap_all.c_end = cap_all.c_begin + 4;
                   cap_inner.c_end = cap_inner.c_begin + 4;
               }
               this->ds_next_offset = cap_all.c_end;
               token_out = DT_HEX_DUMP;
           }
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 83737 "../../lnav/src/data_scanner_re.cc"
yy3070:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych == '%') goto yy1907;
	goto yy3069;
yy3071:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '/') {
		if (yych <= '*') {
			if (yych == '%') goto yy1746;
			goto yy3069;
		} else {
			if (yych == ',') goto yy3069;
			if (yych <= '.') goto yy1746;
			goto yy3069;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '9') goto yy1746;
			if (yych <= '?') goto yy3069;
			if (yych <= 'Z') goto yy1746;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1746;
			goto yy3069;
		}
	}
yy3072:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy3069;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy3069;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy3069;
			goto yy1793;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1793;
			goto yy3069;
		}
	}
yy3073:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy3069;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy3069;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy3069;
			goto yy2159;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1793;
			goto yy3069;
		}
	}
yy3074:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '/') goto yy3069;
		if (yych <= '9') goto yy1912;
		if (yych <= '@') goto yy3069;
		goto yy1912;
	} else {
		if (yych <= '_') {
			if (yych <= '^') goto yy3069;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy3069;
			if (yych <= 'z') goto yy1912;
			goto yy3069;
		}
	}
yy3075:
	yyaccept = 11;
	yych = *(YYMARKER = ++YYCURSOR);
yy3076:
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= '9') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1781;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1781;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '7') goto yy3075;
			}
		}
	} else {
		if (yych <= 'Z') {
			if (yych <= '?') {
				if (yych <= ':') goto yy1927;
				goto yy1781;
			} else {
				if (yych <= '@') goto yy1785;
				if (yych <= 'F') goto yy1931;
				goto yy1790;
			}
		} else {
			if (yych <= '`') {
				if (yych == '_') goto yy1793;
				goto yy1781;
			} else {
				if (yych <= 'f') goto yy1931;
				if (yych <= 'z') goto yy1790;
				goto yy1781;
			}
		}
	}
yy3077:
	yyaccept = 12;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	if (yych <= ':') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy1783;
				goto yy1775;
			} else {
				if (yych == '+') goto yy1745;
				goto yy1783;
			}
		} else {
			if (yych <= '.') {
				if (yych <= '-') goto yy1777;
				goto yy2033;
			} else {
				if (yych <= '/') goto yy1779;
				if (yych <= '9') goto yy3077;
				goto yy1927;
			}
		}
	} else {
		if (yych <= '^') {
			if (yych <= '@') {
				if (yych <= '?') goto yy1783;
				goto yy1785;
			} else {
				if (yych <= 'F') goto yy1931;
				if (yych <= 'Z') goto yy1790;
				goto yy1783;
			}
		} else {
			if (yych <= '`') {
				if (yych <= '_') goto yy1793;
				goto yy1783;
			} else {
				if (yych <= 'f') goto yy1931;
				if (yych <= 'z') goto yy1790;
				goto yy1783;
			}
		}
	}
yy3078:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '/') goto yy1921;
	if (yych == ':') goto yy1779;
	goto yy3069;
yy3079:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '.') {
			if (yych <= ',') goto yy3069;
			goto yy1850;
		} else {
			if (yych <= '/') goto yy3069;
			if (yych <= '9') goto yy1925;
			goto yy3069;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'Z') goto yy1925;
			if (yych <= '^') goto yy3069;
			goto yy1912;
		} else {
			if (yych <= '`') goto yy3069;
			if (yych <= 'z') goto yy1925;
			goto yy3069;
		}
	}
yy3080:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1932;
			goto yy3069;
		} else {
			if (yych == ',') goto yy3069;
			goto yy1932;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= 'Z') goto yy1932;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1932;
			goto yy3069;
		}
	}
yy3081:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1791;
			goto yy3069;
		} else {
			if (yych == ',') goto yy3069;
			goto yy1791;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= 'Z') goto yy1791;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1791;
			goto yy3069;
		}
	}
yy3082:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1794;
			goto yy3069;
		} else {
			if (yych == ',') goto yy3069;
			goto yy1794;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= 'Z') goto yy1794;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1794;
			goto yy3069;
		}
	}
yy3083:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3068;
	goto yy1730;
yy3084:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy3083;
	goto yy1730;
yy3085:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3083;
	goto yy1730;
yy3086:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy3085;
	goto yy1730;
yy3087:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3085;
	goto yy1730;
yy3088:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy3085;
	goto yy1730;
yy3089:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy2986;
			goto yy3111;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy2986;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy2986;
			goto yy2155;
		}
	}
yy3090:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '9') {
		if (yych <= '+') {
			if (yych == '%') goto yy1745;
			if (yych <= '*') goto yy3069;
			goto yy1745;
		} else {
			if (yych <= ',') goto yy3069;
			if (yych <= '.') goto yy1745;
			if (yych <= '/') goto yy3069;
			goto yy2205;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= '@') goto yy1747;
			if (yych <= 'Z') goto yy1793;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1793;
			goto yy3069;
		}
	}
yy3091:
	yyaccept = 3;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yybm_init[1024+yych] & 1) {
		goto yy1773;
	}
	switch (yych) {
		case '%': goto yy1775;
		case '+': goto yy1745;
		case '-': goto yy1777;
		case '.': goto yy2067;
		case '/': goto yy1779;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9': goto yy3091;
		case ':': goto yy1927;
		case '@': goto yy1785;
		case 'A':
		case 'C':
		case 'D':
		case 'F':
		case 'a':
		case 'c':
		case 'd':
		case 'f': goto yy1931;
		case 'B':
		case 'b': goto yy2740;
		case 'E':
		case 'e': goto yy2741;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy1788;
		case 'H':
		case 'I':
		case 'J':
		case 'L':
		case 'N':
		case 'O':
		case 'P':
		case 'Q':
		case 'R':
		case 'T':
		case 'U':
		case 'V':
		case 'W':
		case 'X':
		case 'Y':
		case 'Z':
		case 'h':
		case 'i':
		case 'j':
		case 'l':
		case 'n':
		case 'o':
		case 'q':
		case 'r':
		case 't':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': goto yy1790;
		case 'S':
		case 's': goto yy1792;
		case '_': goto yy1793;
		case 'm': goto yy1795;
		case 'p':
		case 'u': goto yy1796;
		default: goto yy1683;
	}
yy3092:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1932;
			goto yy3069;
		} else {
			if (yych <= '+') goto yy1937;
			if (yych <= ',') goto yy3069;
			if (yych <= '-') goto yy2064;
			goto yy1932;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3069;
			if (yych <= 'Z') goto yy1932;
			goto yy3069;
		} else {
			if (yych == '`') goto yy3069;
			if (yych <= 'z') goto yy1932;
			goto yy3069;
		}
	}
yy3093:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'B') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy3069;
			goto yy1791;
		} else {
			if (yych <= ':') {
				if (yych <= ',') goto yy3069;
				goto yy1791;
			} else {
				if (yych <= '?') goto yy3069;
				if (yych <= 'A') goto yy1791;
				goto yy1792;
			}
		}
	} else {
		if (yych <= 'a') {
			if (yych <= '^') {
				if (yych <= 'Z') goto yy1791;
				goto yy3069;
			} else {
				if (yych == '`') goto yy3069;
				goto yy1791;
			}
		} else {
			if (yych <= 'h') {
				if (yych <= 'b') goto yy1792;
				goto yy1791;
			} else {
				if (yych <= 'i') goto yy1928;
				if (yych <= 'z') goto yy1791;
				goto yy3069;
			}
		}
	}
yy3094:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'S') {
		if (yych <= ',') {
			if (yych <= '%') {
				if (yych <= '$') goto yy3069;
				goto yy1791;
			} else {
				if (yych == '+') goto yy1791;
				goto yy3069;
			}
		} else {
			if (yych <= 'A') {
				if (yych <= ':') goto yy1791;
				if (yych <= '?') goto yy3069;
				goto yy1791;
			} else {
				if (yych <= 'B') goto yy1792;
				if (yych <= 'R') goto yy1791;
				goto yy1792;
			}
		}
	} else {
		if (yych <= 'b') {
			if (yych <= '_') {
				if (yych <= 'Z') goto yy1791;
				if (yych <= '^') goto yy3069;
				goto yy1791;
			} else {
				if (yych <= '`') goto yy3069;
				if (yych <= 'a') goto yy1791;
				goto yy1792;
			}
		} else {
			if (yych <= 'r') {
				if (yych == 'i') goto yy1930;
				goto yy1791;
			} else {
				if (yych <= 's') goto yy1792;
				if (yych <= 'z') goto yy1791;
				goto yy3069;
			}
		}
	}
yy3095:
	yyaccept = 44;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'R') {
		if (yych <= '+') {
			if (yych == '%') goto yy1791;
			if (yych <= '*') goto yy3069;
			goto yy1791;
		} else {
			if (yych <= ',') goto yy3069;
			if (yych <= ':') goto yy1791;
			if (yych <= '?') goto yy3069;
			goto yy1791;
		}
	} else {
		if (yych <= '_') {
			if (yych <= 'S') goto yy1792;
			if (yych <= 'Z') goto yy1791;
			if (yych <= '^') goto yy3069;
			goto yy1791;
		} else {
			if (yych <= 'r') {
				if (yych <= '`') goto yy3069;
				goto yy1791;
			} else {
				if (yych <= 's') goto yy1792;
				if (yych <= 'z') goto yy1791;
				goto yy3069;
			}
		}
	}
yy3096:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3020;
				yyt2 = YYCURSOR;
				goto yy2990;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3020;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3020;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3097:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3112;
				yyt2 = YYCURSOR;
				goto yy2990;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3020;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3020;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3098:
	yych = *++YYCURSOR;
	if (yych <= '`') {
		if (yych <= '9') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '5') goto yy3112;
				goto yy3020;
			}
		} else {
			if (yych <= '@') {
				if (yych <= ':') {
					yyt2 = YYCURSOR;
					goto yy2990;
				}
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 'F') goto yy3020;
				if (yych <= 'Z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			}
		}
	} else {
		if (yych <= 0xDF) {
			if (yych <= 'z') {
				if (yych <= 'f') goto yy3020;
				goto yy1730;
			} else {
				if (yych <= 0x7F) {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 0xC1) goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1948;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xE0) {
					yyt2 = YYCURSOR;
					goto yy1949;
				}
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3099:
	yych = *++YYCURSOR;
	if (yybm_init[0+yych] & 128) {
		goto yy3099;
	}
	if (yych <= 0xE0) {
		if (yych <= '\n') {
			if (yych <= 0x00) goto yy1730;
			goto yy3113;
		} else {
			if (yych <= 0xC1) goto yy1730;
			if (yych >= 0xE0) goto yy3101;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xEF) goto yy3102;
			goto yy3103;
		} else {
			if (yych <= 0xF3) goto yy3104;
			if (yych <= 0xF4) goto yy3105;
			goto yy1730;
		}
	}
yy3100:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3099;
	goto yy1730;
yy3101:
	yych = *++YYCURSOR;
	if (yych <= 0x9F) goto yy1730;
	if (yych <= 0xBF) goto yy3100;
	goto yy1730;
yy3102:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3100;
	goto yy1730;
yy3103:
	yych = *++YYCURSOR;
	if (yych <= 0x8F) goto yy1730;
	if (yych <= 0xBF) goto yy3102;
	goto yy1730;
yy3104:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0xBF) goto yy3102;
	goto yy1730;
yy3105:
	yych = *++YYCURSOR;
	if (yych <= 0x7F) goto yy1730;
	if (yych <= 0x8F) goto yy3102;
	goto yy1730;
yy3106:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3114;
	goto yy1730;
yy3107:
	yych = *++YYCURSOR;
	if (yych == '6') goto yy3115;
	goto yy1730;
yy3108:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3116;
	goto yy1730;
yy3109:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3117;
	goto yy1730;
yy3110:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy3118;
	goto yy1794;
yy3111:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3119;
		if (yych <= ':') goto yy2988;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3119;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3119;
		goto yy1730;
	}
yy3112:
	yych = *++YYCURSOR;
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '.') {
				if (yych <= '-') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt1 = YYCURSOR;
				goto yy2068;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3039;
				yyt2 = YYCURSOR;
				goto yy2990;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3039;
			} else {
				if (yych <= 'Z') goto yy1730;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3039;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy1730;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy1730;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3113:
	++YYCURSOR;
	hunk_heading = yyt1;
#line 373 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_DIFF_HUNK_HEADING);
           cap_inner.c_begin = hunk_heading.val - this->ds_input.udata();
           cap_inner.c_end -= 1;
           this->ds_bol = true;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 84711 "../../lnav/src/data_scanner_re.cc"
yy3114:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3120;
	goto yy1730;
yy3115:
	yych = *++YYCURSOR;
	if (yych == ')') goto yy1769;
	goto yy1730;
yy3116:
	yych = *++YYCURSOR;
	if (yych <= 0xDF) {
		if (yych <= '9') {
			if (yych <= '/') {
				yyt1 = YYCURSOR;
				goto yy3068;
			}
			goto yy1730;
		} else {
			if (yych <= 0x7F) {
				yyt1 = YYCURSOR;
				goto yy3068;
			}
			if (yych <= 0xC1) goto yy1730;
			yyt1 = YYCURSOR;
			goto yy3083;
		}
	} else {
		if (yych <= 0xF0) {
			if (yych <= 0xE0) {
				yyt1 = YYCURSOR;
				goto yy3084;
			}
			if (yych <= 0xEF) {
				yyt1 = YYCURSOR;
				goto yy3085;
			}
			yyt1 = YYCURSOR;
			goto yy3086;
		} else {
			if (yych <= 0xF3) {
				yyt1 = YYCURSOR;
				goto yy3087;
			}
			if (yych <= 0xF4) {
				yyt1 = YYCURSOR;
				goto yy3088;
			}
			goto yy1730;
		}
	}
yy3117:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3121;
	goto yy1730;
yy3118:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy3122;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy3122;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy3122;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy3119:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3123;
		if (yych <= ':') goto yy3018;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3123;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3123;
		goto yy1730;
	}
yy3120:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy2975;
	if (yych <= '9') goto yy3124;
	goto yy2975;
yy3121:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy1730;
	if (yych <= '9') goto yy3125;
	goto yy1730;
yy3122:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3126;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3126;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3126;
		goto yy1794;
	}
yy3123:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '-') {
			if (yych <= ',') goto yy2155;
			goto yy2318;
		} else {
			if (yych <= '/') goto yy2155;
			if (yych <= '9') goto yy3037;
			goto yy3127;
		}
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy2155;
			goto yy3037;
		} else {
			if (yych <= '`') goto yy2155;
			if (yych <= 'f') goto yy3037;
			goto yy2155;
		}
	}
yy3124:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy2975;
	if (yych <= '9') goto yy3128;
	goto yy2975;
yy3125:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy3009;
	if (yych <= '9') goto yy3129;
	goto yy3009;
yy3126:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3130;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3130;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3130;
		goto yy1794;
	}
yy3127:
	yych = *++YYCURSOR;
	if (yych <= '@') {
		if (yych <= '/') goto yy1730;
		if (yych <= '9') goto yy3131;
		if (yych <= ':') goto yy2819;
		goto yy1730;
	} else {
		if (yych <= 'F') goto yy3131;
		if (yych <= '`') goto yy1730;
		if (yych <= 'f') goto yy3131;
		goto yy1730;
	}
yy3128:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy2975;
	if (yych <= '9') goto yy3132;
	goto yy2975;
yy3129:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy3009;
	if (yych <= '9') goto yy3133;
	goto yy3009;
yy3130:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych <= '9') goto yy3134;
		goto yy1794;
	} else {
		if (yych <= 'F') goto yy3134;
		if (yych <= '`') goto yy1794;
		if (yych <= 'f') goto yy3134;
		goto yy1794;
	}
yy3131:
	yych = *++YYCURSOR;
	if (yych <= 'z') {
		if (yych <= '@') {
			if (yych <= '/') {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= '9') goto yy3135;
			if (yych <= ':') {
				yyt2 = YYCURSOR;
				goto yy1945;
			}
			yyt2 = YYCURSOR;
			goto yy1940;
		} else {
			if (yych <= 'Z') {
				if (yych <= 'F') goto yy3135;
				goto yy1730;
			} else {
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= 'f') goto yy3135;
				goto yy1730;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				yyt2 = YYCURSOR;
				goto yy1940;
			}
			if (yych <= 0xC1) goto yy1730;
			if (yych <= 0xDF) {
				yyt2 = YYCURSOR;
				goto yy1948;
			}
			yyt2 = YYCURSOR;
			goto yy1949;
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy1730;
			}
		}
	}
yy3132:
	++YYCURSOR;
	goto yy2975;
yy3133:
	yych = *++YYCURSOR;
	if (yych <= '/') goto yy3009;
	if (yych <= '9') goto yy3136;
	goto yy3009;
yy3134:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych == '-') goto yy3137;
	goto yy1794;
yy3135:
	yyaccept = 30;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'f') {
		if (yych <= ':') {
			if (yych <= '-') {
				if (yych <= ',') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				yyt2 = YYCURSOR;
				goto yy3138;
			} else {
				if (yych <= '/') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				if (yych <= '9') goto yy3056;
				yyt2 = YYCURSOR;
				goto yy3139;
			}
		} else {
			if (yych <= 'F') {
				if (yych <= '@') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3056;
			} else {
				if (yych <= 'Z') goto yy2155;
				if (yych <= '`') {
					yyt2 = YYCURSOR;
					goto yy1940;
				}
				goto yy3056;
			}
		}
	} else {
		if (yych <= 0xE0) {
			if (yych <= 0x7F) {
				if (yych <= 'z') goto yy2155;
				yyt2 = YYCURSOR;
				goto yy1940;
			} else {
				if (yych <= 0xC1) goto yy2155;
				if (yych <= 0xDF) {
					yyt2 = YYCURSOR;
					goto yy1948;
				}
				yyt2 = YYCURSOR;
				goto yy1949;
			}
		} else {
			if (yych <= 0xF0) {
				if (yych <= 0xEF) {
					yyt2 = YYCURSOR;
					goto yy1950;
				}
				yyt2 = YYCURSOR;
				goto yy1951;
			} else {
				if (yych <= 0xF3) {
					yyt2 = YYCURSOR;
					goto yy1952;
				}
				if (yych <= 0xF4) {
					yyt2 = YYCURSOR;
					goto yy1953;
				}
				goto yy2155;
			}
		}
	}
yy3136:
	++YYCURSOR;
	goto yy3009;
yy3137:
	yych = *++YYCURSOR;
	if (yych <= 'Z') {
		if (yych <= '9') {
			if (yych <= '/') goto yy1746;
			goto yy3140;
		} else {
			if (yych <= '@') goto yy1746;
			if (yych <= 'F') goto yy3140;
			goto yy1793;
		}
	} else {
		if (yych <= '`') {
			if (yych == '_') goto yy1793;
			goto yy1746;
		} else {
			if (yych <= 'f') goto yy3140;
			if (yych <= 'z') goto yy1793;
			goto yy1746;
		}
	}
yy3138:
	yyaccept = 24;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1941;
		if (yych <= '9') goto yy2447;
		goto yy1941;
	} else {
		if (yych <= 'F') goto yy2447;
		if (yych <= '`') goto yy1941;
		if (yych <= 'f') goto yy2447;
		goto yy1941;
	}
yy3139:
	yych = *++YYCURSOR;
	if (yych <= '9') {
		if (yych == ' ') goto yy1940;
		if (yych <= '/') goto yy1730;
		goto yy2447;
	} else {
		if (yych <= 'F') {
			if (yych <= '@') goto yy1730;
			goto yy2447;
		} else {
			if (yych <= '`') goto yy1730;
			if (yych <= 'f') goto yy2447;
			goto yy1730;
		}
	}
yy3140:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3141;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3141:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3142;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3142:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3143;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3143:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3144;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3144:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3145;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3145:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3146;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3146:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3147;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3147:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3148;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3148:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3149;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3149:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3150;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3150:
	yyaccept = 13;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= '@') {
		if (yych <= '/') goto yy1794;
		if (yych >= ':') goto yy1794;
	} else {
		if (yych <= 'F') goto yy3151;
		if (yych <= '`') goto yy1794;
		if (yych >= 'g') goto yy1794;
	}
yy3151:
	yyaccept = 45;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= ':') {
		if (yych <= '*') {
			if (yych == '%') goto yy1794;
		} else {
			if (yych != ',') goto yy1794;
		}
	} else {
		if (yych <= '^') {
			if (yych <= '?') goto yy3152;
			if (yych <= 'Z') goto yy1794;
		} else {
			if (yych == '`') goto yy3152;
			if (yych <= 'z') goto yy1794;
		}
	}
yy3152:
#line 404 "../../lnav/src/data_scanner_re.re"
	{ RET(DT_UUID); }
#line 85249 "../../lnav/src/data_scanner_re.cc"
/* *********************************** */
yyc_dbldocstring:
	yych = *YYCURSOR;
	if (yych <= 0x00) goto yy3154;
	if (yych == '"') goto yy3157;
	goto yy3155;
yy3154:
	++YYCURSOR;
#line 212 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 85263 "../../lnav/src/data_scanner_re.cc"
yy3155:
	++YYCURSOR;
yy3156:
#line 217 "../../lnav/src/data_scanner_re.re"
	{
           goto yyc_dbldocstring;
       }
#line 85271 "../../lnav/src/data_scanner_re.cc"
yy3157:
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych != '"') goto yy3156;
	yych = *++YYCURSOR;
	if (yych == '"') goto yy3158;
	YYCURSOR = YYMARKER;
	goto yy3156;
yy3158:
	++YYCURSOR;
#line 206 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           cap_inner.c_end -= 3;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 85287 "../../lnav/src/data_scanner_re.cc"
/* *********************************** */
yyc_sdocstring:
	yych = *YYCURSOR;
	if (yych <= 0x00) goto yy3160;
	if (yych == '\'') goto yy3163;
	goto yy3161;
yy3160:
	++YYCURSOR;
#line 234 "../../lnav/src/data_scanner_re.re"
	{
           CAPTURE(DT_QUOTED_STRING);
           cap_inner.c_end -= 3;
           return tokenize_result{token_out, cap_all, cap_inner, this->ds_input.data()};
       }
#line 85302 "../../lnav/src/data_scanner_re.cc"
yy3161:
	++YYCURSOR;
yy3162:
#line 240 "../../lnav/src/data_scanner_re.re"
	{
           goto yyc_sdocstring;
       }
#line 85310 "../../lnav/src/data_scanner_re.cc"
yy3163:
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych != '\'') goto yy3162;
	yych = *++YYCURSOR;
	if (yych == '\'') goto yy3160;
	YYCURSOR = YYMARKER;
	goto yy3162;
/* *********************************** */
yyc_units:
	yych = *YYCURSOR;
	switch (yych) {
		case 'B':
		case 'S':
		case 'b':
		case 's': goto yy3167;
		case 'G':
		case 'K':
		case 'M':
		case 'g':
		case 'k': goto yy3169;
		case 'm': goto yy3170;
		case 'p':
		case 'u': goto yy3171;
		default: goto yy3165;
	}
yy3165:
	++YYCURSOR;
yy3166:
#line 175 "../../lnav/src/data_scanner_re.re"
	{ return std::nullopt; }
#line 85341 "../../lnav/src/data_scanner_re.cc"
yy3167:
	++YYCURSOR;
yy3168:
#line 438 "../../lnav/src/data_scanner_re.re"
	{
           RET(DT_UNIT);
       }
#line 85349 "../../lnav/src/data_scanner_re.cc"
yy3169:
	yyaccept = 0;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych == 'B') goto yy3167;
		goto yy3166;
	} else {
		if (yych <= 'b') goto yy3167;
		if (yych == 'i') goto yy3172;
		goto yy3166;
	}
yy3170:
	yyaccept = 1;
	yych = *(YYMARKER = ++YYCURSOR);
	if (yych <= 'a') {
		if (yych <= 'B') {
			if (yych <= 'A') goto yy3168;
			goto yy3167;
		} else {
			if (yych == 'S') goto yy3167;
			goto yy3168;
		}
	} else {
		if (yych <= 'i') {
			if (yych <= 'b') goto yy3167;
			if (yych <= 'h') goto yy3168;
			goto yy3174;
		} else {
			if (yych == 's') goto yy3167;
			goto yy3168;
		}
	}
yy3171:
	yych = *++YYCURSOR;
	if (yych == 'S') goto yy3167;
	if (yych == 's') goto yy3167;
	goto yy3166;
yy3172:
	yych = *++YYCURSOR;
	if (yych == 'B') goto yy3167;
	if (yych == 'b') goto yy3167;
yy3173:
	YYCURSOR = YYMARKER;
	if (yyaccept == 0) {
		goto yy3166;
	} else {
		goto yy3168;
	}
yy3174:
	yych = *++YYCURSOR;
	if (yych <= 'a') {
		if (yych == 'B') goto yy3167;
		goto yy3173;
	} else {
		if (yych <= 'b') goto yy3167;
		if (yych == 'n') goto yy3167;
		goto yy3173;
	}
}
#line 493 "../../lnav/src/data_scanner_re.re"


    return std::nullopt;
}
